/* giram.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999 David Odin <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "giram.h"
#include "about_dialog.h"
#include "help_dialog.h"
#include "frame.h"
#include "csgtree.h"
#include "LoadPov.h"
#include "dxf.h"
#ifdef WITH_S3D
#include "s3d.h"
#endif
#include "SavePov.h"
#include "bicubic_patch.h"
#include "blob.h"
#include "box.h"
#include "cone.h"
#include "cubic.h"
#include "cylinder.h"
#include "disc.h"
#include "height_field.h"
#include "julia_fractal.h"
#include "lathe.h"
#include "mesh.h"
#include "plane.h"
#include "poly.h"
#include "polygon.h"
#include "prism.h"
#include "quadric.h"
#include "quartic.h"
#include "sor.h"
#include "sphere.h"
#include "superellipsoid.h"
#include "text.h"
#include "torus.h"
#include "triangle.h"
#include "csg.h"
#include "tool_rotate.h"
#include "tool_scale.h"
#include "tool_options.h"
#include "tools.h"
#include "view.h"
#include "view_camera.h"
#include "parserc.h"
#include "plugins.h"
#include "preferences.h"

gint ModeTool;
gint number_of_shapes;

/*****************************************************************************
*  CreateDefaultTexture
******************************************************************************/
void CreateDefaultTexture(void)
{
  DefaultTexture = g_malloc(sizeof(TextureStruct));
  DefaultTexture->Pigment = g_malloc(sizeof(PigmentStruct));
  DefaultTexture->Pigment->Type = PAT_SOLID_COLOR;
  DefaultTexture->Pigment->Trans = NULL;
  DefaultTexture->Pigment->PigmentList = NULL;
  DefaultTexture->Pigment->Color[0] = 1.0;
  DefaultTexture->Pigment->Color[1] = 1.0;
  DefaultTexture->Pigment->Color[2] = 1.0;
  DefaultTexture->Pigment->Color[3] = 0.0;
  DefaultTexture->Pigment->Color[4] = 0.0;
  DefaultTexture->Pigment->Warp = NULL;
  DefaultTexture->Pigment->ColorMap = NULL;
  DefaultTexture->Normal = NULL;
  DefaultTexture->Finish = NULL;
  DefaultTexture->Halo = NULL;
  DefaultTexture->Trans = NULL;
  DefaultTexture->NextLayer = NULL;
}

/*****************************************************************************
*  giramInit
******************************************************************************/
void giramInit()
{
#ifdef HAVE_GL
  Display *dpy;
  XVisualInfo *vi;
  int attrib[] =
  {
    GLX_RGBA,
    GLX_DOUBLEBUFFER,
    None
  };
#endif
  LibraryPaths = NULL;
  FirstFrame = NULL;
  CSGTreeDialog = NULL;
  ParseGiramRCFile();
  EvenlyScaleFlag = TRUE;
  ConeFunnelFlag = TRUE;
  SnapToAngleFlag = FALSE;
  SnapAngle = 15.0;

  /* End of Parse RC FILEs... */
  InitPlugins();
#ifdef HAVE_GL
  ReallyHaveGL = FALSE;
  if (glXQueryExtension(GDK_DISPLAY(),NULL,NULL))
  {
    printf("It seems you actually have OpenGL installed, Good!\n");
    dpy = GDK_DISPLAY();
    vi = glXChooseVisual(dpy,DefaultScreen(dpy), attrib);
    if (vi)
    {
      GLVisual = gdkx_visual_get(vi->visualid);
      if (GLVisual)
        ReallyHaveGL = TRUE;
      XFree(vi);
    }
  } else
  {
    printf("OpenGL wasn't found, too bad...\n");
  }
#endif
  CreateDefaultTexture();
  CreateToolOptionsDialog();
  ModeTool = MT_SELECT;
  ShowOptionsForTool(ModeTool);
}

/*****************************************************************************
*  comparTri FIXME: this had nothing to do here. -> view.c
******************************************************************************/
int comparTri(const void *Tri1, const void *Tri2)
{
  return (((Triangle2D *)Tri2)->distance - ((Triangle2D *)Tri1)->distance);
}

typedef struct RasterLine
{
  int minX, maxX;
  double MinColor[3];
  double MaxColor[3];
  double MinDistance, MaxDistance;
} RasterLine;

/*****************************************************************************
*  DrawGouraudTriangleZBuffer FIXME: this had nothing to do here. -> view.c
******************************************************************************/
void DrawGouraudTriangleZBuffer(GtkWidget *View, double *Zbuffer,
                                Triangle2DNormCol *Tri)
{
  RasterLine *ListLine;
  int minY, maxY;
  int deltaX, deltaY;
  int offx,offy;
  register int i,j;
  int x,y;
  register int bx;
  double distance;
  GdkColor Color;
  GdkColormap *colormap;
  GdkGC *gc;
  
  gc = gdk_gc_new(View->window);
  colormap = gdk_window_get_colormap(View->window);
  maxY = minY = Tri->P[0][1];
  if (maxY < Tri->P[1][1]) maxY = Tri->P[1][1];
  if (minY > Tri->P[1][1]) minY = Tri->P[1][1];
  if (maxY < Tri->P[2][1]) maxY = Tri->P[2][1];
  if (minY > Tri->P[2][1]) minY = Tri->P[2][1];
  if (minY < 0) return;
  if (maxY >= View->allocation.height) return;
  ListLine = g_malloc((maxY-minY+1)*sizeof(RasterLine));
  for(i=minY;i<=maxY;i++)
  {
    ListLine[i-minY].minX = 2<<29;
    ListLine[i-minY].maxX = -(2<<29);
  }
  /* First, Bresenham algorithm */
  /* Line P[0] -> P[1] */
  deltaX = abs(Tri->P[1][0]-Tri->P[0][0]);
  deltaY = abs(Tri->P[1][1]-Tri->P[0][1]);
  if (Tri->P[1][0] > Tri->P[0][0]) offx=1; else offx=-1;
  if (Tri->P[1][1] > Tri->P[0][1]) offy=1; else offy=-1;
  /* "Trace" du premier point */
  ListLine[Tri->P[0][1]-minY].minX=Tri->P[0][0];
  ListLine[Tri->P[0][1]-minY].maxX=Tri->P[0][0];
  ListLine[Tri->P[0][1]-minY].MinColor[0]=Tri->Color[0][0];
  ListLine[Tri->P[0][1]-minY].MinColor[1]=Tri->Color[0][1];
  ListLine[Tri->P[0][1]-minY].MinColor[2]=Tri->Color[0][2];
  ListLine[Tri->P[0][1]-minY].MaxColor[0]=Tri->Color[0][0];
  ListLine[Tri->P[0][1]-minY].MaxColor[1]=Tri->Color[0][1];
  ListLine[Tri->P[0][1]-minY].MaxColor[2]=Tri->Color[0][2];
  ListLine[Tri->P[0][1]-minY].MinDistance =
  ListLine[Tri->P[0][1]-minY].MaxDistance = Tri->Distance[0];
  if (deltaX > deltaY)
  { /* Plutot horizontal */
    x = Tri->P[0][0];
    y = Tri->P[0][1];
    bx = 2*deltaY-deltaX;
    while(x != Tri->P[1][0])
    {
      x+=offx;
      if (bx<0)
        bx+=2*deltaY;
      else
      {
        bx+=2*(deltaY-deltaX);
        y+=offy;
      }
      /* "Trace" du point */
      if (ListLine[y-minY].minX > x)
      {
        ListLine[y-minY].minX = x;
        ListLine[y-minY].MinColor[0]=Tri->Color[0][0]+((double)x-Tri->P[0][0])/(Tri->P[1][0]-Tri->P[0][0])*(Tri->Color[1][0]-Tri->Color[0][0]);
        ListLine[y-minY].MinColor[1]=Tri->Color[0][1]+((double)x-Tri->P[0][0])/(Tri->P[1][0]-Tri->P[0][0])*(Tri->Color[1][1]-Tri->Color[0][1]);
        ListLine[y-minY].MinColor[2]=Tri->Color[0][2]+((double)x-Tri->P[0][0])/(Tri->P[1][0]-Tri->P[0][0])*(Tri->Color[1][2]-Tri->Color[0][2]);
        ListLine[y-minY].MinDistance=Tri->Distance[0]+((double)x-Tri->P[0][0])/(Tri->P[1][0]-Tri->P[0][0])*(Tri->Distance[1]-Tri->Distance[0]);
      }
      if (ListLine[y-minY].maxX < x)
      {
        ListLine[y-minY].maxX = x;
        ListLine[y-minY].MaxColor[0]=Tri->Color[0][0]+((double)x-Tri->P[0][0])/(Tri->P[1][0]-Tri->P[0][0])*(Tri->Color[1][0]-Tri->Color[0][0]);
        ListLine[y-minY].MaxColor[1]=Tri->Color[0][1]+((double)x-Tri->P[0][0])/(Tri->P[1][0]-Tri->P[0][0])*(Tri->Color[1][1]-Tri->Color[0][1]);
        ListLine[y-minY].MaxColor[2]=Tri->Color[0][2]+((double)x-Tri->P[0][0])/(Tri->P[1][0]-Tri->P[0][0])*(Tri->Color[1][2]-Tri->Color[0][2]);
        ListLine[y-minY].MaxDistance=Tri->Distance[0]+((double)x-Tri->P[0][0])/(Tri->P[1][0]-Tri->P[0][0])*(Tri->Distance[1]-Tri->Distance[0]);
      }
    }
  } else
  { /* Plutot vertical */
    x = Tri->P[0][0];
    y = Tri->P[0][1];
    bx = 2*deltaX-deltaY;
    while (y != Tri->P[1][1])
    {
      y+=offy;
      if (bx<0) bx+=2*deltaX;
      else
      {
        bx+=2*(deltaX-deltaY);
        x+=offx;
      }
      /* "Trace" du point */
      if (ListLine[y-minY].minX > x)
      {
        ListLine[y-minY].minX = x;
        ListLine[y-minY].MinColor[0]=Tri->Color[0][0]+((double)y-Tri->P[0][1])/(Tri->P[1][1]-Tri->P[0][1])*(Tri->Color[1][0]-Tri->Color[0][0]);
        ListLine[y-minY].MinColor[1]=Tri->Color[0][1]+((double)y-Tri->P[0][1])/(Tri->P[1][1]-Tri->P[0][1])*(Tri->Color[1][1]-Tri->Color[0][1]);
        ListLine[y-minY].MinColor[2]=Tri->Color[0][2]+((double)y-Tri->P[0][1])/(Tri->P[1][1]-Tri->P[0][1])*(Tri->Color[1][2]-Tri->Color[0][2]);
        ListLine[y-minY].MinDistance=Tri->Distance[0]+((double)y-Tri->P[0][1])/(Tri->P[1][1]-Tri->P[0][1])*(Tri->Distance[1]-Tri->Distance[0]);
      }
      if (ListLine[y-minY].maxX < x)
      {
        ListLine[y-minY].maxX = x;
        ListLine[y-minY].MaxColor[0]=Tri->Color[0][0]+((double)y-Tri->P[0][1])/(Tri->P[1][1]-Tri->P[0][1])*(Tri->Color[1][0]-Tri->Color[0][0]);
        ListLine[y-minY].MaxColor[1]=Tri->Color[0][1]+((double)y-Tri->P[0][1])/(Tri->P[1][1]-Tri->P[0][1])*(Tri->Color[1][1]-Tri->Color[0][1]);
        ListLine[y-minY].MaxColor[2]=Tri->Color[0][2]+((double)y-Tri->P[0][1])/(Tri->P[1][1]-Tri->P[0][1])*(Tri->Color[1][2]-Tri->Color[0][2]);
        ListLine[y-minY].MaxDistance=Tri->Distance[0]+((double)y-Tri->P[0][1])/(Tri->P[1][1]-Tri->P[0][1])*(Tri->Distance[1]-Tri->Distance[0]);
      }
    }
  }
  /* Line P[1] -> P[2] */
  deltaX = abs(Tri->P[2][0]-Tri->P[1][0]);
  deltaY = abs(Tri->P[2][1]-Tri->P[1][1]);
  if (Tri->P[2][0] > Tri->P[1][0]) offx=1; else offx=-1;
  if (Tri->P[2][1] > Tri->P[1][1]) offy=1; else offy=-1;
  /* "Trace" du premier point */
  if (ListLine[Tri->P[1][1]-minY].minX > Tri->P[1][0])
  {
    ListLine[Tri->P[1][1]-minY].minX=Tri->P[1][0];
    ListLine[Tri->P[1][1]-minY].MinColor[0]=Tri->Color[1][0];
    ListLine[Tri->P[1][1]-minY].MinColor[1]=Tri->Color[1][1];
    ListLine[Tri->P[1][1]-minY].MinColor[2]=Tri->Color[1][2];
    ListLine[Tri->P[1][1]-minY].MinDistance=Tri->Distance[1];
  }
  if (ListLine[Tri->P[1][1]-minY].maxX < Tri->P[1][0])
  {
    ListLine[Tri->P[1][1]-minY].maxX=Tri->P[1][0];
    ListLine[Tri->P[1][1]-minY].MaxColor[0]=Tri->Color[1][0];
    ListLine[Tri->P[1][1]-minY].MaxColor[1]=Tri->Color[1][1];
    ListLine[Tri->P[1][1]-minY].MaxColor[2]=Tri->Color[1][2];
    ListLine[Tri->P[1][1]-minY].MaxDistance=Tri->Distance[1];
  }
  if (deltaX > deltaY)
  { /* Plutot horizontal */
    x = Tri->P[1][0];
    y = Tri->P[1][1];
    bx = 2*deltaY-deltaX;
    while(x != Tri->P[2][0])
    {
      x+=offx;
      if (bx<0) bx+=2*deltaY;
      else
      {
        bx+=2*(deltaY-deltaX);
        y+=offy;
      }
      /* "Trace" du point */
      if (ListLine[y-minY].minX > x)
      {
        ListLine[y-minY].minX = x;
        ListLine[y-minY].MinColor[0]=Tri->Color[1][0]+((double)x-Tri->P[1][0])/(Tri->P[2][0]-Tri->P[1][0])*(Tri->Color[2][0]-Tri->Color[1][0]);
        ListLine[y-minY].MinColor[1]=Tri->Color[1][1]+((double)x-Tri->P[1][0])/(Tri->P[2][0]-Tri->P[1][0])*(Tri->Color[2][1]-Tri->Color[1][1]);
        ListLine[y-minY].MinColor[2]=Tri->Color[1][2]+((double)x-Tri->P[1][0])/(Tri->P[2][0]-Tri->P[1][0])*(Tri->Color[2][2]-Tri->Color[1][2]);
        ListLine[y-minY].MinDistance=Tri->Distance[1]+((double)x-Tri->P[1][0])/(Tri->P[2][0]-Tri->P[1][0])*(Tri->Distance[2]-Tri->Distance[1]);
      }
      if (ListLine[y-minY].maxX < x)
      {
        ListLine[y-minY].maxX = x;
        ListLine[y-minY].MaxColor[0]=Tri->Color[1][0]+((double)x-Tri->P[1][0])/(Tri->P[2][0]-Tri->P[1][0])*(Tri->Color[2][0]-Tri->Color[1][0]);
        ListLine[y-minY].MaxColor[1]=Tri->Color[1][1]+((double)x-Tri->P[1][0])/(Tri->P[2][0]-Tri->P[1][0])*(Tri->Color[2][1]-Tri->Color[1][1]);
        ListLine[y-minY].MaxColor[2]=Tri->Color[1][2]+((double)x-Tri->P[1][0])/(Tri->P[2][0]-Tri->P[1][0])*(Tri->Color[2][2]-Tri->Color[1][2]);
        ListLine[y-minY].MaxDistance=Tri->Distance[1]+((double)x-Tri->P[1][0])/(Tri->P[2][0]-Tri->P[1][0])*(Tri->Distance[2]-Tri->Distance[1]);
      }
    }
  } else
  { /* Plutot vertical */
    x = Tri->P[1][0];
    y = Tri->P[1][1];
    bx = 2*deltaX-deltaY;
    while(y != Tri->P[2][1])
    {
      y+=offy;
      if (bx<0) bx+=2*deltaX;
      else
      {
        bx+=2*(deltaX-deltaY);
        x+=offx;
      }
      /* "Trace" du point */
      if (ListLine[y-minY].minX > x)
      {
        ListLine[y-minY].minX = x;
        ListLine[y-minY].MinColor[0]=Tri->Color[1][0]+((double)y-Tri->P[1][1])/(Tri->P[2][1]-Tri->P[1][1])*(Tri->Color[2][0]-Tri->Color[1][0]);
        ListLine[y-minY].MinColor[1]=Tri->Color[1][1]+((double)y-Tri->P[1][1])/(Tri->P[2][1]-Tri->P[1][1])*(Tri->Color[2][1]-Tri->Color[1][1]);
        ListLine[y-minY].MinColor[2]=Tri->Color[1][2]+((double)y-Tri->P[1][1])/(Tri->P[2][1]-Tri->P[1][1])*(Tri->Color[2][2]-Tri->Color[1][2]);
        ListLine[y-minY].MinDistance=Tri->Distance[1]+((double)y-Tri->P[1][1])/(Tri->P[2][1]-Tri->P[1][1])*(Tri->Distance[2]-Tri->Distance[1]);
      }
      if (ListLine[y-minY].maxX < x)
      {
        ListLine[y-minY].maxX = x;
        ListLine[y-minY].MaxColor[0]=Tri->Color[1][0]+((double)y-Tri->P[1][1])/(Tri->P[2][1]-Tri->P[1][1])*(Tri->Color[2][0]-Tri->Color[1][0]);
        ListLine[y-minY].MaxColor[1]=Tri->Color[1][1]+((double)y-Tri->P[1][1])/(Tri->P[2][1]-Tri->P[1][1])*(Tri->Color[2][1]-Tri->Color[1][1]);
        ListLine[y-minY].MaxColor[2]=Tri->Color[1][2]+((double)y-Tri->P[1][1])/(Tri->P[2][1]-Tri->P[1][1])*(Tri->Color[2][2]-Tri->Color[1][2]);
        ListLine[y-minY].MaxDistance=Tri->Distance[1]+((double)y-Tri->P[1][1])/(Tri->P[2][1]-Tri->P[1][1])*(Tri->Distance[2]-Tri->Distance[1]);
      }
    }
  }
  /* Line P[2] -> P[0] */
  deltaX = abs(Tri->P[0][0]-Tri->P[2][0]);
  deltaY = abs(Tri->P[0][1]-Tri->P[2][1]);
  if (Tri->P[0][0] > Tri->P[2][0]) offx=1; else offx=-1;
  if (Tri->P[0][1] > Tri->P[2][1]) offy=1; else offy=-1;
  /* "Trace" du premier point */
  if (ListLine[Tri->P[2][1]-minY].minX > Tri->P[2][0])
  {
    ListLine[Tri->P[2][1]-minY].minX=Tri->P[2][0];
    ListLine[Tri->P[2][1]-minY].MinColor[0]=Tri->Color[2][0];
    ListLine[Tri->P[2][1]-minY].MinColor[1]=Tri->Color[2][1];
    ListLine[Tri->P[2][1]-minY].MinColor[2]=Tri->Color[2][2];
    ListLine[Tri->P[2][1]-minY].MinDistance=Tri->Distance[2];
  }
  if (ListLine[Tri->P[2][1]-minY].maxX < Tri->P[2][0])
  {
    ListLine[Tri->P[2][1]-minY].maxX=Tri->P[2][0];
    ListLine[Tri->P[2][1]-minY].MaxColor[0]=Tri->Color[2][0];
    ListLine[Tri->P[2][1]-minY].MaxColor[1]=Tri->Color[2][1];
    ListLine[Tri->P[2][1]-minY].MaxColor[2]=Tri->Color[2][2];
    ListLine[Tri->P[2][1]-minY].MaxDistance=Tri->Distance[2];
  }
  if (deltaX > deltaY)
  { /* Plutot horizontal */
    x = Tri->P[2][0];
    y = Tri->P[2][1];
    bx = 2*deltaY-deltaX;
    while(x != Tri->P[0][0])
    {
      x+=offx;
      if (bx<0) bx+=2*deltaY;
      else
      {
        bx+=2*(deltaY-deltaX);
        y+=offy;
      }
      /* "Trace" du point */
      if (ListLine[y-minY].minX > x)
      {
        ListLine[y-minY].minX = x;
        ListLine[y-minY].MinColor[0]=Tri->Color[2][0]+((double)x-Tri->P[2][0])/(Tri->P[0][0]-Tri->P[2][0])*(Tri->Color[0][0]-Tri->Color[2][0]);
        ListLine[y-minY].MinColor[1]=Tri->Color[2][1]+((double)x-Tri->P[2][0])/(Tri->P[0][0]-Tri->P[2][0])*(Tri->Color[0][1]-Tri->Color[2][1]);
        ListLine[y-minY].MinColor[2]=Tri->Color[2][2]+((double)x-Tri->P[2][0])/(Tri->P[0][0]-Tri->P[2][0])*(Tri->Color[0][2]-Tri->Color[2][2]);
        ListLine[y-minY].MinDistance=Tri->Distance[2]+((double)x-Tri->P[2][0])/(Tri->P[0][0]-Tri->P[2][0])*(Tri->Distance[0]-Tri->Distance[2]);
      }
      if (ListLine[y-minY].maxX < x)
      {
        ListLine[y-minY].maxX = x;
        ListLine[y-minY].MaxColor[0]=Tri->Color[2][0]+((double)x-Tri->P[2][0])/(Tri->P[0][0]-Tri->P[2][0])*(Tri->Color[0][0]-Tri->Color[2][0]);
        ListLine[y-minY].MaxColor[1]=Tri->Color[2][1]+((double)x-Tri->P[2][0])/(Tri->P[0][0]-Tri->P[2][0])*(Tri->Color[0][1]-Tri->Color[2][1]);
        ListLine[y-minY].MaxColor[2]=Tri->Color[2][2]+((double)x-Tri->P[2][0])/(Tri->P[0][0]-Tri->P[2][0])*(Tri->Color[0][2]-Tri->Color[2][2]);
        ListLine[y-minY].MaxDistance=Tri->Distance[2]+((double)x-Tri->P[2][0])/(Tri->P[0][0]-Tri->P[2][0])*(Tri->Distance[0]-Tri->Distance[2]);
      }
    }
  } else
  { /* Plutot vertical */
    x = Tri->P[2][0];
    y = Tri->P[2][1];
    bx = 2*deltaX-deltaY;
    while(y != Tri->P[0][1])
    {
      y+=offy;
      if (bx<0) bx+=2*deltaX;
      else
      {
        bx+=2*(deltaX-deltaY);
        x+=offx;
      }
      /* "Trace" du point */
      if (ListLine[y-minY].minX > x)
      {
        ListLine[y-minY].minX = x;
        ListLine[y-minY].MinColor[0]=Tri->Color[2][0]+((double)y-Tri->P[2][1])/(Tri->P[0][1]-Tri->P[2][1])*(Tri->Color[0][0]-Tri->Color[2][0]);
        ListLine[y-minY].MinColor[1]=Tri->Color[2][1]+((double)y-Tri->P[2][1])/(Tri->P[0][1]-Tri->P[2][1])*(Tri->Color[0][1]-Tri->Color[2][1]);
        ListLine[y-minY].MinColor[2]=Tri->Color[2][2]+((double)y-Tri->P[2][1])/(Tri->P[0][1]-Tri->P[2][1])*(Tri->Color[0][2]-Tri->Color[2][2]);
        ListLine[y-minY].MinDistance=Tri->Distance[2]+((double)y-Tri->P[2][1])/(Tri->P[0][1]-Tri->P[2][1])*(Tri->Distance[0]-Tri->Distance[2]);
      }
      if (ListLine[y-minY].maxX < x)
      {
        ListLine[y-minY].maxX = x;
        ListLine[y-minY].MaxColor[0]=Tri->Color[2][0]+((double)y-Tri->P[2][1])/(Tri->P[0][1]-Tri->P[2][1])*(Tri->Color[0][0]-Tri->Color[2][0]);
        ListLine[y-minY].MaxColor[1]=Tri->Color[2][1]+((double)y-Tri->P[2][1])/(Tri->P[0][1]-Tri->P[2][1])*(Tri->Color[0][1]-Tri->Color[2][1]);
        ListLine[y-minY].MaxColor[2]=Tri->Color[2][2]+((double)y-Tri->P[2][1])/(Tri->P[0][1]-Tri->P[2][1])*(Tri->Color[0][2]-Tri->Color[2][2]);
        ListLine[y-minY].MaxDistance=Tri->Distance[2]+((double)y-Tri->P[2][1])/(Tri->P[0][1]-Tri->P[2][1])*(Tri->Distance[0]-Tri->Distance[2]);
      }
    }
  }
  /* Then, Zbuffer drawing */
  for (i=minY;i<=maxY;i++)
  {
    if ((ListLine[i-minY].minX < View->allocation.width) && (ListLine[i-minY].maxX >= 0))
      for (j=ListLine[i-minY].minX ; j<=ListLine[i-minY].maxX ; j++)
      {
        if ((i>=0) && (i<View->allocation.height) && 
            (j>=0) && (j<View->allocation.width))
        {
          if (ListLine[i-minY].maxX != ListLine[i-minY].minX)
            distance = ListLine[i-minY].MinDistance + (j-ListLine[i-minY].minX)/(ListLine[i-minY].maxX-ListLine[i-minY].minX)*(ListLine[i-minY].MaxDistance-ListLine[i-minY].MinDistance);
          else
            distance = ListLine[i-minY].MinDistance;
          if (Zbuffer[i*View->allocation.width+j] > distance)
          {
            Zbuffer[i*View->allocation.width+j] = distance;
            Color.red=(unsigned short)(65535.0*(ListLine[i-minY].MinColor[0]+((double)j-(double)(ListLine[i-minY].minX))/((double)(ListLine[i-minY].maxX)-(double)(ListLine[i-minY].minX))*(ListLine[i-minY].MaxColor[0]-ListLine[i-minY].MinColor[0])));
            Color.green=(unsigned short)(65535.0*(ListLine[i-minY].MinColor[1]+((double)j-(double)(ListLine[i-minY].minX))/((double)(ListLine[i-minY].maxX)-(double)(ListLine[i-minY].minX))*(ListLine[i-minY].MaxColor[1]-ListLine[i-minY].MinColor[1])));
            Color.blue=(unsigned short)(65535.0*(ListLine[i-minY].MinColor[2]+((double)j-(double)(ListLine[i-minY].minX))/((double)(ListLine[i-minY].maxX)-(double)(ListLine[i-minY].minX))*(ListLine[i-minY].MaxColor[2]-ListLine[i-minY].MinColor[2])));
            gdk_color_alloc(colormap, &Color);
            gdk_gc_set_foreground(gc, &Color);
            gdk_draw_point(View->window, gc, j,i); /* FIXME: very ssllooww */
          }
        }
      }
  }
  g_free(ListLine);
  gdk_gc_destroy(gc);
}

/*****************************************************************************
*  NewFrameFunc
******************************************************************************/
void NewFrameFunc(GtkWidget *widget)
{
  NewFrame(_("Untitled"));
}

/*****************************************************************************
*  NewViewOK
******************************************************************************/
FrameStruct *NewViewFrame;
void NewViewOK(GtkObject *Dialog)
{
  GtkToggleButton **Data;

  if (NewViewFrame == NULL)
  {
    NewFrame(_("Untitled"));
    NewViewFrame = FirstFrame;
  }
  Data = gtk_object_get_data(Dialog, "WidgetData");
  if (gtk_toggle_button_get_active(Data[0]))
    NewView(NewViewFrame, ORTHO_XY_CAMERA);
  if (gtk_toggle_button_get_active(Data[1]))
    NewView(NewViewFrame, ORTHO_XZ_CAMERA);
  if (gtk_toggle_button_get_active(Data[2]))
    NewView(NewViewFrame, ORTHO_ZY_CAMERA);
  if (gtk_toggle_button_get_active(Data[3]))
    NewViewCamera(NewViewFrame);
}

/*****************************************************************************
*  NewViewSetFrame
******************************************************************************/
void NewViewSetFrame(GtkWidget *dum, FrameStruct *LocalFrame)
{
  NewViewFrame = LocalFrame;
}

/*****************************************************************************
*  NewViewFunc
******************************************************************************/
void NewViewFunc(GtkWidget *widget, FrameStruct *LocalFrame)
{
  GtkWidget *Dialog;
  GtkWidget *Button;
  GtkWidget *OptionMenu;
  GtkWidget *FrameMenu;
  GtkWidget *MenuItem;
  GSList *MenuGroup;
  FrameStruct *TmpFrame;
  GtkWidget *Separator;
  GtkWidget *Option;
  GSList *OptionGroup;
  GtkWidget **Data;
 
  Dialog = gtk_dialog_new();
  Data = g_malloc(4*sizeof(GtkWidget*));
  gtk_object_set_data(GTK_OBJECT(Dialog), "WidgetData", Data);
  gtk_window_set_title(GTK_WINDOW(Dialog), _("New View"));

  /* The Ok Button */
  Button = gtk_button_new_with_label(_("OK"));
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            (GtkSignalFunc)NewViewOK,
                            GTK_OBJECT(Dialog));
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(Dialog));
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->action_area), Button, TRUE,
                     TRUE, 0);
  gtk_widget_grab_default(Button);
  gtk_widget_show(Button);
  /* The Close Button */
  Button = gtk_button_new_with_label(_("Close"));
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(Dialog));
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->action_area), Button, TRUE,
                     TRUE, 0);
  gtk_widget_show(Button);

  /* The Frame Option Menu */
  OptionMenu = gtk_option_menu_new();
  FrameMenu = gtk_menu_new();
  MenuGroup = NULL;
  NewViewFrame = NULL;
  MenuItem = gtk_radio_menu_item_new_with_label(MenuGroup, _("New Frame"));
  MenuGroup = gtk_radio_menu_item_group(GTK_RADIO_MENU_ITEM(MenuItem));
  if (LocalFrame == NULL)
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(MenuItem), TRUE);
  else
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(MenuItem), FALSE);
  gtk_signal_connect(GTK_OBJECT(MenuItem), "activate",
                     GTK_SIGNAL_FUNC(NewViewSetFrame), NULL);
  gtk_menu_append(GTK_MENU(FrameMenu), MenuItem);
  gtk_widget_show(MenuItem);
  for (TmpFrame = FirstFrame ; TmpFrame ; TmpFrame=TmpFrame->Next)
  {
    MenuItem = gtk_radio_menu_item_new_with_label(MenuGroup, TmpFrame->Name);
    MenuGroup = gtk_radio_menu_item_group(GTK_RADIO_MENU_ITEM(MenuItem));
    if (TmpFrame == LocalFrame)
    {
      NewViewFrame = LocalFrame;
      gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(MenuItem), TRUE);
    } else
      gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(MenuItem), FALSE);
    gtk_signal_connect(GTK_OBJECT(MenuItem), "activate",
                       GTK_SIGNAL_FUNC(NewViewSetFrame),
                       TmpFrame);
    gtk_menu_append(GTK_MENU(FrameMenu), MenuItem);
    gtk_widget_show(MenuItem);
  }
  gtk_option_menu_set_menu(GTK_OPTION_MENU(OptionMenu), FrameMenu);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->vbox), OptionMenu, FALSE,
                     FALSE, 0);
  gtk_widget_show(OptionMenu);
  /* Separator */
  Separator = gtk_hseparator_new();
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->vbox), Separator, FALSE,
                     FALSE, 5);
  gtk_widget_show(Separator);
  /* View Options */ 
  OptionGroup = NULL;
  Option = gtk_radio_button_new_with_label(OptionGroup, _("View X-Y"));
  Data[0] = Option;
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(Option), TRUE);
  OptionGroup = gtk_radio_button_group(GTK_RADIO_BUTTON(Option));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->vbox), Option, FALSE, FALSE,
                     0);
  Option = gtk_radio_button_new_with_label(OptionGroup, _("View X-Z"));
  Data[1] = Option;
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(Option), FALSE);
  OptionGroup = gtk_radio_button_group(GTK_RADIO_BUTTON(Option));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->vbox), Option, FALSE, FALSE,
                     0);
  Option = gtk_radio_button_new_with_label(OptionGroup, _("View Z-Y"));
  Data[2] = Option;
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(Option), FALSE);
  OptionGroup = gtk_radio_button_group(GTK_RADIO_BUTTON(Option));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->vbox), Option, FALSE, FALSE,
                     0);
  Option = gtk_radio_button_new_with_label(OptionGroup, _("Camera View"));
  Data[3] = Option;
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(Option), FALSE);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->vbox), Option, FALSE, FALSE,
                     0);

  gtk_widget_show_all(Dialog);
}

/*****************************************************************************
*  LoadPovFunc
******************************************************************************/
void LoadPovFunc(GtkWidget *dum)
{
  GtkWidget *FileSelector;

  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Load Pov File"));
  /* Connect the ok_button to the LoadPov function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)LoadPov, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}

/*****************************************************************************
*  LoadDXFFunc
******************************************************************************/
void LoadDXFFunc(GtkWidget *dum)
{
  GtkWidget *FileSelector;

  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Load AutoCAD DXF File"));
  /* Connect the ok_button to the LoadPov function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)LoadDXF, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}

/*****************************************************************************
*  InsertDXFFunc
******************************************************************************/
void InsertDXFFunc(GtkWidget *widget, gpointer data)
{
  GtkWidget *FileSelector;

  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Insert DXF objects"));
  gtk_object_set_data(GTK_OBJECT(FileSelector), "frame", data);
  /* Connect the ok_button to file_ok_sel function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)InsertDXF, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}

/*****************************************************************************
*  SavePovFunc
******************************************************************************/
void SavePovFunc(GtkWidget *widget, gpointer data)
{
  GtkWidget *FileSelector;

  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Save Pov File"));
  gtk_object_set_data(GTK_OBJECT(FileSelector), "frame", data);
  /* Connect the ok_button to file_ok_sel function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)SavePov, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}

/*****************************************************************************
*  SaveDXFFunc
******************************************************************************/
void SaveDXFFunc(GtkWidget *widget, gpointer data)
{
  GtkWidget *FileSelector;

  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Save DXF File"));
  gtk_object_set_data(GTK_OBJECT(FileSelector), "frame", data);
  /* Connect the ok_button to file_ok_sel function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)SaveDXF, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}

/*****************************************************************************
*  SaveSelectedDXFFunc
******************************************************************************/
void SaveSelectedDXFFunc(GtkWidget *widget, gpointer data)
{
  GtkWidget *FileSelector;

  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Save DXF File"));
  gtk_object_set_data(GTK_OBJECT(FileSelector), "frame", data);
  /* Connect the ok_button to file_ok_sel function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)SaveSelectedDXF, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}

#ifdef WITH_S3D
/*****************************************************************************
*  SaveS3DFunc
******************************************************************************/
void SaveS3DFunc(GtkWidget *widget, gpointer data)
{
  GtkWidget *FileSelector;

  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Save S3D File"));
  gtk_object_set_data(GTK_OBJECT(FileSelector), "frame", data);
  /* Connect the ok_button to file_ok_sel function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)SaveS3D, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}

/*****************************************************************************
*  SaveSelectedS3DFunc
******************************************************************************/
void SaveSelectedS3DFunc(GtkWidget *widget, gpointer data)
{
  GtkWidget *FileSelector;

  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Save S3D File"));
  gtk_object_set_data(GTK_OBJECT(FileSelector), "frame", data);
  /* Connect the ok_button to file_ok_sel function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)SaveSelectedS3D, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}
#endif

/*****************************************************************************
*  ExitFunc
******************************************************************************/
void ExitFunc(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  gtk_main_quit();
}

/*****************************************************************************
*  main
******************************************************************************/
int main(int argc, char *argv[])
{
  GtkWidget *ToolsWindow;
  GtkWidget *Box;
#ifdef USE_GNOME
static GnomeUIInfo FileMenu[] = 
{
  {
    GNOME_APP_UI_ITEM, N_("New Frame"), N_("Create a new frame"),
    NewFrameFunc, NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL, 0, 0, NULL
  },
  {
    GNOME_APP_UI_ITEM, N_("New View"), N_("Open a new view window"),
    NewViewFunc, NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL, 0, 0, NULL
  },
  {
    GNOME_APP_UI_ITEM, N_("Load Pov"), N_("Load a POVRay file"),
    LoadPovFunc, NULL, NULL, GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_OPEN,
    0, GDK_CONTROL_MASK, NULL
  },
  {
    GNOME_APP_UI_ITEM, N_("Load AutoCAD DXF"), N_("Load an AutoCAD DXF file"),
    LoadDXFFunc, NULL, NULL, GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_OPEN,
    0, GDK_CONTROL_MASK, NULL
  },
  {
    GNOME_APP_UI_ITEM, N_("About..."), N_("About Giram"),
    CreateAboutDialog, NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL, 0, 0, NULL
  },
  {
    GNOME_APP_UI_ITEM, N_("Help..."), N_("Online Help"),
    ShowHelpDialog, NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL, 0, 0, NULL
  },
  {
    GNOME_APP_UI_ITEM, N_("Preferences..."), N_("Preferences setting"),
    GiramPreferences, NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL, 0, 0, NULL
  },
  GNOMEUIINFO_SEPARATOR,
  {
    GNOME_APP_UI_ITEM, N_("Quit"), N_("Quit Giram"),
    ExitFunc, NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL, 0, 0, NULL
  },
  GNOMEUIINFO_END
};

static GnomeUIInfo DialogMenu[] =
{
  {
    GNOME_APP_UI_ITEM, N_("CSG Tree"), N_("Show CSG Tree Window"),
    &ShowCSGTree, NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL, 0, 0, NULL
  },
  {
    GNOME_APP_UI_ITEM, N_("Tool Options"), N_("Show Tool Options Dialog"),
    ShowToolOptionsDialog, NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL, 0, 0, NULL
  },
  GNOMEUIINFO_END
};

static GnomeUIInfo MainMenu[] =
{
  GNOMEUIINFO_SUBTREE(N_("File"), FileMenu),
  GNOMEUIINFO_SUBTREE(N_("Dialogs"), DialogMenu),
  GNOMEUIINFO_END
};
#else
  GtkWidget *menu;
  GtkWidget *menu_items;
  GtkWidget *FileMenu;
  GtkWidget *DialogMenu;
  GtkWidget *menu_bar;
#endif /* USE_GNOME */
  GtkWidget *ToolsTable;
  
#ifdef USE_GNOME
  gnome_init("Giram", VERSION, argc, argv);
#else
#ifdef ENABLE_NLS
  textdomain("giram");
#endif
  gtk_init(&argc, &argv);
#endif
  giramInit();

#ifdef USE_GNOME
  ToolsWindow = gnome_app_new("Giram", "Giram Tools");
#else
  ToolsWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(ToolsWindow), _("Giram Tools"));
#endif
  /* Here we just set a handler for delete_event that immediately
   *        * exits GTK. */
  gtk_signal_connect(GTK_OBJECT(ToolsWindow), "delete_event",
                     GTK_SIGNAL_FUNC(ExitFunc), NULL);
  gtk_container_border_width(GTK_CONTAINER(ToolsWindow), 0);
  gtk_window_set_policy(GTK_WINDOW(ToolsWindow), FALSE, FALSE, TRUE);

  Box = gtk_vbox_new(FALSE, 0);
  /* Put the box into the main window. */
#ifdef USE_GNOME
  gnome_app_set_contents(GNOME_APP(ToolsWindow), Box);
#else
  gtk_container_add(GTK_CONTAINER(ToolsWindow), Box);
#endif
  

#ifdef USE_GNOME
  gnome_app_create_menus(GNOME_APP(ToolsWindow), MainMenu);
#else

  /* Create the Menu */
    menu = gtk_menu_new();
    /* The New Frame Menu Item */
    menu_items = gtk_menu_item_new_with_label(_("New Frame"));
    gtk_menu_append(GTK_MENU(menu), menu_items);
    gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                              GTK_SIGNAL_FUNC(NewFrameFunc), NULL);
    gtk_widget_show(menu_items);
    /* The New View Menu Item */
    menu_items = gtk_menu_item_new_with_label(_("New View"));
    gtk_menu_append(GTK_MENU(menu), menu_items);
    gtk_signal_connect(GTK_OBJECT(menu_items), "activate",
                       GTK_SIGNAL_FUNC(NewViewFunc), NULL);
    gtk_widget_show(menu_items);
    /* The Load Pov Menu Item */
    menu_items = gtk_menu_item_new_with_label(_("Load Pov"));
    gtk_menu_append(GTK_MENU(menu), menu_items);
    gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                              GTK_SIGNAL_FUNC(LoadPovFunc), NULL);
    gtk_widget_show(menu_items);
    /* The Load AutoCAD DXF Menu Item */
    menu_items = gtk_menu_item_new_with_label(_("Load AutoCAD DXF"));
    gtk_menu_append(GTK_MENU(menu), menu_items);
    gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                              GTK_SIGNAL_FUNC(LoadDXFFunc), NULL);
    gtk_widget_show(menu_items);
    /* The About Dialog Menu Item */
    menu_items = gtk_menu_item_new_with_label(_("About..."));
    gtk_menu_append(GTK_MENU(menu), menu_items);
    gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate", 
                              GTK_SIGNAL_FUNC(CreateAboutDialog), NULL);
    gtk_widget_show(menu_items);
    /* The Help Dialog Menu Item */
    menu_items = gtk_menu_item_new_with_label(_("Help..."));
    gtk_menu_append(GTK_MENU(menu), menu_items);
    gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                              GTK_SIGNAL_FUNC(ShowHelpDialog), NULL);
    gtk_widget_show(menu_items);
    /* Separator */
    menu_items = gtk_menu_item_new();
    gtk_menu_append(GTK_MENU(menu), menu_items);
    gtk_widget_show(menu_items);
    /* The Preferences Menu Item */
    menu_items = gtk_menu_item_new_with_label(_("Preferences..."));
    gtk_menu_append(GTK_MENU(menu), menu_items);
    gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                              GTK_SIGNAL_FUNC(GiramPreferences), NULL);
    gtk_widget_show(menu_items);
    /* The Quit Menu Item */
    menu_items = gtk_menu_item_new_with_label(_("Quit"));
    gtk_menu_append(GTK_MENU(menu), menu_items);
    gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                              GTK_SIGNAL_FUNC(ExitFunc), NULL);
    gtk_widget_show(menu_items);
    /* This is the root menu, and will be the label
     * displayed on the menu bar.  There won't be a signal handler attached,
     * as it only pops up the rest of the menu when pressed. */
    FileMenu = gtk_menu_item_new_with_label(_("File"));
    gtk_widget_show(FileMenu);
    /* Now we specify that we want our newly created "menu" to be the menu
     * * for the "root menu" */
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(FileMenu), menu);
    /* Create a menu-bar to hold the menus and add it to our main window */
    menu_bar = gtk_menu_bar_new();
    gtk_box_pack_start(GTK_BOX(Box), menu_bar, FALSE, FALSE, 2);
    gtk_widget_show(menu_bar);
    /* And finally we append the menu-item to the menu-bar */
    gtk_menu_bar_append(GTK_MENU_BAR (menu_bar), FileMenu);

    /* */
    menu = gtk_menu_new();
    /* The CSGTree Menu Item */
    menu_items = gtk_menu_item_new_with_label(_("CSG Tree"));
    gtk_menu_append(GTK_MENU(menu), menu_items);
    gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                              GTK_SIGNAL_FUNC(ShowCSGTree), NULL);
    gtk_widget_show(menu_items);
    /* The Tools Options Menu Item */
    menu_items = gtk_menu_item_new_with_label(_("Tool Options"));
    gtk_menu_append(GTK_MENU(menu), menu_items);
    gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                              GTK_SIGNAL_FUNC(ShowToolOptionsDialog), NULL);
    gtk_widget_show(menu_items);
    /* */
    DialogMenu = gtk_menu_item_new_with_label(_("Dialogs"));
    gtk_widget_show(DialogMenu);
    /* Now we specify that we want our newly created "menu" to be the menu
     * * for the "root menu" */
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(DialogMenu), menu);
    /* And finally we append the menu-item to the menu-bar */
    gtk_menu_bar_append(GTK_MENU_BAR (menu_bar), DialogMenu);
    /* */
#endif /* USE_GNOME */

    
  /* The Tool Table */
  ToolsTable = CreateTools(Box);

  gtk_widget_show(ToolsTable);
  gtk_widget_show(Box);
  gtk_widget_show(ToolsWindow);
  gtk_main();
  return 0;
}

