/* giram.h
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999 David Odin <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* This file should always be the first included. */
#include "config.h"
#include "giramlib/vectors.h"

#include <gtk/gtk.h>
/* Gnome Stuff */
#ifdef USE_GNOME
#include <gnome.h>
#endif

/* OpenGL stuff */
#ifdef HAVE_GL
#include <gdk/gdkx.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#endif

/* This struct is used to store all the paths giram should search when
 * looking for an included file in LoadPov.c
 * it is initialized in parserc.c
 * This should be replaced by a GSList.
 */
typedef struct LibraryPathList
{
  char *PathName;
  struct LibraryPathList *Next;  /* Pointer to the next path if any */
} LibraryPathList;

LibraryPathList *LibraryPaths;

/*****************************************************************************
 *****************************************************************************
 ******************** Beginning of Texture specific code *********************
 *****************************************************************************
 *****************************************************************************/

/* These defines represent all the pattern PovRay can handle, and thus must
 * be handled as well by Giram
 * I'm wondering if it shouldn't be clever to use an enum there.
 */
#define PAT_SOLID_COLOR  1
#define PAT_BRICK        2
#define PAT_CHECKER      3
#define PAT_HEXAGON      4
#define PAT_AGATE        5
#define PAT_AVERAGE      6
#define PAT_BOZO         7
#define PAT_BUMPS        8
#define PAT_CRACKLE      9
#define PAT_DENTS       10
#define PAT_GRADIENT    11
#define PAT_GRANITE     12
#define PAT_LEOPARD     13
#define PAT_MANDEL      14
#define PAT_MARBLE      15
#define PAT_ONION       16
#define PAT_QUILTED     17
#define PAT_RADIAL      18
#define PAT_RIPPLES     19
#define PAT_SPIRAL1     20
#define PAT_SPIRAL2     21
#define PAT_SPOTTED     22
#define PAT_WAVES       23
#define PAT_WOOD        24
#define PAT_WRINKLES    25

/* This structure is used for the pigment-maps */
typedef struct PigmentListStruct
{
  int NbPigment;
  struct PigmentStruct *PigmentArray;
} PigmentListStruct;

/* This define all the values that the wave form can take */
#define WF_RAMP     1            /* / / / / */
#define WF_SINE     2            /* ~~~~~~~~ */
#define WF_SCALLOP  3
#define WF_TRIANGLE 4            /* /\/\/\/\ */

/* This should be used to handle the 'warp' modifiers of a pigment */
typedef struct WarpStruct
{
  void *Dum; /* FIXME */
} WarpStruct;

/* This is one entry of the following array of 'colormap' */
typedef struct ColorMapEntry
{
  double Value;
  Vector Color;
} ColorMapEntry;

typedef struct ColorMapStruct
{
  int NbMapEntries;
  ColorMapEntry *Entries;
} ColorMapStruct;

typedef struct PigmentStruct
{
  int Type;                        /* Type of pattern used */
  Vector Color;                    /* Solid Color */
  TransformStruct *Trans;          /* Transformation for the pigment */
  PigmentListStruct *PigmentList;   /* Only used with Pigment Map */
  Vector BrickSize;
  double Mortar;
  double AgateTurb;
  Vector Gradient;
  int Mandel;
  double Control0;
  double Control1;
  double NumberOfArms;
  double Frequency;
  double Phase;
  int WaveForm;
  Vector Turbulence;
  int Octaves;
  double Lambda;
  double Omega;
  WarpStruct *Warp;
  Vector QuickColor;
  ColorMapStruct *ColorMap; /* The colormap used by many pattern types */
} PigmentStruct;

/* The same as ColorMap Entry, but for pigment... */
typedef struct PigmentMapEntry
{
  double Value;
  PigmentStruct *Pigment;
} PigmentMapEntry;

typedef struct PigmentMapStruct
{
  int NbMapEntries;
  PigmentMapEntry *Entries;
} PigmentMapStruct;

typedef struct NormalStruct
{
  int Type;                      /* The type of pattern used for this normal */
  double AgateTurb;
  Vector BrickSize;
  double Mortar;
  struct NormalStruct *Normal1;  /* Used for brick and checker */
  struct NormalStruct *Normal2;
  struct NormalStruct *Normal3;  /* only with hexagon */
  Vector Gradient;
  int Mandel;
  double Control0, Control1;
  double NumberOfArms;
  double Frequency;
  double Phase;
  int WaveForm;
  int Octaves;
  double Lambda;
  double Omega;
  Vector Turbulence;
  double Amount;
  TransformStruct *Trans;
} NormalStruct;

/* Same as ColorMapEntry, but for Normal... */
typedef struct NormalMapEntry
{
  double Value;
  NormalStruct *Normal;
} NormalMapEntry;

typedef struct NormalMapStruct
{
  int NbMapEntries;
  NormalMapEntry *Entries;
} NormalMapStruct;

/* This struct is used to store the Finish part of a texture */
typedef struct FinishStruct
{
  Vector Ambient;
  double Diffuse;
  double Brilliance;
  double Phong;
  double PhongSize;
  double Specular;
  double Roughness;
  double Metallic;
  Vector Reflection;
  double Refraction;
  double Ior;
  double Caustics;
  double FadeDistance;
  double FadePower;
  struct
  {
    double Amount;
    double Thickness;
    Vector Turbulence;
  } Irid;
  double Crand;
} FinishStruct;

/* Different Halo Types */
#define HALO_ATTENUATING 1
#define HALO_EMITTING    2
#define HALO_GLOWING     3
#define HALO_DUST        4

/* Different Density repartitions */
#define DENSITY_CONSTANT 1
#define DENSITY_LINEAR   2
#define DENSITY_CUBIC    3
#define DENSITY_POLY     4

/* Different mapping shapes */
#define MAP_PLANAR       1
#define MAP_SPHERICAL    2
#define MAP_CYLINDRICAL  3
#define MAP_BOX          4
typedef struct HaloStruct
{
  int Type;
  int Density;
  int Mapping;
  int DustType;
  double Eccentricity;
  double MaxValue;
  double Exponent;
  double Samples;
  int AaLevel;
  double AaThreshold;
  double Jitter;
  Vector Turbulence;
  int Octaves;
  double Omega;
  double Lambda;
  ColorMapStruct *ColorMap;
  double Frequency;
  double Phase;
  TransformStruct *Trans;
  struct HaloStruct *Next;
} HaloStruct;

typedef struct TextureStruct
{
  PigmentStruct *Pigment;
  NormalStruct *Normal;
  FinishStruct *Finish;
  HaloStruct *Halo;
  TransformStruct *Trans;
  struct TextureStruct *NextLayer;
} TextureStruct;

TextureStruct *DefaultTexture;

/*****************************************************************************
 *****************************************************************************
 *********************** End of Texture specific code ************************
 *****************************************************************************
 *****************************************************************************/


typedef struct SkySphereStruct
{
  PigmentStruct **pPigmentList; /* Skysphere, is merely a list of layered
                                   pigments */
  int NbPigment;
  TransformStruct *Trans;
} SkySphereStruct;

/* What type of atmosphere is used in the frame? */
typedef struct AtmosphereStruct
{
  int Type;
  double Distance;
  double Scattering;
  double Eccentricity;
  int Samples;
  double Jitter;
  double AaThreshold;
  int AaLevel;
  Vector Color;                                      
} AtmosphereStruct;

/* Fog */
typedef struct FogStruct
{
  int FogType;
  double Distance;
  Vector Color;
  Vector Turbulence;
  double TurbDepth;
  double Omega;
  double Lambda;
  int Octaves;
  double FogOffset;
  double FogAlt;
  Vector Up;
  TransformStruct *Trans;
  struct FogStruct *Next;
} FogStruct;

/* Povray can even render rainbows... */
typedef struct RainbowStruct
{
  Vector Direction;
  double Angle;
  double Width;
  double Distance;
  ColorMapStruct *ColorMap;
  double Jitter;
  Vector Up;
  double ArcAngle;
  double FalloffAngle;
  struct RainbowStruct *Next;
} RainbowStruct;


/* The Radiosity part of the global settings for the scene */
typedef struct RadiositySetting
{
  double Brightness;
  int Count;
  double DistanceMaximum;
  double ErrorBound;
  double GrayThreshold;
  double LowErrorFactor;
  double MinimumReuse;
  int NearestCount;
  int RecursionLimit;
} RadiositySetting;

/* The global settings for the scene */
typedef struct GlobalSetting
{
  double AssumedGamma;
  int HfGray16;
  Vector AmbientLight;
  RadiositySetting Radiosity;
  int NumberOfWaves;
  double ADCBailout;
  Vector IridWavelength;
  int MaxIntersections;
  int MaxTraceLevel;
} GlobalSetting;

/* The different types of camera. Only the PERSPECTIVE one is used. */
#define CAMERA_PERSPECTIVE      1
#define CAMERA_ORTHOGRAPHIC     2
#define CAMERA_FISHEYE          3
#define CAMERA_ULTRA_WIDE_ANGLE 4
#define CAMERA_OMNIMAX          5
#define CAMERA_PANORAMIC        6
#define CAMERA_CYLINDER         7

typedef struct CameraStruct
{
  int Type;
  int CylinderType;
  Vector Location;
  Vector LookAt;
  Vector Right;
  Vector Up;
  Vector Direction;
  Vector Sky;
  double Angle;
  int BlurSamples;
  double Aperture;
  Vector FocalPoint;
  double Confidence;
  double Variance;
  NormalStruct *Normal;
  /* Private fields. These are used to calculate the Pers3d->2d */
  double x0, y0, z0;
  double PX, PY, PZ;
  double A, B, C, D;
  Vector P00, P01, P10, P11;
  Vector PA, PB, PC, PD;
  Vector Na, Nb, Nc;
  double Du0, Du1, Du2;
  double Dv0, Dv1, Dv2;
  Vector Qux, Quy, Qvx, Qvy;
  double Dux, Duy, Dvx, Dvy;
} CameraStruct;

/* A list of triangles is attached to each object to draw them on the screen */
typedef struct TriangleListStruct
{
  Vector P1,P2,P3;
  Vector N1,N2,N3;
  struct TriangleListStruct *Next;
} TriangleListStruct;

/* THE structure defining an object */
typedef struct ObjectStruct
{
  int Type; /* The shape */
  char *Name;
  TriangleListStruct *FirstTriangle; /* The first triangle of the list */
  struct ObjectStruct *Next;         /* A pointer to the next object of the frame */
  TransformStruct *Trans;            /* The transformation matrix */
  struct ObjectStruct *BoundedBy;    /* Not Currently Used */
  struct ObjectStruct *ClippedBy;    /* NCU */
  TextureStruct *Texture;
  int Inverse;                       /* Is the object inverted ? (CSG) */
  int Sturm;
  double WaterLevel;
  int Smooth;
  int NoShadow;
  int Hierarchy;
  int Hollow;
  gint Selected;                   /* Is the object currently selected? */
  gint Visible;                    /* IS the object currently visible? */
  struct GroupStruct *Group;       /* What's the group the object is attached to? */
  GtkCTreeNode *CSGTreeNode;       /* A pointer to the node of the object in the CSGTree: looks very ugly... */
  int DetailLevel;                 /* NCU: should be used to let the user choose how  detailed is the 
                                      mesh representation of the objects.
                                      This also could be used for the
                                      object->mesh converter. */
  int Id;
} ObjectStruct;

/* Object should be owned by a group */
typedef struct GroupStruct
{
  char *Name; 
  Vector Color;              /* The color of objects of this group when the object is unselected */ 
  Vector SelectedColor;      /* The color of objects of this group when the object is selected */
  struct GroupStruct *Next;  /* Next group of this frame */
  struct GroupOrObjectStruct *FirstObject; /* first object of this group */
} GroupStruct;

/* Group can contain Group or Object... */
typedef struct GroupOrObject
{
  int Type;  /* Group or Object */
  union
  {
    ObjectStruct *Object;
    GroupStruct *Group;
  } GOO;
  struct GroupOrObject *Next; /* The next object/group of this group */
} GroupOrObject;


/* Here begins the differents objects definitions */
#define BICUBIC_PATCH_OBJECT 1
typedef struct BicubicPatchStruct
{
  ObjectStruct Object; /* To respect the OO approach, this is always the
                          first field of every object, so we can cast
                          BicubicPatchStruct to ObjectStruct to have access
                          to the field of the abstract ObjectStruct type */
  int Type;
  double Flatness;
  double USteps, VSteps;
  Vector ControlPoint[4][4];
} BicubicPatchStruct;

#define BLOB_OBJECT 2
typedef struct BlobComponentStruct
{
  ObjectStruct *Object;
  double Strength;
  struct BlobComponentStruct *Next;
} BlobComponentStruct;

typedef struct BlobStruct
{
  ObjectStruct Object;

  double Threshold;
  BlobComponentStruct *FirstComponent;
} BlobStruct;

#define BOX_OBJECT 3
typedef struct BoxStruct
{
  ObjectStruct Object;

  Vector MinCorner;
  Vector MaxCorner;
} BoxStruct;

#define CONE_OBJECT 4
typedef struct ConeStruct
{
  ObjectStruct Object;

  int Open;
  double Radius1;
  double Radius2;
} ConeStruct;

#define CSG_OBJECT 5
#define CSG_DIFFERENCE   1
#define CSG_INTERSECTION 2
#define CSG_MERGE        3
#define CSG_UNION        4
typedef struct CSGStruct
{
  ObjectStruct Object;

  int Type;
  struct ObjectStruct *FirstObject;
} CSGStruct;

#define CUBIC_OBJECT 6
typedef struct CubicStruct
{
  ObjectStruct Object;

  double T[20];
} CubicStruct;

#define CYLINDER_OBJECT 7
typedef struct CylinderStruct
{
  ObjectStruct Object;

  int Open;
} CylinderStruct;

#define DISC_OBJECT 8
typedef struct DiscStruct
{
  ObjectStruct Object;

  double Radius, HoleRadius;
} DiscStruct;

#define HEIGHT_FIELD_OBJECT 9
typedef struct HeightFieldStruct
{
  ObjectStruct Object;

  gushort *Image;
  int FileType;
  int Hierarchy;
  int Smooth;
  double WaterLevel;
} HeightFieldStruct;

#define JULIA_FRACTAL_OBJECT 10
typedef struct JuliaFractalStruct
{
  ObjectStruct Object;

  Vector Parameter;
  int Quaternion;
  int Function;
  int MaxIterations;
  int Precision;
  Vector SliceNormal;
  double SliceDistance;
} JuliaFractalStruct;

#define LATHE_OBJECT 11
typedef struct LatheStruct
{
  ObjectStruct Object;

  int TypeOfSpline;
  int NumberOfPoints;
  Vector *Point;
} LatheStruct;

#define MESH_OBJECT 12
typedef struct MeshStruct
{
  ObjectStruct Object;

  struct TriangleListStruct *FirstTriangle;
} MeshStruct;

#define PLANE_OBJECT 13
typedef struct PlaneStruct
{
  ObjectStruct Object;

  Vector Normal;
  double Distance;
  int NbSquare;
  double SquareSize;
} PlaneStruct;

#define POLY_OBJECT 14
typedef struct PolyStruct
{
  ObjectStruct Object;

  int Order;
  double *T;
} PolyStruct;

#define POLYGON_OBJECT 15
typedef struct PolygonStruct
{
  ObjectStruct Object;

  int NumberOfPoints;
  Vector *Point;
} PolygonStruct;

#define PRISM_OBJECT 16
typedef struct PrismStruct
{
  ObjectStruct Object;

  int Sweep;
  int Spline;
  double Height1, Height2;
  int NumberOfPoints;
  Vector *Point;
  int Open;
} PrismStruct;

#define QUADRIC_OBJECT 17
typedef struct QuadricStruct
{
  ObjectStruct Object;

  double A, B, C, D, E, F, G, H, I, J;
} QuadricStruct;

#define QUARTIC_OBJECT 18
typedef struct QuarticStruct
{
  ObjectStruct Object;

  double T[35];
} QuarticStruct;

#define SOR_OBJECT 19
typedef struct SorStruct
{
  ObjectStruct Object;

  int NumberOfPoints;
  Vector *Point;
  int Open;
} SorStruct;

#define SPHERE_OBJECT 20
typedef struct SphereStruct
{
  ObjectStruct Object;
  
  Vector Center;
  double Radius;
} SphereStruct;

#define SUPERELLIPSOID_OBJECT 21
typedef struct SuperEllipsoidStruct
{
  ObjectStruct Object;

  double E, N;
} SuperEllipsoidStruct;

#define TEXT_OBJECT 22
typedef struct TextStruct
{
  ObjectStruct Object;

  char *FontName;
  char *String;
  double Thickness;
  Vector Offset;
} TextStruct;

#define TORUS_OBJECT 23
typedef struct TorusStruct
{
  ObjectStruct Object;

  double Major, Minor;
} TorusStruct;

#define TRIANGLE_OBJECT 24
typedef struct TriangleStruct
{
  ObjectStruct Object;

  Vector P1, P2, P3;
  Vector N1, N2, N3;
} TriangleStruct;

#define POINT_LIGHTSOURCE       0
#define SPOTLIGHT_LIGHTSOURCE   1
#define CYLINDER_LIGHTSOURCE    2
#define SHADOWLESS_LIGHTSOURCE 16
typedef struct LightSourceStruct
{
  Vector Location;
  Vector Color;
  int Type;
  Vector PointAt;
  double Radius;
  double FallOff;
  double Tightness;
  int AreaLight;
  Vector Area1, Area2;
  double AreaSize1, AreaSize2;
  double Adaptive;
  double Jitter;
  ObjectStruct *LooksLike;
  double FadeDistance;
  double FadePower;
  int AtmosphericAttenuation;
  int Atmosphere;
  struct LightSourceStruct *Next;
} LightSourceStruct;


/****************************************************************************
 * All the '*ConstantStruct' structures are there to hold the constant that *
 *  were created during the parsing of a Povray file.  I guess it would be  *
 *                       useful to use them elsewhere                       *
 ****************************************************************************/
typedef struct FloatConstantStruct
{
  double Float;
  char *Name;
  struct FloatConstantStruct *Next;
} FloatConstantStruct;

typedef struct VectorConstantStruct
{
  Vector Vector;
  char *Name;
  struct VectorConstantStruct *Next;
} VectorConstantStruct;

typedef struct ColorConstantStruct
{
  Vector Color;
  char *Name;
  struct ColorConstantStruct *Next;
} ColorConstantStruct;

typedef struct StringConstantStruct
{
  char *String;
  char *Name;
  struct StringConstantStruct *Next;
} StringConstantStruct;

typedef struct ObjectConstantStruct
{
  ObjectStruct *Object;
  char *Name;
  struct ObjectConstantStruct *Next;
} ObjectConstantStruct;

typedef struct TextureConstantStruct
{
  TextureStruct *Texture;
  char *Name;
  struct TextureConstantStruct *Next;
} TextureConstantStruct;

typedef struct PigmentConstantStruct
{
  PigmentStruct *Pigment;
  char *Name;
  struct PigmentConstantStruct *Next;
} PigmentConstantStruct;

typedef struct NormalConstantStruct
{
  NormalStruct *Normal;
  char *Name;
  struct NormalConstantStruct *Next;
} NormalConstantStruct;

typedef struct FinishConstantStruct
{
  FinishStruct *Finish;
  char *Name;
  struct FinishConstantStruct *Next;
} FinishConstantStruct;

typedef struct HaloConstantStruct
{
  HaloStruct *Halo;
  char *Name;
  struct HaloConstantStruct *Next;
} HaloConstantStruct;

typedef struct ColorMapConstantStruct
{
  ColorMapStruct *ColorMap;
  char *Name;
  struct ColorMapConstantStruct *Next;
} ColorMapConstantStruct;

typedef struct PigmentMapConstantStruct
{
  PigmentMapStruct *PigmentMap;
  char *Name;
  struct PigmentMapConstantStruct *Next;
} PigmentMapConstantStruct;

typedef struct SlopeMapEntry
{
  double Value;
  Vector Slope;
} SlopeMapEntry;

typedef struct SlopeMapStruct
{
  int NbMapEntries;
  SlopeMapEntry *Entries;
} SlopeMapStruct;

typedef struct SlopeMapConstantStruct
{
  SlopeMapStruct *SlopeMap;
  char *Name;
  struct SlopeMapConstantStruct *Next;
} SlopeMapConstantStruct;

typedef struct NormalMapConstantStruct
{
  void *NormalMap;
  char *Name;
  struct NormalMapConstantStruct *Next;
} NormalMapConstantStruct;

typedef struct CameraConstantStruct
{
  CameraStruct *Camera;
  char *Name;
  struct CameraConstantStruct *Next;
} CameraConstantStruct;

typedef struct LightSourceConstantStruct
{
  LightSourceStruct *LightSource;
  char *Name;
  struct LightSourceConstantStruct *Next;
} LightSourceConstantStruct;

typedef struct AtmosphereConstantStruct
{
  void *Atmosphere;
  char *Name;
  struct AtmosphereConstantStruct *Next;
} AtmosphereConstantStruct;

typedef struct FogConstantStruct
{
  void *Fog;
  char *Name;
  struct FogConstantStruct *Next;
} FogConstantStruct;

typedef struct RainbowConstantStruct
{
  void *Rainbow;
  char *Name;
  struct RainbowConstantStruct *Next;
} RainbowConstantStruct;

typedef struct SkySphereConstantStruct
{
  SkySphereStruct *SkySphere;
  char *Name;
  struct SkySphereConstantStruct *Next;
} SkySphereConstantStruct;

typedef struct TransformConstantStruct
{
  TransformStruct *Transform;
  char *Name;
  struct TransformConstantStruct *Next;
} TransformConstantStruct;

typedef struct FrameViewStruct
{
  char *Name;
  GtkWidget *DrawingArea;
  struct FrameStruct *Frame;
  struct FrameViewStruct *Next;
} FrameViewStruct;

/* ObjectSelection is a ringed linked list of the Objects that are
 * currently selected 
 */
typedef struct ObjectSelection
{
  ObjectStruct *Object;
  struct ObjectSelection *Next;
} ObjectSelection;

typedef struct FrameStruct
{
  char *Name; /* Its name */
  int Id;     /* This is used by the plugins interface */
  GlobalSetting Global;
  CameraStruct *Camera;
  ObjectStruct *FirstObject;
  LightSourceStruct *FirstLightSource;
  SkySphereStruct *SkySphere;
  /* All the constants linked list... */
  FloatConstantStruct *FloatConstantList;
  VectorConstantStruct *VectorConstantList;
  ColorConstantStruct *ColorConstantList;
  StringConstantStruct *StringConstantList;
  ObjectConstantStruct *ObjectConstantList;
  TextureConstantStruct *TextureConstantList;
  PigmentConstantStruct *PigmentConstantList;
  NormalConstantStruct *NormalConstantList;
  FinishConstantStruct *FinishConstantList;
  HaloConstantStruct *HaloConstantList;
  ColorMapConstantStruct *ColorMapConstantList;
  PigmentMapConstantStruct *PigmentMapConstantList;
  SlopeMapConstantStruct *SlopeMapConstantList;
  NormalMapConstantStruct *NormalMapConstantList;
  CameraConstantStruct *CameraConstantList;
  LightSourceConstantStruct *LightSourceConstantList;
  AtmosphereConstantStruct *AtmosphereConstantList;
  FogConstantStruct *FogConstantList;
  RainbowConstantStruct *RainbowConstantList;
  SkySphereConstantStruct *SkySphereConstantList;
  TransformConstantStruct *TransformConstantList;
  
  FrameViewStruct *FirstView;
  ObjectSelection *Selection;
  Vector RotationCenter;
  Vector ScaleCenter;
  /* The things for the snap... */
  gboolean SnapToGrid, ShowGrid;
  Vector   SnapGrid;
  struct FrameStruct *Next; /* The Next frame in Giram */
} FrameStruct;

FrameStruct *FirstFrame;

#define FRAME_CAMERA 1
#define ORTHO_XY_CAMERA 2
#define ORTHO_XZ_CAMERA 3
#define ORTHO_ZY_CAMERA 4

/* This is attached to every view, so we know how to render them */
typedef struct ViewUserDataStruct
{
  int CameraStyle;
  FrameStruct *FrameView;
  double ZoomFactor;
  double Xoffset, Yoffset;
  gint RenderStyle;
  GtkWidget *StatusBar;
#ifdef HAVE_GL
  GLXContext glxcontext;
#endif
} ViewUserDataStruct;

#ifdef HAVE_GL
GdkVisual *GLVisual;
int ReallyHaveGL;
#endif

/* This is used for the Gouraud rendering funcs */
typedef struct Triangle2D
{
  Vector P1,P2,P3;
  double distance;
  GdkColor Color;
} Triangle2D;

typedef struct Triangle2DNormCol
{
  int P[3][2];
  double Color[3][3];
  double Distance[3];
} Triangle2DNormCol;

/* The tools of the main tool box. */
extern gint ModeTool;
#define MT_SELECT              1
#define MT_MOVE                2
#define MT_ROTATE              3
#define MT_SCALE               4
#define MT_NEW_BICUBIC_PATCH   5
#define MT_NEW_BLOB            6
#define MT_NEW_BOX             7
#define MT_NEW_CONE            8
#define MT_NEW_CSG             9
#define MT_NEW_CUBIC          10
#define MT_NEW_CYLINDER       11
#define MT_NEW_DISC           12
#define MT_NEW_HEIGHT_FIELD   13
#define MT_NEW_JULIA_FRACTAL  14
#define MT_NEW_LATHE          15
#define MT_NEW_MESH           16
#define MT_NEW_PLANE          17
#define MT_NEW_POLY           18
#define MT_NEW_POLYGON        19
#define MT_NEW_PRISM          20
#define MT_NEW_QUADRIC        21
#define MT_NEW_QUARTIC        22
#define MT_NEW_SOR            23
#define MT_NEW_SPHERE         24
#define MT_NEW_SUPERELLIPSOID 25
#define MT_NEW_TEXT           26
#define MT_NEW_TORUS          27
#define MT_NEW_TRIANGLE       28
#define MT_NEW_LIGHTSOURCE    29

/* i18n things */
#ifdef ENABLE_NLS
#ifndef USE_GNOME
#include<libintl.h>
#define _(String) gettext(String)
#ifdef gettext_noop
#define N_(String) gettext_noop(String)
#else
#define N_(String) (String)
#endif
#endif
#else /* NLS is disabled */
#define _(String) (String)
#define N_(String) (String)
#define textdomain(String) (String)
#define gettext(String) (String)
#ifndef USE_GNOME
#define dgettext(Domain,String) (String)
#define dcgettext(Domain,String,Type) (String)
#endif
#define bindtextdomain(Domain,Directory) (Domain)
#endif


/* The prototypes used in giram.c */
void CreateView();
FrameStruct *NewFrame();
void LoadPovFunc(GtkWidget *dum);
void LoadDXFFunc(GtkWidget *dum);
void InsertDXFFunc(GtkWidget *widget, gpointer data);
void SavePovFunc(GtkWidget *widget, gpointer data);
void SaveDXFFunc(GtkWidget *widget, gpointer data);
void SaveSelectedDXFFunc(GtkWidget *widget, gpointer data);
#ifdef WITH_S3D
void SaveS3DFunc(GtkWidget *widget, gpointer data);
void SaveSelectedS3DFunc(GtkWidget *widget, gpointer data);
#endif
void ExitFunc(GtkWidget *widget, GdkEvent *event, gpointer data);
int comparTri(const void *Tri1, const void *Tri2);
void DrawGouraudTriangleZBuffer(GtkWidget *View, double *Zbuffer,
                                Triangle2DNormCol *Tri);

