/* object.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999 David Odin <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "giram.h"
#include "bicubic_patch.h"
#include "blob.h"
#include "box.h"
#include "cone.h"
#include "csg.h"
#include "cubic.h"
#include "cylinder.h"
#include "disc.h"
#include "height_field.h"
#include "julia_fractal.h"
#include "lathe.h"
#include "mesh.h"
#include "plane.h"
#include "poly.h"
#include "polygon.h"
#include "prism.h"
#include "quadric.h"
#include "quartic.h"
#include "sor.h"
#include "sphere.h"
#include "superellipsoid.h"
#include "text.h"
#include "torus.h"
#include "triangle.h"
#include "copy.h"

/*****************************************************************************
*  InitObject
******************************************************************************/
void InitObject(ObjectStruct *Object)
{
  static Id = 1;

  Object->Name = NULL;
  Object->FirstTriangle = NULL;
  Object->Next = NULL;
  Object->Trans = NULL;
  Object->BoundedBy = NULL;
  Object->ClippedBy = NULL;
  Object->Texture = CopyTexture(DefaultTexture);
  Object->Inverse = 0;
  Object->Sturm = 0;
  Object->WaterLevel = 0.0;
  Object->Smooth = 0;
  Object->NoShadow = 0;
  Object->Hierarchy = 0;
  Object->Hollow = 0;
  Object->Selected = 0;
  Object->Visible = 1;
  Object->Group = NULL;
  Object->CSGTreeNode = NULL;
  Object->Id = Id++;
}

/*****************************************************************************
*  TranslateObject
******************************************************************************/
void TranslateObject(ObjectStruct *Object, Vector Vect)
{
  switch (Object->Type)
  {
    case BICUBIC_PATCH_OBJECT:
      TranslateBicubicPatch(Object, Vect);
      break;
    case BLOB_OBJECT:
      TranslateBlob(Object, Vect);
      break;
    case BOX_OBJECT:
      TranslateBox(Object, Vect);
      break;
    case CONE_OBJECT:
      TranslateCone(Object, Vect);
      break;
    case CSG_OBJECT:
      TranslateCSG(Object, Vect);
      break;
    case CUBIC_OBJECT:
      TranslateCubic(Object, Vect);
      break;
    case CYLINDER_OBJECT:
      TranslateCylinder(Object, Vect);
      break;
    case DISC_OBJECT:
      TranslateDisc(Object, Vect);
      break;
    case HEIGHT_FIELD_OBJECT:
      TranslateHeightField(Object, Vect);
      break;
    case JULIA_FRACTAL_OBJECT:
      TranslateJuliaFractal(Object, Vect);
      break;
    case LATHE_OBJECT:
      TranslateLathe(Object, Vect);
      break;
    case MESH_OBJECT:
      TranslateMesh(Object, Vect);
      break;
    case PLANE_OBJECT:
      TranslatePlane(Object, Vect);
      break;
    case POLY_OBJECT:
      TranslatePoly(Object, Vect);
      break;
    case POLYGON_OBJECT:
      TranslatePolygon(Object, Vect);
      break;
    case PRISM_OBJECT:
      TranslatePrism(Object, Vect);
      break;
    case QUADRIC_OBJECT:
      TranslateQuadric(Object, Vect);
      break;
    case QUARTIC_OBJECT:
      TranslateQuartic(Object, Vect);
      break;
    case SOR_OBJECT:
      TranslateSor(Object, Vect);
      break;
    case SPHERE_OBJECT:
      TranslateSphere(Object, Vect);
      break;
    case SUPERELLIPSOID_OBJECT:
      TranslateSuperEllipsoid(Object, Vect);
      break;
    case TEXT_OBJECT:
      TranslateText(Object, Vect);
      break;
    case TORUS_OBJECT:
      TranslateTorus(Object, Vect);
      break;
    case TRIANGLE_OBJECT:
      TranslateTriangle(Object, Vect);
      break;
  }
}

/*****************************************************************************
*  RotateObject
******************************************************************************/
void RotateObject(ObjectStruct *Object, Vector Vect)
{
  switch (Object->Type)
  {
    case BICUBIC_PATCH_OBJECT:
      RotateBicubicPatch(Object, Vect);
      break;
    case BLOB_OBJECT:
      RotateBlob(Object, Vect);
      break;
    case BOX_OBJECT:
      RotateBox(Object, Vect);
      break;
    case CONE_OBJECT:
      RotateCone(Object, Vect);
      break;
    case CSG_OBJECT:
      RotateCSG(Object, Vect);
      break;
    case CUBIC_OBJECT:
      RotateCubic(Object, Vect);
      break;
    case CYLINDER_OBJECT:
      RotateCylinder(Object, Vect);
      break;
    case DISC_OBJECT:
      RotateDisc(Object, Vect);
      break;
    case HEIGHT_FIELD_OBJECT:
      RotateHeightField(Object, Vect);
      break;
    case JULIA_FRACTAL_OBJECT:
      RotateJuliaFractal(Object, Vect);
      break;
    case LATHE_OBJECT:
      RotateLathe(Object, Vect);
      break;
    case MESH_OBJECT:
      RotateMesh(Object, Vect);
      break;
    case PLANE_OBJECT:
      RotatePlane(Object, Vect);
      break;
    case POLY_OBJECT:
      RotatePoly(Object, Vect);
      break;
    case POLYGON_OBJECT:
      RotatePolygon(Object, Vect);
      break;
    case PRISM_OBJECT:
      RotatePrism(Object, Vect);
      break;
    case QUADRIC_OBJECT:
      RotateQuadric(Object, Vect);
      break;
    case QUARTIC_OBJECT:
      RotateQuartic(Object, Vect);
      break;
    case SOR_OBJECT:
      RotateSor(Object, Vect);
      break;
    case SPHERE_OBJECT:
      RotateSphere(Object, Vect);
      break;
    case SUPERELLIPSOID_OBJECT:
      RotateSuperEllipsoid(Object, Vect);
      break;
    case TEXT_OBJECT:
      RotateText(Object, Vect);
      break;
    case TORUS_OBJECT:
      RotateTorus(Object, Vect);
      break;
    case TRIANGLE_OBJECT:
      RotateTriangle(Object, Vect);
      break;
  }
}

/*****************************************************************************
*  ScaleObject
******************************************************************************/
void ScaleObject(ObjectStruct *Object, Vector Vect)
{
  switch (Object->Type)
  {
    case BICUBIC_PATCH_OBJECT:
      ScaleBicubicPatch(Object, Vect);
      break;
    case BLOB_OBJECT:
      ScaleBlob(Object, Vect);
      break;
    case BOX_OBJECT:
      ScaleBox(Object, Vect);
      break;
    case CONE_OBJECT:
      ScaleCone(Object, Vect);
      break;
    case CSG_OBJECT:
      ScaleCSG(Object, Vect);
      break;
    case CUBIC_OBJECT:
      ScaleCubic(Object, Vect);
      break;
    case CYLINDER_OBJECT:
      ScaleCylinder(Object, Vect);
      break;
    case DISC_OBJECT:
      ScaleDisc(Object, Vect);
      break;
    case HEIGHT_FIELD_OBJECT:
      ScaleHeightField(Object, Vect);
      break;
    case JULIA_FRACTAL_OBJECT:
      ScaleJuliaFractal(Object, Vect);
      break;
    case LATHE_OBJECT:
      ScaleLathe(Object, Vect);
      break;
    case MESH_OBJECT:
      ScaleMesh(Object, Vect);
      break;
    case PLANE_OBJECT:
      ScalePlane(Object, Vect);
      break;
    case POLY_OBJECT:
      ScalePoly(Object, Vect);
      break;
    case POLYGON_OBJECT:
      ScalePolygon(Object, Vect);
      break;
    case PRISM_OBJECT:
      ScalePrism(Object, Vect);
      break;
    case QUADRIC_OBJECT:
      ScaleQuadric(Object, Vect);
      break;
    case QUARTIC_OBJECT:
      ScaleQuartic(Object, Vect);
      break;
    case SOR_OBJECT:
      ScaleSor(Object, Vect);
      break;
    case SPHERE_OBJECT:
      ScaleSphere(Object, Vect);
      break;
    case SUPERELLIPSOID_OBJECT:
      ScaleSuperEllipsoid(Object, Vect);
      break;
    case TEXT_OBJECT:
      ScaleText(Object, Vect);
      break;
    case TORUS_OBJECT:
      ScaleTorus(Object, Vect);
      break;
    case TRIANGLE_OBJECT:
      ScaleTriangle(Object, Vect);
      break;
  }
}

/*****************************************************************************
*  ComputeObjectTriangleMesh
******************************************************************************/
void ComputeObjectTriangleMesh(ObjectStruct *Object)
{
  switch (Object->Type)
  {
    case BICUBIC_PATCH_OBJECT:
      ComputeBicubicPatchTriangleMesh(Object);
      break;
    case BLOB_OBJECT:
      ComputeBlobTriangleMesh(Object);
      break;
    case BOX_OBJECT:
      ComputeBoxTriangleMesh(Object);
      break;
    case CONE_OBJECT:
      ComputeConeTriangleMesh(Object);
      break;
    case CSG_OBJECT:
      ComputeCSGTriangleMesh(Object);
      break;
    case CUBIC_OBJECT:
      ComputeCubicTriangleMesh(Object);
      break;
    case CYLINDER_OBJECT:
      ComputeCylinderTriangleMesh(Object);
      break;
    case DISC_OBJECT:
      ComputeDiscTriangleMesh(Object);
      break;
    case HEIGHT_FIELD_OBJECT:
      ComputeHeightFieldTriangleMesh(Object);
      break;
    case JULIA_FRACTAL_OBJECT:
      ComputeJuliaFractalTriangleMesh(Object);
      break;
    case LATHE_OBJECT:
      ComputeLatheTriangleMesh(Object);
      break;
    case MESH_OBJECT:
      ComputeMeshTriangleMesh(Object);
      break;
    case PLANE_OBJECT:
      ComputePlaneTriangleMesh(Object);
      break;
    case POLY_OBJECT:
      ComputePolyTriangleMesh(Object);
      break;
    case POLYGON_OBJECT:
      ComputePolygonTriangleMesh(Object);
      break;
    case PRISM_OBJECT:
      ComputePrismTriangleMesh(Object);
      break;
    case QUADRIC_OBJECT:
      ComputeQuadricTriangleMesh(Object);
      break;
    case QUARTIC_OBJECT:
      ComputeQuarticTriangleMesh(Object);
      break;
    case SOR_OBJECT:
      ComputeSorTriangleMesh(Object);
      break;
    case SPHERE_OBJECT:
      ComputeSphereTriangleMesh(Object);
      break;
    case SUPERELLIPSOID_OBJECT:
      ComputeSuperEllipsoidTriangleMesh(Object);
      break;
    case TEXT_OBJECT:
      ComputeTextTriangleMesh(Object);
      break;
    case TORUS_OBJECT:
      ComputeTorusTriangleMesh(Object);
      break;
    case TRIANGLE_OBJECT:
      ComputeTriangleTriangleMesh(Object);
      break;
  }
}

/*****************************************************************************
*  InsideObject
******************************************************************************/
int InsideObject(ObjectStruct *Object, double x, double y, double z)
{
  switch (Object->Type)
  {
    case BICUBIC_PATCH_OBJECT:
      return 0; /* We cannot be "inside" a bicubic patch */
    case BLOB_OBJECT:
      return InsideBlob(Object, x, y, z);
    case BOX_OBJECT:
      return InsideBox(Object, x, y, z);
    case CONE_OBJECT:
      return InsideCone(Object, x, y, z);
    case CSG_OBJECT:
      return InsideCSG(Object, x, y, z);
    case CUBIC_OBJECT:
      return InsideCubic(Object, x, y, z);
    case CYLINDER_OBJECT:
      return InsideCylinder(Object, x, y, z);
    case DISC_OBJECT:
      return 0; /* We cannot be "inside" a disc */
    case HEIGHT_FIELD_OBJECT:
      return InsideHeightField(Object, x, y, z);
    case JULIA_FRACTAL_OBJECT:
      return InsideJuliaFractal(Object, x, y, z);
    case LATHE_OBJECT:
      return InsideLathe(Object, x, y, z);
    case MESH_OBJECT:
      return 0; /* We cannot be "inside" a mesh */
    case PLANE_OBJECT:
      return InsidePlane(Object, x, y, z);
    case POLY_OBJECT:
      return InsidePoly(Object, x, y, z);
    case POLYGON_OBJECT:
      return 0; /* We cannot be "inside" a polygon */
    case PRISM_OBJECT:
      return InsidePrism(Object, x, y, z);
    case QUADRIC_OBJECT:
      return InsideQuadric(Object, x, y, z);
    case QUARTIC_OBJECT:
      return InsideQuartic(Object, x, y, z);
    case SOR_OBJECT:
      return InsideSor(Object, x, y, z);
    case SPHERE_OBJECT:
      return InsideSphere(Object, x, y, z);
    case SUPERELLIPSOID_OBJECT:
      return InsideSuperEllipsoid(Object, x, y, z);
    case TEXT_OBJECT:
      return InsideText(Object, x, y, z);
    case TORUS_OBJECT:
      return InsideTorus(Object, x, y, z);
    case TRIANGLE_OBJECT:
      return 0; /* We cannot be "inside" a triangle... */
  }
  return 0; /* keep the compiler happy ;) */
}

/*****************************************************************************
*  IsObjectIntersection
******************************************************************************/
int IsObjectIntersection(ObjectStruct *Object, Vector Origin, Vector Direction)
{
  switch (Object->Type)
  {
    case BICUBIC_PATCH_OBJECT:
      return IsBicubicPatchIntersection(Object, Origin, Direction);
    case BLOB_OBJECT:
      return IsBlobIntersection(Object, Origin, Direction);
    case BOX_OBJECT:
      return IsBoxIntersection(Object, Origin, Direction);
    case CONE_OBJECT:
      return IsConeIntersection(Object, Origin, Direction);
    case CSG_OBJECT:
      return IsCSGIntersection(Object, Origin, Direction);
    case CUBIC_OBJECT:
      return IsCubicIntersection(Object, Origin, Direction);
    case CYLINDER_OBJECT:
      return IsCylinderIntersection(Object, Origin, Direction);
    case DISC_OBJECT:
      return IsDiscIntersection(Object, Origin, Direction);
    case HEIGHT_FIELD_OBJECT:
      return IsHeightFieldIntersection(Object, Origin, Direction);
    case JULIA_FRACTAL_OBJECT:
      return IsJuliaFractalIntersection(Object, Origin, Direction);
    case LATHE_OBJECT:
      return IsLatheIntersection(Object, Origin, Direction);
    case MESH_OBJECT:
      return IsMeshIntersection(Object, Origin, Direction);
    case PLANE_OBJECT:
      return IsPlaneIntersection(Object, Origin, Direction);
    case POLY_OBJECT:
      return IsPolyIntersection(Object, Origin, Direction);
    case POLYGON_OBJECT:
      return IsPolygonIntersection(Object, Origin, Direction);
    case PRISM_OBJECT:
      return IsPrismIntersection(Object, Origin, Direction);
    case QUADRIC_OBJECT:
      return IsQuadricIntersection(Object, Origin, Direction);
    case QUARTIC_OBJECT:
      return IsQuarticIntersection(Object, Origin, Direction);
    case SOR_OBJECT:
      return IsSorIntersection(Object, Origin, Direction);
    case SPHERE_OBJECT:
      return IsSphereIntersection(Object, Origin, Direction);
    case SUPERELLIPSOID_OBJECT:
      return IsSuperEllipsoidIntersection(Object, Origin, Direction);
    case TEXT_OBJECT:
      return IsTextIntersection(Object, Origin, Direction);
    case TORUS_OBJECT:
      return IsTorusIntersection(Object, Origin, Direction);
    case TRIANGLE_OBJECT:
      return IsTriangleIntersection(Object, Origin, Direction);
  }
  return 0; /* keep the compiler happy ;) */
}

/*****************************************************************************
*  ClickInObjectXY
******************************************************************************/
int ClickInObjectXY(ObjectStruct *Object, gdouble x, gdouble y)
{
  switch (Object->Type)
  {
    case BICUBIC_PATCH_OBJECT:
      return ClickInBicubicPatchXY(Object, x, y);
    case BLOB_OBJECT:
      return ClickInBlobXY(Object, x, y);
    case BOX_OBJECT:
      return ClickInBoxXY(Object, x, y);
    case CONE_OBJECT:
      return ClickInConeXY(Object, x, y);
    case CSG_OBJECT:
      return ClickInCSGXY(Object, x, y);
    case CUBIC_OBJECT:
      return ClickInCubicXY(Object, x, y);
    case CYLINDER_OBJECT:
      return ClickInCylinderXY(Object, x, y);
    case DISC_OBJECT:
      return ClickInDiscXY(Object, x, y);
    case HEIGHT_FIELD_OBJECT:
      return ClickInHeightFieldXY(Object, x, y);
    case JULIA_FRACTAL_OBJECT:
      return ClickInJuliaFractalXY(Object, x, y);
    case LATHE_OBJECT:
      return ClickInLatheXY(Object, x, y);
    case MESH_OBJECT:
      return ClickInMeshXY(Object, x, y);
    case PLANE_OBJECT:
      return ClickInPlaneXY(Object, x, y);
    case POLY_OBJECT:
      return ClickInPolyXY(Object, x, y);
    case POLYGON_OBJECT:
      return ClickInPolygonXY(Object, x, y);
    case PRISM_OBJECT:
      return ClickInPrismXY(Object, x, y);
    case QUADRIC_OBJECT:
      return ClickInQuadricXY(Object, x, y);
    case QUARTIC_OBJECT:
      return ClickInQuarticXY(Object, x, y);
    case SOR_OBJECT:
      return ClickInSorXY(Object, x, y);
    case SPHERE_OBJECT:
      return ClickInSphereXY(Object, x, y);
    case SUPERELLIPSOID_OBJECT:
      return ClickInSuperEllipsoidXY(Object, x, y);
    case TEXT_OBJECT:
      return ClickInTextXY(Object, x, y);
    case TORUS_OBJECT:
      return ClickInTorusXY(Object, x, y);
    case TRIANGLE_OBJECT:
      return ClickInTriangleXY(Object, x, y);
  }
  return 0; /* keep the compiler happy ;) */
}

/*****************************************************************************
*  ClickInObjectXZ
******************************************************************************/
int ClickInObjectXZ(ObjectStruct *Object, gdouble x, gdouble y)
{
  switch (Object->Type)
  {
    case BICUBIC_PATCH_OBJECT:
      return ClickInBicubicPatchXZ(Object, x, y);
    case BLOB_OBJECT:
      return ClickInBlobXZ(Object, x, y);
    case BOX_OBJECT:
      return ClickInBoxXZ(Object, x, y);
    case CONE_OBJECT:
      return ClickInConeXZ(Object, x, y);
    case CSG_OBJECT:
      return ClickInCSGXZ(Object, x, y);
    case CUBIC_OBJECT:
      return ClickInCubicXZ(Object, x, y);
    case CYLINDER_OBJECT:
      return ClickInCylinderXZ(Object, x, y);
    case DISC_OBJECT:
      return ClickInDiscXZ(Object, x, y);
    case HEIGHT_FIELD_OBJECT:
      return ClickInHeightFieldXZ(Object, x, y);
    case JULIA_FRACTAL_OBJECT:
      return ClickInJuliaFractalXZ(Object, x, y);
    case LATHE_OBJECT:
      return ClickInLatheXZ(Object, x, y);
    case MESH_OBJECT:
      return ClickInMeshXZ(Object, x, y);
    case PLANE_OBJECT:
      return ClickInPlaneXZ(Object, x, y);
    case POLY_OBJECT:
      return ClickInPolyXZ(Object, x, y);
    case POLYGON_OBJECT:
      return ClickInPolygonXZ(Object, x, y);
    case PRISM_OBJECT:
      return ClickInPrismXZ(Object, x, y);
    case QUADRIC_OBJECT:
      return ClickInQuadricXZ(Object, x, y);
    case QUARTIC_OBJECT:
      return ClickInQuarticXZ(Object, x, y);
    case SOR_OBJECT:
      return ClickInSorXZ(Object, x, y);
    case SPHERE_OBJECT:
      return ClickInSphereXZ(Object, x, y);
    case SUPERELLIPSOID_OBJECT:
      return ClickInSuperEllipsoidXZ(Object, x, y);
    case TEXT_OBJECT:
      return ClickInTextXZ(Object, x, y);
    case TORUS_OBJECT:
      return ClickInTorusXZ(Object, x, y);
    case TRIANGLE_OBJECT:
      return ClickInTriangleXZ(Object, x, y);
  }
  return 0; /* keep the compiler happy ;) */
}

/*****************************************************************************
*  ClickInObjectZY
******************************************************************************/
int ClickInObjectZY(ObjectStruct *Object, gdouble x, gdouble y)
{
  switch (Object->Type)
  {
    case BICUBIC_PATCH_OBJECT:
      return ClickInBicubicPatchZY(Object, x, y);
    case BLOB_OBJECT:
      return ClickInBlobZY(Object, x, y);
    case BOX_OBJECT:
      return ClickInBoxZY(Object, x, y);
    case CONE_OBJECT:
      return ClickInConeZY(Object, x, y);
    case CSG_OBJECT:
      return ClickInCSGZY(Object, x, y);
    case CUBIC_OBJECT:
      return ClickInCubicZY(Object, x, y);
    case CYLINDER_OBJECT:
      return ClickInCylinderZY(Object, x, y);
    case DISC_OBJECT:
      return ClickInDiscZY(Object, x, y);
    case HEIGHT_FIELD_OBJECT:
      return ClickInHeightFieldZY(Object, x, y);
    case JULIA_FRACTAL_OBJECT:
      return ClickInJuliaFractalZY(Object, x, y);
    case LATHE_OBJECT:
      return ClickInLatheZY(Object, x, y);
    case MESH_OBJECT:
      return ClickInMeshZY(Object, x, y);
    case PLANE_OBJECT:
      return ClickInPlaneZY(Object, x, y);
    case POLY_OBJECT:
      return ClickInPolyZY(Object, x, y);
    case POLYGON_OBJECT:
      return ClickInPolygonZY(Object, x, y);
    case PRISM_OBJECT:
      return ClickInPrismZY(Object, x, y);
    case QUADRIC_OBJECT:
      return ClickInQuadricZY(Object, x, y);
    case QUARTIC_OBJECT:
      return ClickInQuarticZY(Object, x, y);
    case SOR_OBJECT:
      return ClickInSorZY(Object, x, y);
    case SPHERE_OBJECT:
      return ClickInSphereZY(Object, x, y);
    case SUPERELLIPSOID_OBJECT:
      return ClickInSuperEllipsoidZY(Object, x, y);
    case TEXT_OBJECT:
      return ClickInTextZY(Object, x, y);
    case TORUS_OBJECT:
      return ClickInTorusZY(Object, x, y);
    case TRIANGLE_OBJECT:
      return ClickInTriangleZY(Object, x, y);
  }
  return 0; /* keep the compiler happy ;) */
}

/*****************************************************************************
*  FindIntersectionObjectSegment
******************************************************************************/
int FindIntersectionObjectSegment(Vector InterPoint, Vector InterNorm,
                                  Vector InPoint, Vector OutPoint,
                                  ObjectStruct *Object)
{
  switch (Object->Type)
  {
    case BICUBIC_PATCH_OBJECT:
      return FindIntersectionBicubicPatchSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case BLOB_OBJECT:
      return FindIntersectionBlobSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case BOX_OBJECT:
      return FindIntersectionBoxSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case CONE_OBJECT:
      return FindIntersectionConeSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case CSG_OBJECT:
      return FindIntersectionCSGSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case CUBIC_OBJECT:
      return FindIntersectionCubicSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case CYLINDER_OBJECT:
      return FindIntersectionCylinderSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case DISC_OBJECT:
      return FindIntersectionDiscSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case HEIGHT_FIELD_OBJECT:
      return FindIntersectionHeightFieldSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case JULIA_FRACTAL_OBJECT:
      return FindIntersectionJuliaFractalSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case LATHE_OBJECT:
      return FindIntersectionLatheSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case MESH_OBJECT:
      return FindIntersectionMeshSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case PLANE_OBJECT:
      return FindIntersectionPlaneSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case POLY_OBJECT:
      return FindIntersectionPolySegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case POLYGON_OBJECT:
      return FindIntersectionPolygonSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case PRISM_OBJECT:
      return FindIntersectionPrismSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case QUADRIC_OBJECT:
      return FindIntersectionQuadricSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case QUARTIC_OBJECT:
      return FindIntersectionQuarticSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case SOR_OBJECT:
      return FindIntersectionSorSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case SPHERE_OBJECT:
      return FindIntersectionSphereSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case SUPERELLIPSOID_OBJECT:
      return FindIntersectionSuperEllipsoidSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case TEXT_OBJECT:
      return FindIntersectionTextSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case TORUS_OBJECT:
      return FindIntersectionTorusSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
    case TRIANGLE_OBJECT:
      return FindIntersectionTriangleSegment(InterPoint, InterNorm, InPoint, OutPoint, Object);
  }
  return 0;
}

/*****************************************************************************
*  PDBTransformObject
******************************************************************************/
int PDBTransformObject(int *FrameId, int *ObjectId, TransformStruct *Trans)
{
  FrameStruct *TmpFrame;
  ObjectStruct *TmpObject;
  int i,j;
  
  for (TmpFrame = FirstFrame ;
       TmpFrame->Id != *FrameId ;
       TmpFrame = TmpFrame->Next)
     ;
  for (TmpObject = TmpFrame->FirstObject ;
       TmpObject->Id != *ObjectId ;
       TmpObject = TmpObject->Next)
     ;
  if (TmpObject->Trans)
  {
    ComposeTrans(TmpObject->Trans, Trans);
  } else
  {
    TmpObject->Trans = g_malloc(sizeof(TransformStruct));
    for (i=0 ; i<4 ; i++)
      for (j=0 ; j<4 ; j++)
      {
        TmpObject->Trans->Direct[i][j] = Trans->Direct[i][j];
        TmpObject->Trans->Inverse[i][j] = Trans->Inverse[i][j];
      }
  }
  ComputeObjectTriangleMesh(TmpObject);
  return *ObjectId;
}

