/* view.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999 David Odin <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include "giram.h"
#include "tool_select.h"
#include "tool_move.h"
#include "tool_rotate.h"
#include "tool_scale.h"
#include "tool_options.h"
#include "bicubic_patch.h"
#include "blob.h"
#include "box.h"
#include "cone.h"
#include "csg.h"
#include "cubic.h"
#include "cylinder.h"
#include "disc.h"
#include "height_field.h"
#include "julia_fractal.h"
#include "lathe.h"
#include "mesh.h"
#include "plane.h"
#include "poly.h"
#include "polygon.h"
#include "prism.h"
#include "quadric.h"
#include "quartic.h"
#include "sor.h"
#include "sphere.h"
#include "superellipsoid.h"
#include "text.h"
#include "torus.h"
#include "triangle.h"
#include "light_source.h"
#include "object.h"
#include "csgtree.h"
#include "viewXY.h"
#include "viewXZ.h"
#include "viewZY.h"
#include "view_camera.h"
#include "utils.h"
#include "camera.h"
#include "SavePov.h"
#include "plugins.h"
#include "edit.h"
#include "snap.h"
#include "dxf.h"
#ifdef WITH_S3D
#include "s3d.h"
#endif

/*****************************************************************************
*  SetWireFrameExpose
******************************************************************************/
void SetWireFrameExpose(GtkWidget *DrawingArea)
{
  ViewUserDataStruct *UserData;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  gtk_signal_disconnect(GTK_OBJECT(DrawingArea), UserData->RenderStyle);
  switch (UserData->CameraStyle)
  {
    case FRAME_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewCamera, NULL);
      break;
    case ORTHO_XY_CAMERA:
      UserData->RenderStyle = 
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewXY, NULL);
      break;
    case ORTHO_XZ_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewXZ, NULL);
      break;
    case ORTHO_ZY_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewZY, NULL);
      break;
  }
  gtk_widget_queue_draw(GTK_WIDGET(DrawingArea));
}

/*****************************************************************************
*  SetHiddenFacesExpose
******************************************************************************/
void SetHiddenFacesExpose(GtkWidget *DrawingArea)
{
  ViewUserDataStruct *UserData;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  gtk_signal_disconnect(GTK_OBJECT(DrawingArea), UserData->RenderStyle);
  switch (UserData->CameraStyle)
  {
    case FRAME_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeHiddenFacesViewCamera, NULL);
      break;
    case ORTHO_XY_CAMERA:
      UserData->RenderStyle = 
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeHiddenFacesViewXY, NULL);
      break;
    case ORTHO_XZ_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeHiddenFacesViewXZ, NULL);
      break;
    case ORTHO_ZY_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeHiddenFacesViewZY, NULL);
      break;
  }
  gtk_widget_queue_draw(GTK_WIDGET(DrawingArea));
}

/*****************************************************************************
*  SetGouraudExpose
******************************************************************************/
void SetGouraudExpose(GtkWidget *DrawingArea)
{
  ViewUserDataStruct *UserData;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  gtk_signal_disconnect(GTK_OBJECT(DrawingArea), UserData->RenderStyle);
  switch (UserData->CameraStyle)
  {
    case FRAME_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeGouraudViewCamera, NULL);
      break;
    case ORTHO_XY_CAMERA:
      UserData->RenderStyle = 
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeGouraudViewXY, NULL);
      break;
    case ORTHO_XZ_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeGouraudViewXZ, NULL);
      break;
    case ORTHO_ZY_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeGouraudViewZY, NULL);
      break;
  }
  gtk_widget_queue_draw(GTK_WIDGET(DrawingArea));
}

#ifdef HAVE_GL
/*****************************************************************************
*  SetOpenGLExpose
******************************************************************************/
void SetOpenGLExpose(GtkWidget *DrawingArea)
{
  ViewUserDataStruct *UserData;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  gtk_signal_disconnect(GTK_OBJECT(DrawingArea), UserData->RenderStyle);
  switch (UserData->CameraStyle)
  {
    case FRAME_CAMERA:
      break;
    case ORTHO_XY_CAMERA:
      UserData->RenderStyle = 
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeGLViewXY, NULL);
      break;
    case ORTHO_XZ_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeGLViewXZ, NULL);
      break;
    case ORTHO_ZY_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeGLViewZY, NULL);
      break;
  }
  /* OpenGL functions can be called only if make_current returns true */
  if (glXMakeCurrent(GDK_DISPLAY(), GDK_WINDOW_XWINDOW(DrawingArea->window),
                     UserData->glxcontext))
  {
    glViewport(0,0, DrawingArea->allocation.width, DrawingArea->allocation.height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0,100, 100,0);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
  }
  gtk_widget_queue_draw(GTK_WIDGET(DrawingArea));
}
#endif

/*****************************************************************************
*  CallPovFront
******************************************************************************/
void CallPovFront(GtkWidget *DrawingArea)
{
  ViewUserDataStruct *UserData;
  FrameStruct *LocalFrame;
  char *TmpFileName;
  gchar *CommandLine;

  TmpFileName = g_malloc(15);
  strcpy(TmpFileName, "/tmp/GiramXXXXXX");
  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  LocalFrame = UserData->FrameView;    
  mktemp(TmpFileName);
  SavePovByName(LocalFrame, TmpFileName);
  CommandLine = g_strdup_printf("povfront %s &", TmpFileName);
  system(CommandLine);
  g_free(CommandLine);
  g_free(TmpFileName);
}

/*****************************************************************************
*  ZoomIn
******************************************************************************/
void ZoomIn(GtkWidget *DrawingArea)
{
  ViewUserDataStruct *UserData;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  UserData->ZoomFactor *= 1.5;
  gtk_widget_queue_draw(GTK_WIDGET(DrawingArea));
  /* FIXME: Rulers */
}

/*****************************************************************************
*  ZoomOut
******************************************************************************/
void ZoomOut(GtkWidget *DrawingArea)
{
  ViewUserDataStruct *UserData;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  UserData->ZoomFactor /= 1.5;
  gtk_widget_queue_draw(GTK_WIDGET(DrawingArea));
  /* FIXME: Rulers */
}

/*****************************************************************************
*  FitToSelectionFunc
******************************************************************************/
void FitToSelectionFunc(GtkWidget *DrawingArea)
{
  ViewUserDataStruct *UserData;
  ObjectSelection *TmpSelect;
  TriangleListStruct *TmpTri;
  Vector MaxExt, MinExt;
  double ZoomX=1.0, ZoomY=1.0;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");

  MinExt[0] = MinExt[1] = MinExt[2] = 1e100;
  MaxExt[0] = MaxExt[1] = MaxExt[2] = -1e100;
  if (UserData->FrameView->Selection == NULL)
    return;
  for (TmpSelect = UserData->FrameView->Selection;
       TmpSelect->Next != UserData->FrameView->Selection ;
       TmpSelect = TmpSelect->Next)
  {
    /* Find the extents */
    for (TmpTri = TmpSelect->Object->FirstTriangle ; 
         TmpTri ; 
         TmpTri = TmpTri->Next)
    {
      if (MinExt[0] > TmpTri->P1[0]) MinExt[0] = TmpTri->P1[0];
      if (MinExt[0] > TmpTri->P2[0]) MinExt[0] = TmpTri->P2[0];
      if (MinExt[0] > TmpTri->P3[0]) MinExt[0] = TmpTri->P3[0];
      if (MinExt[1] > TmpTri->P1[1]) MinExt[1] = TmpTri->P1[1];
      if (MinExt[1] > TmpTri->P2[1]) MinExt[1] = TmpTri->P2[1];
      if (MinExt[1] > TmpTri->P3[1]) MinExt[1] = TmpTri->P3[1];
      if (MinExt[2] > TmpTri->P1[2]) MinExt[2] = TmpTri->P1[2];
      if (MinExt[2] > TmpTri->P2[2]) MinExt[2] = TmpTri->P2[2];
      if (MinExt[2] > TmpTri->P3[2]) MinExt[2] = TmpTri->P3[2];

      if (MaxExt[0] < TmpTri->P1[0]) MaxExt[0] = TmpTri->P1[0];
      if (MaxExt[0] < TmpTri->P2[0]) MaxExt[0] = TmpTri->P2[0];
      if (MaxExt[0] < TmpTri->P3[0]) MaxExt[0] = TmpTri->P3[0];
      if (MaxExt[1] < TmpTri->P1[1]) MaxExt[1] = TmpTri->P1[1];
      if (MaxExt[1] < TmpTri->P2[1]) MaxExt[1] = TmpTri->P2[1];
      if (MaxExt[1] < TmpTri->P3[1]) MaxExt[1] = TmpTri->P3[1];
      if (MaxExt[2] < TmpTri->P1[2]) MaxExt[2] = TmpTri->P1[2];
      if (MaxExt[2] < TmpTri->P2[2]) MaxExt[2] = TmpTri->P2[2];
      if (MaxExt[2] < TmpTri->P3[2]) MaxExt[2] = TmpTri->P3[2];
    }
  }
  for (TmpTri = TmpSelect->Object->FirstTriangle ; 
       TmpTri ; 
       TmpTri = TmpTri->Next)
  {
    if (MinExt[0] > TmpTri->P1[0]) MinExt[0] = TmpTri->P1[0];
    if (MinExt[0] > TmpTri->P2[0]) MinExt[0] = TmpTri->P2[0];
    if (MinExt[0] > TmpTri->P3[0]) MinExt[0] = TmpTri->P3[0];
    if (MinExt[1] > TmpTri->P1[1]) MinExt[1] = TmpTri->P1[1];
    if (MinExt[1] > TmpTri->P2[1]) MinExt[1] = TmpTri->P2[1];
    if (MinExt[1] > TmpTri->P3[1]) MinExt[1] = TmpTri->P3[1];
    if (MinExt[2] > TmpTri->P1[2]) MinExt[2] = TmpTri->P1[2];
    if (MinExt[2] > TmpTri->P2[2]) MinExt[2] = TmpTri->P2[2];
    if (MinExt[2] > TmpTri->P3[2]) MinExt[2] = TmpTri->P3[2];

    if (MaxExt[0] < TmpTri->P1[0]) MaxExt[0] = TmpTri->P1[0];
    if (MaxExt[0] < TmpTri->P2[0]) MaxExt[0] = TmpTri->P2[0];
    if (MaxExt[0] < TmpTri->P3[0]) MaxExt[0] = TmpTri->P3[0];
    if (MaxExt[1] < TmpTri->P1[1]) MaxExt[1] = TmpTri->P1[1];
    if (MaxExt[1] < TmpTri->P2[1]) MaxExt[1] = TmpTri->P2[1];
    if (MaxExt[1] < TmpTri->P3[1]) MaxExt[1] = TmpTri->P3[1];
    if (MaxExt[2] < TmpTri->P1[2]) MaxExt[2] = TmpTri->P1[2];
    if (MaxExt[2] < TmpTri->P2[2]) MaxExt[2] = TmpTri->P2[2];
    if (MaxExt[2] < TmpTri->P3[2]) MaxExt[2] = TmpTri->P3[2];
  }
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      ZoomX = DrawingArea->allocation.width / (MaxExt[0]-MinExt[0]);
      ZoomY = DrawingArea->allocation.height / (MaxExt[1]-MinExt[1]);
      UserData->Xoffset = (MaxExt[0]+MinExt[0]) / 2.0;
      UserData->Yoffset = (MaxExt[1]+MinExt[1]) / 2.0;
      break;
    case ORTHO_XZ_CAMERA:
      ZoomX = DrawingArea->allocation.width / (MaxExt[0]-MinExt[0]);
      ZoomY = DrawingArea->allocation.height / (MaxExt[2]-MinExt[2]);
      UserData->Xoffset = (MaxExt[0]+MinExt[0]) / 2.0;
      UserData->Yoffset = (MaxExt[2]+MinExt[2]) / 2.0;
      break;
    case ORTHO_ZY_CAMERA:
      ZoomX = DrawingArea->allocation.width / (MaxExt[2]-MinExt[2]);
      ZoomY = DrawingArea->allocation.height / (MaxExt[1]-MinExt[1]);
      UserData->Xoffset = (MaxExt[2]+MinExt[2]) / 2.0;
      UserData->Yoffset = (MaxExt[1]+MinExt[1]) / 2.0;
      break;
  }
  UserData->ZoomFactor = MIN(ZoomX, ZoomY) * 0.95;
  gtk_widget_queue_draw(GTK_WIDGET(DrawingArea));
}

/*****************************************************************************
*  FitToSceneFunc
******************************************************************************/
void FitToSceneFunc(GtkWidget *DrawingArea)
{
  ViewUserDataStruct *UserData;
  ObjectStruct *TmpObject;
  TriangleListStruct *TmpTri;
  Vector MaxExt, MinExt;
  double ZoomX=1.0, ZoomY=1.0;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");

  MinExt[0] = MinExt[1] = MinExt[2] = 1e100;
  MaxExt[0] = MaxExt[1] = MaxExt[2] = -1e100;
  if (UserData->FrameView->FirstObject == NULL)
    return;
  for (TmpObject = UserData->FrameView->FirstObject;
       TmpObject ;
       TmpObject = TmpObject->Next)
  {
    /* Find the extents */
    for (TmpTri = TmpObject->FirstTriangle ; 
         TmpTri ; 
         TmpTri = TmpTri->Next)
    {
      if (MinExt[0] > TmpTri->P1[0]) MinExt[0] = TmpTri->P1[0];
      if (MinExt[0] > TmpTri->P2[0]) MinExt[0] = TmpTri->P2[0];
      if (MinExt[0] > TmpTri->P3[0]) MinExt[0] = TmpTri->P3[0];
      if (MinExt[1] > TmpTri->P1[1]) MinExt[1] = TmpTri->P1[1];
      if (MinExt[1] > TmpTri->P2[1]) MinExt[1] = TmpTri->P2[1];
      if (MinExt[1] > TmpTri->P3[1]) MinExt[1] = TmpTri->P3[1];
      if (MinExt[2] > TmpTri->P1[2]) MinExt[2] = TmpTri->P1[2];
      if (MinExt[2] > TmpTri->P2[2]) MinExt[2] = TmpTri->P2[2];
      if (MinExt[2] > TmpTri->P3[2]) MinExt[2] = TmpTri->P3[2];

      if (MaxExt[0] < TmpTri->P1[0]) MaxExt[0] = TmpTri->P1[0];
      if (MaxExt[0] < TmpTri->P2[0]) MaxExt[0] = TmpTri->P2[0];
      if (MaxExt[0] < TmpTri->P3[0]) MaxExt[0] = TmpTri->P3[0];
      if (MaxExt[1] < TmpTri->P1[1]) MaxExt[1] = TmpTri->P1[1];
      if (MaxExt[1] < TmpTri->P2[1]) MaxExt[1] = TmpTri->P2[1];
      if (MaxExt[1] < TmpTri->P3[1]) MaxExt[1] = TmpTri->P3[1];
      if (MaxExt[2] < TmpTri->P1[2]) MaxExt[2] = TmpTri->P1[2];
      if (MaxExt[2] < TmpTri->P2[2]) MaxExt[2] = TmpTri->P2[2];
      if (MaxExt[2] < TmpTri->P3[2]) MaxExt[2] = TmpTri->P3[2];
    }
  }
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      ZoomX = DrawingArea->allocation.width / (MaxExt[0]-MinExt[0]);
      ZoomY = DrawingArea->allocation.height / (MaxExt[1]-MinExt[1]);
      UserData->Xoffset = (MaxExt[0]+MinExt[0]) / 2.0;
      UserData->Yoffset = (MaxExt[1]+MinExt[1]) / 2.0;
      break;
    case ORTHO_XZ_CAMERA:
      ZoomX = DrawingArea->allocation.width / (MaxExt[0]-MinExt[0]);
      ZoomY = DrawingArea->allocation.height / (MaxExt[2]-MinExt[2]);
      UserData->Xoffset = (MaxExt[0]+MinExt[0]) / 2.0;
      UserData->Yoffset = (MaxExt[2]+MinExt[2]) / 2.0;
      break;
    case ORTHO_ZY_CAMERA:
      ZoomX = DrawingArea->allocation.width / (MaxExt[2]-MinExt[2]);
      ZoomY = DrawingArea->allocation.height / (MaxExt[1]-MinExt[1]);
      UserData->Xoffset = (MaxExt[2]+MinExt[2]) / 2.0;
      UserData->Yoffset = (MaxExt[1]+MinExt[1]) / 2.0;
      break;
  }
  UserData->ZoomFactor = MIN(ZoomX, ZoomY) * 0.95;
  gtk_widget_queue_draw(GTK_WIDGET(DrawingArea));
}

/*****************************************************************************
*  SelectAll
******************************************************************************/
void SelectAll(GtkWidget *DrawingArea)
{
  ViewUserDataStruct *UserData;
  ObjectStruct *Tmp;
  FrameViewStruct *TmpView;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  DestroySelection(&(UserData->FrameView->Selection));

  for (Tmp=UserData->FrameView->FirstObject ; Tmp ; Tmp=Tmp->Next)
    AddObjectToSelection(&(UserData->FrameView->Selection), Tmp);
  /* FIXME: should update the CSG Tree selection */
  /* We redraw all the Views */
  for (TmpView = UserData->FrameView->FirstView ; 
       TmpView ; 
       TmpView = TmpView->Next)
    gtk_widget_queue_draw(TmpView->DrawingArea);
}

/*****************************************************************************
*  SelectNone
******************************************************************************/
void SelectNone(GtkWidget *DrawingArea)
{
  ViewUserDataStruct *UserData;
  FrameViewStruct *TmpView;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  DestroySelection(&(UserData->FrameView->Selection));
  /* FIXME: should update the CSG Tree selection */
  /* We redraw all the Views */
  for (TmpView = UserData->FrameView->FirstView ; 
       TmpView ; 
       TmpView = TmpView->Next)
    gtk_widget_queue_draw(TmpView->DrawingArea);
}

/*****************************************************************************
*  InvertSelection
******************************************************************************/
void InvertSelection(GtkWidget *DrawingArea)
{
  ViewUserDataStruct *UserData;
  FrameViewStruct *TmpView;
  ObjectSelection *TmpSelection;
  ObjectSelection *Prev;
  ObjectStruct *Object;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  TmpSelection = NULL;
  for (Object=UserData->FrameView->FirstObject ; Object ; Object=Object->Next)
    AddObjectToSelection(&(TmpSelection), Object);
  if (UserData->FrameView->Selection)
  {
    for (Prev = UserData->FrameView->Selection ;
         Prev->Next != UserData->FrameView->Selection ;
         Prev = Prev->Next)
      RemoveObjectFromSelection(&(TmpSelection), Prev->Object);
    RemoveObjectFromSelection(&(TmpSelection), Prev->Object);
  }
  DestroySelection(&(UserData->FrameView->Selection));
  UserData->FrameView->Selection = TmpSelection;
  /* FIXME: should update the CSG Tree selection */
  /* We redraw all the Views */
  for (TmpView = UserData->FrameView->FirstView ;
       TmpView ;
       TmpView = TmpView->Next)
    gtk_widget_queue_draw(TmpView->DrawingArea);
}

/*****************************************************************************
*  TranslateSelection
******************************************************************************/
void TranslateSelection(GtkWidget *Dialog)
{
  Vector Vect;
  GtkWidget **Data;
  GtkWidget *DrawingArea;
  ViewUserDataStruct *UserData;
  ObjectSelection *ObjectSel;
  FrameViewStruct *TmpView;

  Data = gtk_object_get_data(GTK_OBJECT(Dialog), "WidgetData");
  DrawingArea = Data[3];
  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");

  Vect[0] = atof(gtk_entry_get_text(GTK_ENTRY(Data[0])));
  Vect[1] = atof(gtk_entry_get_text(GTK_ENTRY(Data[1])));
  Vect[2] = atof(gtk_entry_get_text(GTK_ENTRY(Data[2])));
  if (UserData->FrameView->Selection)
  {
    TranslateObject(UserData->FrameView->Selection->Object, Vect);
    ComputeObjectTriangleMesh(UserData->FrameView->Selection->Object);
    for (ObjectSel = UserData->FrameView->Selection ; 
         ObjectSel->Next != UserData->FrameView->Selection ; 
         ObjectSel = ObjectSel->Next)
    {
      TranslateObject(ObjectSel->Object, Vect);
      ComputeObjectTriangleMesh(ObjectSel->Object);
    }
    for (TmpView = UserData->FrameView->FirstView ; 
         TmpView ; 
         TmpView = TmpView->Next)
      gtk_widget_queue_draw(TmpView->DrawingArea);
  }
}

/*****************************************************************************
*  RotateSelection
******************************************************************************/
void RotateSelection(GtkWidget *Dialog)
{
  Vector Vect;
  GtkWidget **Data;
  GtkWidget *DrawingArea;
  ViewUserDataStruct *UserData;
  ObjectSelection *ObjectSel;
  FrameViewStruct *TmpView;

  Data = gtk_object_get_data(GTK_OBJECT(Dialog), "WidgetData");
  DrawingArea = Data[3];
  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");

  Vect[0] = atof(gtk_entry_get_text(GTK_ENTRY(Data[0])));
  Vect[1] = atof(gtk_entry_get_text(GTK_ENTRY(Data[1])));
  Vect[2] = atof(gtk_entry_get_text(GTK_ENTRY(Data[2])));
  if (UserData->FrameView->Selection)
  {
    RotateObject(UserData->FrameView->Selection->Object, Vect);
    ComputeObjectTriangleMesh(UserData->FrameView->Selection->Object);
    for (ObjectSel = UserData->FrameView->Selection ; 
         ObjectSel->Next != UserData->FrameView->Selection ; 
         ObjectSel = ObjectSel->Next)
    {
      RotateObject(ObjectSel->Object, Vect);
      ComputeObjectTriangleMesh(ObjectSel->Object);
    }
    for (TmpView = UserData->FrameView->FirstView ; 
         TmpView ; 
         TmpView = TmpView->Next)
      gtk_widget_queue_draw(TmpView->DrawingArea);
  }
}

/*****************************************************************************
*  ScaleSelection
******************************************************************************/
void ScaleSelection(GtkWidget *Dialog)
{
  Vector Vect;
  GtkWidget **Data;
  GtkWidget *DrawingArea;
  ViewUserDataStruct *UserData;
  ObjectSelection *ObjectSel;
  FrameViewStruct *TmpView;

  Data = gtk_object_get_data(GTK_OBJECT(Dialog), "WidgetData");
  DrawingArea = Data[3];
  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");

  Vect[0] = atof(gtk_entry_get_text(GTK_ENTRY(Data[0])));
  Vect[1] = atof(gtk_entry_get_text(GTK_ENTRY(Data[1])));
  Vect[2] = atof(gtk_entry_get_text(GTK_ENTRY(Data[2])));
  if (fabs(Vect[0]) < 10e-8)
  {
    Vect[0] = 1.0;
    WarningBox(_("Scale"), _("You can't scale by 0, changed to 1"));
  }
  if (fabs(Vect[1]) < 10e-8)
  {
    Vect[1] = 1.0;
    WarningBox(_("Scale"), _("You can't scale by 0, changed to 1"));
  }
  if (fabs(Vect[2]) < 10e-8)
  {
    Vect[2] = 1.0;
    WarningBox(_("Scale"), _("You can't scale by 0, changed to 1"));
  }
  if (UserData->FrameView->Selection)
  {
    ScaleObject(UserData->FrameView->Selection->Object, Vect);
    ComputeObjectTriangleMesh(UserData->FrameView->Selection->Object);
    for (ObjectSel = UserData->FrameView->Selection ; 
         ObjectSel->Next != UserData->FrameView->Selection ; 
         ObjectSel = ObjectSel->Next)
    {
      ScaleObject(ObjectSel->Object, Vect);
      ComputeObjectTriangleMesh(ObjectSel->Object);
    }
    for (TmpView = UserData->FrameView->FirstView ; 
         TmpView ; 
         TmpView = TmpView->Next)
      gtk_widget_queue_draw(TmpView->DrawingArea);
  }
}

/*****************************************************************************
*  TRSSelectionDialog
******************************************************************************/
void TRSSelectionDialog(GtkWidget *DrawingArea, 
                        gchar *Title, 
                        GtkSignalFunc CallBack)
{
  GtkWidget *Dialog;
  GtkWidget *Button;
  GtkWidget *Table;
  GtkWidget *Label;
  GtkWidget *Entry;
  GtkWidget **Data;

  Dialog = gtk_dialog_new();
  Data = g_malloc(4*sizeof(GtkWidget*));
  gtk_object_set_data(GTK_OBJECT(Dialog), "WidgetData", Data);
  gtk_window_set_title(GTK_WINDOW(Dialog), Title);
  /* The "Close" Button */
  Button = gtk_button_new_with_label(_("Close"));
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(Dialog));
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->action_area), Button, TRUE, TRUE, 0);
  gtk_widget_show(Button);
  /* The "Ok" Button */
  Button = gtk_button_new_with_label(_("OK"));
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            CallBack,
                            GTK_OBJECT(Dialog));
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(Dialog));
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->action_area), Button, TRUE, TRUE, 0);
  gtk_widget_grab_default(Button);
  gtk_widget_show(Button);
  
  Table = gtk_table_new(2, 3, FALSE);
  gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(Dialog)->vbox), 5);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(Dialog)->vbox), Table);
  gtk_widget_show(Table);
  /* X */
  Label = gtk_label_new(" x: ");
  gtk_table_attach(GTK_TABLE(Table), Label, 0,1, 0,1,
                   GTK_FILL | GTK_EXPAND,
                   GTK_FILL | GTK_EXPAND, 0, 0);
  gtk_widget_show(Label);
  
  Entry = gtk_entry_new();
  gtk_table_attach(GTK_TABLE(Table), Entry, 1,2, 0,1,
                   GTK_FILL | GTK_EXPAND,
                   GTK_EXPAND, 0, 0);
  gtk_widget_show(Entry);
  Data[0] = Entry;

  /* Y */
  Label = gtk_label_new(" y: ");
  gtk_table_attach(GTK_TABLE(Table), Label, 0,1, 1,2,
                   GTK_FILL | GTK_EXPAND,
                   GTK_FILL | GTK_EXPAND, 0, 0);
  gtk_widget_show(Label);
  
  Entry = gtk_entry_new();
  gtk_table_attach(GTK_TABLE(Table), Entry, 1,2, 1,2,
                   GTK_FILL | GTK_EXPAND,
                   GTK_EXPAND, 0, 0);
  gtk_widget_show(Entry);
  Data[1] = Entry;
  
  /* Z */
  Label = gtk_label_new(" z: ");
  gtk_table_attach(GTK_TABLE(Table), Label, 0,1, 2,3,
                   GTK_FILL | GTK_EXPAND,
                   GTK_FILL | GTK_EXPAND, 0, 0);
  gtk_widget_show(Label);
  
  Entry = gtk_entry_new();
  gtk_table_attach(GTK_TABLE(Table), Entry, 1,2, 2,3,
                   GTK_FILL | GTK_EXPAND,
                   GTK_EXPAND, 0, 0);
  gtk_widget_show(Entry);
  Data[2] = Entry;
  
  Data[3] = DrawingArea;
  
  gtk_widget_show(Dialog);
}

/*****************************************************************************
*  TranslateSelectionDialog
******************************************************************************/
void TranslateSelectionDialog(GtkWidget *DrawingArea)
{
  TRSSelectionDialog(DrawingArea, _("Translate Selection"), TranslateSelection);
}

/*****************************************************************************
*  RotateSelectionDialog
******************************************************************************/
void RotateSelectionDialog(GtkWidget *DrawingArea)
{
  TRSSelectionDialog(DrawingArea, _("Rotate Selection"), RotateSelection);
}

/*****************************************************************************
*  ScaleSelectionDialog
******************************************************************************/
void ScaleSelectionDialog(GtkWidget *DrawingArea)
{
  TRSSelectionDialog(DrawingArea, _("Scale Selection"), ScaleSelection);
}

/*****************************************************************************
*  ChangeColorDialogOK
******************************************************************************/
gint ChangeColorDialogOK(GtkWidget *Button, GtkWidget **Data)
{
  GtkWidget *DrawingArea;
  GtkWidget *ColorSel;
  ViewUserDataStruct *UserData;
  ObjectSelection *ObjectSel;

  DrawingArea = Data[0];
  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  ColorSel = Data[1];

  gtk_color_selection_get_color(GTK_COLOR_SELECTION(ColorSel),
               UserData->FrameView->Selection->Object->Texture->Pigment->Color);
  for (ObjectSel = UserData->FrameView->Selection ; 
       ObjectSel->Next != UserData->FrameView->Selection ; 
       ObjectSel = ObjectSel->Next)
  {
    gtk_color_selection_get_color(GTK_COLOR_SELECTION(ColorSel),
        ObjectSel->Object->Texture->Pigment->Color);
  }
  g_free(Data);
  return FALSE;
}

/*****************************************************************************
*  ChangeColorDialog
******************************************************************************/
void ChangeColorDialog(GtkWidget *DrawingArea)
{
  GtkWidget *Dialog;
  GtkWidget *Button;
  GtkWidget *ColorSel;
  GtkWidget **Data;
  ViewUserDataStruct *UserData;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  if (!(UserData->FrameView->Selection))
    return;
  
  Data = g_malloc(2*sizeof(GtkWidget*));
  Data[0] = DrawingArea;
  Dialog = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(Dialog), _("Change Color"));
    
  /* The Ok Button */
  Button = gtk_button_new_with_label(_("OK"));
  gtk_signal_connect(GTK_OBJECT(Button), "clicked",
                     (GtkSignalFunc)ChangeColorDialogOK,
                     (gpointer)Data);
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(Dialog));
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(Dialog)->action_area), Button);
  gtk_widget_grab_default(Button);
  gtk_widget_show(Button);
  
  ColorSel = gtk_color_selection_new();
  Data[1] = ColorSel;
  gtk_color_selection_set_color(GTK_COLOR_SELECTION(ColorSel),
              UserData->FrameView->Selection->Object->Texture->Pigment->Color);
  
  gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(Dialog)->vbox), ColorSel);
  gtk_widget_show(ColorSel);

  gtk_widget_show(Dialog);
}

/*****************************************************************************
*  DrawSelection
******************************************************************************/
void DrawSelection(GtkWidget *DrawingArea, TriangleListStruct *Tri,
                   GdkGC *gc, double Xoff, double Yoff, double Zoom)
{
  ViewUserDataStruct *UserData;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      DrawSelectionXY(DrawingArea, Tri, gc, Xoff, Yoff, Zoom);
      break;
    case ORTHO_XZ_CAMERA:
      DrawSelectionXZ(DrawingArea, Tri, gc, Xoff, Yoff, Zoom);
      break;
    case ORTHO_ZY_CAMERA:
      DrawSelectionZY(DrawingArea, Tri, gc, Xoff, Yoff, Zoom);
      break;
  }
}

/*****************************************************************************
*  DrawUnSelected
******************************************************************************/
void DrawUnSelected(GtkWidget *DrawingArea)
{
  ViewUserDataStruct *UserData;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      DrawUnSelectedXY(DrawingArea);
      break;
    case ORTHO_XZ_CAMERA:
      DrawUnSelectedXZ(DrawingArea);
      break;
    case ORTHO_ZY_CAMERA:
      DrawUnSelectedZY(DrawingArea);
      break;
  }
}

/*****************************************************************************
*  event_dispatcher
******************************************************************************/
gint event_dispatcher(GtkWidget *DrawingArea, GdkEvent *event, gpointer Menu)
{
  switch (event->type)
  {
    case GDK_BUTTON_PRESS:
    {
      GdkEventButton *bevent = (GdkEventButton *) event;
      if (bevent->button == 3) /* Right Button ==> Popup Menu */
        gtk_menu_popup(GTK_MENU(Menu), NULL, NULL, NULL, NULL,
                                              bevent->button, bevent->time);
      else if (bevent->button == 1)
      { /* Left Button : this is the big game */
        switch (ModeTool)
        {
          case MT_SELECT:       /* The Select Tool */
            ToggleSelect(DrawingArea, bevent);
            break;
          case MT_MOVE:         /* The Move Tool */
            ToolInitMoveSelection(DrawingArea, bevent);
            break;
          case MT_ROTATE:       /* The Rotate Tool */
            ToolInitRotateSelection(DrawingArea, bevent);
            break;
          case MT_SCALE:        /* The Scale Tool */
            ToolInitScaleSelection(DrawingArea, bevent);
            break;
          case MT_NEW_BICUBIC_PATCH:
            ToolNewBicubicPatch(DrawingArea, bevent);
            break;
          case MT_NEW_BLOB:
            ToolNewBlob(DrawingArea, bevent);
            break;
          case MT_NEW_BOX:      /* The New Box Tool */
            ToolInitNewBox(DrawingArea, bevent);
            break;
          case MT_NEW_CONE:     /* The New Cone Tool */
            ToolInitNewCone(DrawingArea, bevent);
            break;
/* FIXME          case MT_NEW_CSG:
            ToolNewCSG(DrawingArea, bevent);
            break;*/
          case MT_NEW_CUBIC:
            ToolNewCubic(DrawingArea, bevent);
            break;
          case MT_NEW_CYLINDER: /* The New Cylinder Tool */
            ToolInitNewCylinder(DrawingArea, bevent);
            break;
          case MT_NEW_DISC:
            ToolInitNewDisc(DrawingArea, bevent);
            break;
          case MT_NEW_HEIGHT_FIELD:
            ToolNewHeightField(DrawingArea, bevent);
            break;
          case MT_NEW_JULIA_FRACTAL:
            ToolNewJuliaFractal(DrawingArea, bevent);
            break;
          case MT_NEW_LATHE:
            ToolNewLathe(DrawingArea, bevent);
            break;
          case MT_NEW_MESH:
            ToolNewMesh(DrawingArea, bevent);
            break;
          case MT_NEW_PLANE:    /* The New Plane Tool */
            ToolNewPlane(DrawingArea, bevent);
            break;
          case MT_NEW_POLY:
            ToolNewPoly(DrawingArea, bevent);
            break;
          case MT_NEW_POLYGON:
            ToolNewPolygon(DrawingArea, bevent);
            break;
          case MT_NEW_PRISM:
            ToolNewPrism(DrawingArea, bevent);
            break;
          case MT_NEW_QUADRIC:  /* The New Quadric Tool */
            ToolNewQuadric(DrawingArea, bevent);
            break;
          case MT_NEW_QUARTIC:
            ToolNewQuartic(DrawingArea, bevent);
            break;
          case MT_NEW_SOR:
            ToolNewSor(DrawingArea, bevent);
            break;
          case MT_NEW_SPHERE:   /* The New Sphere Tool */
            ToolInitNewSphere(DrawingArea, bevent);
            break;
          case MT_NEW_SUPERELLIPSOID: /* The New SuperEllipsoid Tool */
            ToolNewSuperEllipsoid(DrawingArea, bevent);
            break;
          case MT_NEW_TEXT:
            ToolNewText(DrawingArea, bevent);
            break;
          case MT_NEW_TORUS:    /* The New Torus Tool */
            ToolInitNewTorus(DrawingArea, bevent);
            break;
          case MT_NEW_TRIANGLE:
            ToolNewTriangle(DrawingArea, bevent);
            break;
          case MT_NEW_LIGHTSOURCE:
            ToolNewLightSource(DrawingArea, bevent);
            break;
        }   
      }
      return TRUE;
    }
    case GDK_MOTION_NOTIFY:
    {
      switch (ModeTool)
      {
        case MT_MOVE:
          ToolMotionMoveSelection(DrawingArea, event);
          break;
        case MT_ROTATE:
          ToolMotionRotateSelection(DrawingArea, event);
          break;
        case MT_SCALE:
          ToolMotionScaleSelection(DrawingArea, event);
          break;
        case MT_NEW_BOX:
          ToolMotionNewBox(DrawingArea, event);
          break;
        case MT_NEW_CONE:
          ToolMotionNewCone(DrawingArea, event);
          break;
        case MT_NEW_CYLINDER:
          ToolMotionNewCylinder(DrawingArea, event);
          break;
        case MT_NEW_DISC:
          ToolMotionNewDisc(DrawingArea, event);
          break;
        case MT_NEW_SPHERE:
          ToolMotionNewSphere(DrawingArea, event);
          break;
        case MT_NEW_TORUS:
          ToolMotionNewTorus(DrawingArea, event);
          break;
      }
      return FALSE;
    }
    case GDK_BUTTON_RELEASE:
    {
      switch (ModeTool)
      {
        case MT_MOVE:
          ToolReleaseMoveSelection(DrawingArea, event);
          break;
        case MT_ROTATE:
          ToolReleaseRotateSelection(DrawingArea, event);
          break;
        case MT_SCALE:
          ToolReleaseScaleSelection(DrawingArea, event);
          break;
        case MT_NEW_BOX:
          ToolReleaseNewBox(DrawingArea, event);
          break;
        case MT_NEW_CONE:
          ToolReleaseNewCone(DrawingArea, event);
          break;
        case MT_NEW_CYLINDER:
          ToolReleaseNewCylinder(DrawingArea, event);
          break;
        case MT_NEW_DISC:
          ToolReleaseNewDisc(DrawingArea, event);
          break;
        case MT_NEW_SPHERE:
          ToolReleaseNewSphere(DrawingArea, event);
          break;
        case MT_NEW_TORUS:
          ToolReleaseNewTorus(DrawingArea, event);
          break;
      }
      return TRUE;
    }
    case GDK_KEY_PRESS:
      printf("a key pressed event has been catched...");
      break;
    default:
      return FALSE;
  }
  return FALSE;
}

/*****************************************************************************
*  HRulerConfigure
******************************************************************************/
void HRulerConfigure(GtkWidget *DrawingArea, gpointer data, GtkWidget *HRuler)
{
  ViewUserDataStruct *UserData;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  gtk_ruler_set_range(GTK_RULER(HRuler),
                      UserData->Xoffset-DrawingArea->allocation.width / 2.0 / UserData->ZoomFactor,
                      UserData->Xoffset+DrawingArea->allocation.width / 2.0 / UserData->ZoomFactor,
                      0, 50);
}

/*****************************************************************************
*  VRulerConfigure
******************************************************************************/
void VRulerConfigure(GtkWidget *DrawingArea, gpointer data, GtkWidget *VRuler)
{
  ViewUserDataStruct *UserData;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  gtk_ruler_set_range(GTK_RULER(VRuler),
                      UserData->Yoffset-DrawingArea->allocation.height / 2.0 / UserData->ZoomFactor,
                      UserData->Yoffset+DrawingArea->allocation.height / 2.0 / UserData->ZoomFactor,
                      0, 50);
}

#ifdef HAVE_GL
/*****************************************************************************
*  XVisualInfo: borrowed from gtkglarea.
******************************************************************************/
static XVisualInfo *get_xvisualinfo(GdkVisual *visual)
{
  Display *dpy;
  XVisualInfo vinfo_template;
  XVisualInfo *vi;
  int nitems_return;

  dpy = GDK_DISPLAY();

  /* TODO: is this right way to get VisualInfo from Visual ?? */
  /* AFAIK VisualID and depth should be enough to uniquely identify visual */
  vinfo_template.visual   = GDK_VISUAL_XVISUAL(visual);
  vinfo_template.visualid = XVisualIDFromVisual(vinfo_template.visual);
  vinfo_template.depth    = visual->depth;
  vi = XGetVisualInfo(dpy, VisualIDMask|VisualDepthMask, &vinfo_template, &nitems_return);
  g_assert( vi!=0  && nitems_return==1 ); /* visualinfo needs to be unique */

  /* remember to XFree returned XVisualInfo !!! */
  return vi;
}
#endif

/*****************************************************************************
*  GiramDeleteView
******************************************************************************/
gint GiramDeleteView(GtkWidget *window, GdkEvent *ev, FrameViewStruct *FrameView)
{
  FrameViewStruct *TmpFrameView;
  ViewUserDataStruct *UserData;

  printf("The deletion of an XYZ view has been catched!\n");
  g_free(FrameView->Name);
  if (FrameView == FrameView->Frame->FirstView)
  {
    FrameView->Frame->FirstView = FrameView->Next;
  } else
  {
    for (TmpFrameView = FrameView->Frame->FirstView ;
         TmpFrameView->Next != FrameView ;
         TmpFrameView = TmpFrameView->Next)
      ;
    TmpFrameView->Next = TmpFrameView->Next->Next;
  }
  g_free(FrameView);
  UserData = gtk_object_get_data(GTK_OBJECT(window), "ViewData");
#ifdef HAVE_GL
  glXMakeCurrent(GDK_DISPLAY(), None, NULL);
  glXDestroyContext(GDK_DISPLAY(), UserData->glxcontext);
#endif
  g_free(UserData);
  return FALSE;
}

/*****************************************************************************
*  NewView
******************************************************************************/
void NewView(FrameStruct *LocalFrame, int CameraStyle)
{
  GtkWidget *window;
  GtkWidget *table;
  GtkWidget *ruler;
  /*  GtkWidget *Arrow; */
  GtkWidget *DrawingArea;
  ViewUserDataStruct *UserData;
  GtkWidget *menu;
  GtkWidget *menu_items;
  GtkWidget *SubMenu;
  GtkWidget *FileMenu;
  GtkWidget *EditMenu;
  GtkWidget *RenderMenu;
  GtkWidget *DialogMenu;
  GtkWidget *ViewMenu;
  GtkWidget *SelectionMenu;
  GtkWidget *OperationMenu;
  GtkWidget *StatusBar;
  FrameViewStruct *View;
#ifdef HAVE_GL
  XVisualInfo *vi;
#endif

  if (LocalFrame->FirstView)
  {
    for (View = LocalFrame->FirstView ; View->Next ; View = View->Next)
          ;
    View->Next = g_malloc(sizeof(FrameViewStruct));
    View = View->Next;
  } else
  {
    View = LocalFrame->FirstView = g_malloc(sizeof(FrameViewStruct));
  }
  View->Next = NULL;
  View->Name = g_malloc(strlen(LocalFrame->Name)+7);
  switch(CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      strcpy(View->Name, "(X-Y) ");
      break;
    case ORTHO_XZ_CAMERA:
      strcpy(View->Name, "(X-Z) ");
      break;
    case ORTHO_ZY_CAMERA:
      strcpy(View->Name, "(Z-Y) ");
      break;
  }
  strcat(View->Name, LocalFrame->Name);
  View->Frame = LocalFrame;
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

  gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, FALSE);
  
  gtk_window_set_title(GTK_WINDOW(window), View->Name);
  gtk_widget_set_usize(window, 512, 384); /* XXX: should be setable */
  gtk_widget_set_events(window, GDK_POINTER_MOTION_MASK
                              | GDK_POINTER_MOTION_HINT_MASK
                              | GDK_KEY_PRESS_MASK);
  gtk_container_set_border_width(GTK_CONTAINER(window), 0);
  gtk_signal_connect(GTK_OBJECT(window), "delete_event",
                     GTK_SIGNAL_FUNC(GiramDeleteView), View);
        
  /* The Table */
  table = gtk_table_new(4, 5, FALSE);
  gtk_container_add(GTK_CONTAINER(window), table);
  gtk_widget_show(table);

  UserData = g_malloc(sizeof(ViewUserDataStruct));
#ifdef HAVE_GL
  if (ReallyHaveGL)
  {
    vi = get_xvisualinfo(GLVisual);  
    UserData->glxcontext = glXCreateContext(GDK_DISPLAY(), vi, 0, True);
    XFree(vi);
/*    gtk_widget_push_colormap(gdk_colormap_new(GLVisual,TRUE));
    gtk_widget_push_visual(GLVisual);
    This should be done so the drawing area is created with a suitable
    visual and colormap for OpenGL rendering, but this cause Giram to
    segfault on the second close of an XYZ window. Don't know why...
 */
  }
#endif
  DrawingArea = gtk_drawing_area_new();

#ifdef HAVE_GL
  if (ReallyHaveGL)
  {
/*    gtk_widget_pop_visual();
    gtk_widget_pop_colormap();*/
  }
#endif
  View->DrawingArea = DrawingArea;
  /* The Horizontal Ruler */
  ruler = gtk_hruler_new();
  gtk_signal_connect_after(GTK_OBJECT(DrawingArea), "configure_event",
                           (GtkSignalFunc)HRulerConfigure, ruler);
  gtk_signal_connect_object(GTK_OBJECT(window), "motion_notify_event",
                            GTK_SIGNAL_FUNC(GTK_WIDGET_CLASS(GTK_OBJECT(ruler)->klass)->motion_notify_event),
                            GTK_OBJECT(ruler));
  gtk_table_attach(GTK_TABLE (table), ruler, 2,3, 0,1,
                   GTK_FILL | GTK_EXPAND,
                   GTK_FILL, 0, 0);
  gtk_widget_show(ruler);
  /* The Vertical Ruler */
  ruler = gtk_vruler_new();
  gtk_signal_connect_after(GTK_OBJECT(DrawingArea), "configure_event",
                           (GtkSignalFunc)VRulerConfigure, ruler);
  gtk_signal_connect_object(GTK_OBJECT(window), "motion_notify_event",
                            GTK_SIGNAL_FUNC(GTK_WIDGET_CLASS(GTK_OBJECT(ruler)->klass)->motion_notify_event),
                            GTK_OBJECT(ruler));
  gtk_table_attach(GTK_TABLE(table), ruler, 0,1, 2,3,
                   GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_widget_show(ruler);
#if 0
  /* The Arrows */
  Arrow = gtk_arrow_new(GTK_ARROW_LEFT, GTK_SHADOW_OUT);  /* Left Arrow */
  gtk_table_attach(GTK_TABLE(table), Arrow, 1,2, 0,1,
                   GTK_FILL, GTK_FILL, 0, 0);
  gtk_widget_show(Arrow);
  Arrow = gtk_arrow_new(GTK_ARROW_RIGHT, GTK_SHADOW_OUT); /* Right Arrow */
  gtk_table_attach(GTK_TABLE(table), Arrow, 3,4, 0,1,
                   GTK_FILL, GTK_FILL, 0, 0);
  gtk_widget_show(Arrow);
  Arrow = gtk_arrow_new(GTK_ARROW_UP, GTK_SHADOW_OUT);    /* Up Arrow */
  gtk_table_attach(GTK_TABLE(table), Arrow, 0,1, 1,2,
                   GTK_FILL, GTK_FILL, 0, 0);
  gtk_widget_show(Arrow);
  Arrow = gtk_arrow_new(GTK_ARROW_DOWN, GTK_SHADOW_OUT);  /* Down Arrow */
  gtk_table_attach(GTK_TABLE(table), Arrow, 0,1, 3,4,
                   GTK_FILL, GTK_FILL, 0, 0);
  gtk_widget_show(Arrow);
#endif
  /* The Drawing Area */
  gtk_table_attach(GTK_TABLE(table), DrawingArea, 1, 4, 1, 4,
                   GTK_FILL,
                   GTK_FILL, 0,0);
  switch (CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      UserData->RenderStyle = 
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewXY, NULL);
      break;
    case ORTHO_XZ_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewXZ, NULL);
      break;
    case ORTHO_ZY_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewZY, NULL);
      break;
  }
  gtk_widget_set_events(DrawingArea, GDK_EXPOSURE_MASK
                                   | GDK_LEAVE_NOTIFY_MASK
                                   | GDK_BUTTON_PRESS_MASK
                                   | GDK_BUTTON_RELEASE_MASK
                                   | GDK_POINTER_MOTION_MASK
                                   | GDK_POINTER_MOTION_HINT_MASK
                                   | GDK_KEY_PRESS_MASK);
  UserData->CameraStyle = CameraStyle;
  UserData->FrameView = LocalFrame;
  UserData->ZoomFactor = 50.0;
  UserData->Xoffset = 0.0;
  UserData->Yoffset = 0.0;
  gtk_object_set_data(GTK_OBJECT(DrawingArea), "ViewData", UserData);
  gtk_object_set_data(GTK_OBJECT(window), "ViewData", UserData);
  /* The Status Bar */
  StatusBar = gtk_statusbar_new();
  gtk_table_attach(GTK_TABLE(table), StatusBar, 0,4, 4,5,
                   GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0);
  gtk_widget_show(StatusBar);
  UserData->StatusBar = StatusBar;
  /* Create the Menu */
  menu = gtk_menu_new();
  /* The "File" SubMenu */
  SubMenu = gtk_menu_new();
  /* The Load Pov Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Load Pov"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(LoadPovFunc), NULL);
  gtk_widget_show(menu_items);
  /* The Load AutoCAD DXF Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Load AutoCAD DXF"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(LoadDXFFunc), NULL);
  gtk_widget_show(menu_items);
  /* Separator */
  menu_items = gtk_menu_item_new();
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_widget_show(menu_items);
  /* The Save Pov Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Save Pov"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect(GTK_OBJECT(menu_items), "activate",
                     GTK_SIGNAL_FUNC(SavePovFunc), LocalFrame);
  gtk_widget_show(menu_items);
  /* The Save AutoCAD DXF Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Save AutoCAD DXF"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect(GTK_OBJECT(menu_items), "activate",
                     GTK_SIGNAL_FUNC(SaveDXFFunc), LocalFrame);
  gtk_widget_show(menu_items);
#ifdef WITH_S3D
  /* The Save Bishop S3D Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Save Bishop S3D"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect(GTK_OBJECT(menu_items), "activate",
                     GTK_SIGNAL_FUNC(SaveS3DFunc), LocalFrame);
  gtk_widget_show(menu_items);
#endif
  /* Separator */
  menu_items = gtk_menu_item_new();
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_widget_show(menu_items);
  /* The Quit Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Quit"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(ExitFunc), NULL);
  gtk_widget_show(menu_items);
  /* */
  FileMenu = gtk_menu_item_new_with_label(_("Files"));
  gtk_widget_show(FileMenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(FileMenu), SubMenu);
  gtk_menu_append(GTK_MENU(menu), FileMenu);
  /* The "Edit" SubMenu */
  SubMenu = gtk_menu_new();
  /* The Cut Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Cut"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(CutSelection), 
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The Copy Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Copy"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(CopySelection),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The Paste Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Paste"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(PasteSelection),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The Snap Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Snap"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(SnapDialogShow),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* */
  EditMenu = gtk_menu_item_new_with_label(_("Edit"));
  gtk_widget_show(EditMenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(EditMenu), SubMenu);
  gtk_menu_append(GTK_MENU(menu), EditMenu);
  /* The "View" SubMenu */
  SubMenu = gtk_menu_new();
  /* The Zoom In Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Zoom in"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(ZoomIn),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The Zoom Out Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Zoom out"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(ZoomOut),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The Fit To Selection Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Fit To Selection"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(FitToSelectionFunc),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The Fit To Scene Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Fit To Scene"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(FitToSceneFunc),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* */
  ViewMenu = gtk_menu_item_new_with_label(_("View"));
  gtk_widget_show(ViewMenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(ViewMenu), SubMenu);
  gtk_menu_append(GTK_MENU(menu), ViewMenu);
  /* The "Selection" SubMenu */
  SubMenu = gtk_menu_new();
  /* The "Select All" Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Select All"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(SelectAll),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The "Select None" Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Select None"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(SelectNone),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The "Invert Selection" Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Invert Selection"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(InvertSelection),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The Export to DXF Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Export to DXF"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect(GTK_OBJECT(menu_items), "activate",
                     GTK_SIGNAL_FUNC(SaveSelectedDXFFunc), LocalFrame);
  gtk_widget_show(menu_items);
#ifdef WITH_S3D
  /* The Export to S3D Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Export to S3D"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect(GTK_OBJECT(menu_items), "activate",
                     GTK_SIGNAL_FUNC(SaveSelectedS3DFunc), LocalFrame);
  gtk_widget_show(menu_items);
#endif
  /* The Import DXF Objects Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Import DXF Objects"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect(GTK_OBJECT(menu_items), "activate",
                     GTK_SIGNAL_FUNC(InsertDXFFunc), LocalFrame);
  gtk_widget_show(menu_items);
  /* */
  SelectionMenu = gtk_menu_item_new_with_label(_("Selection"));
  gtk_widget_show(SelectionMenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(SelectionMenu), SubMenu);
  gtk_menu_append(GTK_MENU(menu), SelectionMenu);
  /* The "Operations" SubMenu */
  SubMenu = gtk_menu_new();
  /* The "Translate" Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Translate"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(TranslateSelectionDialog),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The "Rotate" Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Rotate"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(RotateSelectionDialog),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The "Scale" Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Scale"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(ScaleSelectionDialog),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The "Change Color" Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Change Color"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(ChangeColorDialog),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The "Edit Camera" Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Edit Camera"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(EditCamera),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* */
  OperationMenu = gtk_menu_item_new_with_label(_("Operations"));
  gtk_widget_show(OperationMenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(OperationMenu), SubMenu);
  gtk_menu_append(GTK_MENU(menu), OperationMenu);
  /* The "Render" SubMenu */
  /* Create the SubMenu */
  SubMenu = gtk_menu_new();
  /* The Wire Frame Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Wire Frame"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(SetWireFrameExpose),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The Hidden Faces Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Hidden Faces"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(SetHiddenFacesExpose),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The Gouraud Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Gouraud"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(SetGouraudExpose),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
#ifdef HAVE_GL
  if (ReallyHaveGL)
  {
    /* The OpenGL Menu Item */
    menu_items = gtk_menu_item_new_with_label(_("OpenGL WireFrame"));
    gtk_menu_append(GTK_MENU(SubMenu), menu_items);
    gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                              GTK_SIGNAL_FUNC(SetOpenGLExpose),
                              GTK_OBJECT(DrawingArea));
    gtk_widget_show(menu_items);
  }
#endif
  /* The Povfront MenuItem */
  menu_items = gtk_menu_item_new_with_label(_("POVRay Rendering"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(CallPovFront),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* */
  RenderMenu = gtk_menu_item_new_with_label(_("Render"));
  gtk_widget_show(RenderMenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(RenderMenu), SubMenu);
  gtk_menu_append(GTK_MENU(menu), RenderMenu);
  /* The "Dialogs" SubMenu */
  /* Create the SubMenu */
  SubMenu = gtk_menu_new();
  /* The CSG Tree Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("CSG Tree"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(ShowCSGTreeGTKWrapper),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The Tool Options Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Tool Options"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(ShowToolOptionsDialog),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* */
  DialogMenu = gtk_menu_item_new_with_label(_("Dialogs"));
  gtk_widget_show(DialogMenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(DialogMenu), SubMenu);
  gtk_menu_append(GTK_MENU(menu), DialogMenu);
  /* The "Plugins" SubMenu */
  gtk_menu_append(GTK_MENU(menu), CreatePlugInsMenu(DrawingArea));
  /*  */
  gtk_signal_connect(GTK_OBJECT(DrawingArea), "event",
                     GTK_SIGNAL_FUNC(event_dispatcher), GTK_OBJECT(menu));
  gtk_widget_show(DrawingArea);
  gtk_widget_show(window);
}

