// wm.cc
//
// Frank Hale 
// frankhale@yahoo.com
#include "wm.hh"

WindowManager *wm;

void signalhandler(int i) {
  static int re_enter = 0;
  
  fprintf(stderr, "%s: toplevel:\n\t[ signal %d caught ]\n", __FILE__, i);
  if (! re_enter) {
    re_enter = 1;  
    fprintf(stderr, "\t[ shutting down ]\n");
    wm->quit_nicely();
  }

  fprintf(stderr, "\t[ exiting ]\n");
  abort();
}

int handle_xerror(Display *dpy, XErrorEvent *e)
{
	if (e->error_code == BadAccess && e->resourceid == wm->root) {
		err("root window unavailible (maybe another wm is running?)");
		exit(1);
	}
#ifdef DEBUG
	else {
		char msg[80];
		XGetErrorText(dpy, e->error_code, msg, sizeof(msg));
		err("X error (%#x): %s", e->resourceid, msg);
	}
#endif
	return 0;
}

WindowManager::WindowManager(int argc, char **argv) 
{
	opt_display = "";
	opt_font = DEF_FONT;
	opt_fg = DEF_FG;
	opt_bg = DEF_BG;
	opt_bd = DEF_BD;
	opt_bg2 = DEF_BG2;

	::wm = this;

	head_client = new Client();

	ClientList = new LinkedList<Client>;

	int i;

	for (i = 1; i < argc; i++) {
		if (streq(argv[i], "-fn") && i+1<argc)
			opt_font = argv[++i];
		else if (streq(argv[i], "-display") && i+1<argc)
			opt_display = argv[++i];
		else if (streq(argv[i], "-fg") && i+1<argc)
			opt_fg = argv[++i];
		else if (streq(argv[i], "-bg") && i+1<argc)
			opt_bg = argv[++i];
		else if (streq(argv[i], "-bd") && i+1<argc)
			opt_bd = argv[++i];
		else if (streq(argv[i], "-version") && i<argc) {
			printf("version %s\nby: Frank Hale\nemail: frankhale@yahoo.com\n", windowmanager_version);
			exit(0);
		}
		else {
			printf("usage: %s [-display <display>] [-fn <font>]\n", windowmanager_name);
			printf("  [-fg foreground] [-bg background] [-bd border] [-version]\n");
			exit(2);
		}
	}

	theme = new Theme();
	
	setup_display();
	scan_windows();
	do_event_loop();
	exit(1);
}

WindowManager::~WindowManager() 
{
	delete theme;
	
	LinkedListIterator<Menu> sm(subs);
				
	for(; sm.current(); sm++) 
	{
		subs->remove(sm.current());
	}
	delete subs;
	delete t;
	delete rm;
	delete head_client;
	delete ClientList;
}

void WindowManager::setup_display()
{
	XGCValues gv;
	XSetWindowAttributes attr;
	XColor dummy;
#ifdef SHAPE
	int dummy2;
#endif

	XSetErrorHandler(handle_xerror);
	signal(SIGTERM, (void (*)(int)) signalhandler);
	signal(SIGINT,  (void (*)(int)) signalhandler);
	signal(SIGHUP,  (void (*)(int)) signalhandler);

	dpy = XOpenDisplay(opt_display);

	if (!dpy) { 
		err("can't open display %s", opt_display);
		exit(1);
	}

	screen = DefaultScreen(dpy);
	root = RootWindow(dpy, screen);
	
	XDefineCursor(dpy, root, XCreateFontCursor(wm->dpy, XC_left_ptr));
	
	xa_wm_state = XInternAtom(dpy, "WM_STATE", False);
	xa_wm_change_state = XInternAtom(dpy, "WM_CHANGE_STATE", False);
	xa_wm_protos = XInternAtom(dpy, "WM_PROTOCOLS", False);
	xa_wm_delete = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
	xa_wm_cmapwins = XInternAtom(dpy, "WM_COLORMAP_WINDOWS", False);

	XAllocNamedColor(dpy, DefaultColormap(dpy, screen), opt_fg, &fg, &dummy);
	XAllocNamedColor(dpy, DefaultColormap(dpy, screen), opt_bg, &bg, &dummy);
	XAllocNamedColor(dpy, DefaultColormap(dpy, screen), opt_bd, &bd, &dummy);
	XAllocNamedColor(dpy, DefaultColormap(dpy, screen), opt_bg2, &bg2, &dummy);
	
	font = XLoadQueryFont(dpy, opt_font);
	if (!font) font = XLoadQueryFont(dpy, DEF_FONT);
	iheight = font->ascent + font->descent + 2*SPACE;

#ifdef SHAPE
	shape = XShapeQueryExtension(dpy, &shape_event, &dummy2);
#endif
	move_curs = XCreateFontCursor(dpy, XC_fleur);
	resize_curs = XCreateFontCursor(dpy, XC_plus);
	menu_curs = XCreateFontCursor(dpy, XC_sb_left_arrow);

	gv.function = GXcopy;
	gv.foreground = fg.pixel;
	gv.font = font->fid;
	string_gc = XCreateGC(dpy, root, GCFunction | GCForeground | GCFont, &gv);

	gv.foreground = bd.pixel;
	gv.line_width = BORDER;
	border_gc = XCreateGC(dpy, root, GCFunction | GCForeground | GCLineWidth, &gv);

	gv.function = GXinvert;
	gv.subwindow_mode = IncludeInferiors;
	invert_gc = XCreateGC(dpy, root, GCFunction | GCSubwindowMode | GCLineWidth, &gv);

	attr.event_mask = ChildMask | ColormapChangeMask | ButtonMask | PropertyChangeMask;
	XChangeWindowAttributes(dpy, root, CWEventMask, &attr);
	
	xres = WidthOfScreen(ScreenOfDisplay(wm->dpy, wm->screen));
	yres = HeightOfScreen(ScreenOfDisplay(wm->dpy, wm->screen));
		
	char themePath[strlen("default")+strlen(HOME)+11];
	sprintf(themePath, "%s/.sapphire/%s", HOME, "default");
	
	char WMPath[strlen("wmconfig")+strlen(HOME)+11];
	sprintf(WMPath, "%s/.sapphire/%s", HOME, "wmconfig");	
	
	WMLexer *wml = new WMLexer(WMPath);
	wml->analyze();
	
	if(! wml->isError()) {
		
		ThemeLexer *l = new ThemeLexer(alternateThemePath);
		l->analyze();
		delete l;
		
	} else {
		
		printf("using default theme...\n");
		ThemeLexer *l = new ThemeLexer(themePath);
		l->analyze();
		delete l;	
	
		WMConfig *wmc = new WMConfig(WMPath);
		char startUpTheme[256];
		sprintf(startUpTheme, "startUpTheme { \"%s\" }", themePath);

		wmc->write("# This file is edited by Sapphire #");
		wmc->write("# Please don't edit #");
		wmc->write(startUpTheme);
	
		delete wmc;	
	}
	delete wml;
	
	t = new Toolbar();
	rm = new RootMenu();
	rm->insertMenuItems();
	
	subs = new LinkedList<Menu>; 
	subs = rm->getSubMenus();
}

void WindowManager::scan_windows()
{
	unsigned int i, nwins;
	Window dw1, dw2, *wins;
	XWindowAttributes attr;

	XQueryTree(dpy, root, &dw1, &dw2, &wins, &nwins);
	for (i = 0; i < nwins; i++) {
		XGetWindowAttributes(dpy, wins[i], &attr);
		if (!attr.override_redirect && attr.map_state == IsViewable)
			head_client->make_new_client(wins[i]);
	}
	XFree(wins);

}

void WindowManager::do_event_loop()
{
	XEvent ev;

	int xfd = ConnectionNumber(wm->dpy);

	for (;;) {

		if(XPending(wm->dpy))
		{
			XNextEvent(wm->dpy, &ev);

			#ifdef DEBUB
			show_event(ev);
			#endif

			switch (ev.type) {
				case ButtonPress:
				{
					head_client->handle_button_press_event(&ev.xbutton); 
					t->do_button_press_event(&ev.xbutton);
				
					rm->handle_button_event(&ev.xbutton);
				
					LinkedListIterator<Menu> sm(subs);
				
					for(; sm.current(); sm++) 
					{
						sm.current()->handle_button_event(&ev.xbutton);
					}
				
				
					switch(ev.xbutton.button)
					{
				
						case 1: {
						button1pressed=True; break;
						}
					
						case 2: {
						button2pressed=True; break;
						}

						case 3: {
							button3pressed=True;
				
					 		if(ev.xbutton.window == root)
					  		{
								rm->show(True);
					  		} 
						break;	
						}
					}
				break;
				}
			
				case ButtonRelease:
				{
					head_client->handle_button_release_event(&ev.xbutton); 
					t->do_button_release_event(&ev.xbutton);
										
					switch(ev.xbutton.button)
					{
				
					case 1: {
					button1pressed=False; break;
					}
				
					case 2: {
					button2pressed=False; break;
					}

					case 3: {
					button3pressed=False; break;
					}
				
				}
				break;
				}
				
				case ConfigureRequest:
				{
					head_client->handle_configure_request(&ev.xconfigurerequest); 
				
				break;
				}
				
				case ConfigureNotify:
				{
					rm->handle_config_notify_event(&ev.xconfigure);
					LinkedListIterator<Menu> sm(subs);
					for(; sm.current(); sm++) 
					{
						sm.current()->handle_config_notify_event(&ev.xconfigure);
					}
				
				break;
				}
				
				case MapRequest:
				{
					head_client->handle_map_request(&ev.xmaprequest);
				
				break;
				}
			
				case UnmapNotify:
				{
					head_client->handle_unmap_event(&ev.xunmap); 
				break;
				}
			
				case ClientMessage:
				{
					head_client->handle_client_message(&ev.xclient); 
				break;
				}
			
				case ColormapNotify:
				{
					head_client->handle_colormap_change(&ev.xcolormap); 
				break;
				}
			
				case PropertyNotify:
				{
					head_client->handle_property_change(&ev.xproperty); 
				break;
				}

				case LeaveNotify:
				{
					head_client->handle_leave_event(&ev.xcrossing);
				
					rm->handle_leave_notify(&ev.xcrossing);
					
					LinkedListIterator<Menu> sm(subs);
					for(; sm.current(); sm++) 
					{
						sm.current()->handle_leave_notify(&ev.xcrossing);
					}

				break;
				}
			
				case EnterNotify:
				{
					head_client->handle_enter_event(&ev.xcrossing); 
					rm->handle_enter_notify(&ev.xcrossing);
				
					LinkedListIterator<Menu> sm(subs);

					for(; sm.current(); sm++) 
					{
						sm.current()->handle_enter_notify(&ev.xcrossing);
					}
				break;
				}

				case MotionNotify:
				{
					rm->handle_motion_notify_event(&ev.xmotion);
					LinkedListIterator<Menu> sm(subs);

					for(; sm.current(); sm++) 
					{
						sm.current()->handle_motion_notify_event(&ev.xmotion);
					}
				break;
				}
			
				case Expose:
				{
					head_client->handle_expose_event(&ev.xexpose);
				
					rm->handle_expose_event(&ev.xexpose); 
					t->handle_expose_event(&ev.xexpose); 

					LinkedListIterator<Menu> sm(subs);

					for(; sm.current(); sm++) 
					{
						sm.current()->handle_expose_event(&ev.xexpose);
					}
				
					break;
				}
			
				case ReparentNotify:
				{
					head_client->handle_reparent_event(&ev.xreparent); 
				
				break;
				}

				default: 
				{				
					#ifdef SHAPE
					if (wm->shape && ev.type == wm->shape_event) 
					head_client->handle_shape_change((XShapeEvent *)&ev);
					#endif
					 	 			
				break;		 
				}	
			
			
			}
		
		} else {
			
			// put a wait on the network file descriptor for the X connection...
			// this saves blackbox from eating all available cpu
			fd_set rfds;
			FD_ZERO(&rfds);
			FD_SET(xfd, &rfds);
      
			struct timeval tv;
			tv.tv_sec = 0;
			tv.tv_usec = 1000;
      
			select(xfd + 1, &rfds, 0, 0, &tv);
			t->checkClock();
		}

	}
	

}



void WindowManager::quit_nicely()
{
	XFreeFont(dpy, font);

	XSetInputFocus(dpy, PointerRoot, RevertToPointerRoot, CurrentTime);
	instcmap(DefaultColormap(dpy, screen));
	XCloseDisplay(dpy);
	
	printf("\nbye bye....\n");
	
	exit(0);
}

void WindowManager::doTheme(char* path)
{
	ThemeLexer *l = new ThemeLexer(path);
	l->analyze();
	delete l;
	
	t->reconfigure();
	rm->reconfigure();
		
	LinkedListIterator<Client> c(ClientList);
	for(; c.current(); c++) 
	{
		head_client->reconfigure(c.current());
	}
	
	LinkedListIterator<Menu> sm(subs);
	for(; sm.current(); sm++) 
	{
		sm.current()->reconfigure();
	}
	
	char WMPath[strlen("wmconfig")+strlen(HOME)+11];
	sprintf(WMPath, "%s/.sapphire/%s", HOME, "wmconfig");	
	
	WMConfig *wmc = new WMConfig(WMPath);
	char startUpTheme[256];
	sprintf(startUpTheme, "startUpTheme { \"%s\" }", path);

	wmc->write("# This file is edited by Sapphire #");
	wmc->write("# Please don't edit #");
	wmc->write(startUpTheme);
	
	delete wmc;	
}

void WindowManager::closeRootMenu() { rm->show(False); }
