/* LoadPov.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999 David Odin <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>  
#include "giram.h"
#include "utils.h"
#include "frame.h"
#include "bicubic_patch.h"
#include "blob.h"
#include "box.h"
#include "cone.h"
#include "cubic.h"
#include "cylinder.h"
#include "disc.h"
#include "height_field.h"
#include "julia_fractal.h"
#include "lathe.h"
#include "mesh.h"
#include "plane.h"
#include "poly.h"
#include "polygon.h"
#include "prism.h"
#include "quadric.h"
#include "quartic.h"
#include "sor.h"
#include "sphere.h"
#include "superellipsoid.h"
#include "text.h"
#include "torus.h"
#include "triangle.h"
#include "csg.h"
#include "object.h"
#include "camera.h"
#include "csgtree.h"
#include "view.h"
#include "view_camera.h"
#include "copy.h"
#include "LoadPov.h"

#define SMALL_VALUE    10e-10

#define LastWord       263
struct { int Token; char *String; } ReservedWordTable[LastWord] = {
  { Taa_level, "aa_level" },
  { Taa_threshold, "aa_threshold" },
  { Tabs, "abs" },
  { Tacos, "acos" },  
  { Tadaptive, "adaptive" },
  { Tadc_bailout, "adc_bailout" },
  { Tagate, "agate" },
  { Tagate_turb, "agate_turb" },
  { Talpha, "alpha" },
  { Tambient, "ambient" },
  { Tambient_light, "ambient_light" },
  { Tangle, "angle" },
  { Taperture, "aperture" },
  { Tarc_angle, "arc_angle" },
  { Tarea_light, "area_light" },
  { Tasc, "asc" },
  { Tasin, "asin" },
  { Tassumed_gamma, "assumed_gamma" },
  { Tatan2, "atan2" },
  { Tatmosphere, "atmosphere" },
  { Tatmospheric_attenuation, "atmospheric_attenuation" },
  { Tattenuating, "attenuating" },
  { Taverage, "average" },
  { Tbicubic_patch, "bicubic_patch" },
  { Tblob, "blob" },
  { Tblue, "blue" },
  { Tblur_samples, "blur_samples" },
  { Tbounded_by, "bounded_by" },
  { Tbox, "box" },
  { Tbox_mapping, "box_mapping" },
  { Tbozo, "bozo" },
  { Tbrick, "brick" },
  { Tbrick_size, "brick_size" },
  { Tbrightness, "brightness" },
  { Tbrilliance, "brilliance" },
  { Tbump_map, "bump_map" },
  { Tbump_size, "bump_size" },
  { Tbumps, "bumps" },
  { Tcamera, "camera" },
  { Tcaustics, "caustics" },
  { Tceil, "ceil" },
  { Tchecker, "checker" },
  { Tchr, "chr" },
  { Tclipped_by, "clipped_by" },
  { Tclock, "clock" },
  { Tcolor, "color" },
  { Tcolor, "colour" },
  { Tcolor_map, "color_map" },
  { Tcolor_map, "colour_map" },
  { Tconcat, "concat" },
  { Tcone, "cone" },
  { Tconfidence, "confidence" },
  { Tconstant, "constant" },
  { Tcontrol0, "control0" },
  { Tcontrol1, "control1" },
  { Tcos, "cos" },
  { Tcount, "count" },
  { Tcrackle, "crackle" },
  { Tcrand, "crand" },
  { Tcubic, "cubic" },
  { Tcylinder, "cylinder" },
  { Tcylindrical_mapping, "cylindrical_mapping" },
  { Tdegrees, "degrees" },
  { Tdents, "dents" },
  { Tdifference, "difference" },
  { Tdiffuse, "diffuse" },
  { Tdirection, "direction" },
  { Tdisc, "disc" },
  { Tdistance, "distance" },
  { Tdistance_maximum, "distance_maximum" },
  { Tdiv, "div" },
  { Tdust, "dust" },
  { Tdust_type, "dust_type" },
  { Teccentricity, "eccentricity" },
  { Temitting, "emitting" },
  { Terror_bound, "error_bound" },
  { Texp, "exp" },
  { Texponent, "exponent" },
  { Tfade_distance, "fade_distance" },
  { Tfade_power, "fade_power" },
  { Tfalloff, "falloff" },
  { Tfalloff_angle, "falloff_angle" },
  { Tfalse, "false" },
  { Tfile_exists, "file_exists" },
  { Tfilter, "filter" },
  { Tfinish, "finish" },
  { Tfisheye, "fisheye" },
  { Tfloor, "floor" },
  { Tfocal_point, "focal_point" },
  { Tfog, "fog" },
  { Tfog_alt, "fog_alt" },
  { Tfog_offset, "fog_offset" },
  { Tfog_type, "fog_type" },
  { Tfrequency, "frequency" },
  { Tglobal_settings, "global_settings" },
  { Tglowing, "glowing" },
  { Tgradient, "gradient" },
  { Tgranite, "granite" },
  { Tgray_threshold, "gray_threshold" },
  { Tgreen, "green" },
  { Thalo, "halo" },
  { Theight_field, "height_field" },
  { Thexagon, "hexagon" },
  { Thf_gray_16, "hf_gray_16" },
  { Thierarchy, "hierarchy" },
  { Thollow, "hollow" },
  { Timage_map, "image_map" },
  { Tint, "int" },
  { Tintersection, "intersection" },
  { Tinverse, "inverse" },
  { Tior, "ior" },
  { Tirid, "irid" },
  { Tirid_wavelength, "irid_wavelength" },
  { Tjitter, "jitter" },
  { Tjulia_fractal, "julia_fractal" },
  { Tlambda, "lambda" },
  { Tlathe, "lathe" },
  { Tleopard, "leopard" },
  { Tlight_source, "light_source" },
  { Tlinear, "linear" },
  { Tlocation, "location" },
  { Tlog, "log" },
  { Tlow_error_factor, "low_error_factor" },
  { Tlook_at, "look_at" },
  { Tlooks_like, "looks_like" },
  { Tmandel, "mandel" },
  { Tmarble, "marble" },
  { Tmatrix, "matrix" },
  { Tmax, "max" },
  { Tmax_intersections, "max_intersections" },
  { Tmax_trace_level, "max_trace_level" },
  { Tmax_value, "max_value" },
  { Tmerge, "merge" },
  { Tmesh, "mesh" },
  { Tmetallic, "metallic" },
  { Tmin, "min" },
  { Tminimum_reuse, "minimum_reuse" },
  { Tmod, "mod" },
  { Tmortar, "mortar" },
  { Tnearest_count, "nearest_count" },
  { Tno, "no" },
  { Tno_shadow, "no_shadow" },
  { Tnormal, "normal" },
  { Tnormal_map, "normal_map" },
  { Tnumber_of_waves, "number_of_waves" },
  { Tobject, "object" },
  { Toctaves, "octaves" },
  { Toff, "off" },
  { Tomega, "omega" },
  { Tomnimax, "omnimax" },
  { Ton, "on" },
  { Tonion, "onion" },
  { Topen, "open" },
  { Torthographic, "orthographic" },
  { Tpanoramic, "panoramic" },
  { Tperspective, "perspective" },
  { Tphase, "phase" },
  { Tphong, "phong" },
  { Tphong_size, "phong_size" },
  { Tpi, "pi" },
  { Tpigment, "pigment" },
  { Tpigment_map, "pigment_map" },
  { Tplanar_mapping, "planar_mapping" },
  { Tplane, "plane" },
  { Tpoint_at, "point_at" },
  { Tpoly, "poly" },
  { Tpolygon, "polygon" },
  { Tpow, "pow" },
  { Tprism, "prism" },
  { Tquadric, "quadric" },
  { Tquartic, "quartic" },
  { Tquick_color, "quick_color" },
  { Tquilted, "quilted" },
  { Tradial, "radial" },
  { Tradians, "radians" },
  { Tradiosity, "radiosity" },
  { Tradius, "radius" },
  { Trainbow, "rainbow" },
  { Tramp_wave, "ramp_wave" },
  { Trand, "rand" },
  { Trecursion_limit, "recursion_limit" },
  { Tred, "red" },
  { Treflection, "reflection" },
  { Trefraction, "refraction" },
  { Trgb, "rgb" },
  { Trgbf, "rgbf" },
  { Trgbft, "rgbft" },
  { Trgbt, "rgbt" },
  { Tright, "right" },
  { Tripples, "ripples" },
  { Trotate, "rotate" },
  { Troughness, "roughness" },
  { Tsamples, "samples" },
  { Tscale, "scale" },
  { Tscallop_wave, "scallop_wave" },
  { Tscattering, "scattering" },
  { Tseed, "seed" },
  { Tshadowless, "shadowless" },
  { Tsin, "sin" },
  { Tsine_wave, "sine_wave" },
  { Tsky, "sky" },
  { Tsky_sphere, "sky_sphere" },
  { Tslope_map, "slope_map" },
  { Tsmooth, "smooth" },
  { Tsmooth_triangle, "smooth_triangle" },
  { Tsor, "sor" },
  { Tspecular, "specular" },
  { Tsphere, "sphere" },
  { Tspherical_mapping, "spherical_mapping" },
  { Tspiral1, "spiral1" },
  { Tspiral2, "spiral2" },
  { Tspotlight, "spotlight" },
  { Tspotted, "spotted" },
  { Tsqrt, "sqrt" },
  { Tstr, "str" },
  { Tstrcmp, "strcmp" },
  { Tstrlen, "strlen" },
  { Tstrlwr, "strlwr" },
  { Tstrupr, "strupr" },
  { Tsturm, "sturm" },
  { Tsubstr, "substr" },
  { Tsuperellipsoid, "superellipsoid" },
  { Tt, "t" },
  { Ttan, "tan" },
  { Ttext, "text" },
  { Ttexture, "texture" },
  { Tthickness, "thickness" },
  { Ttightness, "tightness" },
  { Ttorus, "torus" },
  { Ttransform, "transform" },
  { Ttranslate, "translate" },
  { Ttransmit, "transmit" },
  { Ttriangle, "triangle" },
  { Ttriangle_wave, "triangle_wave" },
  { Ttrue, "true" },
  { Tttf, "ttf" },
  { Tturb_depth, "turb_depth" },
  { Tturbulence, "turbulence" },
  { Ttype, "type" },
  { Tu, "u" },
  { Tunion, "union" },
  { Tup, "up" },
  { Tultra_wide_angle, "ultra_wide_angle" },
  { Tv, "v" },
  { Tval, "val" },
  { Tvariance, "variance" },
  { Tvaxis_rotate, "vaxis_rotate" },
  { Tvcross, "vcross" },
  { Tvdot, "vdot" },
  { Tversion, "version" },
  { Tvlength, "vlength" },
  { Tvnormalize, "vnormalize" },
  { Tvrotate, "vrotate" },
  { Twarp, "warp" },
  { Twater_level, "water_level" },
  { Twaves, "waves" },
  { Twidth, "width" },
  { Twood, "wood" },
  { Twrinkles, "wrinkles" },
  { Tx, "x" },
  { Ty, "y" },
  { Tyes, "yes" },
  { Tz, "z" }
};

ListFile *PovFile;
int OldChar, IsOldChar;
int OldToken, IsOldToken;
double PovVersionLangage;
double LastFloat;
char *LastString = NULL;
FloatConstantStruct *LastFloatConstant;
VectorConstantStruct *LastVectorConstant;
ColorConstantStruct *LastColorConstant;
StringConstantStruct *LastStringConstant;
ObjectConstantStruct *LastObjectConstant;
TextureConstantStruct *LastTextureConstant;
PigmentConstantStruct *LastPigmentConstant;
NormalConstantStruct *LastNormalConstant;
FinishConstantStruct *LastFinishConstant;
HaloConstantStruct *LastHaloConstant;
ColorMapConstantStruct *LastColorMapConstant;
PigmentMapConstantStruct *LastPigmentMapConstant;
SlopeMapConstantStruct *LastSlopeMapConstant;
NormalMapConstantStruct *LastNormalMapConstant;
CameraConstantStruct *LastCameraConstant;
LightSourceConstantStruct *LastLightSourceConstant;
AtmosphereConstantStruct *LastAtmosphereConstant;
FogConstantStruct *LastFogConstant;
RainbowConstantStruct *LastRainbowConstant;
SkySphereConstantStruct *LastSkySphereConstant;
TransformConstantStruct *LastTransformConstant;
double PovClock = 0.0;
static unsigned int NbPovRandomTable;
static unsigned long int *PovRandomTable;
int ParsingDirective;
int DontParse;

typedef struct ParsedVector
{
  Vector Val;
  int nbTerms;
} ParsedVector;

#define IF_TRUE_IF_ELSE           1
#define IF_TRUE_ELSE_END          2
#define IF_FALSE_IF_ELSE          3
#define IF_FALSE_ELSE_END         4
#define WHILE                     5
#define SWITCH_SWITCH_CASE        6
#define SWITCH_CASE_BREAK_TRUE    7
#define SWITCH_CASE_BREAK_FALSE   8
#define SWITCH_BREAK_END          9
#define UNPARSED                 10

typedef struct CondDirective
{
  struct CondDirective *Next;
  int Style;
  double SwitchValue;
  int WhileLineNo;
  ListFile *WhileFile;
  long WhilePos;
} CondDirective;

CondDirective *ConditionalStack;
FrameStruct *CurrentFrame;


int NextToken(int *pToken);
int ParseFloat(double *pLocalFloat);
int ParseColor(Vector TmpVector);
int ParsePigment(PigmentStruct **pLocalPigment);
int ParseRelFactor(ParsedVector *TmpParsedVector);
int ParseObject(ObjectStruct **pLocalObject);
int ParseLightSource(LightSourceStruct **pLocalLightSource);
int ParseAtmosphere(AtmosphereStruct **pLocalAtmosphere);
int ParseFog(FogStruct **pLocalFog);
int ParseRainbow(RainbowStruct **pLocalRainbow);
int ParseSkySphere(SkySphereStruct **pLocalSkysphere);
int ParseTexture(TextureStruct **pLocalTexture);
int ParseNormal(NormalStruct **pLocalNormal);
int ParseFinish(FinishStruct **pLocalFinish);
int ParseHalo(HaloStruct **pLocalHalo);
int ParseColorMap(ColorMapStruct **pLocalColorMap);
int ParsePigmentMap(PigmentMapStruct **pLocalPigmentMap);
int ParseSlopeMap(SlopeMapStruct **pLocalSlopeMap);
int ParseNormalMap(NormalMapStruct **pLocalNormalMap);
int ParseCamera(CameraStruct **pLocalCamera);
int ParseTransform(TransformStruct **pLocalTransform);

/*****************************************************************************
*  NextChar
******************************************************************************/
int NextChar()
{
  if (IsOldChar)
  {
    IsOldChar = 0;
    return OldChar;
  }
  return getc(PovFile->handle);
}

/*****************************************************************************
*  IgnoreSpace
******************************************************************************/
int IgnoreSpace(void)
{
  int c;
  
  do
  {
    c = NextChar(); 
    if (c=='\n')
    {
      while (c=='\n')
      {
        PovFile->Line++;
        c = NextChar();
        if (c=='\r') c = NextChar();
      }
    } else if (c=='\r')
    {
      while (c=='\r')
      {
        PovFile->Line++;
        c = NextChar();
        if (c=='\n') c = NextChar();
      }
    }
  } while (c==' ' || c=='\t' || c=='' || c=='\r' || c=='\n');
  OldChar = c; IsOldChar = 1;
  return 0;
}

/*****************************************************************************
*  ParseComma
******************************************************************************/
int ParseComma(void)
{
  int t;
  
  if (NextToken(&t))
    return 1;
  if (t != TComma)
  {
    OldToken = t ; IsOldToken = 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseFloatLiteral
******************************************************************************/
int ParseFloatLiteral(double *pReturnedFloat)
{
  char *FloatText;
  int FloatTextLength, FloatTextMaxLength;
  int c;

  FloatTextMaxLength = 20;
  FloatText = g_malloc(21);
  FloatTextLength = 0;

  c = NextChar();
  if (c == '.')
  {
    strcpy(FloatText,"0.");
    FloatTextLength = 2;
    c = NextChar();
    if (c<'0' || c>'9')
    {
      WarningBox(_("Error"), _("Misformed float\n"));
      return 1;
    }
    while (c>='0' && c<='9')
    {
      FloatText[FloatTextLength++] = (char) c;
      if (FloatTextLength==FloatTextMaxLength)
      {
        FloatTextMaxLength *= 2;
        FloatText = g_realloc(FloatText, FloatTextMaxLength);
      }
      c = NextChar();
    }
    OldChar = c; IsOldChar = 1;
  } else
  { /* c = [0-9] */
    while (c>='0' && c<='9')
    {
      FloatText[FloatTextLength++] = (char) c;
      if (FloatTextLength==FloatTextMaxLength)
      {
        FloatTextMaxLength *= 2;
        FloatText = g_realloc(FloatText, FloatTextMaxLength);
      }
      c = NextChar();
    }
    if (c=='.')
    {
      FloatText[FloatTextLength++] = (char) c;
      if (FloatTextLength==FloatTextMaxLength)
      {
        FloatTextMaxLength *= 2;
        FloatText = g_realloc(FloatText, FloatTextMaxLength);
      }
      c = NextChar();
      if (c<'0' || c>'9')
      {
        FloatText[FloatTextLength++] = '0';
        if (FloatTextLength==FloatTextMaxLength)
        {
          FloatTextMaxLength *= 2;
          FloatText = g_realloc(FloatText, FloatTextMaxLength);
        }
      }
      while (c>='0' && c<='9')
      {
        FloatText[FloatTextLength++] = (char) c;
        if (FloatTextLength==FloatTextMaxLength)
        {
          FloatTextMaxLength *= 2;
          FloatText = g_realloc(FloatText, FloatTextMaxLength);
        }
        c = NextChar();
      }
    }
    OldChar = c; IsOldChar = 1;
  }
  /* At this point, FloatText is either 
     0.[0-9]+, [0-9]+, [0-9]+.0 or [0-9]+.[0-9]+ */
  c = NextChar();
  if (c=='e' || c=='E')
  {
    FloatText[FloatTextLength++] = 'e';
    if (FloatTextLength==FloatTextMaxLength)
    {
      FloatTextMaxLength *= 2;
      FloatText = g_realloc(FloatText, FloatTextMaxLength);
    }
    c = NextChar();
    if (c=='+' || c=='-')
    {
      FloatText[FloatTextLength++] = (char) c;
      if (FloatTextLength==FloatTextMaxLength)
      {
        FloatTextMaxLength *= 2;
        FloatText = g_realloc(FloatText, FloatTextMaxLength);
      }
      c = NextChar();
    }
    if (c<'0' || c>'9')
    {
      WarningBox(_("Error"), _("Misformed float\n"));
      return 1;
    }
    while (c>='0' && c<='9')
    {
      FloatText[FloatTextLength++] = (char) c;
      if (FloatTextLength==FloatTextMaxLength)
      {
        FloatTextMaxLength *= 2;
        FloatText = g_realloc(FloatText, FloatTextMaxLength);
      }
      c = NextChar();
    }
  }
  OldChar = c; IsOldChar = 1;
  FloatText[FloatTextLength]=0; 
  *pReturnedFloat = atof(FloatText);
  g_free(FloatText);
  return 0;
}

/*****************************************************************************
*  ParenError
******************************************************************************/
int ParenError(char *Function, int Token)
{
  int t;
  if (NextToken(&t))
    return 1;
  if (t != Token)
  {
    WarningBox(_("Error"), _("The argument of '%s' must be enclosed between parentheses...\n"), Function);
    return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseStringLiteral
******************************************************************************/
int ParseStringLiteral(char **pString)
{
  int StringLength, StringMaxLength;
  int c;
  
  StringMaxLength = 20;
  *pString = g_malloc(21);
  StringLength = 0;
  c=NextChar();
  while(c!='"')
  {
    if (c == EOF) return 1;
    (*pString)[StringLength++]=(char)c;
    if (StringLength==StringMaxLength)
    {
      StringMaxLength *= 2;
      *pString = g_realloc(*pString, StringMaxLength);
    }
    c=NextChar();
  }
  (*pString)[StringLength] = 0;
  return 0;
}

/*****************************************************************************
*  ParseString
******************************************************************************/
int ParseString(char **pLocalString)
{
  char TmpString1[100],TmpString2[100];
  char *AppendString;
  int t,i;
  double Val1, Val2, Val3;

  if (NextToken(&t))
    return 1;
  switch (t)
  {
    case TStringLiteral:
      *pLocalString = LastString;
      break;
    case Tstr:
      if (ParenError("str", TLeftParen))
        return 1;
      if (ParseFloat(&Val1))
        return 1;
      if (ParseComma())
        return 1;
      if (ParseFloat(&Val2))
        return 1;
      if (ParseComma())
        return 1;
      if (ParseFloat(&Val3))
        return 1;
      if (ParenError("str", TRightParen))
        return 1;
      if (Val3>=0)
      {
        if (Val2>0)
          sprintf(TmpString1, "%%%d.%df", (int)Val2, (int)Val3);
        else if (Val2 < 0)
          sprintf(TmpString1, "%%0%d.%df", (int)-Val2, (int)Val3);
        else
          sprintf(TmpString1, "%%.%df", (int)Val3);
      } else
      {
        if (Val2>0)
          sprintf(TmpString1, "%%%df",(int)Val2);
        else if (Val2 < 0)
          sprintf(TmpString1, "%%0%df",(int)-Val2);
        else
          strcpy(TmpString1, "%f");
      }
      sprintf(TmpString2, TmpString1, Val1);
      *pLocalString = g_malloc(strlen(TmpString2)+1);
      strcpy(*pLocalString, TmpString2);
      break;
    case Tconcat:
      if (ParenError("concat",TLeftParen))
        return 1;
      if (ParseString(pLocalString))
        return 1;
      if (NextToken(&t))
        return 1;
      while(t != TRightParen)
      {
        if (t != TComma)
        {
          OldToken = t; IsOldToken = 1;
        }
        ParseString(&AppendString);
        *pLocalString = g_realloc(*pLocalString, strlen(*pLocalString)+strlen(AppendString)+1);
        strcat(*pLocalString, AppendString);
        g_free(AppendString);
        if (NextToken(&t))
          return 1;
      }
      break;
    case Tchr:
      if (ParenError("chr", TLeftParen))
        return 1;
      if (ParseFloat(&Val1))
        return 1;
      if (ParenError("chr", TRightParen))
        return 1;
      if (Val1 < 0.0 || Val1 >= 256.0)
      {
        WarningBox(_("Error"), _("Bad value for chr() : %g...\n"), Val1);
        return 1;
      }
      *pLocalString = g_malloc(2);
      *pLocalString[0] = (char)((int)Val1);
      *pLocalString[1] = 0;
      break;
    case Tsubstr:
      if (ParenError("substr", TLeftParen))
        return 1;
      if (ParseString(&AppendString))
        return 1;
      if (ParseComma())
        return 1;
      if (ParseFloat(&Val1))
        return 1;
      if (ParseComma())
        return 1;
      if (ParseFloat(&Val2))
        return 1;
      if (ParenError("substr", TRightParen))
        return 1;
      if (((Val1+Val2-1.0) > strlen(AppendString)) ||
         (Val1<0.0) ||
         (Val2<1.0))
      {
        WarningBox(_("Error"), _("Bad params in substr(%s,%g,%g)...\n"), AppendString, Val1, Val2);
        return 1;
      }
      *pLocalString = g_malloc((int)Val2+1);
      strncpy(*pLocalString,&(AppendString[(int)Val1-1]),(int)Val2);
      *pLocalString[(int)Val2]='\0';
      g_free(AppendString);
      break;
    case Tstrupr:
      if (ParenError("strupr", TLeftParen))
        return 1;
      if (ParseString(pLocalString))
        return 1;
      if (ParenError("strupr", TRightParen))
        return 1;
      for (i=0 ; i<strlen(*pLocalString) ; i++)
        if (*pLocalString[i]>='a' && *pLocalString[i] <='z')
          *pLocalString[i]-=32;
      break;
    case Tstrlwr:
      if (ParenError("strlwr", TLeftParen))
        return 1;
      if (ParseString(pLocalString))
        return 1;
      if (ParenError("strlwr", TRightParen))
        return 1;
      for (i=0 ; i<strlen(*pLocalString) ; i++)
        if (*pLocalString[i]>='A' && *pLocalString[i] <='Z')
          *pLocalString[i]+=32;
      break;
    case TStringID:
      *pLocalString = g_strdup(LastStringConstant->String);
      break;
    default:
      WarningBox(_("Error"), _("A String was expected here...\n"));
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseVector
******************************************************************************/
int ParseVector(ParsedVector *TmpParsedVector)
{
  if (ParseRelFactor(TmpParsedVector))
    return 1;
  if (TmpParsedVector->nbTerms>3)
  {
    WarningBox(_("Error"), _("Expecting a vector but got a color expression...\n"));
    return 1;
  }
  if (TmpParsedVector->nbTerms==1)
  {
    TmpParsedVector->Val[1] = TmpParsedVector->Val[2] = TmpParsedVector->Val[0];
  } else if (TmpParsedVector->nbTerms==2)
  {
    TmpParsedVector->Val[2] = 0.0;
  }
  TmpParsedVector->nbTerms = 3;
  return 0;
}

/*****************************************************************************
*  ParseRelTerm
******************************************************************************/
int ParseRelTerm(ParsedVector *TmpParsedVector)
{
  ParsedVector SecondFactor;
  int t;
  
  if (ParseRelFactor(TmpParsedVector))
    return 1;
  if (NextToken(&t))
    return 1;
  while ((t==TLeftAngleBracket) || (t==TLowerOrEqual) ||
         (t==TEqual) || (t==TUnequal) ||
         (t==TUpperOrEqual) || (t==TRightAngleBracket))
  {
    if (ParseRelFactor(&SecondFactor))
      return 1;
    TmpParsedVector->nbTerms = 1;
    switch(t)
    {
      case TLeftAngleBracket:
        TmpParsedVector->Val[0] = 
               (TmpParsedVector->Val[0] < SecondFactor.Val[0]) ? 1.0 : 0.0;
        break;
      case TLowerOrEqual:
        TmpParsedVector->Val[0] = 
               (TmpParsedVector->Val[0] <= SecondFactor.Val[0]) ? 1.0 : 0.0;
        break;
      case TEqual:
        TmpParsedVector->Val[0] = 
               (TmpParsedVector->Val[0] == SecondFactor.Val[0]) ? 1.0 : 0.0;
        break;
      case TUnequal:
        TmpParsedVector->Val[0] = 
               (TmpParsedVector->Val[0] != SecondFactor.Val[0]) ? 1.0 : 0.0;
        break;
      case TUpperOrEqual:
        TmpParsedVector->Val[0] = 
               (TmpParsedVector->Val[0] >= SecondFactor.Val[0]) ? 1.0 : 0.0;
        break;
      case TRightAngleBracket:
        TmpParsedVector->Val[0] = 
               (TmpParsedVector->Val[0] > SecondFactor.Val[0]) ? 1.0 : 0.0;
        break;
    }
    if (NextToken(&t))
      return 1;
  }
  OldToken = t;
  IsOldToken = 1;
  return 0;
}

/*****************************************************************************
*  ParseLogical
******************************************************************************/
int ParseLogical(ParsedVector *TmpParsedVector)
{
  ParsedVector SecondTerm;
  int t;
  
  if (ParseRelTerm(TmpParsedVector))
    return 1;
  if (NextToken(&t))
    return 1;
  while ((t==TAmpersand) || (t==TBar))
  {
    if (ParseRelTerm(&SecondTerm))
      return 1;
    TmpParsedVector->nbTerms = 1;
    if (t==TAmpersand)
    {
      TmpParsedVector->Val[0] = TmpParsedVector->Val[0] && SecondTerm.Val[0];
    } else
    {
      TmpParsedVector->Val[0] = TmpParsedVector->Val[0] || SecondTerm.Val[0];
    }
    if (NextToken(&t))
      return 1;
  }
  OldToken = t;
  IsOldToken = 1;
  return 0;
}

/*****************************************************************************
*  ParseExpress
******************************************************************************/
int ParseExpress(ParsedVector *TmpParsedVector)
{
  ParsedVector LocalVector;
  ParsedVector LocalVector1;
  ParsedVector LocalVector2;
  int t,i;

  if (ParseLogical(&LocalVector))
    return 1;
  if (NextToken(&t))
    return 1;
  if (t == TQuestion)
  {
    if (LocalVector.nbTerms != 1)
    {
      WarningBox(_("Error"), _("Conditional must evaluate to a float...\n"));
      return 1;
    }
    if (ParseExpress(&LocalVector1))
      return 1;
    if (NextToken(&t))
      return 1;
    if (t != TColon)
    {
      WarningBox(_("Error"), _("Conditional must be followed by a colon...\n"));
      return 1;
    }
    if (ParseExpress(&LocalVector2))
      return 1;
    if (LocalVector.Val[0])
    {
      TmpParsedVector->nbTerms = LocalVector1.nbTerms;
      for (i=0 ; i<TmpParsedVector->nbTerms ; i++)
        TmpParsedVector->Val[i] = LocalVector1.Val[i];
    } else
    {
      TmpParsedVector->nbTerms = LocalVector2.nbTerms;
      for (i=0 ; i<TmpParsedVector->nbTerms ; i++)
        TmpParsedVector->Val[i] = LocalVector2.Val[i];
    }
  } else
  {
    OldToken = t; IsOldToken = 1;
    TmpParsedVector->nbTerms = LocalVector.nbTerms;
    for (i=0 ; i<TmpParsedVector->nbTerms ; i++)
      TmpParsedVector->Val[i] = LocalVector.Val[i];
  }
  return 0;
}

/*****************************************************************************
*  ParseNumFactor
******************************************************************************/
int ParseNumFactor(ParsedVector *TmpParsedVector)
{
  int t,c,i;
  ParsedVector LocalVector;
  ParsedVector LocalVector2;
  double Val1, Val2;
  FILE *f;
  char *LocalString, *LocalString2;

  if (NextToken(&t))
    return 1;
  while (t == TPlus)
  {
    if (NextToken(&t))
      return 1;
  }
  switch (t)
  {
    case Tabs:
      if (ParenError("abs", TLeftParen))
        return 1;
      if (ParseExpress(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        WarningBox(_("Error"), _("The argument of 'abs' must be a float...\n"));
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = abs(LocalVector.Val[0]);
      if (ParenError("abs", TRightParen))
        return 1;
      break;
    case Tacos:
      if (ParenError("acos", TLeftParen))
        return 1;
      if (ParseExpress(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        WarningBox(_("Error"), _("The argument of 'acos' must be a float...\n"));
        return 1;
      }
      if (LocalVector.Val[0]>1.0 || LocalVector.Val[0]<-1.0)
      {
        WarningBox(_("Error"), _("The argument of 'acos' must be between -1 and 1...\n"));
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = acos(LocalVector.Val[0]);
      if (ParenError("acos", TRightParen))
        return 1;
      break;
    case Tval:
      if (ParenError("val", TLeftParen))
        return 1;
      if (ParseString(&LocalString))
        return 1;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = atof(LocalString);
      g_free(LocalString);
      if (ParenError("val", TRightParen))
        return 1;
      break;
    case Tasc:
      if (ParenError("asc", TLeftParen))
        return 1;
      if (ParseString(&LocalString))
        return 1;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = LocalString[0];
      g_free(LocalString);
      if (ParenError("asc", TRightParen))
        return 1;
      break;
    case Tasin:
      if (ParenError("asin", TLeftParen))
        return 1;
      if (ParseExpress(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        WarningBox(_("Error"), _("The argument of 'asin' must be a float...\n"));
        return 1;
      }
      if (LocalVector.Val[0]>1.0 || LocalVector.Val[0]<-1.0)
      {
        WarningBox(_("Error"), _("The argument of 'asin' must be between -1 and 1...\n"));
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = asin(LocalVector.Val[0]);
      if (ParenError("asin", TRightParen))
        return 1;
      break;
    case Tatan2:
      if (ParenError("atan2", TLeftParen))
        return 1;
      if (ParseFloat(&Val1))
        return 1;
      if (ParseComma())
        return 1;
      if (ParseFloat(&Val2))
        return 1;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = atan2(Val1,Val2);
      if (ParenError("atan2", TRightParen))
        return 1;
      break;
    case Tceil:
      if (ParenError("ceil", TLeftParen))
        return 1;
      if (ParseExpress(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        WarningBox(_("Error"), _("The argument of 'ceil' must be a float...\n"));
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = ceil(LocalVector.Val[0]);
      if (ParenError("ceil", TRightParen))
        return 1;
      break;
    case Tclock: /* FIXME: should be special... */
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = PovClock;
      break;
    case Tcos:
      if (ParenError("cos", TLeftParen))
        return 1;
      if (ParseExpress(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        WarningBox(_("Error"), _("The argument of 'cos' must be a float...\n"));
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = cos(LocalVector.Val[0]);
      if (ParenError("cos", TRightParen))
        return 1;
      break;
    case Tdegrees:
      if (ParenError("degrees", TLeftParen))
        return 1;
      if (ParseExpress(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        WarningBox(_("Error"), _("The argument of 'degrees' must be a float...\n"));
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = LocalVector.Val[0]/M_PI*180.0;
      if (ParenError("degrees", TRightParen))
        return 1;
      break;
    case Tdiv:
      if (ParenError("div", TLeftParen))
        return 1;
      if (ParseFloat(&Val1))
        return 1;
      if (ParseComma())
        return 1;
      if (ParseFloat(&Val2))
        return 1;
      if (abs(Val2) <= 1.0e-6)
      {
        WarningBox(_("Error"), _("The second argument of 'div' mustn't be nul...\n"));
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = (double)((int)(Val1 / Val2));
      if (ParenError("div", TRightParen))
        return 1;
      break;
    case Texp:
      if (ParenError("exp", TLeftParen))
        return 1;
      if (ParseExpress(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        WarningBox(_("Error"), _("The argument of 'exp' must be a float...\n"));
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = exp(LocalVector.Val[0]);
      if (ParenError("exp", TRightParen))
        return 1;
      break;
    case Tfile_exists:
      if (ParenError("file_exists", TLeftParen))
        return 1;
      if (ParseString(&LocalString))
        return 1;
      TmpParsedVector->nbTerms = 1;
      f = LibOpen(LocalString);
      if (f)
        TmpParsedVector->Val[0] = 1.0;
      else
        TmpParsedVector->Val[0] = 0.0;
      fclose(f);
      g_free(LocalString);
      if (ParenError("file_exists", TRightParen))
        return 1;
      break;
    case TFloatID:
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = LastFloatConstant->Float;
      break;
    case TFloatLiteral:
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = LastFloat;
      break;
    case Tfloor:
      if (ParenError("floor", TLeftParen))
        return 1;
      if (ParseExpress(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        WarningBox(_("Error"), _("The argument of 'floor' must be a float...\n"));
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = floor(LocalVector.Val[0]);
      if (ParenError("floor", TRightParen))
        return 1;
      break;
    case Tint:
      if (ParenError("int", TLeftParen))
        return 1;
      if (ParseExpress(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        WarningBox(_("Error"), _("The argument of 'int' must be a float...\n"));
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = (double)((int)LocalVector.Val[0]);
      if (ParenError("int", TRightParen))
        return 1;
      break;
    case Tlog:
      if (ParenError("log", TLeftParen))
        return 1;
      if (ParseExpress(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        WarningBox(_("Error"), _("The argument of 'log' must be a float...\n"));
        return 1;
      }
      if (TmpParsedVector->Val[0] <= 0.0)
      {
        WarningBox(_("Error"), _("The argument of 'log' must be strictly positive...\n"));
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = log(LocalVector.Val[0]);
      if (ParenError("log", TRightParen))
        return 1;
      break;
    case Tmin:
      if (ParenError("min", TLeftParen))
        return 1;
      if (ParseFloat(&Val1))
        return 1;
      if (ParseComma())
        return 1;
      if (ParseFloat(&Val2))
        return 1;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = Val1>Val2?Val2:Val1;
      if (ParenError("min", TRightParen))
        return 1;
      break;
    case Tmax:
      if (ParenError("max", TLeftParen))
        return 1;
      if (ParseFloat(&Val1))
        return 1;
      if (ParseComma())
        return 1;
      if (ParseFloat(&Val2))
        return 1;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = Val1<Val2?Val2:Val1;
      if (ParenError("max", TRightParen))
        return 1;
      break;
    case Tmod:
      if (ParenError("mod", TLeftParen))
        return 1;
      if (ParseFloat(&Val1))
        return 1;
      if (ParseComma())
        return 1;
      if (ParseFloat(&Val2))
        return 1;
      if (abs(Val2) <= 1.0e-6)
      {
        WarningBox(_("Error"), _("The second argument of 'mod' mustn't be nul...\n"));
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = fmod(Val1,Val2);
      if (ParenError("mod", TRightParen))
        return 1;
      break;
    case Tpi:
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = M_PI;
      break;
    case Tpow:
      if (ParenError("pow", TLeftParen))
        return 1;
      if (ParseFloat(&Val1))
        return 1;
      if (ParseComma())
        return 1;
      if (ParseFloat(&Val2))
        return 1;
      if (Val1 <= 0.0 )
        TmpParsedVector->Val[0] = pow(Val1, (int) Val2);
      else
        TmpParsedVector->Val[0] = pow(Val1, Val2);
      TmpParsedVector->nbTerms = 1;
      if (ParenError("pow", TRightParen))
        return 1;
      break;
    case Tradians:
      if (ParenError("radians", TLeftParen))
        return 1;
      if (ParseExpress(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        WarningBox(_("Error"), _("The argument of 'radians' must be a float...\n"));
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = LocalVector.Val[0] * M_PI / 180.0;
      if (ParenError("radians", TRightParen))
        return 1;
      break;
    case Tsin:
      if (ParenError("sin", TLeftParen))
        return 1;
      if (ParseExpress(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        WarningBox(_("Error"), _("The argument of 'sin' must be a float...\n"));
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = sin(LocalVector.Val[0]);
      if (ParenError("sin", TRightParen))
        return 1;
      break;
    case Tsqrt:
      if (ParenError("sqrt", TLeftParen))
        return 1;
      if (ParseExpress(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        WarningBox(_("Error"), _("The argument of 'sqrt' must be a float...\n"));
        return 1;
      }
      if (LocalVector.Val[0] < 0.0)
      {
        WarningBox(_("Error"), _("The argument of 'sqrt' must be postive...\n"));
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = sqrt(LocalVector.Val[0]);
      if (ParenError("sqrt", TRightParen))
        return 1;
      break;
    case Tstrcmp:
      if (ParenError("strcmp", TLeftParen))
        return 1;
      if (ParseString(&LocalString))
        return 1;
      if (ParseComma())
        return 1;
      if (ParseString(&LocalString2))
        return 1;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = strcmp(LocalString, LocalString2);
      g_free(LocalString);
      g_free(LocalString2);
      if (ParenError("strcmp", TRightParen))
        return 1;
      break;
    case Tstrlen:
      if (ParenError("strlen", TLeftParen))
        return 1;
      if (ParseString(&LocalString))
        return 1;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = strlen(LocalString);
      g_free(LocalString);
      if (ParenError("strlen", TRightParen))
        return 1;
      break;
    case Ttan:
      if (ParenError("tan", TLeftParen))
        return 1;
      if (ParseExpress(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        WarningBox(_("Error"), _("The argument of 'tan' must be a float...\n"));
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = tan(LocalVector.Val[0]);
      if (ParenError("tan", TRightParen))
        return 1;
      break;
    case Tvdot:
      if (ParenError("vdot", TLeftParen))
        return 1;
      if (ParseVector(&LocalVector))
        return 1;
      if (ParseComma())
        return 1;
      if (ParseVector(&LocalVector2))
        return 1;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = VDot(LocalVector.Val, LocalVector2.Val);
      if (ParenError("vdot", TRightParen))
        return 1;
      break;
    case Tvlength:
      if (ParenError("vlength", TLeftParen))
        return 1;
      ParseVector(&LocalVector);
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = VLength(LocalVector.Val);
      if (ParenError("vlength", TRightParen))
        return 1;
      break;
    case Tversion:
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = PovVersionLangage;
      break;
    case Ttrue:
    case Tyes:
    case Ton:
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = 1.0;
      break;
    case Tfalse:
    case Tno:
    case Toff:
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = 0.0;
      break;
    case Tseed:
      if (ParenError("seed", TLeftParen))
        return 1;
      if (ParseExpress(&LocalVector))
        return 1;
      if (ParenError("seed", TRightParen))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        WarningBox(_("Error"), _("The argument of 'seed' must be a float...\n"));
        return 1;
      }
      PovRandomTable = g_realloc(PovRandomTable, 
                                 (NbPovRandomTable+1)*sizeof(unsigned long));
      PovRandomTable[NbPovRandomTable] = (unsigned long)LocalVector.Val[0];
      NbPovRandomTable++;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = NbPovRandomTable-1;
      break;
    case Trand:
      if (ParenError("rand", TLeftParen))
        return 1;
      if (ParseExpress(&LocalVector))
        return 1;
      if (ParenError("rand", TRightParen))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        WarningBox(_("Error"), _("The argument of 'rand' must be a float...\n"));
        return 1;
      }
      PovRandomTable[(int)LocalVector.Val[0]] = 
             PovRandomTable[(int)LocalVector.Val[0]] * 1812433253L + 12345L;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = 
             ((double)(PovRandomTable[(int)LocalVector.Val[0]] & 0xFFFFFFFFUL) 
              / 0xFFFFFFFFUL);
      break;
    case Tvaxis_rotate:
      if (ParenError("vaxis_rotate", TLeftParen))
        return 1;
      if (ParseVector(&LocalVector))
        return 1;
      if (ParseComma())
        return 1;
      if (ParseVector(&LocalVector2))
        return 1;
      if (ParseComma())
        return 1;
      if (ParseFloat(&Val1))
        return 1;
      if (ParenError("vaxis_rotate", TRightParen))
        return 1;
      /* FIXME */
      WarningBox(_("Not Yet!"), _("The function 'vaxis_rotate' not yet implemented\n"));
      return 1;
      break;
    case Tvcross:
      if (ParenError("vcross", TLeftParen))
        return 1;
      if (ParseVector(&LocalVector))
        return 1;
      if (ParseComma())
        return 1;
      if (ParseVector(&LocalVector2))
        return 1;
      if (ParenError("vcross", TRightParen))
        return 1;
      TmpParsedVector->nbTerms = 3;
      VCross(TmpParsedVector->Val, LocalVector.Val, LocalVector2.Val);
      break;
    case TVectorID:
      TmpParsedVector->nbTerms = 3;
      TmpParsedVector->Val[0] = LastVectorConstant->Vector[0];
      TmpParsedVector->Val[1] = LastVectorConstant->Vector[1];
      TmpParsedVector->Val[2] = LastVectorConstant->Vector[2];
      break;
    case Tvnormalize:
      if (ParenError("vnormalize", TLeftParen))
        return 1;
      if (ParseVector(TmpParsedVector))
        return 1;
      if (ParenError("vnormalize", TRightParen))
        return 1;
      Val1 = VLength(TmpParsedVector->Val);
      if (Val1 != 0.0);
      {
        TmpParsedVector->Val[0] /= Val1;
        TmpParsedVector->Val[1] /= Val1;
        TmpParsedVector->Val[2] /= Val1;
      }
      break;
    case Tvrotate:
      WarningBox(_("Not Yet!"), _("The function 'vrotate' not yet implemented\n"));
      return 1;
      /* FIXME */
      break;
    case Tx:
      TmpParsedVector->nbTerms = 3;
      TmpParsedVector->Val[0] = 1.0;
      TmpParsedVector->Val[1] = 0.0;
      TmpParsedVector->Val[2] = 0.0;
      break;
    case Ty:
      TmpParsedVector->nbTerms = 3;
      TmpParsedVector->Val[0] = 0.0;
      TmpParsedVector->Val[1] = 1.0;
      TmpParsedVector->Val[2] = 0.0;
      break;
    case Tz:
      TmpParsedVector->nbTerms = 3;
      TmpParsedVector->Val[0] = 0.0;
      TmpParsedVector->Val[1] = 0.0;
      TmpParsedVector->Val[2] = 1.0;
      break;
    case TColorID:
      TmpParsedVector->nbTerms = 5;
      TmpParsedVector->Val[0] = LastColorConstant->Color[0];
      TmpParsedVector->Val[1] = LastColorConstant->Color[1];
      TmpParsedVector->Val[2] = LastColorConstant->Color[2];
      TmpParsedVector->Val[3] = LastColorConstant->Color[3];
      TmpParsedVector->Val[4] = LastColorConstant->Color[4];
      break;
    case Tt:
      TmpParsedVector->nbTerms = 4;
      TmpParsedVector->Val[0] = 0.0;
      TmpParsedVector->Val[1] = 0.0;
      TmpParsedVector->Val[2] = 0.0;
      TmpParsedVector->Val[3] = 1.0;
      break;
    case Tu:
      TmpParsedVector->nbTerms = 2;
      TmpParsedVector->Val[0] = 1.0;
      TmpParsedVector->Val[1] = 0.0;
      break;
    case Tv:
      TmpParsedVector->nbTerms = 2;
      TmpParsedVector->Val[0] = 0.0;
      TmpParsedVector->Val[1] = 1.0;
      break;
    case TMinus:
      if (ParseNumFactor(TmpParsedVector))
        return 1;
      for (i=0 ; i<5 ; i++)
        TmpParsedVector->Val[i] = -TmpParsedVector->Val[i];
      break;
    case TExclamation:
      if (ParseNumFactor(TmpParsedVector))
        return 1;
      for (i=0 ; i<TmpParsedVector->nbTerms ; i++)
        TmpParsedVector->Val[i] = fabs(TmpParsedVector->Val[i])>10e-6?0.0:1.0;
      break;
    case TLeftParen:
      if (ParseExpress(TmpParsedVector))
        return 1;
      if (ParenError(_("Left Parenthese"), TRightParen))
        return 1;
      break;
    case TLeftAngleBracket:
      if (ParseFloat(&(TmpParsedVector->Val[0])))
        return 1;
      if (ParseComma())
        return 1;
      if (NextToken(&t))
        return 1;
      if (t==TRightAngleBracket)
      {
        TmpParsedVector->nbTerms = 1;
      } else
      { 
        IsOldToken = 1; OldToken = t;
        if (ParseFloat(&(TmpParsedVector->Val[1])))
          return 1;
        if (ParseComma())
          return 1;
        if (NextToken(&t))
          return 1;
        if (t==TRightAngleBracket)
        {
          TmpParsedVector->nbTerms = 2;
        } else
        {
          IsOldToken = 1; OldToken = t;
          if (ParseFloat(&(TmpParsedVector->Val[2])))
            return 1;
          if (ParseComma())
            return 1;
          if (NextToken(&t))
            return 1;
          if (t==TRightAngleBracket)
          {
            TmpParsedVector->nbTerms = 3;
          } else
          {
            IsOldToken = 1; OldToken = t;
            if (ParseFloat(&(TmpParsedVector->Val[3])))
              return 1;
            if (ParseComma())
              return 1;
            if (NextToken(&t))
              return 1;
            if (t==TRightAngleBracket)
            {
              TmpParsedVector->nbTerms = 4;
            } else
            {
              IsOldToken = 1; OldToken = t;
              if (ParseFloat(&(TmpParsedVector->Val[4])))
                return 1;
              if (ParseComma())
                return 1;
              if (NextToken(&t))
                return 1;
              if (t==TRightAngleBracket)
              {
                TmpParsedVector->nbTerms = 5;
              } else
              {
                WarningBox(_("Error"), _("Vectors can only have 5 terms or less...\n"));
                return 1;
              }
            }
          }
        }
      }
      break;
    default:
      WarningBox(_("Error"), _("%s:%d : Numeric Expression expected (%d)...\n"), 
                 PovFile->Name, PovFile->Line, t);
      return 1;
  }
  if (IgnoreSpace())
    return 1;
  c = NextChar();
  if (c != '.')
  {
    IsOldChar = 1; OldChar = c;
    return 0;
  }
  if (NextToken(&t))
    return 1;
  switch (t)
  {
    case Tx:
      if (TmpParsedVector->nbTerms != 3)
      {
        WarningBox(_("Error"), _("The 'x' suffix can only be used with vectors...\n"));
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
      }
      break;
    case Ty:
      if (TmpParsedVector->nbTerms != 3)
      {
        WarningBox(_("Error"), _("The 'y' suffix can only be used with vectors...\n"));
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
        TmpParsedVector->Val[0] = TmpParsedVector->Val[1];
      }
      break;
    case Tz:
      if (TmpParsedVector->nbTerms != 3)
      {
        WarningBox(_("Error"), _("The 'z' suffix can only be used with vectors...\n"));
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
        TmpParsedVector->Val[0] = TmpParsedVector->Val[2];
      }
      break;
    case Tred:
      if (TmpParsedVector->nbTerms != 5)
      {
        WarningBox(_("Error"), _("The 'red' suffix can only be used with colors...\n"));
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
      }
      break;
    case Tgreen:
      if (TmpParsedVector->nbTerms != 5)
      {
        WarningBox(_("Error"), _("The 'green' suffix can only be used with colors...\n"));
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
        TmpParsedVector->Val[0] = TmpParsedVector->Val[1];
      }
      break;
    case Tblue:
      if (TmpParsedVector->nbTerms != 5)
      {
        WarningBox(_("Error"), _("The 'blue' suffix can only be used with colors...\n"));
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
        TmpParsedVector->Val[0] = TmpParsedVector->Val[2];
      }
      break;
    case Tfilter:
      if (TmpParsedVector->nbTerms != 5)
      {
        WarningBox(_("Error"), _("The 'filter' suffix can only be used with colors...\n"));
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
        TmpParsedVector->Val[0] = TmpParsedVector->Val[3];
      }
      break;
    case Ttransmit:
      if (TmpParsedVector->nbTerms != 5)
      {
        WarningBox(_("Error"), _("The 'transmit' suffix can only be used with colors...\n"));
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
        TmpParsedVector->Val[0] = TmpParsedVector->Val[4];
      }
      break;
    case Tu:
      TmpParsedVector->nbTerms = 1;
      break;
    case Tv:
      if (TmpParsedVector->nbTerms == 1)
      {
        WarningBox(_("Error"), _("The 'v' suffix can't be used with float...\n"));
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
        TmpParsedVector->Val[0] = TmpParsedVector->Val[1];
      }
      break;
    case Tt:
      if (TmpParsedVector->nbTerms < 4)
      {
        WarningBox(_("Error"), _("The 't' suffix can't be used there...\n"));
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
        TmpParsedVector->Val[0] = TmpParsedVector->Val[3];
      }
      break;
  }
  return 0;
}

/*****************************************************************************
*  ParseNumTerm
******************************************************************************/
int ParseNumTerm(ParsedVector *TmpParsedVector)
{
  ParsedVector SecondFactor;
  int t,i;
  
  if (ParseNumFactor(TmpParsedVector))
    return 1;
  if (NextToken(&t))
    return 1;
  while ((t==TSlash) || (t==TStar))
  {
    if (ParseNumFactor(&SecondFactor))
      return 1;
    if (SecondFactor.nbTerms != TmpParsedVector->nbTerms)
    {
      if (SecondFactor.nbTerms > TmpParsedVector->nbTerms)
      {
        if (TmpParsedVector->nbTerms == 1)
          for (i=1 ; i<SecondFactor.nbTerms ; i++)
            TmpParsedVector->Val[i] = TmpParsedVector->Val[0];
        else
          for (i=TmpParsedVector->nbTerms ; i<SecondFactor.nbTerms ; i++)
            TmpParsedVector->Val[i] = 0.0;
        TmpParsedVector->nbTerms = SecondFactor.nbTerms;
      } else
      {
        if (SecondFactor.nbTerms == 1)
          for (i=1 ; i<TmpParsedVector->nbTerms ; i++)
            SecondFactor.Val[i] = SecondFactor.Val[0];
        else
          for (i=SecondFactor.nbTerms ; i<TmpParsedVector->nbTerms ; i++)
            SecondFactor.Val[i] = 0.0;
      }
    }
    if (t==TStar)
    {
      for (i=0 ; i<TmpParsedVector->nbTerms ; i++)
        TmpParsedVector->Val[i] *= SecondFactor.Val[i];
    } else
    {
      for (i=0 ; i<TmpParsedVector->nbTerms ; i++)
        TmpParsedVector->Val[i] /= SecondFactor.Val[i];
    }
    if (NextToken(&t))
      return 1;
  }
  OldToken = t;
  IsOldToken = 1;
  return 0;
}

/*****************************************************************************
*  ParseRelFactor
******************************************************************************/
int ParseRelFactor(ParsedVector *TmpParsedVector)
{
  ParsedVector SecondTerm;
  int t,i;
  
  if (ParseNumTerm(TmpParsedVector))
    return 1;
  if (NextToken(&t))
    return 1;
  while ((t==TPlus) || (t==TMinus))
  {
    if (ParseNumTerm(&SecondTerm))
      return 1;
    if (SecondTerm.nbTerms != TmpParsedVector->nbTerms)
    {
      if (SecondTerm.nbTerms > TmpParsedVector->nbTerms)
      {
        if (TmpParsedVector->nbTerms == 1)
          for (i=1 ; i<SecondTerm.nbTerms ; i++)
            TmpParsedVector->Val[i] = TmpParsedVector->Val[0];
        else
          for (i=TmpParsedVector->nbTerms ; i<SecondTerm.nbTerms ; i++)
            TmpParsedVector->Val[i] = 0.0;
        TmpParsedVector->nbTerms = SecondTerm.nbTerms;
      } else
      {
        if (SecondTerm.nbTerms == 1)
          for (i=1 ; i<TmpParsedVector->nbTerms ; i++)
            SecondTerm.Val[i] = SecondTerm.Val[0];
        else
          for (i=SecondTerm.nbTerms ; i<TmpParsedVector->nbTerms ; i++)
            SecondTerm.Val[i] = 0.0;
      }
    }
    if (t==TPlus)
    {
      for (i=0 ; i<TmpParsedVector->nbTerms ; i++)
        TmpParsedVector->Val[i] += SecondTerm.Val[i];
    } else
    {
      for (i=0 ; i<TmpParsedVector->nbTerms ; i++)
        TmpParsedVector->Val[i] -= SecondTerm.Val[i];
    }
    if (NextToken(&t))
      return 1;
  }
  OldToken = t;
  IsOldToken = 1;
  return 0;
}

/*****************************************************************************
*  ParseFloat
******************************************************************************/
int ParseFloat(double *pLocalFloat)
{
  
  ParsedVector TmpParsedVector;

  ParseRelFactor(&TmpParsedVector);
  if (TmpParsedVector.nbTerms != 1)
  {
    WarningBox(_("Error"), _("Expecting a Float but got a vector...\n"));
    return 1;
  }
  *pLocalFloat = TmpParsedVector.Val[0];
  return 0;
}

/*****************************************************************************
*  ParseWord
******************************************************************************/
int ParseWord(int *pToken)
{
  char *Word;
  int WordLength, WordMaxLength;
  int c,i;
  FloatConstantStruct *LocalFloatConstant;
  VectorConstantStruct *LocalVectorConstant;
  ColorConstantStruct *LocalColorConstant;
  StringConstantStruct *LocalStringConstant;
  ObjectConstantStruct *LocalObjectConstant;
  TextureConstantStruct *LocalTextureConstant;
  PigmentConstantStruct *LocalPigmentConstant;
  NormalConstantStruct *LocalNormalConstant;
  FinishConstantStruct *LocalFinishConstant;
  HaloConstantStruct *LocalHaloConstant;
  ColorMapConstantStruct *LocalColorMapConstant;
  PigmentMapConstantStruct *LocalPigmentMapConstant;
  SlopeMapConstantStruct *LocalSlopeMapConstant;
  NormalMapConstantStruct *LocalNormalMapConstant;
  CameraConstantStruct *LocalCameraConstant;
  LightSourceConstantStruct *LocalLightSourceConstant;
  AtmosphereConstantStruct *LocalAtmosphereConstant;
  FogConstantStruct *LocalFogConstant;
  RainbowConstantStruct *LocalRainbowConstant;
  SkySphereConstantStruct *LocalSkySphereConstant;
  TransformConstantStruct *LocalTransformConstant;
  
  WordMaxLength = 20;
  Word = g_malloc(21);
  WordLength = 0;
  c = NextChar();
  while((c>='a' && c<='z') || (c>='A' && c<='Z') || (c>='0' && c<='9') 
      || c=='_')
  {
    Word[WordLength++]=(char)c;
    if (WordLength==WordMaxLength)
    {
      WordMaxLength *= 2;
      Word = g_realloc(Word, WordMaxLength);
    }
    c = NextChar();
  }
  OldChar = c; IsOldChar = 1;
  Word[WordLength] = 0;
 
  for (i=0 ; i<LastWord ; i++)
  {
    if (!strcmp(ReservedWordTable[i].String,Word))
    {
      *pToken = ReservedWordTable[i].Token;
      g_free(Word);
      return 0;
    }
  }
  for (LocalFloatConstant = CurrentFrame->FloatConstantList ;
       LocalFloatConstant ;
       LocalFloatConstant=LocalFloatConstant->Next)
  {
    if (!strcmp(LocalFloatConstant->Name, Word))
    {
      *pToken = TFloatID;
      LastFloatConstant = LocalFloatConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalVectorConstant = CurrentFrame->VectorConstantList ;
       LocalVectorConstant ;
       LocalVectorConstant=LocalVectorConstant->Next)
  {
    if (!strcmp(LocalVectorConstant->Name, Word))
    {
      *pToken = TVectorID;
      LastVectorConstant = LocalVectorConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalColorConstant = CurrentFrame->ColorConstantList ;
       LocalColorConstant ;
       LocalColorConstant=LocalColorConstant->Next)
  {
    if (!strcmp(LocalColorConstant->Name, Word))
    {
      *pToken = TColorID;
      LastColorConstant = LocalColorConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalStringConstant = CurrentFrame->StringConstantList ;
       LocalStringConstant ;
       LocalStringConstant = LocalStringConstant->Next)
  {
    if (!strcmp(LocalStringConstant->Name, Word))
    {
      *pToken = TStringID;
      LastStringConstant = LocalStringConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalObjectConstant = CurrentFrame->ObjectConstantList ;
       LocalObjectConstant ;
       LocalObjectConstant = LocalObjectConstant->Next)
  {
    if (!strcmp(LocalObjectConstant->Name, Word))
    {
      *pToken = TObjectID;
      LastObjectConstant = LocalObjectConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalTextureConstant = CurrentFrame->TextureConstantList ;
       LocalTextureConstant ;
       LocalTextureConstant = LocalTextureConstant->Next)
  {
    if (!strcmp(LocalTextureConstant->Name, Word))
    {
      *pToken = TTextureID;
      LastTextureConstant = LocalTextureConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalPigmentConstant = CurrentFrame->PigmentConstantList ;
       LocalPigmentConstant ;
       LocalPigmentConstant = LocalPigmentConstant->Next)
  {
    if (!strcmp(LocalPigmentConstant->Name, Word))
    {
      *pToken = TPigmentID;
      LastPigmentConstant = LocalPigmentConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalNormalConstant = CurrentFrame->NormalConstantList ;
       LocalNormalConstant ;
       LocalNormalConstant = LocalNormalConstant->Next)
  {
    if (!strcmp(LocalNormalConstant->Name, Word))
    {
      *pToken = TNormalID;
      LastNormalConstant = LocalNormalConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalFinishConstant = CurrentFrame->FinishConstantList ;
       LocalFinishConstant ;
       LocalFinishConstant = LocalFinishConstant->Next)
  {
    if (!strcmp(LocalFinishConstant->Name, Word))
    {
      *pToken = TFinishID;
      LastFinishConstant = LocalFinishConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalHaloConstant = CurrentFrame->HaloConstantList ;
       LocalHaloConstant ;
       LocalHaloConstant = LocalHaloConstant->Next)
  {
    if (!strcmp(LocalHaloConstant->Name, Word))
    {
      *pToken = THaloID;
      LastHaloConstant = LocalHaloConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalColorMapConstant = CurrentFrame->ColorMapConstantList ;
       LocalColorMapConstant ;
       LocalColorMapConstant = LocalColorMapConstant->Next)
  {
    if (!strcmp(LocalColorMapConstant->Name, Word))
    {
      *pToken = TColorMapID;
      LastColorMapConstant = LocalColorMapConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalPigmentMapConstant = CurrentFrame->PigmentMapConstantList ;
       LocalPigmentMapConstant ;
       LocalPigmentMapConstant = LocalPigmentMapConstant->Next)
  {
    if (!strcmp(LocalPigmentMapConstant->Name, Word))
    {
      *pToken = TPigmentMapID;
      LastPigmentMapConstant = LocalPigmentMapConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalSlopeMapConstant = CurrentFrame->SlopeMapConstantList;
       LocalSlopeMapConstant ;
       LocalSlopeMapConstant = LocalSlopeMapConstant->Next)
  {
    if (!strcmp(LocalSlopeMapConstant->Name, Word))
    {
      *pToken = TSlopeMapID;
      LastSlopeMapConstant = LocalSlopeMapConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalNormalMapConstant = CurrentFrame->NormalMapConstantList;
       LocalNormalMapConstant ;
       LocalNormalMapConstant = LocalNormalMapConstant->Next)
  {
    if (!strcmp(LocalNormalMapConstant->Name, Word))
    {
      *pToken = TNormalMapID;
      LastNormalMapConstant = LocalNormalMapConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalCameraConstant = CurrentFrame->CameraConstantList;
       LocalCameraConstant ;
       LocalCameraConstant = LocalCameraConstant->Next)
  {
    if (!strcmp(LocalCameraConstant->Name, Word))
    {
      *pToken = TCameraID;
      LastCameraConstant = LocalCameraConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalLightSourceConstant = CurrentFrame->LightSourceConstantList;
       LocalLightSourceConstant ;
       LocalLightSourceConstant = LocalLightSourceConstant->Next)
  {
    if (!strcmp(LocalLightSourceConstant->Name, Word))
    {
      *pToken = TLightSourceID;
      LastLightSourceConstant = LocalLightSourceConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalAtmosphereConstant = CurrentFrame->AtmosphereConstantList;
       LocalAtmosphereConstant ;
       LocalAtmosphereConstant = LocalAtmosphereConstant->Next)
  {
    if (!strcmp(LocalAtmosphereConstant->Name, Word))
    {
      *pToken = TAtmosphereID;
      LastAtmosphereConstant = LocalAtmosphereConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalFogConstant = CurrentFrame->FogConstantList;
       LocalFogConstant ;
       LocalFogConstant = LocalFogConstant->Next)
  {
    if (!strcmp(LocalFogConstant->Name, Word))
    {
      *pToken = TFogID;
      LastFogConstant = LocalFogConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalRainbowConstant = CurrentFrame->RainbowConstantList;
       LocalRainbowConstant ;
       LocalRainbowConstant = LocalRainbowConstant->Next)
  {
    if (!strcmp(LocalRainbowConstant->Name, Word))
    {
      *pToken = TRainbowID;
      LastRainbowConstant = LocalRainbowConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalSkySphereConstant = CurrentFrame->SkySphereConstantList;
       LocalSkySphereConstant ;
       LocalSkySphereConstant = LocalSkySphereConstant->Next)
  {
    if (!strcmp(LocalSkySphereConstant->Name, Word))
    {
      *pToken = TSkySphereID;
      LastSkySphereConstant = LocalSkySphereConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalTransformConstant = CurrentFrame->TransformConstantList;
       LocalTransformConstant ;
       LocalTransformConstant = LocalTransformConstant->Next)
  {
    if (!strcmp(LocalTransformConstant->Name, Word))
    {
      *pToken = TTransformID;
      LastTransformConstant = LocalTransformConstant;
      g_free(Word);
      return 0;
    }
  }
  if (*pToken == -1)
  {
    WarningBox(_("Error"), _("%s:%d : Unknown keyword : '%s'...\n"), 
               PovFile->Name, PovFile->Line, Word);
    return 1;
  }
  g_free(Word);
  return 0;
}

/*****************************************************************************
*  GetTypeId
******************************************************************************/
int GetTypeId(char *Name)
{
  FloatConstantStruct *LocalFloatConstant;
  VectorConstantStruct *LocalVectorConstant;
  ColorConstantStruct *LocalColorConstant;
  StringConstantStruct *LocalStringConstant;
  ObjectConstantStruct *LocalObjectConstant;
  TextureConstantStruct *LocalTextureConstant;
  PigmentConstantStruct *LocalPigmentConstant;
  NormalConstantStruct *LocalNormalConstant;
  FinishConstantStruct *LocalFinishConstant;
  HaloConstantStruct *LocalHaloConstant;
  ColorMapConstantStruct *LocalColorMapConstant;
  PigmentMapConstantStruct *LocalPigmentMapConstant;
  SlopeMapConstantStruct *LocalSlopeMapConstant;
  NormalMapConstantStruct *LocalNormalMapConstant;
  CameraConstantStruct *LocalCameraConstant;
  LightSourceConstantStruct *LocalLightSourceConstant;
  AtmosphereConstantStruct *LocalAtmosphereConstant;
  FogConstantStruct *LocalFogConstant;
  RainbowConstantStruct *LocalRainbowConstant;
  SkySphereConstantStruct *LocalSkySphereConstant;
  TransformConstantStruct *LocalTransformConstant;

  for (LocalFloatConstant = CurrentFrame->FloatConstantList ;
       LocalFloatConstant ;
       LocalFloatConstant=LocalFloatConstant->Next)
    if (!strcmp(LocalFloatConstant->Name, Name))
      return TFloatID;
  for (LocalVectorConstant = CurrentFrame->VectorConstantList ;
       LocalVectorConstant ;
       LocalVectorConstant=LocalVectorConstant->Next)
    if (!strcmp(LocalVectorConstant->Name, Name))
      return TVectorID;
  for (LocalColorConstant = CurrentFrame->ColorConstantList ;
       LocalColorConstant ;
       LocalColorConstant=LocalColorConstant->Next)
    if (!strcmp(LocalColorConstant->Name, Name))
      return TColorID;
  for (LocalStringConstant = CurrentFrame->StringConstantList ;
       LocalStringConstant ;
       LocalStringConstant = LocalStringConstant->Next)
    if (!strcmp(LocalStringConstant->Name, Name))
      return TStringID;
  for (LocalObjectConstant = CurrentFrame->ObjectConstantList ;
       LocalObjectConstant ;
       LocalObjectConstant = LocalObjectConstant->Next)
    if (!strcmp(LocalObjectConstant->Name, Name))
      return TObjectID;
  for (LocalTextureConstant = CurrentFrame->TextureConstantList ;
       LocalTextureConstant ;
       LocalTextureConstant = LocalTextureConstant->Next)
    if (!strcmp(LocalTextureConstant->Name, Name))
      return TTextureID;
  for (LocalPigmentConstant = CurrentFrame->PigmentConstantList ;
       LocalPigmentConstant ;
       LocalPigmentConstant = LocalPigmentConstant->Next)
    if (!strcmp(LocalPigmentConstant->Name, Name))
      return TPigmentID;
  for (LocalNormalConstant = CurrentFrame->NormalConstantList ;
       LocalNormalConstant ;
       LocalNormalConstant = LocalNormalConstant->Next)
    if (!strcmp(LocalNormalConstant->Name, Name))
      return TNormalID;
  for (LocalFinishConstant = CurrentFrame->FinishConstantList ;
       LocalFinishConstant ;
       LocalFinishConstant = LocalFinishConstant->Next)
    if (!strcmp(LocalFinishConstant->Name, Name))
      return TFinishID;
  for (LocalHaloConstant = CurrentFrame->HaloConstantList ;
       LocalHaloConstant ;
       LocalHaloConstant = LocalHaloConstant->Next)
    if (!strcmp(LocalHaloConstant->Name, Name))
      return THaloID;
  for (LocalColorMapConstant = CurrentFrame->ColorMapConstantList ;
       LocalColorMapConstant ;
       LocalColorMapConstant = LocalColorMapConstant->Next)
    if (!strcmp(LocalColorMapConstant->Name, Name))
      return TColorMapID;
  for (LocalPigmentMapConstant = CurrentFrame->PigmentMapConstantList ;
       LocalPigmentMapConstant ;
       LocalPigmentMapConstant = LocalPigmentMapConstant->Next)
    if (!strcmp(LocalPigmentMapConstant->Name, Name))
      return TPigmentMapID;
  for (LocalSlopeMapConstant = CurrentFrame->SlopeMapConstantList;
       LocalSlopeMapConstant ;
       LocalSlopeMapConstant = LocalSlopeMapConstant->Next)
    if (!strcmp(LocalSlopeMapConstant->Name, Name))
      return TSlopeMapID;
  for (LocalNormalMapConstant = CurrentFrame->NormalMapConstantList;
       LocalNormalMapConstant ;
       LocalNormalMapConstant = LocalNormalMapConstant->Next)
    if (!strcmp(LocalNormalMapConstant->Name, Name))
      return TNormalMapID;
  for (LocalCameraConstant = CurrentFrame->CameraConstantList;
       LocalCameraConstant ;
       LocalCameraConstant = LocalCameraConstant->Next)
    if (!strcmp(LocalCameraConstant->Name, Name))
      return TCameraID;
  for (LocalLightSourceConstant = CurrentFrame->LightSourceConstantList;
       LocalLightSourceConstant ;
       LocalLightSourceConstant = LocalLightSourceConstant->Next)
    if (!strcmp(LocalLightSourceConstant->Name, Name))
      return TLightSourceID;
  for (LocalAtmosphereConstant = CurrentFrame->AtmosphereConstantList;
       LocalAtmosphereConstant ;
       LocalAtmosphereConstant = LocalAtmosphereConstant->Next)
    if (!strcmp(LocalAtmosphereConstant->Name, Name))
      return TAtmosphereID;
  for (LocalFogConstant = CurrentFrame->FogConstantList;
       LocalFogConstant ;
       LocalFogConstant = LocalFogConstant->Next)
    if (!strcmp(LocalFogConstant->Name, Name))
      return TFogID;
  for (LocalRainbowConstant = CurrentFrame->RainbowConstantList;
       LocalRainbowConstant ;
       LocalRainbowConstant = LocalRainbowConstant->Next)
    if (!strcmp(LocalRainbowConstant->Name, Name))
      return TRainbowID;
  for (LocalSkySphereConstant = CurrentFrame->SkySphereConstantList;
       LocalSkySphereConstant ;
       LocalSkySphereConstant = LocalSkySphereConstant->Next)
    if (!strcmp(LocalSkySphereConstant->Name, Name))
      return TSkySphereID;
  for (LocalTransformConstant = CurrentFrame->TransformConstantList;
       LocalTransformConstant ;
       LocalTransformConstant = LocalTransformConstant->Next)
    if (!strcmp(LocalTransformConstant->Name, Name))
      return TTransformID;
  return 0;
}

/*****************************************************************************
*  ParseDeclare
*  Parsing of "#declare Name = blah" sort of things 
******************************************************************************/
int ParseDeclare(void)
{
  char *ConstantName;
  int ConstantNameMaxLength, ConstantNameLength;
  int c,t,i;
  ParsedVector LocalExpress;
  Vector LocalColor;
  char *LocalString;
  ObjectStruct *LocalObject = NULL;
  TextureStruct *LocalTexture = NULL;
  PigmentStruct *LocalPigment = NULL;
  NormalStruct *LocalNormal = NULL;
  FinishStruct *LocalFinish = NULL;
  HaloStruct *LocalHalo = NULL;
  ColorMapStruct *LocalColorMap = NULL;
  PigmentMapStruct *LocalPigmentMap = NULL;
  SlopeMapStruct *LocalSlopeMap = NULL;
  NormalMapStruct *LocalNormalMap = NULL;
  CameraStruct *LocalCamera = NULL;
  LightSourceStruct *LocalLightSource = NULL;
  AtmosphereStruct *LocalAtmosphere = NULL;
  FogStruct *LocalFog = NULL;
  RainbowStruct *LocalRainbow = NULL;
  SkySphereStruct *LocalSkySphere = NULL;
  TransformStruct *LocalTransform = NULL;
  int PreviousTypeId;
  FloatConstantStruct *TmpFloatConstant;
  VectorConstantStruct *TmpVectorConstant;
  ColorConstantStruct *TmpColorConstant;
  StringConstantStruct *TmpStringConstant;
  ObjectConstantStruct *TmpObjectConstant;
  TextureConstantStruct *TmpTextureConstant;
  PigmentConstantStruct *TmpPigmentConstant;
  NormalConstantStruct *TmpNormalConstant;
  FinishConstantStruct *TmpFinishConstant;
  HaloConstantStruct *TmpHaloConstant;
  ColorMapConstantStruct *TmpColorMapConstant;
  PigmentMapConstantStruct *TmpPigmentMapConstant;
  SlopeMapConstantStruct *TmpSlopeMapConstant;
  NormalMapConstantStruct *TmpNormalMapConstant;
  CameraConstantStruct *TmpCameraConstant;
  LightSourceConstantStruct *TmpLightSourceConstant;
  AtmosphereConstantStruct *TmpAtmosphereConstant;
  FogConstantStruct *TmpFogConstant;
  RainbowConstantStruct *TmpRainbowConstant;
  SkySphereConstantStruct *TmpSkySphereConstant;
  TransformConstantStruct *TmpTransformConstant;
 
  /* First, get the Constant Name ([a-zA-Z0-9_]+) before the '=' */
  if (IgnoreSpace())
    return 1;
  ConstantNameMaxLength = 20;
  ConstantName = g_malloc(21);
  ConstantNameLength = 0;
  c = NextChar();
  while((c>='a' && c<='z') || (c>='A' && c<='Z') || 
        (c>='0' && c<='9') || c=='_')
  {
    ConstantName[ConstantNameLength++]=(char)c;
    if (ConstantNameLength==ConstantNameMaxLength)
    {
      ConstantNameMaxLength *= 2;
      ConstantName = g_realloc(ConstantName, ConstantNameMaxLength);
    }
    c = NextChar();
  }
  OldChar = c; IsOldChar = 1;
  ConstantName[ConstantNameLength] = 0;
  PreviousTypeId = GetTypeId(ConstantName);
  if (NextToken(&t))
    return 1;
  if (t != TEqual)
  {
    WarningBox(_("Error"), _("%s:%d : A '=' sign was expected here...\n"),
               PovFile->Name, PovFile->Line);
    return 1;
  }
  /* Get the first Token */
  if (NextToken(&t))
    return 1;
  OldToken = t; IsOldToken = 1;
  /* And guess the type of the declaration with the first token */
  switch (t)
  {
    case TFloatID: case TVectorID: case TPlus: case TMinus: case Tabs: 
    case Tacos: case Tval: case Tasc: case Tasin: case Tatan2: case Tceil:
    case Tclock: case Tcos: case Tdegrees: case Tdiv: case Texp: 
    case Tfile_exists: case TFloatLiteral: case Tfloor: case Tint: case Tlog:
    case Tmin: case Tmax: case Tmod: case Tpi: case Tpow: case Tradians:
    case Tsin: case Tsqrt: case Tstrcmp: case Tstrlen: case Ttan: case Tvdot:
    case Tvlength: case Tversion: case Ttrue: case Tyes: case Ton: case Tfalse:
    case Tno: case Toff: case Tseed: case Trand: case Tvaxis_rotate:
    case Tvcross: case Tvnormalize: case Tvrotate: case Tx: case Ty: case Tz:
    case Tt: case Tu: case Tv: case TExclamation: case TLeftParen:
    case TLeftAngleBracket:
      /* Expression Constant */
      if (ParseExpress(&LocalExpress))
        return 1;
      if (LocalExpress.nbTerms == 1)
      { /* Float */
        switch(PreviousTypeId)
        {
          case 0: /* New Float Constant */
            TmpFloatConstant = g_malloc(sizeof(FloatConstantStruct));
            TmpFloatConstant->Next = CurrentFrame->FloatConstantList;
            TmpFloatConstant->Name = g_malloc(strlen(ConstantName)+1);
            strcpy(TmpFloatConstant->Name, ConstantName);
            TmpFloatConstant->Float = LocalExpress.Val[0];
            CurrentFrame->FloatConstantList = TmpFloatConstant;
            break;
          case TFloatID: /* Redefinition of a Float Constant */
            for (TmpFloatConstant = CurrentFrame->FloatConstantList ;
                 TmpFloatConstant ;
                 TmpFloatConstant = TmpFloatConstant->Next)
              if (!strcmp(TmpFloatConstant->Name, ConstantName))
                TmpFloatConstant->Float = LocalExpress.Val[0];
            break;
          default: /* Redefinition of a non-Float Constant to a Float Constant */
            /* FIXME */
            WarningBox(_("Not Yet!"), _("Can't deals (yet) with redefinitions...(%s)\n"), ConstantName);
            return 1;
        }
      } else
      { /* Vector */
        switch (PreviousTypeId)
        {
          case 0: /* New Vector Constant */
            TmpVectorConstant = g_malloc(sizeof(VectorConstantStruct));
            TmpVectorConstant->Next = CurrentFrame->VectorConstantList;
            TmpVectorConstant->Name = g_malloc(strlen(ConstantName)+1);
            strcpy(TmpVectorConstant->Name, ConstantName);
            for (i=0; i < LocalExpress.nbTerms ; i++)
              TmpVectorConstant->Vector[i] = LocalExpress.Val[i];
            CurrentFrame->VectorConstantList = TmpVectorConstant;
          case TVectorID: /* Redefinition of a Vector Constant */
            for (TmpVectorConstant = CurrentFrame->VectorConstantList ;
                 TmpVectorConstant ;
                 TmpVectorConstant = TmpVectorConstant->Next)
              if (!strcmp(TmpVectorConstant->Name, ConstantName))
              {
                TmpVectorConstant->Vector[0] = LocalExpress.Val[0];
                TmpVectorConstant->Vector[1] = LocalExpress.Val[1];
                TmpVectorConstant->Vector[2] = LocalExpress.Val[2];
                TmpVectorConstant->Vector[3] = LocalExpress.Val[3];
                TmpVectorConstant->Vector[4] = LocalExpress.Val[4];
              }
            return 1;
            break;
          default: /* Redefinition of a non-vector constant to a vector constant */
            /* FIXME */
            WarningBox(_("Not Yet!"), _("Can't deals (yet) with redefinitions...(%s)\n"), ConstantName);
            return 1;
        }
      }
      break;
    case TColorID: case Tred: case Tgreen: case Tblue: case Tfilter: 
    case Ttransmit: case Tcolor: case Talpha: case Trgb: case Trgbt:
    case Trgbf: case Trgbft:
      /* Color Constant */
      if (ParseColor(LocalColor))
        return 1;
      switch(PreviousTypeId)
      {
        case 0: /* New Color Constant */
          TmpColorConstant = g_malloc(sizeof(ColorConstantStruct));
          TmpColorConstant->Next = CurrentFrame->ColorConstantList;
          TmpColorConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpColorConstant->Name, ConstantName);
          for (i=0 ; i<5 ; i++)
            TmpColorConstant->Color[i] = LocalColor[i];
          CurrentFrame->ColorConstantList = TmpColorConstant;
          break;
        case TColorID: /* Redefinition of a Color Constant */
          for (TmpColorConstant = CurrentFrame->ColorConstantList ;
               TmpColorConstant ;
               TmpColorConstant = TmpColorConstant->Next)
            if (!strcmp(TmpColorConstant->Name, ConstantName))
            {
              TmpColorConstant->Color[0] = LocalColor[0];
              TmpColorConstant->Color[1] = LocalColor[1];
              TmpColorConstant->Color[2] = LocalColor[2];
              TmpColorConstant->Color[3] = LocalColor[3];
              TmpColorConstant->Color[4] = LocalColor[4];
            }
          break;
        default: /* Redefinition of a non-Color Constant to a Color Constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deals (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TStringID: case Tstr: case Tconcat: case Tchr: case Tsubstr:
    case Tstrupr: case Tstrlwr: case TStringLiteral:
      /* String Constant */
      if (ParseString(&LocalString))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New String Constant */
          TmpStringConstant = g_malloc(sizeof(StringConstantStruct));
          TmpStringConstant->Next = CurrentFrame->StringConstantList;
          TmpStringConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpStringConstant->Name, ConstantName);
          TmpStringConstant->String = g_malloc(strlen(LocalString)+1);
          strcpy(TmpStringConstant->String, LocalString);
          CurrentFrame->StringConstantList = TmpStringConstant;
          break;
        case TStringID: /* Redefinition of a String Constant */
          for (TmpStringConstant = CurrentFrame->StringConstantList ;
               TmpStringConstant ;
               TmpStringConstant = TmpStringConstant->Next)
            if (!strcmp(TmpStringConstant->Name, ConstantName))
            {
              g_free(TmpStringConstant->String);
              TmpStringConstant->String = g_strdup(LocalString);
            }
          break;
        default: /* Redefinition of a non-String Constant to a String Constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deals (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TObjectID: case Tobject: case Tbicubic_patch: case Tblob: case Tbox:
    case Tcone: case Tcubic: case Tcylinder: case Tdifference: case Tdisc:
    case Theight_field: case Tintersection: case Tjulia_fractal: case Tlathe:
    case Tmerge: case Tmesh: case Tplane: case Tpoly: case Tpolygon:
    case Tprism: case Tquadric: case Tquartic: case Tsmooth_triangle: case Tsor:
    case Tsphere: case Tsuperellipsoid: case Ttext: case Ttorus: case Ttriangle:
    case Tunion:
      /* Object Constant */
      if (ParseObject(&LocalObject))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New Object Constant */
          TmpObjectConstant = g_malloc(sizeof(ObjectConstantStruct));
          TmpObjectConstant->Next = CurrentFrame->ObjectConstantList;
          TmpObjectConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpObjectConstant->Name, ConstantName);
          TmpObjectConstant->Object = CopyObject(LocalObject);
          CurrentFrame->ObjectConstantList = TmpObjectConstant;
          break;
        case TObjectID: /* Redefinition of an object constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with object redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-Object Constant to an object constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TTextureID: case Ttexture:
      /* Texture Constant */
      IsOldToken = 0;
      if (ParseTexture(&LocalTexture))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New Texture Constant */
          TmpTextureConstant = g_malloc(sizeof(TextureConstantStruct));
          TmpTextureConstant->Next = CurrentFrame->TextureConstantList;
          TmpTextureConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpTextureConstant->Name, ConstantName);
          TmpTextureConstant->Texture = CopyTexture(LocalTexture);
          CurrentFrame->TextureConstantList = TmpTextureConstant;
          break;
        case TTextureID: /* Redefinition of a texture constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with texture redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-Texture Constant to a texture constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TPigmentID: case Tpigment:
      /* Pigment Constant */
      IsOldToken = 0;
      if (ParsePigment(&LocalPigment))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New Pigment Constant */
          TmpPigmentConstant = g_malloc(sizeof(PigmentConstantStruct));
          TmpPigmentConstant->Next = CurrentFrame->PigmentConstantList;
          TmpPigmentConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpPigmentConstant->Name, ConstantName);
          TmpPigmentConstant->Pigment = CopyPigment(LocalPigment);
          CurrentFrame->PigmentConstantList = TmpPigmentConstant;
          break;
        case TPigmentID: /* Redefinition of a pigment constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with pigment redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-pigment Constant to a pigment constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TNormalID: case Tnormal:
      /* Normal Constant */
      IsOldToken = 0;
      if (ParseNormal(&LocalNormal))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New Normal Constant */
          TmpNormalConstant = g_malloc(sizeof(NormalConstantStruct));
          TmpNormalConstant->Next = CurrentFrame->NormalConstantList;
          TmpNormalConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpNormalConstant->Name, ConstantName);
          TmpNormalConstant->Normal = CopyNormal(LocalNormal);
          CurrentFrame->NormalConstantList = TmpNormalConstant;
          break;
        case TNormalID: /* Redefinition of a normal constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with normal redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-normal constant to a normal constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TFinishID: case Tfinish:
      /* Finish Constant */
      IsOldToken = 0;
      if (ParseFinish(&LocalFinish))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New finish constant */
          TmpFinishConstant = g_malloc(sizeof(FinishConstantStruct));
          TmpFinishConstant->Next = CurrentFrame->FinishConstantList;
          TmpFinishConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpFinishConstant->Name, ConstantName);
          TmpFinishConstant->Finish = CopyFinish(LocalFinish);
          CurrentFrame->FinishConstantList = TmpFinishConstant;
          break;
        case TFinishID: /* Redefinition of a finish constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with finish redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-finish constant to a finish constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case THaloID: case Thalo:
      /* Halo Constant */
      IsOldToken = 0;
      if (ParseHalo(&LocalHalo))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New halo constant */
          TmpHaloConstant = g_malloc(sizeof(HaloConstantStruct));
          TmpHaloConstant->Next = CurrentFrame->HaloConstantList;
          TmpHaloConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpHaloConstant->Name, ConstantName);
          TmpHaloConstant->Halo = CopyHalo(LocalHalo);
          CurrentFrame->HaloConstantList = TmpHaloConstant;
          break;
        case THaloID: /* Redefinition of a halo constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with halo redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-halo constant to a halo constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TColorMapID: case Tcolor_map:
      /* Color Map Constant */
      IsOldToken = 0;
      if (ParseColorMap(&LocalColorMap))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New colormap constant */
          TmpColorMapConstant = g_malloc(sizeof(ColorMapConstantStruct));
          TmpColorMapConstant->Next = CurrentFrame->ColorMapConstantList;
          TmpColorMapConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpColorMapConstant->Name, ConstantName);
          TmpColorMapConstant->ColorMap = CopyColorMap(LocalColorMap);
          CurrentFrame->ColorMapConstantList = TmpColorMapConstant;
          break;
        case TColorMapID: /* Redefinition of a colormap constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with colormap redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-colormap constant to a colormap constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TPigmentMapID: case Tpigment_map:
      /* Pigment Map Constant */
      IsOldToken = 0;
      if (ParsePigmentMap(&LocalPigmentMap))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New pigmentmap constant */
          TmpPigmentMapConstant = g_malloc(sizeof(PigmentMapConstantStruct));
          TmpPigmentMapConstant->Next = CurrentFrame->PigmentMapConstantList;
          TmpPigmentMapConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpPigmentMapConstant->Name, ConstantName);
          TmpPigmentMapConstant->PigmentMap = CopyPigmentMap(LocalPigmentMap);
          CurrentFrame->PigmentMapConstantList = TmpPigmentMapConstant;
          break;
        case TPigmentMapID: /* Redefinition of a pigmentmap constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with pigment map redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-pigmentmap constant to a pigmentmap constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TSlopeMapID: case Tslope_map:
      /* Slope Map Constant */
      IsOldToken = 0;
      if (ParseSlopeMap(&LocalSlopeMap))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New slopemap constant */
          TmpSlopeMapConstant = g_malloc(sizeof(SlopeMapConstantStruct));
          TmpSlopeMapConstant->Next = CurrentFrame->SlopeMapConstantList;
          TmpSlopeMapConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpSlopeMapConstant->Name, ConstantName);
          TmpSlopeMapConstant->SlopeMap = CopySlopeMap(LocalSlopeMap);
          CurrentFrame->SlopeMapConstantList = TmpSlopeMapConstant;
          break;
        case TSlopeMapID: /* Redefinition of a slopemap constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with slope map redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-slopemap constant to a slopemap constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TNormalMapID: case Tnormal_map:
      /* Normal Map Constant */
      IsOldToken = 0;
      if (ParseNormalMap(&LocalNormalMap))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New normalmap constant */
          TmpNormalMapConstant = g_malloc(sizeof(NormalMapConstantStruct));
          TmpNormalMapConstant->Next = CurrentFrame->NormalMapConstantList;
          TmpNormalMapConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpNormalMapConstant->Name, ConstantName);
          TmpNormalMapConstant->NormalMap = CopyNormalMap(LocalNormalMap);
          CurrentFrame->NormalMapConstantList = TmpNormalMapConstant;
          break;
        case TNormalMapID: /* Redefinition of a normalmap constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with normal map redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-normalmap constant to a normalmap constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TCameraID: case Tcamera:
      /* Camera Constant */
      IsOldToken = 0;
      if (ParseCamera(&LocalCamera))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New camera constant */
          TmpCameraConstant = g_malloc(sizeof(CameraConstantStruct));
          TmpCameraConstant->Next = CurrentFrame->CameraConstantList;
          TmpCameraConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpCameraConstant->Name, ConstantName);
          TmpCameraConstant->Camera = CopyCamera(LocalCamera);
          CurrentFrame->CameraConstantList = TmpCameraConstant;
          break;
        case TCameraID: /* Redefinition of a camera constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with camera redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-camera constant to a camera constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TLightSourceID: case Tlight_source:
      /* Light Source Constant */
      IsOldToken = 0;
      if (ParseLightSource(&LocalLightSource))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New lightsource constant */
          TmpLightSourceConstant = g_malloc(sizeof(LightSourceConstantStruct));
          TmpLightSourceConstant->Next = CurrentFrame->LightSourceConstantList;
          TmpLightSourceConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpLightSourceConstant->Name, ConstantName);
          TmpLightSourceConstant->LightSource = CopyLightSource(LocalLightSource);
          CurrentFrame->LightSourceConstantList = TmpLightSourceConstant;
          break;
        case TLightSourceID: /* Redefinition of a lightsource constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with light source redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-lightsource constant to a lightsource constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TAtmosphereID: case Tatmosphere:
      /* Atmosphere Constant */
      IsOldToken = 0;
      if (ParseAtmosphere(&LocalAtmosphere))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New atmosphere constant */
          TmpAtmosphereConstant = g_malloc(sizeof(AtmosphereConstantStruct));
          TmpAtmosphereConstant->Next = CurrentFrame->AtmosphereConstantList;
          TmpAtmosphereConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpAtmosphereConstant->Name, ConstantName);
          TmpAtmosphereConstant->Atmosphere = CopyAtmosphere(LocalAtmosphere);
          CurrentFrame->AtmosphereConstantList = TmpAtmosphereConstant;
          break;
        case TAtmosphereID: /* Redefinition of a atmosphere constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with atmosphere redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-atmosphere constant to a atmosphere constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TFogID: case Tfog:
      /* Fog Constant */
      IsOldToken = 0;
      if (ParseFog(&LocalFog))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New fog constant */
          TmpFogConstant = g_malloc(sizeof(FogConstantStruct));
          TmpFogConstant->Next = CurrentFrame->FogConstantList;
          TmpFogConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpFogConstant->Name, ConstantName);
          TmpFogConstant->Fog = CopyFog(LocalFog);
          CurrentFrame->FogConstantList = TmpFogConstant;
          break;
        case TFogID: /* Redefinition of a fog constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with fog redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-fog constant to a fog constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TRainbowID: case Trainbow:
      /* Rainbow Constant */
      IsOldToken = 0;
      if (ParseRainbow(&LocalRainbow))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New atmosphere constant */
          TmpRainbowConstant = g_malloc(sizeof(RainbowConstantStruct));
          TmpRainbowConstant->Next = CurrentFrame->RainbowConstantList;
          TmpRainbowConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpRainbowConstant->Name, ConstantName);
          TmpRainbowConstant->Rainbow = CopyRainbow(LocalRainbow);
          CurrentFrame->RainbowConstantList = TmpRainbowConstant;
          break;
        case TRainbowID: /* Redefinition of a rainbow constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with rainbow redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-rainbow constant to a rainbow constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TSkySphereID: case Tsky_sphere:
      /* Sky Sphere Constant */
      IsOldToken = 0;
      if (ParseSkySphere(&LocalSkySphere))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New skysphere constant */
          TmpSkySphereConstant = g_malloc(sizeof(SkySphereConstantStruct));
          TmpSkySphereConstant->Next = CurrentFrame->SkySphereConstantList;
          TmpSkySphereConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpSkySphereConstant->Name, ConstantName);
          TmpSkySphereConstant->SkySphere = CopySkySphere(LocalSkySphere);
          CurrentFrame->SkySphereConstantList = TmpSkySphereConstant;
          break;
        case TSkySphereID: /* Redefinition of a skysphere constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with skysphere redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-skysphere constant to a skysphere constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    case TTransformID: case Ttransform:
      /* Transform Constant */
      IsOldToken = 0;
      if (ParseTransform(&LocalTransform))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New transform constant */
          TmpTransformConstant = g_malloc(sizeof(TransformConstantStruct));
          TmpTransformConstant->Next = CurrentFrame->TransformConstantList;
          TmpTransformConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpTransformConstant->Name, ConstantName);
          TmpTransformConstant->Transform = CopyTransform(LocalTransform);
          CurrentFrame->TransformConstantList = TmpTransformConstant;
          break;
        case TTransformID: /* Redefinition of a transform constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with transform redefinitions...(%s)\n"), ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-transform constant to a transform constant */
          /* FIXME */
          WarningBox(_("Not Yet!"), _("Can't deal (yet) with redefinitions...(%s)\n"), ConstantName);
          return 1;
      }
      break;
    default:
      WarningBox(_("Error"), _("%s:%d : Parse Error...\n"), PovFile->Name, PovFile->Line);
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseDirective
******************************************************************************/
int ParseDirective(void)
{
  char *Directive;
  int DirectiveLength, DirectiveMaxLength;
  int c;
  char *FileName;
  ParsedVector LocalExpress;
  CondDirective *TmpCond; 
  int LocalToken;
  char *LocalString;
 
  if (IgnoreSpace()) /* Are Spaces really allowed between Hash and Directive ?!? */
    return 1;
  ParsingDirective = 1;
  DirectiveMaxLength = 20;
  Directive = g_malloc(21);
  DirectiveLength = 0;
  c=NextChar();
  while(c>='a' && c<='z')
  {
    Directive[DirectiveLength++]=(char)c;
    if (DirectiveLength==DirectiveMaxLength)
    {
      DirectiveMaxLength *= 2;
      Directive = g_realloc(Directive, DirectiveMaxLength);
    }
    c=NextChar();
  }
  OldChar = c; IsOldChar = 1;
  Directive[DirectiveLength] = 0;

  if (!strcmp(Directive,"version"))
  {
    if (ParseFloat(&PovVersionLangage))
      return 1;
  } else if (!strcmp(Directive, "include"))
  {
    if (ParseString(&FileName))
      return 1;
    PovFile->Next = g_malloc(sizeof(ListFile));
    PovFile->Next->Prev = PovFile;
    PovFile = PovFile->Next;
    PovFile->Next = NULL;
    PovFile->handle = LibOpen(FileName);
    if (PovFile->handle == NULL)
    {
      WarningBox(_("Error"), _("%s:%d : Can't open file \"%s\"...\n"),
                 PovFile->Prev->Name, PovFile->Prev->Line, FileName);
      return 1;
    }
    PovFile->Name = FileName;
    PovFile->Line = 1;
  } else if (!strcmp(Directive, "ifdef"))
  {
    TmpCond = g_malloc(sizeof(CondDirective));
    TmpCond->Next = ConditionalStack;
    ConditionalStack = TmpCond;
    if (DontParse)
    {
      ConditionalStack->Style = UNPARSED;
    } else
    {
      if (ParenError("#ifdef", TLeftParen))
        return 1;
      LocalToken = 0;
      if (ParseWord(&LocalToken))
        return 1;
      if (ParenError("#ifdef", TRightParen))
        return 1;
      if (LocalToken)
      {
        ConditionalStack->Style = IF_TRUE_IF_ELSE;
      } else
      {
        ConditionalStack->Style = IF_FALSE_IF_ELSE;
        DontParse = TRUE;
      }
    }
  } else if (!strcmp(Directive, "ifndef"))
  {
    TmpCond = g_malloc(sizeof(CondDirective));
    TmpCond->Next = ConditionalStack;
    ConditionalStack = TmpCond;
    if (DontParse)
    {
      ConditionalStack->Style = UNPARSED;
    } else
    {
      if (ParenError("#ifndef", TLeftParen))
        return 1;
      LocalToken = 0;
      if (ParseWord(&LocalToken))
        return 1;
      if (ParenError("#ifndef", TRightParen))
        return 1;
      if (!LocalToken)
      {
        ConditionalStack->Style = IF_TRUE_IF_ELSE;
      } else
      {
        ConditionalStack->Style = IF_FALSE_IF_ELSE;
        DontParse = TRUE;
      }
    }
  } else if (!strcmp(Directive, "if"))
  {
    TmpCond = g_malloc(sizeof(CondDirective));
    TmpCond->Next = ConditionalStack;
    ConditionalStack = TmpCond;
    if (DontParse)
    {
      ConditionalStack->Style = UNPARSED;
    } else
    {
      if (ParenError("#if", TLeftParen))
        return 1;
      if (ParseExpress(&LocalExpress))
        return 1;
      if (ParenError("#if", TRightParen))
        return 1;
      if (LocalExpress.nbTerms != 1)
      {
        WarningBox(_("Error"), _("%s:%d : Conditions must evaluate to a float...\n"),
                   PovFile->Name, PovFile->Line);
        return 1;
      }
      if (fabs(LocalExpress.Val[0] > SMALL_VALUE))
      {
        ConditionalStack->Style = IF_TRUE_IF_ELSE;
      } else
      {
        ConditionalStack->Style = IF_FALSE_IF_ELSE;
        DontParse = TRUE;
      }
    }
  } else if (!strcmp(Directive, "else"))
  {
    if (!ConditionalStack)
    {
      WarningBox(_("Error"), _("%s:%d : Misplaced #else...\n"), PovFile->Name, PovFile->Line);
      return 1;
    }
    switch (ConditionalStack->Style)
    {
      case IF_TRUE_IF_ELSE :
        DontParse = TRUE;
        ConditionalStack->Style = IF_TRUE_ELSE_END;
        break;
      case IF_TRUE_ELSE_END :
      case IF_FALSE_ELSE_END :
      case WHILE :
      case SWITCH_SWITCH_CASE :
        WarningBox(_("Error"), _("%s:%d : Misplaced #else...\n"), PovFile->Name, PovFile->Line);
        return 1;
      case IF_FALSE_IF_ELSE :
      case SWITCH_CASE_BREAK_FALSE :
        DontParse = FALSE;
        ConditionalStack->Style = IF_FALSE_ELSE_END;
        break;
      case SWITCH_CASE_BREAK_TRUE :
        DontParse = TRUE;
        ConditionalStack->Style = SWITCH_BREAK_END;
        break;
      case SWITCH_BREAK_END:
      case UNPARSED:
        /* Nothing */
        break;
    }
  } else if (!strcmp(Directive,"end"))
  {
    if (!ConditionalStack)
    {
      WarningBox(_("Error"), _("%s:%d : Misplaced #end...\n"), PovFile->Name, PovFile->Line);
      return 1;
    }
    switch (ConditionalStack->Style)
    {
      case IF_TRUE_IF_ELSE:
      case IF_TRUE_ELSE_END:
      case IF_FALSE_IF_ELSE:
      case IF_FALSE_ELSE_END:
      case SWITCH_BREAK_END:
      case SWITCH_CASE_BREAK_TRUE:
      case SWITCH_CASE_BREAK_FALSE:
      case SWITCH_SWITCH_CASE:
        TmpCond = ConditionalStack->Next;
        g_free(ConditionalStack);
        ConditionalStack = TmpCond;
        DontParse = FALSE;
        break;
      case WHILE:
        if (ConditionalStack->WhileFile != PovFile)
        {
          WarningBox(_("Error"), _("%s:%d : #while loop didn't end in file where it started...\n"),
                     PovFile->Name, PovFile->Line);
          return 1;
        }
        fseek(PovFile->handle, ConditionalStack->WhilePos, SEEK_SET);
        PovFile->Line = ConditionalStack->WhileLineNo;
        if (ParenError("#while", TLeftParen))
          return 1;
        if (ParseExpress(&LocalExpress))
          return 1;
        if (ParenError("#while", TRightParen))
          return 1;
        if (fabs(LocalExpress.Val[0]) > SMALL_VALUE)
        {
          /* Nothing ! ? */
        } else
        {
          DontParse = TRUE;
          ConditionalStack->Style = IF_TRUE_ELSE_END;
        }
        break;
      case UNPARSED:
        TmpCond = ConditionalStack->Next;
        g_free(ConditionalStack);
        ConditionalStack = TmpCond;
        break;
    }
  } else if (!strcmp(Directive, "declare"))
  {
    if (ParseDeclare())
      return 1;
  } else if (!strcmp(Directive, "debug"))
  {
    if (!DontParse)
    {
      if (ParseString(&LocalString))
        return 1;
      printf(LocalString); /* FIXME Should be redirected on the debug stream */
      g_free(LocalString);
    }
  } else if (!strcmp(Directive, "error"))
  {
    if (ParseString(&LocalString))
      return 1;
    if (!DontParse)
    {
      WarningBox(_("Error"), LocalString);
      return 1;
    }
    g_free(LocalString);
  } else if (!strcmp(Directive, "render"))
  {
    if (ParseString(&LocalString))
      return 1;
    if (!DontParse)
    {
      printf(LocalString); /* FIXME */
    }
    g_free(LocalString);
  } else if (!strcmp(Directive, "statistics"))
  {
    if (ParseString(&LocalString))
      return 1;
    if (!DontParse)
    {
      printf(LocalString); /* FIXME */
    }
    g_free(LocalString);
  } else if (!strcmp(Directive, "warning"))
  {
    if (ParseString(&LocalString))
      return 1;
    if (!DontParse)
    {
      printf(LocalString); /* FIXME */
    }
    g_free(LocalString);
  } else 
  {
    WarningBox(_("Error"), _("%s:%d : Unknown directive : %s\n"),
               PovFile->Name, PovFile->Line, Directive);
    return 1;
  }
  g_free(Directive);
  ParsingDirective = 0;
  return 0;
}

/*****************************************************************************
*  NextToken
******************************************************************************/
int NextToken(int *pToken)
{
  int c, c1, c2;

  *pToken = 0;
  while (*pToken == 0)
  {
    if (IsOldToken)
    {
      IsOldToken = 0;
      if (OldToken != -1)
      {
        *pToken = OldToken;
        return 0; 
      }  
    }
    if (IgnoreSpace())
      return 1;
    c = NextChar();
    switch (c)
    {
      case EOF :
        if (PovFile->Prev)
        {
          fclose(PovFile->handle);
          g_free(PovFile->Name);
          PovFile = PovFile->Prev;
          g_free(PovFile->Next);
        } else *pToken = TEOF;
        break;
      case '!' :
        c2 = NextChar();
        if (c2 == '=')
        {
          *pToken = TUnequal; 
        } else
        {
          *pToken = TExclamation;
          IsOldChar = 1 ; OldChar = c2;
        }
        break;
      case '"' :
        if (ParseStringLiteral(&LastString))
          return 1;
        *pToken = TStringLiteral;
        if (DontParse)
          g_free(LastString);
        break;
      case '#' :
        if (ParsingDirective)
        {
          IsOldChar = 1; OldChar = c;  
          *pToken = -1;
        } else
        {
          if (ParseDirective())
            return 1;
          if (OldToken == -1)
            IsOldToken = 0;
        }
        break;
      case '&' :
        *pToken = TAmpersand;
        break;
      case '(' :
        *pToken = TLeftParen;
        break;
      case ')' :
        *pToken = TRightParen;
        break;
      case '*' :
        *pToken = TStar;
        break;
      case '+' :
        *pToken = TPlus;
        break;
      case ',' :
        *pToken = TComma;
        break;
      case '-' :
        *pToken = TMinus;
        break;
      case '0' ... '9' :
      case '.' :
        OldChar = c; IsOldChar = 1;
        if (ParseFloatLiteral(&LastFloat))
          return 1;
        *pToken = TFloatLiteral;
        break;
      case '/' :
        c = NextChar();
        if (c == '/')
        { /* comments ala C++ */
          while (c!='\r' && c!='\n')
          {
            c=NextChar();
          }
          if (c=='\r')
          {
            c=NextChar();
            if (c=='\n') c=NextChar();
          }
          if (c=='\n')
          {
            c=NextChar();
            if (c=='\r') c=NextChar();
          }
          PovFile->Line++;
          OldChar = c; IsOldChar = 1;
          break;
        }
        if (c == '*')
        { /* comments ala C */
          c1 = c2 = NextChar();
          do
          {
            if (c2=='\n')
            {
              PovFile->Line++;
              c2 = NextChar();
              if (c2=='\r') c2 = NextChar();
            } else if (c2=='\r')
            {
              PovFile->Line++;
              c2 = NextChar();
              if (c2=='\n') c2 = NextChar();
            } else
            {
              c1 = c2;
              c2 = NextChar();
            }
          } while (c1!='*' || c2!='/');
          break;
        }
        OldChar = c; IsOldChar = 1;
        *pToken = TSlash;
        break;
      case ':' :
        *pToken = TColon;
        break;
      case '<' :
        c2 = NextChar();
        if (c2 == '=')
        {
          *pToken = TLowerOrEqual;
        } else
        {
          IsOldChar = 1; OldChar = c2;
          *pToken = TLeftAngleBracket;
        }
        break;
      case '=' :
        *pToken = TEqual;
        break;
      case '>' :
        c2 = NextChar();
        if (c2 == '=')
        {
          *pToken = TUpperOrEqual;
        } else
        {
          IsOldChar = 1; OldChar = c2;
          *pToken = TRightAngleBracket;
        }
        break;
      case '?' :
        *pToken = TQuestion;
        break;
      case 'A' ... 'Z' :
      case 'a' ... 'z' :
      case '_' :
        OldChar = c; IsOldChar = 1;
        if (!DontParse)
          *pToken = -1;
        if (ParseWord(pToken))
          return 1;
        break;
      case '{' :
        *pToken = TLeftCurlyBracket;
        break;
      case '|' :
        *pToken = TBar;
        break;
      case '}' :
        *pToken = TRightCurlyBracket;
        break;
      case '[' :
        *pToken = TLeftSquareBracket;
        break;
      case ']' :
        *pToken = TRightSquareBracket;
        break;
      default :
        WarningBox(_("Error"), _("Unexpected character : '%c'\n"), c);
        return 1;
    }
    if (DontParse)
      *pToken = 0;
  }
  return 0;
}

/*****************************************************************************
*  ParseColor
******************************************************************************/
int ParseColor(Vector LocalColor)
{
  int t, i, Done;
  ParsedVector TmpColor;
  
  if (NextToken(&t))
    return 1;
  if (t != Tcolor)
  {
    IsOldToken = 1; OldToken = t;
  }
  Done = FALSE;
  TmpColor.nbTerms = 5;
  for (i=0 ; i<5 ; i++)
    LocalColor[i] = TmpColor.Val[i]=0.0;
  while (!Done)
  {
    if (NextToken(&t))
      return 1;
    switch (t)
    {
      case Tred:
        if (ParseFloat(&(TmpColor.Val[0])))
          return 1;
        LocalColor[0] = TmpColor.Val[0];
        break;
      case Tgreen:
        if (ParseFloat(&(TmpColor.Val[1])))
          return 1;
        LocalColor[1] = TmpColor.Val[1];
        break;
      case Tblue:
        if (ParseFloat(&(TmpColor.Val[2])))
          return 1;
        LocalColor[2] = TmpColor.Val[2];
        break;
      case Talpha: case Tfilter:
        if (ParseFloat(&(TmpColor.Val[3])))
          return 1;
        LocalColor[3] = TmpColor.Val[3];
        break;
      case Ttransmit:
        if (ParseFloat(&(TmpColor.Val[4])))
          return 1;
        LocalColor[4] = TmpColor.Val[4];
        break;
      case Trgb:
        if (ParseExpress(&TmpColor))
          return 1;
        if (TmpColor.nbTerms == 1)
        {
          TmpColor.Val[2] = TmpColor.Val[1] = TmpColor.Val[0];
        } else if (TmpColor.nbTerms < 3)
        {
          for (i=TmpColor.nbTerms ; i<3 ; i++)
            TmpColor.Val[i] = 0.0;
        }
        for (i=0 ; i<3 ; i++)
          LocalColor[i] = TmpColor.Val[i];
        LocalColor[3] = LocalColor[4] = 0.0;
        break;
      case Trgbf:
        if (ParseExpress(&TmpColor))
          return 1;
        if (TmpColor.nbTerms == 1)
        {
          TmpColor.Val[3] = TmpColor.Val[2] = 
                            TmpColor.Val[1] = 
                            TmpColor.Val[0];
        } else if (TmpColor.nbTerms < 4)
        {
          for (i=TmpColor.nbTerms ; i<4 ; i++)
            TmpColor.Val[i] = 0.0;
        }
        for (i=0 ; i<4 ; i++)
          LocalColor[i] = TmpColor.Val[i];
        LocalColor[4] = 0.0;
        break;
      case Trgbft:
        if (ParseExpress(&TmpColor))
          return 1;
        if (TmpColor.nbTerms == 1)
        {
          TmpColor.Val[4] = TmpColor.Val[3] = TmpColor.Val[2] = 
                            TmpColor.Val[1] = TmpColor.Val[0];
        } else if (TmpColor.nbTerms < 5)
        {
          for (i=TmpColor.nbTerms ; i<5 ; i++)
            TmpColor.Val[i] = 0.0;
        }
        for (i=0 ; i<5 ; i++)
          LocalColor[i] = TmpColor.Val[i];
        break;
      case Trgbt:
        if (ParseExpress(&TmpColor))
          return 1;
        if (TmpColor.nbTerms == 1)
        {
          TmpColor.Val[3] = TmpColor.Val[2] = 
                            TmpColor.Val[1] = 
                            TmpColor.Val[0];
        } else if (TmpColor.nbTerms < 4)
        {
          for (i=TmpColor.nbTerms ; i<4 ; i++)
            TmpColor.Val[i] = 0.0;
        }
        for (i=0 ; i<3 ; i++)
          LocalColor[i] = TmpColor.Val[i];
        LocalColor[3] = 0.0;
        LocalColor[4] = TmpColor.Val[3];
        break;
      case TPlus: case TMinus:
      case Tabs: case Tacos: case Tval: case Tasc: case Tasin: case Tatan2:
      case Tceil: case Tclock: case Tcos: case Tdegrees: case Tdiv: 
      case Texp: case Tfile_exists: case TFloatID: case TFloatLiteral:
      case Tfloor: case Tint: case Tlog: case Tmin: case Tmax: case Tmod:
      case Tpi: case Tpow: case Tradians: case Tsin: case Tsqrt: 
      case Tstrcmp: case Tstrlen: case Ttan: case Tvdot: case Tvlength:
      case Tversion: case Ttrue: case Tyes: case Ton: case Tfalse: 
      case Tno: case Toff: case Tseed: case Trand: case Tvaxis_rotate:
      case Tvcross: case TVectorID: case Tvnormalize: case Tvrotate:
      case Tx: case Ty: case Tz: case TColorID: case Tt: case Tu: case Tv:
      case TExclamation: case TLeftParen: case TLeftAngleBracket:
        OldToken = t; IsOldToken = 1;
        if (ParseExpress(&TmpColor))
          return 1;
        if (TmpColor.nbTerms == 1)
        {
          TmpColor.Val[4] = TmpColor.Val[3] = TmpColor.Val[2] = 
                            TmpColor.Val[1] = TmpColor.Val[0];
        } else if (TmpColor.nbTerms < 5)
        {
          for (i=TmpColor.nbTerms ; i<5 ; i++)
            TmpColor.Val[i] = 0.0;
        }
        for (i=0 ; i<5 ; i++)
          LocalColor[i] = TmpColor.Val[i];
        break;
      default:
        IsOldToken = 1; OldToken = t;
        Done = TRUE;
    }
  }
  return 0;
}

/*****************************************************************************
*  ParsePigmentOrColor
******************************************************************************/
int ParsePigmentOrColor(PigmentStruct **pLocalPigment)
{
  int t;
  
  if (NextToken(&t))
    return 1;
  if (t == Tpigment)
  {
    if (ParsePigment(pLocalPigment))
      return 1;
  } else
  {
    (*pLocalPigment)->Type = PAT_SOLID_COLOR;
    (*pLocalPigment)->Trans = NULL;
    (*pLocalPigment)->PigmentList = NULL;
    if (ParseColor((*pLocalPigment)->Color))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseColorMap
******************************************************************************/
int ParseColorMap(ColorMapStruct **pLocalColorMap)
{
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (NextToken(&t))
    return 1;
  if (t == TColorMapID)
  {
    *pLocalColorMap = CopyColorMap(LastColorMapConstant->ColorMap);
    if (NextToken(&t))
      return 1;
    if (t != TRightCurlyBracket)
    {
      WarningBox(_("Error"), _("%s:%d a '}' was expected here...\n"), PovFile->Name, PovFile->Line);
      return 1;
    }
    return 0;
  } else
  {
    if (*pLocalColorMap)
    {
      g_free((*pLocalColorMap)->Entries);
    } else
    {
      *pLocalColorMap = g_malloc(sizeof(ColorMapStruct));
    }
    (*pLocalColorMap)->NbMapEntries = 0;
    (*pLocalColorMap)->Entries = NULL;
    OldToken = t; IsOldToken = 1;
  }
  if (NextToken(&t))
    return 1;
  while (t==TLeftSquareBracket)
  {
    if ((*pLocalColorMap)->Entries == NULL)
    {
      (*pLocalColorMap)->Entries = g_malloc(sizeof(ColorMapEntry));
    } else
    {
      (*pLocalColorMap)->Entries = g_realloc((*pLocalColorMap)->Entries,
                                         ((*pLocalColorMap)->NbMapEntries+1)*
                                         sizeof(ColorMapEntry));
    }
    if (ParseFloat(&((*pLocalColorMap)->Entries[(*pLocalColorMap)->NbMapEntries].Value)))
      return 1;
    if (ParseComma())
      return 1;
    if (NextToken(&t))
      return 1;
    OldToken = t; IsOldToken = 1;
    switch (t)
    {
      case TColorID: case Tred: case Tgreen: case Tblue: case Tfilter: 
      case Ttransmit: case Tcolor: case Talpha: case Trgb: case Trgbt:
      case Trgbf: case Trgbft: /* Normal Syntax */
        if (ParseColor((*pLocalColorMap)->Entries[(*pLocalColorMap)->NbMapEntries].Color))
          return 1;
        (*pLocalColorMap)->NbMapEntries++;
        break;
      case TFloatID: case TVectorID: case TPlus: case TMinus: case Tabs: 
      case Tacos: case Tval: case Tasc: case Tasin: case Tatan2: case Tceil:
      case Tclock: case Tcos: case Tdegrees: case Tdiv: case Texp: 
      case Tfile_exists: case TFloatLiteral: case Tfloor: case Tint: case Tlog:
      case Tmin: case Tmax: case Tmod: case Tpi: case Tpow: case Tradians:
      case Tsin: case Tsqrt: case Tstrcmp: case Tstrlen: case Ttan: case Tvdot:
      case Tvlength: case Tversion: case Ttrue: case Tyes: case Ton: case Tfalse:
      case Tno: case Toff: case Tseed: case Trand: case Tvaxis_rotate:
      case Tvcross: case Tvnormalize: case Tvrotate: case Tx: case Ty: case Tz:
      case Tt: case Tu: case Tv: case TExclamation: case TLeftParen:
      case TLeftAngleBracket: /* old syntax */
        (*pLocalColorMap)->Entries = g_realloc((*pLocalColorMap)->Entries,
                                           ((*pLocalColorMap)->NbMapEntries+2)*
                                            sizeof(ColorMapEntry));
        if (ParseFloat(&((*pLocalColorMap)->Entries[(*pLocalColorMap)->NbMapEntries+1].Value)))
          return 1;
        if (ParseComma())
          return 1;
        if (ParseColor((*pLocalColorMap)->Entries[(*pLocalColorMap)->NbMapEntries].Color))
          return 1;
        if (ParseComma())
          return 1;
        if (ParseColor((*pLocalColorMap)->Entries[(*pLocalColorMap)->NbMapEntries+1].Color))
          return 1;
        (*pLocalColorMap)->NbMapEntries += 2;
       break;
    }
    if (NextToken(&t))
      return 1;
    if (t != TRightSquareBracket)
    {
      WarningBox(_("Error"), _("%s:%d a ']' was expected here...\n"), PovFile->Name, PovFile->Line);
      return 1;
    }
    if (NextToken(&t))
      return 1;
  }
  if (t != TRightCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '}' was expected here...\n"), PovFile->Name, PovFile->Line);
    return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseMatrix
******************************************************************************/
int ParseMatrix(TransformStruct *pTrans)
{
  int t;

  pTrans->Direct[0][3] = 0.0;
  pTrans->Direct[1][3] = 0.0;
  pTrans->Direct[2][3] = 0.0;
  pTrans->Direct[3][3] = 1.0;
  if (NextToken(&t))
    return 1;
  if (t != TLeftAngleBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '<' was expected here...\n"), PovFile->Name, PovFile->Line);
    return 1;
  }
  if (ParseFloat(&(pTrans->Direct[0][0])))
    return 1;
  if (ParseComma()) return 1;
  if (ParseFloat(&(pTrans->Direct[0][1])))
    return 1;
  if (ParseComma()) return 1;
  if (ParseFloat(&(pTrans->Direct[0][2])))
    return 1;
  if (ParseComma()) return 1;
  if (ParseFloat(&(pTrans->Direct[1][0])))
    return 1;
  if (ParseComma()) return 1;
  if (ParseFloat(&(pTrans->Direct[1][1])))
    return 1;
  if (ParseComma()) return 1;
  if (ParseFloat(&(pTrans->Direct[1][2])))
    return 1;
  if (ParseComma()) return 1;
  if (ParseFloat(&(pTrans->Direct[2][0])))
    return 1;
  if (ParseComma()) return 1;
  if (ParseFloat(&(pTrans->Direct[2][1])))
    return 1;
  if (ParseComma()) return 1;
  if (ParseFloat(&(pTrans->Direct[2][2])))
    return 1;
  if (ParseComma()) return 1;
  if (ParseFloat(&(pTrans->Direct[3][0])))
    return 1;
  if (ParseComma()) return 1;
  if (ParseFloat(&(pTrans->Direct[3][1])))
    return 1;
  if (ParseComma()) return 1;
  if (ParseFloat(&(pTrans->Direct[3][2])))
    return 1;
  if (NextToken(&t))
    return 1;
  if (t != TRightAngleBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '>' was expected here...\n"), PovFile->Name, PovFile->Line);
    return 1;
  }
  MInverse(pTrans->Inverse, pTrans->Direct);
  return 0;
}

/*****************************************************************************
*  ParsePigment
******************************************************************************/
int ParsePigment(PigmentStruct **pLocalPigment)
{
  int t;
  ParsedVector LocalVector;
  TransformStruct Trans;
  double LocalFloat;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (NextToken(&t))
    return 1;
  if (t == TPigmentID)
  {
    *pLocalPigment = CopyPigment(LastPigmentConstant->Pigment);
  } else
  {
    if (*pLocalPigment == NULL)
    {
      (*pLocalPigment) = g_malloc(sizeof(PigmentStruct));
      (*pLocalPigment)->Type = PAT_SOLID_COLOR;
      (*pLocalPigment)->Trans = NULL;
      (*pLocalPigment)->PigmentList = NULL;
      (*pLocalPigment)->Color[0] = 0.0;
      (*pLocalPigment)->Color[1] = 0.0;
      (*pLocalPigment)->Color[2] = 0.0;
      (*pLocalPigment)->Warp = NULL;
      (*pLocalPigment)->ColorMap = NULL;
    }
    OldToken = t; IsOldToken = 1;
  }
  if (NextToken(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch(t)
    {
      case TColorID: case Tred: case Tgreen: case Tblue: case Tfilter: 
      case Ttransmit: case Tcolor: case Talpha: case Trgb: case Trgbt:
      case Trgbf: case Trgbft:
        OldToken = t;IsOldToken = 1;
        if (ParseColor((*pLocalPigment)->Color))
          return 1;
        (*pLocalPigment)->QuickColor[0] = (*pLocalPigment)->Color[0];
        (*pLocalPigment)->QuickColor[1] = (*pLocalPigment)->Color[1];
        (*pLocalPigment)->QuickColor[2] = (*pLocalPigment)->Color[2];
        (*pLocalPigment)->QuickColor[3] = (*pLocalPigment)->Color[3];
        (*pLocalPigment)->QuickColor[4] = (*pLocalPigment)->Color[4];
        (*pLocalPigment)->Type = PAT_SOLID_COLOR;
        break;
      case Tbrick:
/* FIXME LocalPigment->PigmentList = g_malloc(sizeof(PigmentListStruct));
        LocalPigment->PigmentList->NbPigment = 2;
        LocalPigment->PigmentList->PigmentArray = g_malloc(2*sizeof(PigmentStruct));
        ParsePigmentOrColor(&(LocalPigment->PigmentList->PigmentArray[0]));
        ParsePigmentOrColor(&(LocalPigment->PigmentList->PigmentArray[1]));
        LocalPigment->Type = PAT_BRICK;*/
        break;
      case Tbrick_size:
        if (ParseVector(&LocalVector))
          return 1;
        (*pLocalPigment)->BrickSize[0] = LocalVector.Val[0];
        (*pLocalPigment)->BrickSize[1] = LocalVector.Val[1];
        (*pLocalPigment)->BrickSize[2] = LocalVector.Val[2];
        break;
      case Tmortar:
        if (ParseFloat(&((*pLocalPigment)->Mortar)))
          return 1;
        break;
      case Tchecker:
/* FIXME LocalPigment->PigmentList = g_malloc(sizeof(PigmentListStruct));
        LocalPigment->PigmentList->NbPigment = 2;
        LocalPigment->PigmentList->PigmentArray = g_malloc(2*sizeof(PigmentStruct));
        ParsePigmentOrColor(&(LocalPigment->PigmentList->PigmentArray[0]));
        ParsePigmentOrColor(&(LocalPigment->PigmentList->PigmentArray[1]));
        LocalPigment->Type = PAT_CHECKER;*/
        break;
      case Thexagon:
/* FIXME LocalPigment->PigmentList = g_malloc(sizeof(PigmentListStruct));
        LocalPigment->PigmentList->NbPigment = 3;
        LocalPigment->PigmentList->PigmentArray = g_malloc(3*sizeof(PigmentStruct));
        ParsePigmentOrColor(&(LocalPigment->PigmentList->PigmentArray[0]));
        ParsePigmentOrColor(&(LocalPigment->PigmentList->PigmentArray[1]));
        ParsePigmentOrColor(&(LocalPigment->PigmentList->PigmentArray[2]));
        LocalPigment->Type = PAT_HEXAGON;*/
        break;
      case Tagate:
        (*pLocalPigment)->Type = PAT_AGATE;
        break;
      case Tagate_turb:
        if (ParseFloat(&((*pLocalPigment)->AgateTurb)))
          return 1;
        break;
      case Taverage:
        (*pLocalPigment)->Type = PAT_AVERAGE;
        break;
      case Tbozo:
        (*pLocalPigment)->Type = PAT_BOZO;
        break;
      case Tbumps:
        (*pLocalPigment)->Type = PAT_BUMPS;
        break;
      case Tcrackle:
        (*pLocalPigment)->Type = PAT_CRACKLE;
        break;
      case Tdents:
        (*pLocalPigment)->Type = PAT_DENTS;
        break;
      case Tgradient:
        (*pLocalPigment)->Type = PAT_GRADIENT;
        if (ParseVector(&LocalVector))
          return 1;
        (*pLocalPigment)->Gradient[0] = LocalVector.Val[0];
        (*pLocalPigment)->Gradient[1] = LocalVector.Val[1];
        (*pLocalPigment)->Gradient[2] = LocalVector.Val[2];
        break;
      case Tgranite:
        (*pLocalPigment)->Type = PAT_GRANITE;
        break;
      case Tleopard:
        (*pLocalPigment)->Type = PAT_LEOPARD;
        break;
      case Tmandel:
        (*pLocalPigment)->Type = PAT_MANDEL;
        if (ParseFloat(&LocalFloat))
          return 1;
        (*pLocalPigment)->Mandel = (int)LocalFloat;
        break;
      case Tmarble:
        (*pLocalPigment)->Type = PAT_MARBLE;
        break;
      case Tonion:
        (*pLocalPigment)->Type = PAT_ONION;
        break;
      case Tquilted:
        (*pLocalPigment)->Type = PAT_QUILTED;
        break;
      case Tcontrol0:
        if (ParseFloat(&((*pLocalPigment)->Control0)))
          return 1;
        break;
      case Tcontrol1:
        if (ParseFloat(&((*pLocalPigment)->Control1)))
          return 1;
        break;
      case Tradial:
        (*pLocalPigment)->Type = PAT_RADIAL;
        break;
      case Tripples:
        (*pLocalPigment)->Type = PAT_RIPPLES;
        break;
      case Tspiral1:
        (*pLocalPigment)->Type = PAT_SPIRAL1;
        if (ParseFloat(&((*pLocalPigment)->NumberOfArms)))
          return 1;
        break;
      case Tspiral2:
        (*pLocalPigment)->Type = PAT_SPIRAL2;
        if (ParseFloat(&((*pLocalPigment)->NumberOfArms)))
          return 1;
        break;
      case Tspotted:
        (*pLocalPigment)->Type = PAT_SPOTTED;
        break;
      case Twaves:
        (*pLocalPigment)->Type = PAT_WAVES;
        break;
      case Tfrequency:
        if (ParseFloat(&((*pLocalPigment)->Frequency)))
          return 1;
        break;
      case Tphase:
        if (ParseFloat(&((*pLocalPigment)->Phase)))
          return 1;
        break;
      case Twood:
        (*pLocalPigment)->Type = PAT_WOOD;
        break;
      case Twrinkles:
        (*pLocalPigment)->Type = PAT_WRINKLES;
        break;
      case Tcolor_map:
        ParseColorMap(&((*pLocalPigment)->ColorMap));
        break;
      case Tpigment_map:
        /* FIXME */
        break;
      case Timage_map:
        /* FIXME */
        break;
      case Tquick_color:
        if (ParseColor((*pLocalPigment)->QuickColor))
          return 1;
        break;
      case Ttranslate:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalPigment)->Trans)
        {
          ComputeTranslateTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalPigment)->Trans, &Trans);
        } else
        {
          (*pLocalPigment)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeTranslateTrans((*pLocalPigment)->Trans, LocalVector.Val);
        }
        break;
      case Trotate:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalPigment)->Trans)
        {
          ComputeRotateTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalPigment)->Trans, &Trans);
        } else
        {
          (*pLocalPigment)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeRotateTrans((*pLocalPigment)->Trans, LocalVector.Val);
        }
        break;
      case Tscale:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalPigment)->Trans)
        {
          ComputeScaleTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalPigment)->Trans, &Trans);
        } else
        {
          (*pLocalPigment)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeScaleTrans((*pLocalPigment)->Trans, LocalVector.Val);
        }
        break;
      case Tmatrix:
        if ((*pLocalPigment)->Trans)
        {
          if (ParseMatrix(&Trans))
            return 1;
          ComposeTrans((*pLocalPigment)->Trans, &Trans);
        } else
        {
          (*pLocalPigment)->Trans = g_malloc(sizeof(TransformStruct));
          if (ParseMatrix((*pLocalPigment)->Trans))
            return 1;
        }
        break;
      case Tramp_wave:
        (*pLocalPigment)->WaveForm = WF_RAMP;
        break;
      case Tsine_wave:
        (*pLocalPigment)->WaveForm = WF_SINE;
        break;
      case Tscallop_wave:
        (*pLocalPigment)->WaveForm = WF_SCALLOP;
        break;
      case Toctaves:
        if (ParseFloat(&LocalFloat))
          return 1;
        (*pLocalPigment)->Octaves = (int)LocalFloat;
        break;
      case Tlambda:
        if (ParseFloat(&((*pLocalPigment)->Lambda)))
          return 1;
        break;
      case Tomega:
        if (ParseFloat(&((*pLocalPigment)->Omega)))
          return 1;
        break;
      case Twarp:
        /* FIXME ParseWarp(&((*pLocalPigment)->Warp));*/
        break;
      case Tturbulence:
        ParseVector(&LocalVector);
        (*pLocalPigment)->Turbulence[0] = LocalVector.Val[0];
        (*pLocalPigment)->Turbulence[1] = LocalVector.Val[1];
        (*pLocalPigment)->Turbulence[2] = LocalVector.Val[2];
        break;
      default:
        WarningBox(_("Error"), _("%s:%d : A pigment modifier was expected here...\n"),
                   PovFile->Name, PovFile->Line);
        return 1;
    }
    if (NextToken(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseNormal
******************************************************************************/
int ParseNormal(NormalStruct **pLocalNormal)
{
  int t;
  ParsedVector LocalVector;
  TransformStruct Trans;
  double LocalFloat;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (NextToken(&t))
    return 1;
  if (t == TNormalID)
  {
    *pLocalNormal = CopyNormal(LastNormalConstant->Normal);
  } else
  {
    if (*pLocalNormal == NULL)
    {
      (*pLocalNormal) = g_malloc(sizeof(NormalStruct));
      (*pLocalNormal)->Type = PAT_BUMPS;
      (*pLocalNormal)->AgateTurb = 1.0;
      (*pLocalNormal)->BrickSize[0] = 8.0;
      (*pLocalNormal)->BrickSize[1] = 3.0;
      (*pLocalNormal)->BrickSize[2] = 4.5;
      (*pLocalNormal)->Mortar = 0.5;
      (*pLocalNormal)->Normal1 = NULL;
      (*pLocalNormal)->Normal2 = NULL;
      (*pLocalNormal)->Normal3 = NULL;
      (*pLocalNormal)->Gradient[0] = 0.0;
      (*pLocalNormal)->Gradient[1] = 1.0;
      (*pLocalNormal)->Gradient[2] = 0.0;
      (*pLocalNormal)->Mandel = 25;
      (*pLocalNormal)->Control0 = 1.0;
      (*pLocalNormal)->Control1 = 1.0;
      (*pLocalNormal)->NumberOfArms = 8.0;
      (*pLocalNormal)->Frequency = 1.0;
      (*pLocalNormal)->Phase = 0.0;
      (*pLocalNormal)->WaveForm = WF_RAMP;
      (*pLocalNormal)->Octaves = 6;
      (*pLocalNormal)->Lambda = 2.0;
      (*pLocalNormal)->Omega = 0.5;
      (*pLocalNormal)->Turbulence[0] = 0.0;
      (*pLocalNormal)->Turbulence[1] = 0.0;
      (*pLocalNormal)->Turbulence[2] = 0.0;
      (*pLocalNormal)->Amount = 1.0;
      (*pLocalNormal)->Trans = NULL;
    }
    OldToken = t; IsOldToken = 1;
  }
  if (NextToken(&t))
    return 1;
  switch (t)
  {
    case Tagate:
      (*pLocalNormal)->Type = PAT_AGATE;
      (*pLocalNormal)->AgateTurb = 1.0;
      break;
    case Taverage:
      (*pLocalNormal)->Type = PAT_AVERAGE;
      break;
    case Tbozo:
      (*pLocalNormal)->Type = PAT_BOZO;
      break;
    case Tbrick:
      (*pLocalNormal)->Type = PAT_BRICK;
      (*pLocalNormal)->Mortar = 0.5;
      (*pLocalNormal)->BrickSize[0] = 8.0;
      (*pLocalNormal)->BrickSize[1] = 3.0;
      (*pLocalNormal)->BrickSize[2] = 4.5;
      if (NextToken(&t))
        return 1;
      if (t == Tnormal)
      {
        if (ParseNormal(&((*pLocalNormal)->Normal1)))
          return 1;
        if (ParseComma())
          return 1;
        if (NextToken(&t))
          return 1;
        if (t != Tnormal)
        {
          WarningBox(_("Error"), _("%s:%d : a 'normal' keyword was expected here...\n"),
                     PovFile->Name, PovFile->Line);
          return 1;
        }
        if (ParseNormal(&((*pLocalNormal)->Normal2)))
          return 1;
      } else
      {
        OldToken = t; IsOldToken = 1;
      }
      break;
    case Tbumps:
      (*pLocalNormal)->Type = PAT_BUMPS;
      break;
    case Tchecker:
      (*pLocalNormal)->Type = PAT_CHECKER;
      if (NextToken(&t))
        return 1;
      if (t == Tnormal)
      {
        if (ParseNormal(&((*pLocalNormal)->Normal1)))
          return 1;
        if (ParseComma())
          return 1;
        if (NextToken(&t))
          return 1;
        if (t != Tnormal)
        {
          WarningBox(_("Error"), _("%s:%d : a 'normal' keyword was expected here...\n"),
                     PovFile->Name, PovFile->Line);
          return 1;
        }
        if (ParseNormal(&((*pLocalNormal)->Normal2)))
          return 1;
      } else
      {
        OldToken = t; IsOldToken = 1;
      }
      break;
    case Tcrackle:
      (*pLocalNormal)->Type = PAT_CRACKLE;
      break;
    case Tdents:
      (*pLocalNormal)->Type = PAT_DENTS;
      break;
    case Tgradient:
      (*pLocalNormal)->Type = PAT_GRADIENT;
      if (ParseVector(&LocalVector))
        return 1;
      (*pLocalNormal)->Gradient[0] = LocalVector.Val[0];
      (*pLocalNormal)->Gradient[1] = LocalVector.Val[1];
      (*pLocalNormal)->Gradient[2] = LocalVector.Val[2];
      break;
    case Tgranite:
      (*pLocalNormal)->Type = PAT_GRANITE;
      break;
    case Thexagon:
      (*pLocalNormal)->Type = PAT_HEXAGON;
      if (NextToken(&t))
        return 1;
      if (t == Tnormal)
      {
        if (ParseNormal(&((*pLocalNormal)->Normal1)))
          return 1;
        if (ParseComma())
          return 1;
        if (NextToken(&t))
          return 1;
        if (t != Tnormal)
        {
          WarningBox(_("Error"), _("%s:%d : a 'normal' keyword was expected here...\n"),
                     PovFile->Name, PovFile->Line);
          return 1;
        }
        if (ParseNormal(&((*pLocalNormal)->Normal2)))
          return 1;
        if (ParseComma())
          return 1;
        if (NextToken(&t))
          return 1;
        if (t != Tnormal)
        {
          WarningBox(_("Error"), _("%s:%d : a 'normal' keyword was expected here...\n"),
                     PovFile->Name, PovFile->Line);
          return 1;
        }
        if (ParseNormal(&((*pLocalNormal)->Normal2)))
          return 1;
      } else
      {
        OldToken = t; IsOldToken = 1;
      }
      break;
    case Tleopard:
      (*pLocalNormal)->Type = PAT_LEOPARD;
      break;
    case Tmandel:
      (*pLocalNormal)->Type = PAT_MANDEL;
      if (ParseFloat(&LocalFloat))
        return 1;
      (*pLocalNormal)->Mandel = (int)LocalFloat;
      break;
    case Tmarble:
      (*pLocalNormal)->Type = PAT_MARBLE;
      break;
    case Tonion:
      (*pLocalNormal)->Type = PAT_ONION;
      break;
    case Tquilted:
      (*pLocalNormal)->Type = PAT_QUILTED;
      (*pLocalNormal)->Control0 = 1.0;
      (*pLocalNormal)->Control1 = 1.0;
      break;
    case Tradial:
      (*pLocalNormal)->Type = PAT_RADIAL;
      break;
    case Tripples:
      (*pLocalNormal)->Type = PAT_RIPPLES;
      break;
    case Tspiral1:
      (*pLocalNormal)->Type = PAT_SPIRAL1;
      if (ParseFloat(&LocalFloat))
        return 1;
      (*pLocalNormal)->NumberOfArms = (int)LocalFloat;
      break;
    case Tspiral2:
      (*pLocalNormal)->Type = PAT_SPIRAL2;
      if (ParseFloat(&LocalFloat))
        return 1;
      (*pLocalNormal)->NumberOfArms = (int)LocalFloat;
      break;
    case Tspotted:
      (*pLocalNormal)->Type = PAT_SPOTTED;
      break;
    case Twaves:
      (*pLocalNormal)->Type = PAT_WAVES;
      break;
    case Twood:
      (*pLocalNormal)->Type = PAT_WOOD;
      break;
    case Twrinkles:
      (*pLocalNormal)->Type = PAT_WRINKLES;
      break;
    case Tbump_map:
      /* FIXME */
      break;
  }
  if (ParseComma())
    return 1;
  if (NextToken(&t))
    return 1;
  OldToken = t; IsOldToken = 1;
  switch (t)
  {
    case TFloatID: case TVectorID: case TPlus: case TMinus: case Tabs: 
    case Tacos: case Tval: case Tasc: case Tasin: case Tatan2: case Tceil:
    case Tclock: case Tcos: case Tdegrees: case Tdiv: case Texp: 
    case Tfile_exists: case TFloatLiteral: case Tfloor: case Tint: case Tlog:
    case Tmin: case Tmax: case Tmod: case Tpi: case Tpow: case Tradians:
    case Tsin: case Tsqrt: case Tstrcmp: case Tstrlen: case Ttan: case Tvdot:
    case Tvlength: case Tversion: case Ttrue: case Tyes: case Ton: case Tfalse:
    case Tno: case Toff: case Tseed: case Trand: case Tvaxis_rotate:
    case Tvcross: case Tvnormalize: case Tvrotate: case Tx: case Ty: case Tz:
    case Tt: case Tu: case Tv: case TExclamation: case TLeftParen:
    case TLeftAngleBracket:
      if (ParseFloat(&((*pLocalNormal)->Amount)))
        return 1;
  }
  if (NextToken(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch(t)
    {
      case Tbrick_size:
        if ((*pLocalNormal)->Type != PAT_BRICK)
        {
          WarningBox(_("Error"), _("%s:%d : The 'brick_size' modifier can only be \n  used with 'brick' pattern...\n"), PovFile->Name, PovFile->Line);
          return 1;
        }
        if (ParseVector(&LocalVector))
          return 1;
        (*pLocalNormal)->BrickSize[0] = LocalVector.Val[0];
        (*pLocalNormal)->BrickSize[1] = LocalVector.Val[1];
        (*pLocalNormal)->BrickSize[2] = LocalVector.Val[2];
        break;
      case Tmortar:
        if ((*pLocalNormal)->Type != PAT_BRICK)
        {
          WarningBox(_("Error"), _("%s:%d : The 'brick_size' modifier can only be \n  used with 'brick' pattern...\n"), PovFile->Name, PovFile->Line);
          return 1;
        }
        if (ParseFloat(&((*pLocalNormal)->Mortar)))
          return 1;
        break;
      case Tagate_turb:
        if ((*pLocalNormal)->Type != PAT_AGATE)
        {
          WarningBox(_("Error"), _("%s:%d : The 'agate_turb' modifier can only be \n  used with 'agate' pattern...\n"), PovFile->Name, PovFile->Line);
          return 1;
        }
        if (ParseFloat(&((*pLocalNormal)->AgateTurb)))
          return 1;
        break;
      case Tcontrol0:
        if ((*pLocalNormal)->Type != PAT_QUILTED)
        {
          WarningBox(_("Error"), _("%s:%d : The 'control0' modifier can only be \n  used with 'quilted' pattern...\n"), PovFile->Name, PovFile->Line);
          return 1;
        }
        if (ParseFloat(&((*pLocalNormal)->Control0)))
          return 1;
        break;
      case Tcontrol1:
        if ((*pLocalNormal)->Type != PAT_QUILTED)
        {
          WarningBox(_("Error"), _("%s:%d : The 'control1' modifier can only be \n  used with 'quilted' pattern...\n"), PovFile->Name, PovFile->Line);
          return 1;
        }
        if (ParseFloat(&((*pLocalNormal)->Control1)))
          return 1;
        break;
      case Tfrequency:
        if (ParseFloat(&((*pLocalNormal)->Frequency)))
          return 1;
        break;
      case Tphase:
        if (ParseFloat(&((*pLocalNormal)->Phase)))
          return 1;
        break;
      case Tramp_wave:
        (*pLocalNormal)->WaveForm = WF_RAMP;
        break;
      case Tsine_wave:
        (*pLocalNormal)->WaveForm = WF_SINE;
        break;
      case Tscallop_wave:
        (*pLocalNormal)->WaveForm = WF_SCALLOP;
        break;
      case Ttriangle_wave:
        (*pLocalNormal)->WaveForm = WF_TRIANGLE;
        break;
      case Toctaves:
        if (ParseFloat(&LocalFloat))
          return 1;
        (*pLocalNormal)->Octaves = (int)LocalFloat;
        break;
      case Tlambda:
        if (ParseFloat(&((*pLocalNormal)->Lambda)))
          return 1;
        break;
      case Tomega:
        if (ParseFloat(&((*pLocalNormal)->Omega)))
          return 1;
        break;
      case Twarp:
        /* FIXME ParseWarp(&((*pLocalNormal)->Warp));*/
        break;
      case Tturbulence:
        if (ParseVector(&LocalVector))
          return 1;
        (*pLocalNormal)->Turbulence[0] = LocalVector.Val[0];
        (*pLocalNormal)->Turbulence[1] = LocalVector.Val[1];
        (*pLocalNormal)->Turbulence[2] = LocalVector.Val[2];
        break;
      case Tbump_size:
        if (ParseFloat(&((*pLocalNormal)->Amount)))
          return 1;
        break;
      case Tnormal_map:
        /* FIXME */
        break;
      case Tslope_map:
        /* FIXME */
        break;
      case Ttranslate:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalNormal)->Trans)
        {
          ComputeTranslateTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalNormal)->Trans, &Trans);
        } else
        {
          (*pLocalNormal)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeTranslateTrans((*pLocalNormal)->Trans, LocalVector.Val);
        }
        break;
      case Trotate:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalNormal)->Trans)
        {
          ComputeRotateTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalNormal)->Trans, &Trans);
        } else
        {
          (*pLocalNormal)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeRotateTrans((*pLocalNormal)->Trans, LocalVector.Val);
        }
        break;
      case Tscale:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalNormal)->Trans)
        {
          ComputeScaleTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalNormal)->Trans, &Trans);
        } else
        {
          (*pLocalNormal)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeScaleTrans((*pLocalNormal)->Trans, LocalVector.Val);
        }
        break;
      case Tmatrix:
        if ((*pLocalNormal)->Trans)
        {
          if (ParseMatrix(&Trans))
            return 1;
          ComposeTrans((*pLocalNormal)->Trans, &Trans);
        } else
        {
          (*pLocalNormal)->Trans = g_malloc(sizeof(TransformStruct));
          if (ParseMatrix((*pLocalNormal)->Trans))
            return 1;
        }
        break;
      case Ttransform:
        /* FIXME */
        break;
      default:
        WarningBox(_("Error"), _("%s:%d : A Normal modifier was expected here...\n"),
                   PovFile->Name, PovFile->Line);
        return 1;
    }
    if (NextToken(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseFinish
******************************************************************************/
int ParseFinish(FinishStruct **pLocalFinish)
{
  int t;
  ParsedVector LocalVector;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (NextToken(&t))
    return 1;
  if (t == TFinishID)
  {
    *pLocalFinish = CopyFinish(LastFinishConstant->Finish);
  } else
  {
    if (*pLocalFinish == NULL)
    {
      (*pLocalFinish) = g_malloc(sizeof(FinishStruct));
      (*pLocalFinish)->Ambient[0] = 0.1;
      (*pLocalFinish)->Ambient[1] = 0.1;
      (*pLocalFinish)->Ambient[2] = 0.1;
      (*pLocalFinish)->Diffuse = 0.6;
      (*pLocalFinish)->Brilliance = 1.0;
      (*pLocalFinish)->Phong = 0.0;
      (*pLocalFinish)->PhongSize = 40.0;
      (*pLocalFinish)->Specular = 0.0;
      (*pLocalFinish)->Roughness = 20.0;
      (*pLocalFinish)->Metallic = 0.0;
      (*pLocalFinish)->Reflection[0] = 0.0;
      (*pLocalFinish)->Reflection[1] = 0.0;
      (*pLocalFinish)->Reflection[2] = 0.0;
      (*pLocalFinish)->Refraction = 0.0;
      (*pLocalFinish)->Ior = 1.0;
      (*pLocalFinish)->Caustics = 0.0;
      (*pLocalFinish)->FadeDistance = 0.0;
      (*pLocalFinish)->FadePower = 0.0;
      (*pLocalFinish)->Irid.Amount = 0.0;
      (*pLocalFinish)->Irid.Thickness = 0.0;
      (*pLocalFinish)->Irid.Turbulence[0] = 0.0;
      (*pLocalFinish)->Irid.Turbulence[1] = 0.0;
      (*pLocalFinish)->Irid.Turbulence[2] = 0.0;
      (*pLocalFinish)->Crand = 0.0;
    }
    OldToken = t; IsOldToken = 1;
  }
  if (NextToken(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch(t)
    {
      case Tambient:
        if (ParseColor((*pLocalFinish)->Ambient))
          return 1;
        break;
      case Tdiffuse:
        if (ParseFloat(&((*pLocalFinish)->Diffuse)))
          return 1;
        break;
      case Tbrilliance:
        if (ParseFloat(&((*pLocalFinish)->Brilliance)))
          return 1;
        break;
      case Tphong:
        if (ParseFloat(&((*pLocalFinish)->Phong)))
          return 1;
        break;
      case Tphong_size:
        if (ParseFloat(&((*pLocalFinish)->PhongSize)))
          return 1;
        break;
      case Tspecular:
        if (ParseFloat(&((*pLocalFinish)->Specular)))
          return 1;
        break;
      case Troughness:
        if (ParseFloat(&((*pLocalFinish)->Roughness)))
          return 1;
        break;
      case Tmetallic:
        if (NextToken(&t))
          return 1;
        OldToken = t; IsOldToken = 1;
        switch (t)
        {
          case TFloatID: case TVectorID: case TPlus: case TMinus: case Tabs: 
          case Tacos: case Tval: case Tasc: case Tasin: case Tatan2: case Tceil:
          case Tclock: case Tcos: case Tdegrees: case Tdiv: case Texp: 
          case Tfile_exists: case TFloatLiteral: case Tfloor: case Tint: case Tlog:
          case Tmin: case Tmax: case Tmod: case Tpi: case Tpow: case Tradians:
          case Tsin: case Tsqrt: case Tstrcmp: case Tstrlen: case Ttan: case Tvdot:
          case Tvlength: case Tversion: case Ttrue: case Tyes: case Ton: case Tfalse:
          case Tno: case Toff: case Tseed: case Trand: case Tvaxis_rotate:
          case Tvcross: case Tvnormalize: case Tvrotate: case Tx: case Ty: case Tz:
          case Tt: case Tu: case Tv: case TExclamation: case TLeftParen:
          case TLeftAngleBracket:
            if (ParseFloat(&((*pLocalFinish)->Metallic)))
              return 1;
            break;
          default:
            (*pLocalFinish)->Metallic = 1.0;
        }
        break;
      case Treflection:
        if (ParseColor((*pLocalFinish)->Reflection))
          return 1;
        break;
      case Trefraction:
        if (ParseFloat(&((*pLocalFinish)->Refraction)))
          return 1;
        break;
      case Tior:
        if (ParseFloat(&((*pLocalFinish)->Ior)))
          return 1;
        break;
      case Tcaustics:
        if (ParseFloat(&((*pLocalFinish)->Caustics)))
          return 1;
        break;
      case Tfade_distance:
        if (ParseFloat(&((*pLocalFinish)->FadeDistance)))
          return 1;
        break;
      case Tfade_power:
        if (ParseFloat(&((*pLocalFinish)->FadePower)))
          return 1;
        break;
      case Tirid:
        if (NextToken(&t))
          return 1;
        if (t != TLeftCurlyBracket)
        {
          WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"),
                     PovFile->Name, PovFile->Line); 
          return 1;
        }
        if (ParseFloat(&((*pLocalFinish)->Irid.Amount)))
          return 1;
        if (NextToken(&t))
          return 1;
        if (t != Tthickness)
        {
          WarningBox(_("Error"), _("%s:%d : the 'thickness' keyword was expected here...\n"),
                     PovFile->Name, PovFile->Line); 
          return 1;
        }
        if (ParseFloat(&((*pLocalFinish)->Irid.Thickness)))
          return 1;
        if (NextToken(&t))
          return 1;
        if (t != Tturbulence)
        {
          WarningBox(_("Error"), _("%s:%d : the 'turbulence' keyword was expected here...\n"),
                     PovFile->Name, PovFile->Line); 
          return 1;
        }
        if (ParseVector(&LocalVector))
          return 1;
        (*pLocalFinish)->Irid.Turbulence[0] = LocalVector.Val[0];
        (*pLocalFinish)->Irid.Turbulence[1] = LocalVector.Val[1];
        (*pLocalFinish)->Irid.Turbulence[2] = LocalVector.Val[2];
        if (NextToken(&t))
          return 1;
        if (t != TRightCurlyBracket)
        {
          WarningBox(_("Error"), _("%s:%d : a '}' was expected here...\n"),
                     PovFile->Name, PovFile->Line); 
          return 1;
        }
        break;
      case Tcrand:
        if (ParseFloat(&((*pLocalFinish)->Crand)))
          return 1;
        break;
      default:
        WarningBox(_("Error"), _("%s:%d : A finish modifier was expected here...\n"),
                   PovFile->Name, PovFile->Line);
        return 1;
    }
    if (NextToken(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseHalo
******************************************************************************/
int ParseHalo(HaloStruct **pLocalHalo)
{
  int t;
  ParsedVector LocalVector;
  TransformStruct Trans;
  double LocalFloat;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (NextToken(&t))
    return 1;
  if (t == THaloID)
  {
    *pLocalHalo = CopyHalo(LastHaloConstant->Halo);
  } else
  {
    if (*pLocalHalo == NULL)
    {
      (*pLocalHalo) = g_malloc(sizeof(HaloStruct));
      (*pLocalHalo)->Type = HALO_ATTENUATING; 
      (*pLocalHalo)->Density = DENSITY_CONSTANT;
      (*pLocalHalo)->Mapping = MAP_PLANAR;
      (*pLocalHalo)->DustType = 1;
      (*pLocalHalo)->Eccentricity = 1.0;
      (*pLocalHalo)->MaxValue = 1.0;
      (*pLocalHalo)->Exponent = 1;
      (*pLocalHalo)->Samples = 10;
      (*pLocalHalo)->AaLevel = 3;
      (*pLocalHalo)->AaThreshold = 0.3;
      (*pLocalHalo)->Jitter = 0.0;
      (*pLocalHalo)->Turbulence[0] = 0.0;
      (*pLocalHalo)->Turbulence[1] = 0.0; 
      (*pLocalHalo)->Turbulence[2] = 0.0;
      (*pLocalHalo)->Octaves = 1;
      (*pLocalHalo)->Omega = 0.0;
      (*pLocalHalo)->Lambda = 0.0;
      (*pLocalHalo)->ColorMap = NULL;
      (*pLocalHalo)->Frequency = 1.0;
      (*pLocalHalo)->Phase = 0.0;
      (*pLocalHalo)->Trans = NULL;
      (*pLocalHalo)->Next = NULL;
    }
    OldToken = t; IsOldToken = 1;
  }
  if (NextToken(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch(t)
    {
      case Tattenuating:
        (*pLocalHalo)->Type = HALO_ATTENUATING;
        break;
      case Temitting:
        (*pLocalHalo)->Type = HALO_EMITTING;
        break;
      case Tglowing:
        (*pLocalHalo)->Type = HALO_GLOWING;
        break;
      case Tdust:
        (*pLocalHalo)->Type = HALO_DUST;
        break;
      case Tconstant:
        (*pLocalHalo)->Density = DENSITY_CONSTANT;
        break;
      case Tlinear:
        (*pLocalHalo)->Density = DENSITY_LINEAR;
        break;
      case Tcubic:
        (*pLocalHalo)->Density = DENSITY_CUBIC;
        break;
      case Tpoly:
        (*pLocalHalo)->Density = DENSITY_POLY;
        break;
      case Tplanar_mapping:
        (*pLocalHalo)->Mapping = MAP_PLANAR;
        break;
      case Tspherical_mapping:
        (*pLocalHalo)->Mapping = MAP_SPHERICAL;
        break;
      case Tcylindrical_mapping:
        (*pLocalHalo)->Mapping = MAP_CYLINDRICAL;
        break;
      case Tbox_mapping:
        (*pLocalHalo)->Mapping = MAP_BOX;
        break;
      case Tdust_type:
        if (ParseFloat(&LocalFloat))
          return 1;
        (*pLocalHalo)->DustType = (int)LocalFloat;
        break;
      case Teccentricity:
        if (ParseFloat(&((*pLocalHalo)->Eccentricity)))
          return 1;
        break;
      case Tmax_value:
        if (ParseFloat(&((*pLocalHalo)->MaxValue)))
          return 1;
        break;
      case Texponent:
        if (ParseFloat(&((*pLocalHalo)->Exponent)))
          return 1;
        break;
      case Tsamples:
        if (ParseFloat(&((*pLocalHalo)->Samples)))
          return 1;
        break;
      case Taa_level:
        if (ParseFloat(&LocalFloat))
          return 1;
        (*pLocalHalo)->AaLevel = (int)LocalFloat;
        break;
      case Taa_threshold:
        if (ParseFloat(&((*pLocalHalo)->AaThreshold)))
          return 1;
        break;
      case Tjitter:
        if (ParseFloat(&((*pLocalHalo)->Jitter)))
          return 1;
        break;
      case Tturbulence:
        if (ParseVector(&LocalVector))
          return 1;
        (*pLocalHalo)->Turbulence[0] = LocalVector.Val[0];
        (*pLocalHalo)->Turbulence[1] = LocalVector.Val[1];
        (*pLocalHalo)->Turbulence[2] = LocalVector.Val[2];
        break;
      case Toctaves:
        if (ParseFloat(&LocalFloat))
          return 1;
        (*pLocalHalo)->Octaves = (int)LocalFloat;
        break;
      case Tomega:
        if (ParseFloat(&((*pLocalHalo)->Omega)))
          return 1;
        break;
      case Tlambda:
        if (ParseFloat(&((*pLocalHalo)->Lambda)))
          return 1;
        break;
      case Tcolor_map:
        if (ParseColorMap(&((*pLocalHalo)->ColorMap)))
          return 1;
        break;
      case Tfrequency:
        if (ParseFloat(&((*pLocalHalo)->Frequency)))
          return 1;
        break;
      case Tphase:
        if (ParseFloat(&((*pLocalHalo)->Phase)))
          return 1;
        break;
      case Tscale:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalHalo)->Trans)
        {
          ComputeScaleTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalHalo)->Trans, &Trans);
        } else
        {
          (*pLocalHalo)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeScaleTrans((*pLocalHalo)->Trans, LocalVector.Val);
        }
        break;
      case Trotate:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalHalo)->Trans)
        {
          ComputeRotateTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalHalo)->Trans, &Trans);
        } else
        {
          (*pLocalHalo)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeRotateTrans((*pLocalHalo)->Trans, LocalVector.Val);
        }
        break;
      case Ttranslate:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalHalo)->Trans)
        {
          ComputeTranslateTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalHalo)->Trans, &Trans);
        } else
        {
          (*pLocalHalo)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeTranslateTrans((*pLocalHalo)->Trans, LocalVector.Val);
        }
        break;
      case Tmatrix:
        if ((*pLocalHalo)->Trans)
        {
          if (ParseMatrix(&Trans))
            return 1;
          ComposeTrans((*pLocalHalo)->Trans, &Trans);
        } else
        {
          (*pLocalHalo)->Trans = g_malloc(sizeof(TransformStruct));
          if (ParseMatrix((*pLocalHalo)->Trans))
            return 1;
        }
        break;
      case Ttransform:
        /* FIXME */
        break;
      default:
        WarningBox(_("Error"), _("%s:%d : A halo modifier was expected here...\n"),
                   PovFile->Name, PovFile->Line);
        return 1;
    }
    if (NextToken(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseTexture
******************************************************************************/
int ParseTexture(TextureStruct **pLocalTexture)
{ /* FIXME: Only plain texture for now */
  int t;
  ParsedVector LocalVector;
  TransformStruct Trans;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (NextToken(&t))
    return 1;
  if (t == TTextureID)
  {
    *pLocalTexture = CopyTexture(LastTextureConstant->Texture);
  } else
  {
    if (*pLocalTexture == NULL)
    {
      (*pLocalTexture) = g_malloc(sizeof(TextureStruct));
      (*pLocalTexture)->Pigment = NULL;
      (*pLocalTexture)->Normal = NULL;
      (*pLocalTexture)->Finish = NULL;
      (*pLocalTexture)->Halo = NULL;
      (*pLocalTexture)->Trans = NULL;
      (*pLocalTexture)->NextLayer = NULL;
    }
    OldToken = t; IsOldToken = 1;
  }
  if (NextToken(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch(t)
    {
      case Tpigment:
        if ( ParsePigment(&((*pLocalTexture)->Pigment)) )
          return 1;
        break;
      case Tnormal:
        if ( ParseNormal(&((*pLocalTexture)->Normal)) )
          return 1;
        break;
      case Tfinish:
        if ( ParseFinish(&((*pLocalTexture)->Finish)) )
          return 1;
        break;
      case Thalo:
        if ( ParseHalo(&((*pLocalTexture)->Halo)) )
          return 1;
        break;
      case Ttranslate:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalTexture)->Trans)
        {
          ComputeTranslateTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalTexture)->Trans, &Trans);
        } else
        {
          (*pLocalTexture)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeTranslateTrans((*pLocalTexture)->Trans, LocalVector.Val);
        }
        break;
      case Trotate:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalTexture)->Trans)
        {
          ComputeRotateTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalTexture)->Trans, &Trans);
        } else
        {
          (*pLocalTexture)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeRotateTrans((*pLocalTexture)->Trans, LocalVector.Val);
        }
        break;
      case Tscale:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalTexture)->Trans)
        {
          ComputeScaleTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalTexture)->Trans, &Trans);
        } else
        {
          (*pLocalTexture)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeScaleTrans((*pLocalTexture)->Trans, LocalVector.Val);
        }
        break;
      case Tmatrix:
        if ((*pLocalTexture)->Trans)
        {
          if (ParseMatrix(&Trans))
            return 1;
          ComposeTrans((*pLocalTexture)->Trans, &Trans);
        } else
        {
          (*pLocalTexture)->Trans = g_malloc(sizeof(TransformStruct));
          if (ParseMatrix((*pLocalTexture)->Trans))
            return 1;
        }
        break;
      default:
        WarningBox(_("Error"), _("%s:%d : A texture modifier was expected here...\n"),
                   PovFile->Name, PovFile->Line);
        return 1;
    }
    if (NextToken(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParsePigmentMap FIXME
******************************************************************************/
int ParsePigmentMap(PigmentMapStruct **pLocalPigmentMap)
{

  return 1;
}

/*****************************************************************************
*  ParseSlopeMap FIXME
******************************************************************************/
int ParseSlopeMap(SlopeMapStruct **pLocalSlopeMap)
{
  return 1;
}

/*****************************************************************************
*  ParseNormalMap FIXME
******************************************************************************/
int ParseNormalMap(NormalMapStruct **pLocalNormalMap)
{

  return 1;
}

/*****************************************************************************
*  ParseTransform FIXME
******************************************************************************/
int ParseTransform(TransformStruct **pTransform)
{

  return 1;
}

/*****************************************************************************
*  ParseObjectModifier
******************************************************************************/
int ParseObjectModifier(ObjectStruct *LocalObject)
{
  int t;
  ParsedVector LocalVector;
  ObjectStruct *TmpObject;
  TransformStruct Trans;

  if (NextToken(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch (t)
    {
      case Ttranslate:
        if (ParseVector(&LocalVector))
          return 1;
        TranslateObject(LocalObject, LocalVector.Val);
        break;
      case Trotate:
        if (ParseVector(&LocalVector))
          return 1;
        RotateObject(LocalObject, LocalVector.Val);
        break;
      case Tscale:
        if (ParseVector(&LocalVector))
          return 1;
        if (LocalVector.Val[0] == 0.0)
        { 
          WarningBox(_("Warning"), _("You can't scale by 0.0\nchanged to 1.0"));
          LocalVector.Val[0] = 1.0;
        }
        if (LocalVector.Val[1] == 0.0)
        {
          WarningBox(_("Warning"), _("You can't scale by 0.0\nchanged to 1.0"));
          LocalVector.Val[1] = 1.0;
        }
        if (LocalVector.Val[2] == 0.0)
        {
          WarningBox(_("Warning"), _("You can't scale by 0.0\nchanged to 1.0"));
          LocalVector.Val[2] = 1.0;
        }
        ScaleObject(LocalObject, LocalVector.Val);
        break;
      case Ttransform:
        /* FIXME */
        break;
      case Tmatrix:
        if (LocalObject->Trans)
        {
          if (ParseMatrix(&Trans))
            return 1;
          ComposeTrans(LocalObject->Trans, &Trans);
        } else
        {
          LocalObject->Trans = g_malloc(sizeof(TransformStruct));
          if (ParseMatrix(LocalObject->Trans))
            return 1;
        }
        break;
      case Tbounded_by:
        TmpObject = NULL;
        if (NextToken(&t))
          return 1;
        if (t != TLeftCurlyBracket)
        {
          WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
          return 1;
        }
        if (ParseObject(&TmpObject))
          return 1;
        if (NextToken(&t))
          return 1;
        if (t != TRightCurlyBracket)
        {
          WarningBox(_("Error"), _("%s:%d a '}' was expected here...\n"), PovFile->Name, PovFile->Line); 
          return 1;
        }
        break;
      case Tclipped_by:
        if (NextToken(&t))
          return 1;
        if (t != TLeftCurlyBracket)
        {
          WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
          return 1;
        }
        if (ParseObject(&(LocalObject->ClippedBy)))
          return 1;
        if (NextToken(&t))
          return 1;
        if (t != TRightCurlyBracket)
        {
          WarningBox(_("Error"), _("%s:%d a '}' was expected here...\n"), PovFile->Name, PovFile->Line); 
          return 1;
        }
        break;
      case Ttexture:
        if (ParseTexture(&(LocalObject->Texture)))
          return 1;
        break;
      case Tpigment:
        if (LocalObject->Texture == NULL)
        {
          LocalObject->Texture = g_malloc(sizeof(TextureStruct));
          LocalObject->Texture->Pigment = NULL;
          LocalObject->Texture->Normal = NULL;
          LocalObject->Texture->Finish = NULL;
          LocalObject->Texture->Halo = NULL;
          LocalObject->Texture->Trans = NULL;
          LocalObject->Texture->NextLayer = NULL;
        }
        if (ParsePigment(&(LocalObject->Texture->Pigment)))
          return 1;
        break;
      case Tnormal:
        if (LocalObject->Texture == NULL)
        {
          LocalObject->Texture = g_malloc(sizeof(TextureStruct));
          LocalObject->Texture->Pigment = NULL;
          LocalObject->Texture->Normal = NULL;
          LocalObject->Texture->Finish = NULL;
          LocalObject->Texture->Halo = NULL;
          LocalObject->Texture->Trans = NULL;
          LocalObject->Texture->NextLayer = NULL;
        }
        if (ParseNormal(&(LocalObject->Texture->Normal)))
          return 1;
        break;
      case Tfinish:
        if (LocalObject->Texture == NULL)
        {
          LocalObject->Texture = g_malloc(sizeof(TextureStruct));
          LocalObject->Texture->Pigment = NULL;
          LocalObject->Texture->Normal = NULL;
          LocalObject->Texture->Finish = NULL;
          LocalObject->Texture->Halo = NULL;
          LocalObject->Texture->Trans = NULL;
          LocalObject->Texture->NextLayer = NULL;
        }
        if (ParseFinish(&(LocalObject->Texture->Finish)))
          return 1;
        break;
      case Thalo:
        if (LocalObject->Texture == NULL)
        {
          LocalObject->Texture = g_malloc(sizeof(TextureStruct));
          LocalObject->Texture->Pigment = NULL;
          LocalObject->Texture->Normal = NULL;
          LocalObject->Texture->Finish = NULL;
          LocalObject->Texture->Halo = NULL;
          LocalObject->Texture->Trans = NULL;
          LocalObject->Texture->NextLayer = NULL;
        }
        if (ParseHalo(&(LocalObject->Texture->Halo)))
          return 1;
        break;
      case Tinverse: /* A test for Flat Object could be welcomed here */
        LocalObject->Inverse = ! (LocalObject->Inverse);
        break;
      case Tsturm:
        LocalObject->Sturm = 1;
        break;
      case Twater_level:
        if (ParseFloat(&(LocalObject->WaterLevel)))
          return 1;
        break;
      case Tsmooth:
        LocalObject->Smooth = 1;
        break;
      case Tno_shadow:
        LocalObject->NoShadow = 1;
        break;
      case Thierarchy:
        LocalObject->Hierarchy = 1;
        break;
      case Thollow:
        LocalObject->Hollow = 1;
        break;
      case Topen:
        switch (LocalObject->Type) {
	  case CYLINDER_OBJECT:
            ((CylinderStruct *)LocalObject)->Open = TRUE;
	    break;
          case SOR_OBJECT:
            ((SorStruct *)LocalObject)->Open = TRUE;
	    break;
	  case CONE_OBJECT:
	    ((ConeStruct *)LocalObject)->Open = TRUE;
	    break;
	  default:
            WarningBox(_("Error"), _("%s:%d : The 'open' keyword can only be used	with some objects\n"),
                  PovFile->Name, PovFile->Line);
            return 1;
        }
        break;
      default:
        WarningBox(_("Error"), _("%s:%d : An object modifier was expected here...\n"),
                   PovFile->Name, PovFile->Line);
        return 1;
    }
    if (NextToken(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseBicubicPatch
*  Parsing of bicubic_patch object
******************************************************************************/
int ParseBicubicPatch(ObjectStruct **pBicubicPatch)
{
  BicubicPatchStruct *BBicubicPatch;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  BBicubicPatch = g_malloc(sizeof(BicubicPatchStruct));
  *pBicubicPatch = (ObjectStruct *)BBicubicPatch;
  InitObject(*pBicubicPatch);
  (*pBicubicPatch)->Type = BICUBIC_PATCH_OBJECT;
  /* FIXME */  
  if (ParseObjectModifier(*pBicubicPatch))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseBlob
*  Parsing of blob object
******************************************************************************/
int ParseBlob(ObjectStruct **pBlob)
{
  BlobStruct *BBlob;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  BBlob = g_malloc(sizeof(BlobStruct));
  *pBlob = (ObjectStruct *)BBlob;
  InitObject(*pBlob);
  (*pBlob)->Type = BLOB_OBJECT;
  /* FIXME */
  if (ParseObjectModifier(*pBlob))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseBox
*  Parsing of box object
******************************************************************************/
int ParseBox(ObjectStruct **pBox)
{
  BoxStruct *BBox;
  ParsedVector LocalVector1, LocalVector2;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  BBox = g_malloc(sizeof(BoxStruct));
  *pBox = (ObjectStruct *)BBox;
  InitObject(*pBox);
  (*pBox)->Type = BOX_OBJECT;
  if (ParseVector(&LocalVector1))
    return 1;
  if (ParseComma())
    return 1;
  if (ParseVector(&LocalVector2))
    return 1;
  BBox->MinCorner[0] = MIN(LocalVector1.Val[0], LocalVector2.Val[0]);
  BBox->MaxCorner[0] = MAX(LocalVector1.Val[0], LocalVector2.Val[0]);
  BBox->MinCorner[1] = MIN(LocalVector1.Val[1], LocalVector2.Val[1]);
  BBox->MaxCorner[1] = MAX(LocalVector1.Val[1], LocalVector2.Val[1]);
  BBox->MinCorner[2] = MIN(LocalVector1.Val[2], LocalVector2.Val[2]);
  BBox->MaxCorner[2] = MAX(LocalVector1.Val[2], LocalVector2.Val[2]);
  if (ParseObjectModifier(*pBox))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseCone
*  Parsing of cone object
******************************************************************************/
int ParseCone(ObjectStruct **pCone)
{
  ConeStruct *CCone;
  ParsedVector LocalVector1, LocalVector2;
  Vector Translate, Scale;
  Vector N, U1, U2;
  double ULength, Length;
  TransformStruct Transform;
  int t;
  double TmpRad;
  Vector TmpVect;
  int i;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  CCone = g_malloc(sizeof(ConeStruct));
  *pCone = (ObjectStruct *)CCone;
  InitObject(*pCone);
  (*pCone)->Type = CONE_OBJECT;
  if (ParseVector(&LocalVector1))
    return 1;
  if (ParseComma())
    return 1;
  if (ParseFloat(&(CCone->Radius1)))
    return 1;
  if (ParseComma())
    return 1;
  if (ParseVector(&LocalVector2))
    return 1;
  if (ParseComma())
    return 1;
  if (ParseFloat(&(CCone->Radius2)))
    return 1;
  if (CCone->Radius1 > CCone->Radius2)
  { /* We want Radius1 < Radius2, no matter what sign */
    TmpRad = CCone->Radius1;
    CCone->Radius1 = CCone->Radius2;
    CCone->Radius2 = TmpRad;
    for (i=0 ; i<3 ; i++)
    {
      TmpVect[i] = LocalVector1.Val[i];
      LocalVector1.Val[i] = LocalVector2.Val[i];
      LocalVector2.Val[i] = TmpVect[i];
    }
  }
  N[0] = LocalVector2.Val[0] - LocalVector1.Val[0];
  N[1] = LocalVector2.Val[1] - LocalVector1.Val[1];
  N[2] = LocalVector2.Val[2] - LocalVector1.Val[2];
  Length = VLength(N);
  if (Length == 0.0)
  {
    WarningBox(_("Error"), _("The Length of cone mustn't be nul.\n"));
    return 1;
  }
  if (CCone->Radius2 - CCone->Radius1 < 10e-6)
  {
    WarningBox(_("Error"), _("The two Radius of cones must be differents, use cylinder instead.\n"));
    return 1;
  }
  Scale[0] = 1.0;
  Scale[1] = Length / (CCone->Radius2-CCone->Radius1);
  Scale[2] = 1.0;
  ScaleCone(*pCone, Scale);

  N[0] /= Length;
  N[1] /= Length;
  N[2] /= Length;
  if ( (fabs(N[0])>fabs(N[1])) && (fabs(N[0])>fabs(N[2])) )
  { /* The cone axis is mainly upon the X axis */
    U1[0] = -N[1]/N[0];
    U1[1] = 1;
    U1[2] = 0;
  } else if ( (fabs(N[1])>fabs(N[0])) && (fabs(N[1])>fabs(N[2])) )
  { /* The cone axis is mainly upon the Y axis */
    U1[0] = 1;
    U1[1] = -N[0]/N[1];
    U1[2] = 0;
  } else
  { /* The cone axis is mainly upon the Z axis */
    U1[0] = 0;
    U1[1] = 1;
    U1[2] = -N[1]/N[2];
  }
  ULength = VLength(U1);
  U1[0]/=ULength; U1[1]/=ULength; U1[2]/=ULength;
  VCross(U2, N, U1);
  ULength = VLength(U2);
  U2[0]/=ULength; U2[1]/=ULength; U2[2]/=ULength;
                   
  Transform.Direct[0][0] = U1[0];
  Transform.Direct[0][1] = U1[1];
  Transform.Direct[0][2] = U1[2];
  Transform.Direct[0][3] = 0.0;
  Transform.Direct[1][0] = N[0];
  Transform.Direct[1][1] = N[1];
  Transform.Direct[1][2] = N[2];
  Transform.Direct[1][3] = 0.0;
  Transform.Direct[2][0] = U2[0];
  Transform.Direct[2][1] = U2[1];
  Transform.Direct[2][2] = U2[2];
  Transform.Direct[2][3] = 0.0;
  Transform.Direct[3][0] = 0.0;
  Transform.Direct[3][1] = 0.0;
  Transform.Direct[3][2] = 0.0;
  Transform.Direct[3][3] = 1.0;
  MInverse(Transform.Inverse, Transform.Direct);
  
  ComposeTrans((*pCone)->Trans, &Transform);
  Translate[0] = LocalVector1.Val[0]+(LocalVector2.Val[0]-LocalVector1.Val[0])*
                                CCone->Radius1/(CCone->Radius1-CCone->Radius2);
  Translate[1] = LocalVector1.Val[1]+(LocalVector2.Val[1]-LocalVector1.Val[1])*
                                CCone->Radius1/(CCone->Radius1-CCone->Radius2);
  Translate[2] = LocalVector1.Val[2]+(LocalVector2.Val[2]-LocalVector1.Val[2])*
                                CCone->Radius1/(CCone->Radius1-CCone->Radius2);
  TranslateCone(*pCone, Translate);

  if (ParseObjectModifier(*pCone))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseCubic
*  Parsing of cubic object
******************************************************************************/
int ParseCubic(ObjectStruct **pCubic)
{
  CubicStruct *CCubic;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  CCubic = g_malloc(sizeof(CubicStruct));
  *pCubic = (ObjectStruct *)CCubic;
  InitObject(*pCubic);
  (*pCubic)->Type = CUBIC_OBJECT;
  /* FIXME */
  if (ParseObjectModifier(*pCubic))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseCylinder
*  Parsing of cylinder object
******************************************************************************/
int ParseCylinder(ObjectStruct **pCylinder)
{
  ParsedVector LocalVector1, LocalVector2;
  double Radius;
  Vector TmpVect;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (ParseVector(&LocalVector1))
    return 1;
  if (ParseComma())
    return 1;
  if (ParseVector(&LocalVector2))
    return 1;
  if (ParseComma())
    return 1;
  if (ParseFloat(&Radius))
    return 1;
  if (Radius <= 0.0)
  {
    WarningBox(_("Error"), _("The Radius of cylinder must be strictly positive.\n"));
    return 1;
  }
  TmpVect[0] = LocalVector2.Val[0] - LocalVector1.Val[0];
  TmpVect[1] = LocalVector2.Val[1] - LocalVector1.Val[1];
  TmpVect[2] = LocalVector2.Val[2] - LocalVector1.Val[2];
  if (abs(VLength(TmpVect)) < 10e-6)
  {
    WarningBox(_("Error"), _("The Length of cylinder mustn't be nul.\n"));
    return 1;
  }
  *pCylinder = CreateCylinder(LocalVector1.Val, LocalVector2.Val, Radius);
  if (ParseObjectModifier(*pCylinder))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseDisc
*  Parsing of disc object
******************************************************************************/
int ParseDisc(ObjectStruct **pDisc)
{
  DiscStruct *DDisc;
  ParsedVector LocalVector;
  Vector Center, Normal;
  double Length, ULength;
  Vector U1, U2;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  DDisc = g_malloc(sizeof(DiscStruct));
  *pDisc = (ObjectStruct *)DDisc;
  InitObject(*pDisc);
  (*pDisc)->Type = DISC_OBJECT;
  if (ParseVector(&LocalVector))
    return 1;
  Center[0] = LocalVector.Val[0];
  Center[1] = LocalVector.Val[1];
  Center[2] = LocalVector.Val[2];
  if (ParseComma())
    return 1;
  if (ParseVector(&LocalVector))
    return 1;
  Normal[0] = LocalVector.Val[0];
  Normal[1] = LocalVector.Val[1];
  Normal[2] = LocalVector.Val[2];
  if (ParseComma())
    return 1;
  if (ParseFloat(&(DDisc->Radius)))
    return 1;
  if (ParseComma())
    return 1;
  DDisc->HoleRadius = 0.0;
  if (NextToken(&t))
    return 1;
  OldToken = t; IsOldToken = 1;
  switch (t)
  {
    case TFloatID: case TVectorID: case TPlus: case TMinus: case Tabs: 
    case Tacos: case Tval: case Tasc: case Tasin: case Tatan2: case Tceil:
    case Tclock: case Tcos: case Tdegrees: case Tdiv: case Texp: 
    case Tfile_exists: case TFloatLiteral: case Tfloor: case Tint: case Tlog:
    case Tmin: case Tmax: case Tmod: case Tpi: case Tpow: case Tradians:
    case Tsin: case Tsqrt: case Tstrcmp: case Tstrlen: case Ttan: case Tvdot:
    case Tvlength: case Tversion: case Ttrue: case Tyes: case Ton: case Tfalse:
    case Tno: case Toff: case Tseed: case Trand: case Tvaxis_rotate:
    case Tvcross: case Tvnormalize: case Tvrotate: case Tx: case Ty: case Tz:
    case Tt: case Tu: case Tv: case TExclamation: case TLeftParen:
    case TLeftAngleBracket:
      if (ParseFloat(&(DDisc->HoleRadius)))
        return 1;
      break;
  }
  (*pDisc)->Trans = g_malloc(sizeof(TransformStruct));
  Length = VLength(Normal);
  Normal[0] /= Length;
  Normal[1] /= Length;
  Normal[2] /= Length;
  if ( (fabs(Normal[0])>fabs(Normal[1])) && (fabs(Normal[0])>fabs(Normal[2])) )
  { /* The normal is mainly upon the X axis */
    U1[0] = -Normal[1]/Normal[0];
    U1[1] = 1;
    U1[2] = 0;
  } else if ( (fabs(Normal[1])>fabs(Normal[0])) && 
              (fabs(Normal[1])>fabs(Normal[2])) )
  { /* The normal is mainly upon the Y axis */
    U1[0] = 1;
    U1[1] = -Normal[0]/Normal[1];
    U1[2] = 0;
  } else
  { /* The normal axis is mainly upon the Z axis */
    U1[0] = 0;
    U1[1] = 1;
    U1[2] = -Normal[1]/Normal[2];
  }
  ULength = VLength(U1);
  U1[0]/=ULength; U1[1]/=ULength; U1[2]/=ULength;
  VCross(U2, Normal, U1);
  ULength = VLength(U2);
  U2[0]/=ULength; U2[1]/=ULength; U2[2]/=ULength;
                   
  (*pDisc)->Trans->Direct[0][0] = U1[0];
  (*pDisc)->Trans->Direct[0][1] = U1[1];
  (*pDisc)->Trans->Direct[0][2] = U1[2];
  (*pDisc)->Trans->Direct[0][3] = 0.0;
  (*pDisc)->Trans->Direct[1][0] = Normal[0];
  (*pDisc)->Trans->Direct[1][1] = Normal[1];
  (*pDisc)->Trans->Direct[1][2] = Normal[2];
  (*pDisc)->Trans->Direct[1][3] = 0.0;
  (*pDisc)->Trans->Direct[2][0] = U2[0];
  (*pDisc)->Trans->Direct[2][1] = U2[1];
  (*pDisc)->Trans->Direct[2][2] = U2[2];
  (*pDisc)->Trans->Direct[2][3] = 0.0;
  (*pDisc)->Trans->Direct[3][0] = 0.0;
  (*pDisc)->Trans->Direct[3][1] = 0.0;
  (*pDisc)->Trans->Direct[3][2] = 0.0;
  (*pDisc)->Trans->Direct[3][3] = 1.0;
  MInverse((*pDisc)->Trans->Inverse, (*pDisc)->Trans->Direct);
  TranslateDisc(*pDisc, Center);
  if (ParseObjectModifier(*pDisc))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseHeightField
*  Parsing of heightfield object
******************************************************************************/
int ParseHeightField(ObjectStruct **pHeightField)
{
  HeightFieldStruct *HHeightField;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  HHeightField = g_malloc(sizeof(HeightFieldStruct));
  *pHeightField = (ObjectStruct *)HHeightField;
  InitObject(*pHeightField);
  (*pHeightField)->Type = HEIGHT_FIELD_OBJECT;
  /* FIXME */
  if (ParseObjectModifier(*pHeightField))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseJuliaFractal
*  Parsing of julia_fractal object
******************************************************************************/
int ParseJuliaFractal(ObjectStruct **pJuliaFractal)
{
  JuliaFractalStruct *JJuliaFractal;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  JJuliaFractal = g_malloc(sizeof(JuliaFractalStruct));
  *pJuliaFractal = (ObjectStruct *)JJuliaFractal;
  InitObject(*pJuliaFractal);
  (*pJuliaFractal)->Type = JULIA_FRACTAL_OBJECT;
  /* FIXME */
  if (ParseObjectModifier(*pJuliaFractal))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseLathe
*  Parsing of lathe object
******************************************************************************/
int ParseLathe(ObjectStruct **pLathe)
{
  LatheStruct *LLathe;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  LLathe = g_malloc(sizeof(LatheStruct));
  *pLathe = (ObjectStruct *)LLathe;
  InitObject(*pLathe);
  (*pLathe)->Type = LATHE_OBJECT;
  /* FIXME */
  if (ParseObjectModifier(*pLathe))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseMesh
*  Parsing of mesh object
******************************************************************************/
int ParseMesh(ObjectStruct **pMesh)
{
  MeshStruct *MMesh;
  ParsedVector LocalVector;
  TriangleListStruct *TmpTri;
  Vector V1, V2;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  MMesh = g_malloc(sizeof(MeshStruct));
  *pMesh = (ObjectStruct *)MMesh;
  InitObject(*pMesh);
  (*pMesh)->Type = MESH_OBJECT;
  MMesh->FirstTriangle = NULL;
  if (NextToken(&t))
    return 1;
  while (t != TRightCurlyBracket)
  { 
    switch(t)
    {
      case Tsmooth_triangle:
        if (NextToken(&t))
          return 1;
        if (t != TLeftCurlyBracket)
        {
          WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
          return 1;
        }
        TmpTri = g_malloc(sizeof(TriangleListStruct));
        TmpTri->Next = MMesh->FirstTriangle;
        MMesh->FirstTriangle = TmpTri;
        if (ParseVector(&LocalVector)) return 1;
        TmpTri->P1[0] = LocalVector.Val[0];
        TmpTri->P1[1] = LocalVector.Val[1];
        TmpTri->P1[2] = LocalVector.Val[2];
        if (ParseComma()) return 1;
        if (ParseVector(&LocalVector)) return 1;
        TmpTri->N1[0] = LocalVector.Val[0];
        TmpTri->N1[1] = LocalVector.Val[1];
        TmpTri->N1[2] = LocalVector.Val[2];
        if (ParseComma()) return 1;
        if (ParseVector(&LocalVector)) return 1;
        TmpTri->P2[0] = LocalVector.Val[0];
        TmpTri->P2[1] = LocalVector.Val[1];
        TmpTri->P2[2] = LocalVector.Val[2];
        if (ParseComma()) return 1;
        if (ParseVector(&LocalVector)) return 1;
        TmpTri->N2[0] = LocalVector.Val[0];
        TmpTri->N2[1] = LocalVector.Val[1];
        TmpTri->N2[2] = LocalVector.Val[2];
        if (ParseComma()) return 1;
        if (ParseVector(&LocalVector)) return 1;
        TmpTri->P3[0] = LocalVector.Val[0];
        TmpTri->P3[1] = LocalVector.Val[1];
        TmpTri->P3[2] = LocalVector.Val[2];
        if (ParseComma()) return 1;
        if (ParseVector(&LocalVector)) return 1;
        TmpTri->N3[0] = LocalVector.Val[0];
        TmpTri->N3[1] = LocalVector.Val[1];
        TmpTri->N3[2] = LocalVector.Val[2];
        if (NextToken(&t))
          return 1;
        if (t != TRightCurlyBracket)
        {
          WarningBox(_("Error"), _("%s:%d a '}' was expected here...\n"), PovFile->Name, PovFile->Line); 
          return 1;
        }
        break;
      case Ttriangle:
        if (NextToken(&t))
          return 1;
        if (t != TLeftCurlyBracket)
        {
          WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
          return 1;
        }
        TmpTri = g_malloc(sizeof(TriangleListStruct));
        TmpTri->Next = MMesh->FirstTriangle;
        MMesh->FirstTriangle = TmpTri;
        if (ParseVector(&LocalVector)) return 1;
        TmpTri->P1[0] = LocalVector.Val[0];
        TmpTri->P1[1] = LocalVector.Val[1];
        TmpTri->P1[2] = LocalVector.Val[2];
        if (ParseComma()) return 1;
        if (ParseVector(&LocalVector)) return 1;
        TmpTri->P2[0] = LocalVector.Val[0];
        TmpTri->P2[1] = LocalVector.Val[1];
        TmpTri->P2[2] = LocalVector.Val[2];
        if (ParseComma()) return 1;
        if (ParseVector(&LocalVector)) return 1;
        TmpTri->P3[0] = LocalVector.Val[0];
        TmpTri->P3[1] = LocalVector.Val[1];
        TmpTri->P3[2] = LocalVector.Val[2];
        if (NextToken(&t))
          return 1;
        if (t != TRightCurlyBracket)
        {
          WarningBox(_("Error"), _("%s:%d a '}' was expected here...\n"), PovFile->Name, PovFile->Line); 
          return 1;
        }
        V1[0] = TmpTri->P2[0] - TmpTri->P1[0];
        V1[1] = TmpTri->P2[1] - TmpTri->P1[1];
        V1[2] = TmpTri->P2[2] - TmpTri->P1[2];
        V2[0] = TmpTri->P3[0] - TmpTri->P1[0];
        V2[1] = TmpTri->P3[1] - TmpTri->P1[1];
        V2[2] = TmpTri->P3[2] - TmpTri->P1[2];
        VCross(TmpTri->N1, V1, V2);
        VCross(TmpTri->N2, V1, V2);
        VCross(TmpTri->N3, V1, V2);
        break;
      default:
        OldToken = t; IsOldToken = 1;
        if (ParseObjectModifier(*pMesh))
          return 1;
        IsOldToken = 1;
        OldToken = TRightCurlyBracket;
    }
    if (NextToken(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParsePlane
*  Parsing of plane object
******************************************************************************/
int ParsePlane(ObjectStruct **pPlane)
{
  PlaneStruct *PPlane;
  ParsedVector LocalVector;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  PPlane = g_malloc(sizeof(PlaneStruct));
  *pPlane = (ObjectStruct *)PPlane;
  InitObject(*pPlane);
  (*pPlane)->Type = PLANE_OBJECT;
  if (ParseVector(&LocalVector))
    return 1;
  PPlane->Normal[0] = LocalVector.Val[0];
  PPlane->Normal[1] = LocalVector.Val[1];
  PPlane->Normal[2] = LocalVector.Val[2];
  if (ParseComma())
    return 1;
  if (ParseFloat(&(PPlane->Distance)))
    return 1;
  PPlane->NbSquare = 4;
  PPlane->SquareSize = 2.5;
  if (ParseObjectModifier(*pPlane))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParsePoly
*  Parsing of poly object
******************************************************************************/
int ParsePoly(ObjectStruct **pPoly)
{
  PolyStruct *PPoly;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  PPoly = g_malloc(sizeof(PolyStruct));
  *pPoly = (ObjectStruct *)PPoly;
  InitObject(*pPoly);
  (*pPoly)->Type = POLY_OBJECT;
  /* FIXME */
  if (ParseObjectModifier(*pPoly))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParsePolygon
*  Parsing of polygon object
******************************************************************************/
int ParsePolygon(ObjectStruct **pPolygon)
{
  PolygonStruct *PPolygon;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  PPolygon = g_malloc(sizeof(PolygonStruct));
  *pPolygon = (ObjectStruct *)PPolygon;
  InitObject(*pPolygon);
  (*pPolygon)->Type = POLYGON_OBJECT;
  /* FIXME */
  if (ParseObjectModifier(*pPolygon))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParsePrism
*  Parsing of prism object
******************************************************************************/
int ParsePrism(ObjectStruct **pPrism)
{
  PrismStruct *PPrism;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  PPrism = g_malloc(sizeof(PrismStruct));
  *pPrism = (ObjectStruct *)PPrism;
  InitObject(*pPrism);
  (*pPrism)->Type = PRISM_OBJECT;
  /* FIXME */
  if (ParseObjectModifier(*pPrism))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseQuadric
*  Parsing of quadric object
******************************************************************************/
int ParseQuadric(ObjectStruct **pQuadric)
{
  QuadricStruct *QQuadric;
  ParsedVector LocalVector;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  QQuadric = g_malloc(sizeof(QuadricStruct));
  *pQuadric = (ObjectStruct *)QQuadric;
  InitObject(*pQuadric);
  (*pQuadric)->Type = QUADRIC_OBJECT;
  if (ParseVector(&LocalVector))
    return 1;
  QQuadric->A = LocalVector.Val[0];
  QQuadric->B = LocalVector.Val[1];
  QQuadric->C = LocalVector.Val[2];
  if (ParseComma())
    return 1;
  if (ParseVector(&LocalVector))
    return 1;
  QQuadric->D = LocalVector.Val[0];
  QQuadric->E = LocalVector.Val[1];
  QQuadric->F = LocalVector.Val[2];
  if (ParseComma())
    return 1;
  if (ParseVector(&LocalVector))
    return 1;
  QQuadric->G = LocalVector.Val[0];
  QQuadric->H = LocalVector.Val[1];
  QQuadric->I = LocalVector.Val[2];
  if (ParseComma())
    return 1;
  if (ParseFloat(&(QQuadric->J)))
    return 1;
  if (ParseObjectModifier(*pQuadric))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseQuartic
*  Parsing of quartic object
******************************************************************************/
int ParseQuartic(ObjectStruct **pQuartic)
{
  QuarticStruct *QQuartic;
  int t, i;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  QQuartic = g_malloc(sizeof(QuarticStruct));
  *pQuartic = (ObjectStruct *)QQuartic;
  InitObject(*pQuartic);
  (*pQuartic)->Type = QUARTIC_OBJECT;
  if (NextToken(&t))
    return 1;
  if (t != TLeftAngleBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '<' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  for (i=0 ; i<35 ; i++)
  {
    if (ParseFloat(&(QQuartic->T[i])))
      return 1;
  }
  if (NextToken(&t))
    return 1;
  if (t != TRightAngleBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '>' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (ParseObjectModifier(*pQuartic))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseSmoothTriangle
*  Parsing of smooth triangles
******************************************************************************/
int ParseSmoothTriangle(ObjectStruct **pTriangle)
{
  TriangleStruct *TTriangle;
  ParsedVector LocalVector;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  TTriangle = g_malloc(sizeof(TriangleStruct));
  *pTriangle = (ObjectStruct *)TTriangle;
  InitObject(*pTriangle);
  (*pTriangle)->Type = TRIANGLE_OBJECT;
  if (ParseVector(&LocalVector))
    return 1;
  TTriangle->P1[0] = LocalVector.Val[0];
  TTriangle->P1[1] = LocalVector.Val[1];
  TTriangle->P1[2] = LocalVector.Val[2];
  if (ParseComma())
    return 1;
  if (ParseVector(&LocalVector))
    return 1;
  TTriangle->N1[0] = LocalVector.Val[0];
  TTriangle->N1[1] = LocalVector.Val[1];
  TTriangle->N1[2] = LocalVector.Val[2];
  if (ParseComma())
    return 1;
  if (ParseVector(&LocalVector))
    return 1;
  TTriangle->P2[0] = LocalVector.Val[0];
  TTriangle->P2[1] = LocalVector.Val[1];
  TTriangle->P2[2] = LocalVector.Val[2];
  if (ParseComma())
    return 1;
  if (ParseVector(&LocalVector))
    return 1;
  TTriangle->N2[0] = LocalVector.Val[0];
  TTriangle->N2[1] = LocalVector.Val[1];
  TTriangle->N2[2] = LocalVector.Val[2];
  if (ParseComma())
    return 1;
  if (ParseVector(&LocalVector))
    return 1;
  TTriangle->P3[0] = LocalVector.Val[0];
  TTriangle->P3[1] = LocalVector.Val[1];
  TTriangle->P3[2] = LocalVector.Val[2];
  if (ParseComma())
    return 1;
  if (ParseVector(&LocalVector))
    return 1;
  TTriangle->N3[0] = LocalVector.Val[0];
  TTriangle->N3[1] = LocalVector.Val[1];
  TTriangle->N3[2] = LocalVector.Val[2];
  if (ParseObjectModifier(*pTriangle))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseSor
*  Parsing of sor object
******************************************************************************/
int ParseSor(ObjectStruct **pSor)
{
  SorStruct *SSor;
  double LocalFloat;
  int t, i;
  ParsedVector LocalVector;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  SSor = g_malloc(sizeof(SorStruct));
  *pSor = (ObjectStruct *)SSor;
  InitObject(*pSor);
  (*pSor)->Type = SOR_OBJECT;
  if (ParseFloat(&(LocalFloat)))
    return 1;
  SSor->NumberOfPoints = (int)LocalFloat;
  if (SSor->NumberOfPoints < 4)
  {
    WarningBox(_("Error"), _("%s:%d Illegal number of point in a SOR object...\n"), PovFile->Name, PovFile->Line);
    return 1;
  }
  SSor->Point = g_malloc(SSor->NumberOfPoints*sizeof(Vector));
  for (i=0 ; i<SSor->NumberOfPoints ; i++)
  {
    if (ParseComma())
      return 1;
    if (ParseVector(&LocalVector))
      return 1;
    SSor->Point[i][0] = LocalVector.Val[0];
    SSor->Point[i][1] = LocalVector.Val[1];
  }
  if (ParseObjectModifier(*pSor))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseSphere
*  Parsing of sphere object
******************************************************************************/
int ParseSphere(ObjectStruct **pSphere)
{
  SphereStruct *SSphere;
  ParsedVector LocalVector;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  SSphere = g_malloc(sizeof(SphereStruct));
  *pSphere = (ObjectStruct *)SSphere;
  InitObject(*pSphere);
  (*pSphere)->Type = SPHERE_OBJECT;
  if (ParseVector(&LocalVector))
    return 1;
  SSphere->Center[0] = LocalVector.Val[0];
  SSphere->Center[1] = LocalVector.Val[1];
  SSphere->Center[2] = LocalVector.Val[2];
  if (ParseComma())
    return 1;
  if (ParseFloat(&(SSphere->Radius)))
    return 1;
  if (ParseObjectModifier(*pSphere))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseSuperEllipsoid
*  Parsing of superellipsoid object
******************************************************************************/
int ParseSuperEllipsoid(ObjectStruct **pSuperEllipsoid)
{
  SuperEllipsoidStruct *SSuperEllipsoid;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  SSuperEllipsoid = g_malloc(sizeof(SuperEllipsoidStruct));
  *pSuperEllipsoid = (ObjectStruct *)SSuperEllipsoid;
  InitObject(*pSuperEllipsoid);
  (*pSuperEllipsoid)->Type = SUPERELLIPSOID_OBJECT;
  if (NextToken(&t))
    return 1;
  if (t != TLeftAngleBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '<' was expected here...\n"), PovFile->Name, PovFile->Line);
    return 1;
  }
  if (ParseFloat(&(SSuperEllipsoid->E)))
    return 1;
  if (ParseComma())
    return 1;
  if (ParseFloat(&(SSuperEllipsoid->N)))
    return 1;
  if (NextToken(&t))
    return 1;
  if (t != TRightAngleBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '>' was expected here...\n"), PovFile->Name, PovFile->Line);
    return 1;
  }
  if (ParseObjectModifier(*pSuperEllipsoid))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseText
*  Parsing of text object
******************************************************************************/
int ParseText(ObjectStruct **pText)
{
  TextStruct *TText;
  ParsedVector LocalVector;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  TText = g_malloc(sizeof(TextStruct));
  *pText = (ObjectStruct *)TText;
  InitObject(*pText);
  (*pText)->Type = TEXT_OBJECT;
  if (NextToken(&t))
    return 1;
  if (t != Tttf)
  {
    WarningBox(_("Error"), _("%s:%d the 'ttf' keyword was expected here...\n"), PovFile->Name, PovFile->Line);
    return 1;
  }
  if (ParseString(&(TText->FontName))) return 1;
  if (ParseComma()) return 1;
  if (ParseString(&(TText->String))) return 1;
  if (ParseComma()) return 1;
  if (ParseFloat(&(TText->Thickness))) return 1;
  if (ParseComma()) return 1;
  if (ParseVector(&LocalVector))
    return 1;
  TText->Offset[0] = LocalVector.Val[0];
  TText->Offset[1] = LocalVector.Val[1];
  TText->Offset[2] = LocalVector.Val[2];
  if (ParseObjectModifier(*pText))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseTorus
*  Parsing of torus object
******************************************************************************/
int ParseTorus(ObjectStruct **pTorus)
{
  TorusStruct *TTorus;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  TTorus = g_malloc(sizeof(TorusStruct));
  *pTorus = (ObjectStruct *)TTorus;
  InitObject(*pTorus);
  (*pTorus)->Type = TORUS_OBJECT;
  if (ParseFloat(&(TTorus->Major)))
    return 1;
  if (ParseComma())
    return 1;
  if (ParseFloat(&(TTorus->Minor)))
    return 1;
  if (ParseObjectModifier(*pTorus))
    return 1;
  return 0;
}

/*****************************************************************************
*  ParseTriangle
*  Parsing of triangle object
******************************************************************************/
int ParseTriangle(ObjectStruct **pTriangle)
{
  TriangleStruct *TTriangle;
  ParsedVector LocalVector;
  Vector V1, V2;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  TTriangle = g_malloc(sizeof(TriangleStruct));
  *pTriangle = (ObjectStruct *)TTriangle;
  InitObject(*pTriangle);
  (*pTriangle)->Type = TRIANGLE_OBJECT;
  if (ParseVector(&LocalVector))
    return 1;
  TTriangle->P1[0] = LocalVector.Val[0];
  TTriangle->P1[1] = LocalVector.Val[1];
  TTriangle->P1[2] = LocalVector.Val[2];
  if (ParseComma())
    return 1;
  if (ParseVector(&LocalVector))
    return 1;
  TTriangle->P2[0] = LocalVector.Val[0];
  TTriangle->P2[1] = LocalVector.Val[1];
  TTriangle->P2[2] = LocalVector.Val[2];
  if (ParseComma())
    return 1;
  if (ParseVector(&LocalVector))
    return 1;
  TTriangle->P3[0] = LocalVector.Val[0];
  TTriangle->P3[1] = LocalVector.Val[1];
  TTriangle->P3[2] = LocalVector.Val[2];
  V1[0] = TTriangle->P2[0] - TTriangle->P1[0];
  V1[1] = TTriangle->P2[1] - TTriangle->P1[1];
  V1[2] = TTriangle->P2[2] - TTriangle->P1[2];
  V2[0] = TTriangle->P3[0] - TTriangle->P1[0];
  V2[1] = TTriangle->P3[1] - TTriangle->P1[1];
  V2[2] = TTriangle->P3[2] - TTriangle->P1[2];
  VCross(TTriangle->N1, V1, V2);
  VCross(TTriangle->N2, V1, V2);
  VCross(TTriangle->N3, V1, V2);
  if (ParseObjectModifier(*pTriangle))
    return 1;
  return 0;
}

/*****************************************************************************
*  AddObjectToCSG
******************************************************************************/
void AddObjectToCSG(ObjectStruct *CSG, ObjectStruct *LocalObject)
{
  CSGStruct *CCSG = (CSGStruct *)CSG;
  ObjectStruct *TmpObject;

  if (CCSG->FirstObject)
  {
    for (TmpObject = CCSG->FirstObject ;
         TmpObject->Next ;
         TmpObject = TmpObject->Next)
      ;
    TmpObject->Next = LocalObject;
  } else
  {
    CCSG->FirstObject = LocalObject;
  }
}

/*****************************************************************************
*  ParseCSG
*  Parsing of CSG object
******************************************************************************/
int ParseCSG(int Type, ObjectStruct **pCSG)
{
  CSGStruct *CCSG;
  ObjectStruct *TmpObject;
  int t;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  CCSG = g_malloc(sizeof(CSGStruct));
  *pCSG = (ObjectStruct *)CCSG;
  InitObject(*pCSG);
  (*pCSG)->Type = CSG_OBJECT;
  CCSG->Type = Type;
  CCSG->FirstObject = NULL;
  if (NextToken(&t))
    return 1;
  while (t != TRightCurlyBracket)
  { 
    switch(t)
    {
      case Tbicubic_patch:
        if (ParseBicubicPatch(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tblob:
        if (ParseBlob(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tbox:
        if (ParseBox(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tcone:
        if (ParseCone(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tcubic:
        if (ParseCubic(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tcylinder:
        if (ParseCylinder(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tdifference:
        if (ParseCSG(CSG_DIFFERENCE, &TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tdisc:
        if (ParseDisc(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Theight_field:
        if (ParseHeightField(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tintersection:
        if (ParseCSG(CSG_INTERSECTION, &TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tjulia_fractal:
        if (ParseJuliaFractal(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tlathe:
        if (ParseLathe(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tmerge:
        if (ParseCSG(CSG_MERGE, &TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tmesh:
        if (ParseMesh(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tobject:
        OldToken = t; IsOldToken = 1;
        TmpObject = NULL;
        if (ParseObject(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tplane:
        if (ParsePlane(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tpoly:
        if (ParsePoly(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tpolygon:
        if (ParsePolygon(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tprism:
        if (ParsePrism(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tquadric:
        if (ParseQuadric(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tquartic:
        if (ParseQuartic(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tsmooth_triangle:
        if (ParseSmoothTriangle(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tsor:
        if (ParseSor(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tsphere:
        if (ParseSphere(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tsuperellipsoid:
        if (ParseSuperEllipsoid(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Ttext:
        if (ParseText(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Ttorus:
        if (ParseTorus(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Ttriangle:
        if (ParseTriangle(&TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      case Tunion:
        if (ParseCSG(CSG_UNION, &TmpObject))
          return 1;
        AddObjectToCSG(*pCSG, TmpObject);
        break;
      default:
        OldToken = t; IsOldToken = 1;
        if (ParseObjectModifier(*pCSG))
          return 1;
        IsOldToken = 1;
        OldToken = TRightCurlyBracket;
    }
    if (NextToken(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseObject
******************************************************************************/
int ParseObject(ObjectStruct **pLocalObject)
{
  int t;

  if (NextToken(&t))
    return 1;
  switch (t)
  {
    case TObjectID:
      *pLocalObject = CopyObject(LastObjectConstant->Object);
      break;
    case Tobject:
      if (NextToken(&t))
        return 1;
      if (t != TLeftCurlyBracket)
      {
        WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
        return 1;
      }
      if (ParseObject(pLocalObject))
        return 1;
      if (ParseObjectModifier(*pLocalObject))
        return 1;
      break;
    case Tbicubic_patch:
      if (ParseBicubicPatch(pLocalObject))
        return 1;
      break;
    case Tblob:
      if (ParseBlob(pLocalObject))
        return 1;
      break;
    case Tbox:
      if (ParseBox(pLocalObject))
        return 1;
      break;
    case Tcone:
      if (ParseCone(pLocalObject))
        return 1;
      break;
    case Tcubic:
      if (ParseCubic(pLocalObject))
        return 1;
      break;
    case Tcylinder:
      if (ParseCylinder(pLocalObject))
        return 1;
      break;
    case Tdifference:
      if (ParseCSG(CSG_DIFFERENCE, pLocalObject))
        return 1;
      break;
    case Tdisc:
      if (ParseDisc(pLocalObject))
        return 1;
      break;
    case Theight_field:
      if (ParseHeightField(pLocalObject))
        return 1;
      break;
    case Tintersection:
      if (ParseCSG(CSG_INTERSECTION, pLocalObject))
        return 1;
      break;
    case Tjulia_fractal:
      if (ParseJuliaFractal(pLocalObject))
        return 1;
      break;
    case Tlathe:
      if (ParseLathe(pLocalObject))
        return 1;
      break;
    case Tmerge:
      if (ParseCSG(CSG_MERGE, pLocalObject))
        return 1;
      break;
    case Tmesh:
      if (ParseMesh(pLocalObject))
        return 1;
      break;
    case Tplane:
      if (ParsePlane(pLocalObject))
        return 1;
      break;
    case Tpoly:
      if (ParsePoly(pLocalObject))
        return 1;
      break;
    case Tpolygon:
      if (ParsePolygon(pLocalObject))
        return 1;
      break;
    case Tprism:
      if (ParsePrism(pLocalObject))
        return 1;
      break;
    case Tquadric:
      if (ParseQuadric(pLocalObject))
        return 1;
      break;
    case Tquartic:
      if (ParseQuartic(pLocalObject))
        return 1;
      break;
    case Tsmooth_triangle:
      if (ParseSmoothTriangle(pLocalObject))
        return 1;
      break;
    case Tsor:
      if (ParseSor(pLocalObject))
        return 1;
      break;
    case Tsphere:
      if (ParseSphere(pLocalObject))
        return 1;
      break;
    case Tsuperellipsoid:
      if (ParseSuperEllipsoid(pLocalObject))
        return 1;
      break;
    case Ttext:
      if (ParseText(pLocalObject))
        return 1;
      break;
    case Ttorus:
      if (ParseTorus(pLocalObject))
        return 1;
      break;
    case Ttriangle:
      if (ParseTriangle(pLocalObject))
        return 1;
      break;
    case Tunion:
      if (ParseCSG(CSG_UNION, pLocalObject))
        return 1;
      break;
    default:
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseGlobalSettings
******************************************************************************/
int ParseGlobalSettings(void)
{
  int t,t1;
  double LocalFloat;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (NextToken(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch(t)
    {
      case Tassumed_gamma:
        if (ParseFloat(&(CurrentFrame->Global.AssumedGamma)))
          return 1;
        break;
      case Thf_gray_16:
        if (NextToken(&t1))
          return 1;
        OldToken = t1;
        IsOldToken = 1;
        switch (t1)
        {
          case TPlus: case TMinus:
          case Tabs: case Tacos: case Tval: case Tasc: case Tasin: case Tatan2:
          case Tceil: case Tclock: case Tcos: case Tdegrees: case Tdiv: 
          case Texp: case Tfile_exists: case TFloatID: case TFloatLiteral:
          case Tfloor: case Tint: case Tlog: case Tmin: case Tmax: case Tmod:
          case Tpi: case Tpow: case Tradians: case Tsin: case Tsqrt: 
          case Tstrcmp: case Tstrlen: case Ttan: case Tvdot: case Tvlength:
          case Tversion: case Ttrue: case Tyes: case Ton: case Tfalse: 
          case Tno: case Toff: case Tseed: case Trand: case Tvaxis_rotate:
          case Tvcross: case TVectorID: case Tvnormalize: case Tvrotate:
          case Tx: case Ty: case Tz: case TColorID: case Tt: case Tu: case Tv:
          case TExclamation: case TLeftParen: case TLeftAngleBracket:
            if (ParseFloat(&LocalFloat))
              return 1;
            CurrentFrame->Global.HfGray16 = (int)LocalFloat;
            break;
          default:
            CurrentFrame->Global.HfGray16 = 1.0;
            break;
        }
        break;
      case Tambient_light:
        if (ParseColor(CurrentFrame->Global.AmbientLight))
          return 1;
        break;
      case Tnumber_of_waves:
        if (ParseFloat(&LocalFloat))
          return 1;
        CurrentFrame->Global.NumberOfWaves = (int)LocalFloat;
        break;
      case Tadc_bailout:
        if (ParseFloat(&(CurrentFrame->Global.ADCBailout)))
          return 1;
        break;
      case Tirid_wavelength:
        if (ParseColor(CurrentFrame->Global.IridWavelength))
          return 1;
        break;
      case Tmax_intersections:
        if (ParseFloat(&LocalFloat))
          return 1;
        CurrentFrame->Global.MaxIntersections = (int)LocalFloat;
        break;
      case Tmax_trace_level:
        if (ParseFloat(&LocalFloat))
          return 1;
        CurrentFrame->Global.MaxTraceLevel = (int)LocalFloat;
        break;
      case Tradiosity:
        if (NextToken(&t))
          return 1;
        if (t != TLeftCurlyBracket)
        {
          WarningBox(_("Error"), _("a '{' was expected here...\n"));
          return 1;
        }
        if (NextToken(&t1))
          return 1;
        while (t1 != TRightCurlyBracket)
        {
          switch(t1)
          {
            case Tbrightness:
              if (ParseFloat(&(CurrentFrame->Global.Radiosity.Brightness)))
                return 1;
              break;
            case Tcount:
              if (ParseFloat(&LocalFloat))
                return 1;
              CurrentFrame->Global.Radiosity.Count = (int)LocalFloat;
              break;
            case Tdistance_maximum:
              if (ParseFloat(&(CurrentFrame->Global.Radiosity.DistanceMaximum)))
                return 1;
              break;
            case Terror_bound:
              if (ParseFloat(&(CurrentFrame->Global.Radiosity.ErrorBound)))
                return 1;
              break;
            case Tgray_threshold:
              if (ParseFloat(&(CurrentFrame->Global.Radiosity.GrayThreshold)))
                return 1;
              break;
            case Tlow_error_factor:
              if (ParseFloat(&(CurrentFrame->Global.Radiosity.LowErrorFactor)))
                return 1;
              break;
            case Tminimum_reuse:
              if (ParseFloat(&(CurrentFrame->Global.Radiosity.MinimumReuse)))
                return 1;
              break;
            case Tnearest_count:
              if (ParseFloat(&LocalFloat))
                return 1;
              CurrentFrame->Global.Radiosity.NearestCount = (int)LocalFloat;
              break;
            case Trecursion_limit:
              if (ParseFloat(&LocalFloat))
                return 1;
              CurrentFrame->Global.Radiosity.RecursionLimit = (int)LocalFloat;
              break;
            default:
              WarningBox(_("Error"), _("A radiosity statement was expected here...\n"));
              return 1;
          }
          if (NextToken(&t1))
            return 1;
        }
        break;
      default:
        WarningBox(_("Error"), _("A global statement was expected here...\n"));
        return 1;
    }
    if (NextToken(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseLightSource
******************************************************************************/
int ParseLightSource(LightSourceStruct **pLocalLightSource)
{
  int t;
  ParsedVector LocalVector;
  
  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d : a '{' was expected here...\n"), PovFile->Name, PovFile->Line);
    return 1;
  }
  /* Creation & initialization */
  if (NextToken(&t))
    return 1;
  if (t == TLightSourceID)
  {
    *pLocalLightSource = CopyLightSource(LastLightSourceConstant->LightSource);
    return 0; /* XXX: Is it the 'Right Thing To Do' ? */
  } else
  {
    (*pLocalLightSource) = g_malloc(sizeof(LightSourceStruct));
    (*pLocalLightSource)->Next = NULL;
    (*pLocalLightSource)->Location[0] = 0.0;
    (*pLocalLightSource)->Location[1] = 0.0;
    (*pLocalLightSource)->Location[2] = 0.0;
    (*pLocalLightSource)->Color[0] = 1.0;
    (*pLocalLightSource)->Color[1] = 1.0;
    (*pLocalLightSource)->Color[2] = 1.0;
    (*pLocalLightSource)->Color[3] = 0.0;
    (*pLocalLightSource)->Color[4] = 0.0;
    (*pLocalLightSource)->Type = POINT_LIGHTSOURCE;
    (*pLocalLightSource)->PointAt[0] = 0.0;
    (*pLocalLightSource)->PointAt[1] = 0.0;
    (*pLocalLightSource)->PointAt[2] = 1.0;
    (*pLocalLightSource)->Radius = 0.35;
    (*pLocalLightSource)->FallOff = 0.35;
    (*pLocalLightSource)->Tightness = 10.0;
    (*pLocalLightSource)->AreaLight = FALSE;
    (*pLocalLightSource)->Area1[0] = 0.0;
    (*pLocalLightSource)->Area1[1] = 0.0;
    (*pLocalLightSource)->Area1[2] = 1.0;
    (*pLocalLightSource)->Area2[0] = 0.0;
    (*pLocalLightSource)->Area2[1] = 1.0;
    (*pLocalLightSource)->Area2[2] = 0.0;
    (*pLocalLightSource)->AreaSize1 = 0.0;
    (*pLocalLightSource)->AreaSize2 = 0.0;
    (*pLocalLightSource)->Adaptive = 100.0;
    (*pLocalLightSource)->Jitter = FALSE;
    (*pLocalLightSource)->LooksLike = NULL;
    (*pLocalLightSource)->FadeDistance = 0.0;
    (*pLocalLightSource)->FadePower = 0.0;
    (*pLocalLightSource)->AtmosphericAttenuation = FALSE;
    (*pLocalLightSource)->Atmosphere = FALSE;
    OldToken = t; IsOldToken = 1;
  }
  /* Parsing the mandatory parameters */
  if (ParseVector(&LocalVector))
    return 1;
  (*pLocalLightSource)->Location[0] = LocalVector.Val[0];
  (*pLocalLightSource)->Location[1] = LocalVector.Val[1];
  (*pLocalLightSource)->Location[2] = LocalVector.Val[2];
  if (ParseComma())
    return 1;
  if (ParseColor((*pLocalLightSource)->Color))
    return 1;
  /* Parsing the optionnal parameters */
  if (NextToken(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch (t)
    {
      case Tspotlight:
        (*pLocalLightSource)->Type = SPOTLIGHT_LIGHTSOURCE;
        break;
      case Tcylinder:
        (*pLocalLightSource)->Type = CYLINDER_LIGHTSOURCE;
        break;
      case Tshadowless:
        (*pLocalLightSource)->Type |= SHADOWLESS_LIGHTSOURCE;
        break;
      case Tpoint_at:
        if (ParseVector(&LocalVector))
          return 1;
        (*pLocalLightSource)->PointAt[0] = LocalVector.Val[0];
        (*pLocalLightSource)->PointAt[1] = LocalVector.Val[1];
        (*pLocalLightSource)->PointAt[2] = LocalVector.Val[2];
        break;
      case Tradius:
        if (ParseFloat(&((*pLocalLightSource)->Radius)))
          return 1;
        break;
      case Tfalloff:
        if (ParseFloat(&((*pLocalLightSource)->FallOff)))
          return 1;
        break;
      case Ttightness:
        if (ParseFloat(&((*pLocalLightSource)->Tightness)))
          return 1;
        break;
      case Tarea_light:
        /* FIXME */
        break;
      case Tadaptive:
        if (ParseFloat(&((*pLocalLightSource)->Adaptive)))
          return 1;
        break;
      case Tjitter:
        if (ParseFloat(&((*pLocalLightSource)->Jitter)))
          return 1;
        break;
      case Tlooks_like:
        /* FIXME */
        break;
      case Tfade_distance:
        if (ParseFloat(&((*pLocalLightSource)->FadeDistance)))
          return 1;
        break;
      case Tfade_power:
        if (ParseFloat(&((*pLocalLightSource)->FadePower)))
          return 1;
        break;
      case Tatmospheric_attenuation:
        /* FIXME */
        break;
      case Tatmosphere:
        /* FIXME */
        break;
      case Ttranslate:
        if (ParseVector(&LocalVector))
          return 1;
        /* FIXME */
        break;
      case Trotate:
        if (ParseVector(&LocalVector))
          return 1;
        /* FIXME */
        break;
      case Tscale:
        if (ParseVector(&LocalVector))
          return 1;
        /* FIXME */
        break;
      case Ttransform:
        /* FIXME */
        break;
      case Tmatrix:
        /* FIXME */
        break;
      default:
        WarningBox(_("Error"), _("%s:%d : A Light Source modifier was expected here...\n"),
                   PovFile->Name, PovFile->Line);
        return 1;
    }
    if (NextToken(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseCamera
******************************************************************************/
int ParseCamera(CameraStruct **pCamera)
{
  int t;
  double LocalFloat;
  ParsedVector LocalVector;
  Vector Vect;
  double k1, k2;
  double Norme;
  double DirectionLength, UpLength, RightLength;
  double Handedness;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d a '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (NextToken(&t))
    return 1;
  if (t == TCameraID)
  {
    *pCamera = CopyCamera(LastCameraConstant->Camera);
  } else
  {
    if (*pCamera == NULL)
    {
      (*pCamera) = g_malloc(sizeof(CameraStruct));
      (*pCamera)->Type = CAMERA_PERSPECTIVE;
      (*pCamera)->CylinderType = 1;
      (*pCamera)->Location[0] = 0.0;
      (*pCamera)->Location[1] = 0.0;
      (*pCamera)->Location[2] = 0.0;
      (*pCamera)->LookAt[0] = 0.0;
      (*pCamera)->LookAt[1] = 0.0;
      (*pCamera)->LookAt[2] = 1.0;
      (*pCamera)->Right[0] = 1.33;
      (*pCamera)->Right[1] = 0.0;
      (*pCamera)->Right[2] = 0.0;
      (*pCamera)->Up[0] = 0.0;
      (*pCamera)->Up[1] = 1.0;
      (*pCamera)->Up[2] = 0.0;
      (*pCamera)->Direction[0] = 0.0;
      (*pCamera)->Direction[1] = 0.0;
      (*pCamera)->Direction[2] = 1.0;
      (*pCamera)->Sky[0] = 0.0;
      (*pCamera)->Sky[1] = 1.0;
      (*pCamera)->Sky[2] = 0.0;
      (*pCamera)->Angle = 90.0;
      (*pCamera)->BlurSamples = 0;
      (*pCamera)->Aperture = 0.0;
      (*pCamera)->FocalPoint[0] = 0.0;
      (*pCamera)->FocalPoint[1] = 0.0;
      (*pCamera)->FocalPoint[2] = 1.0;
      (*pCamera)->Confidence = 0.9;
      (*pCamera)->Variance = 1.0/128.0;
      (*pCamera)->Normal = NULL;
    }
    OldToken = t; IsOldToken = 1;
  }
  if (NextToken(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch(t)
    {
      case Tperspective:
        (*pCamera)->Type = CAMERA_PERSPECTIVE;
        break;
      case Torthographic:
        (*pCamera)->Type = CAMERA_ORTHOGRAPHIC;
        Vect[0] = (*pCamera)->LookAt[0] - (*pCamera)->Location[0];
        Vect[1] = (*pCamera)->LookAt[1] - (*pCamera)->Location[1];
        Vect[2] = (*pCamera)->LookAt[2] - (*pCamera)->Location[2];
        k1 = VLength(Vect);
        k2 = VLength((*pCamera)->Direction);
        if ((k1 > 10e-6) && (k2 > 10e-6))
        {
          (*pCamera)->Right[0] *= k1/k2;
          (*pCamera)->Right[1] *= k1/k2;
          (*pCamera)->Right[2] *= k1/k2;
          (*pCamera)->Up[0] *= k1/k2;
          (*pCamera)->Up[1] *= k1/k2;
          (*pCamera)->Up[2] *= k1/k2;
        }
        break;
      case Tfisheye:
        (*pCamera)->Type = CAMERA_FISHEYE;
        break;
      case Tultra_wide_angle:
        (*pCamera)->Type = CAMERA_ULTRA_WIDE_ANGLE;
        break;
      case Tomnimax:
        (*pCamera)->Type = CAMERA_OMNIMAX;
        break;
      case Tpanoramic:
        (*pCamera)->Type = CAMERA_PANORAMIC;
        break;
      case Tcylinder:
        (*pCamera)->Type = CAMERA_CYLINDER;
        if (ParseFloat(&LocalFloat))
          return 1;
        (*pCamera)->CylinderType = (int)LocalFloat;
        if (((*pCamera)->CylinderType < 1) ||
            ((*pCamera)->CylinderType > 4))
        {
          WarningBox(_("Error"), _("%s:%d : There's only 4 cylinder type of camera...\n"),
                     PovFile->Name, PovFile->Line);
          return 1;
        }
        break;
      case Tlocation:
        if (ParseVector(&LocalVector))
          return 1;
        (*pCamera)->Location[0] = LocalVector.Val[0];
        (*pCamera)->Location[1] = LocalVector.Val[1];
        (*pCamera)->Location[2] = LocalVector.Val[2];
        break;
      case Tlook_at:
        DirectionLength = VLength((*pCamera)->Direction);
        UpLength = VLength((*pCamera)->Up);
        RightLength = VLength((*pCamera)->Right);
        VCross(Vect, (*pCamera)->Up, (*pCamera)->Direction);
        Handedness = VDot(Vect, (*pCamera)->Right);
        if (ParseVector(&LocalVector))
          return 1;
        (*pCamera)->LookAt[0] = (*pCamera)->Direction[0] = LocalVector.Val[0];
        (*pCamera)->LookAt[1] = (*pCamera)->Direction[1] = LocalVector.Val[1];
        (*pCamera)->LookAt[2] = (*pCamera)->Direction[2] = LocalVector.Val[2];
        (*pCamera)->Direction[0] -= (*pCamera)->Location[0];
        (*pCamera)->Direction[1] -= (*pCamera)->Location[1];
        (*pCamera)->Direction[2] -= (*pCamera)->Location[2];
        /* Check for zero length direction vector. */
        if (VLength((*pCamera)->Direction) < 10e-6)
        {
          WarningBox(_("Error"), _("%s:%d : Camera location and look_at point must be different.\n"),
                     PovFile->Name, PovFile->Line);
          return 1;
        }
        Norme = VLength((*pCamera)->Direction);
        (*pCamera)->Direction[0] /= Norme;
        (*pCamera)->Direction[1] /= Norme;
        (*pCamera)->Direction[2] /= Norme;
        VCross((*pCamera)->Right, (*pCamera)->Sky, (*pCamera)->Direction);
        Norme = VLength((*pCamera)->Right);
        (*pCamera)->Right[0] /= Norme;
        (*pCamera)->Right[1] /= Norme;
        (*pCamera)->Right[2] /= Norme;
        VCross((*pCamera)->Up, (*pCamera)->Direction, (*pCamera)->Right);
        (*pCamera)->Direction[0] *= DirectionLength;
        (*pCamera)->Direction[1] *= DirectionLength;
        (*pCamera)->Direction[2] *= DirectionLength;
        if (Handedness > 0.0)
        {
          (*pCamera)->Right[0] *= RightLength;
          (*pCamera)->Right[1] *= RightLength;
          (*pCamera)->Right[2] *= RightLength;
        }
        else
        {
          (*pCamera)->Right[0] *= -RightLength;
          (*pCamera)->Right[1] *= -RightLength;
          (*pCamera)->Right[2] *= -RightLength;
        }
        (*pCamera)->Up[0] *= UpLength;
        (*pCamera)->Up[1] *= UpLength;
        (*pCamera)->Up[2] *= UpLength;
        break;
      case Tright:
        if (ParseVector(&LocalVector))
          return 1;
        (*pCamera)->Right[0] = LocalVector.Val[0];
        (*pCamera)->Right[1] = LocalVector.Val[1];
        (*pCamera)->Right[2] = LocalVector.Val[2];
        break;
      case Tup:
        if (ParseVector(&LocalVector))
          return 1;
        (*pCamera)->Up[0] = LocalVector.Val[0];
        (*pCamera)->Up[1] = LocalVector.Val[1];
        (*pCamera)->Up[2] = LocalVector.Val[2];
        break;
      case Tdirection:
        if (ParseVector(&LocalVector))
          return 1;
        (*pCamera)->Direction[0] = LocalVector.Val[0];
        (*pCamera)->Direction[1] = LocalVector.Val[1];
        (*pCamera)->Direction[2] = LocalVector.Val[2];
        break;
      case Tsky:
        if (ParseVector(&LocalVector))
          return 1;
        (*pCamera)->Sky[0] = LocalVector.Val[0];
        (*pCamera)->Sky[1] = LocalVector.Val[1];
        (*pCamera)->Sky[2] = LocalVector.Val[2];
        break;
      case Tangle:
        if (ParseFloat(&((*pCamera)->Angle)))
          return 1;
        if ((*pCamera)->Angle < 0.0)
        {
          WarningBox(_("Error"), _("%s:%d : Negative viewing angle."), 
                     PovFile->Name, PovFile->Line);
          return 1;
        }
        if ((*pCamera)->Type == CAMERA_PERSPECTIVE)
        {
          if ((*pCamera)->Angle >= 180.0)
          {
            WarningBox(_("Error"), _("%s:%d : Viewing angle has to be smaller than 180 degrees."),
                       PovFile->Name, PovFile->Line);
          }
          Norme = VLength((*pCamera)->Direction);
          (*pCamera)->Direction[0] /= Norme;
          (*pCamera)->Direction[1] /= Norme;
          (*pCamera)->Direction[2] /= Norme;
          Norme = VLength((*pCamera)->Right) / tan((*pCamera)->Angle/2.0 * M_PI/180.0)/2.0;
          (*pCamera)->Direction[0] *= Norme;
          (*pCamera)->Direction[1] *= Norme;
          (*pCamera)->Direction[2] *= Norme;
        }
        break;
      case Tblur_samples:
        if (ParseFloat(&LocalFloat))
          return 1;
        (*pCamera)->BlurSamples = (int)LocalFloat;
        if ((*pCamera)->BlurSamples <=0 )
        {
          WarningBox(_("Error"), _("%s:%d : The number of focal blur samples must be strictly positive...\n"),
                     PovFile->Name, PovFile->Line);
          return 1;
        }
        break;
      case Taperture:
        if (ParseFloat(&((*pCamera)->Aperture)))
          return 1;
        break;
      case Tfocal_point:
        if (ParseVector(&LocalVector))
          return 1;
        (*pCamera)->FocalPoint[0] = LocalVector.Val[0];
        (*pCamera)->FocalPoint[1] = LocalVector.Val[1];
        (*pCamera)->FocalPoint[2] = LocalVector.Val[2];
        break;
      case Tnormal:
        if (ParseNormal(&((*pCamera)->Normal)))
          return 1;
        break;
      case Tconfidence:
        if (ParseFloat(&LocalFloat))
          return 1;
        if ((LocalFloat > 0.0) && (LocalFloat < 1.0))
          (*pCamera)->Confidence = LocalFloat;
        else
          WarningBox(_("Warning"), _("Illegal value of confidence last one will be use...\n"));
        break;
      case Tvariance:
        if (ParseFloat(&LocalFloat))
          return 1;
        if ((LocalFloat >= 0.0) && (LocalFloat <= 1.0))
          (*pCamera)->Variance = LocalFloat;
        else
          WarningBox(_("Warning"), _("%s:%d : Illegal value of variance last one will be use...\n"),
                     PovFile->Name, PovFile->Line);
        break;
      case Tscale:
        if (ParseVector(&LocalVector))
          return 1;
        /* FIXME */
        break;
      case Trotate:
        if (ParseVector(&LocalVector))
          return 1;
        /* FIXME */
        break;
      case Ttranslate:
        ParseVector(&LocalVector);
        (*pCamera)->Location[0] += LocalVector.Val[0];
        (*pCamera)->Location[1] += LocalVector.Val[1];
        (*pCamera)->Location[2] += LocalVector.Val[2];
        (*pCamera)->LookAt[0] += LocalVector.Val[0];
        (*pCamera)->LookAt[1] += LocalVector.Val[1];
        (*pCamera)->LookAt[2] += LocalVector.Val[2];
        break;
      case Tmatrix:
        /* FIXME */
        break;
      case Ttransform:
        /* FIXME */
        break;
      default:
        WarningBox(_("Error"), _("%s:%d : A Camera modifier was expected here...\n"),
                   PovFile->Name, PovFile->Line);
        return 1;
    }
    if (NextToken(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
* ParseAtmosphere
******************************************************************************/
int ParseAtmosphere(AtmosphereStruct **pLocalAtmosphere)
{
  int t;
  double LocalFloat;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d : a '{' was expected here...\n"),
               PovFile->Name, PovFile->Line);
    return 1;
  }
  if (NextToken(&t))
    return 1;
  if (t == TAtmosphereID)
  {
    *pLocalAtmosphere = CopyAtmosphere(LastAtmosphereConstant->Atmosphere);
  } else
  {
    *pLocalAtmosphere = g_malloc(sizeof(AtmosphereStruct));
    (*pLocalAtmosphere)->Type = 1;
    (*pLocalAtmosphere)->Distance = 0.0;
    (*pLocalAtmosphere)->Scattering = 1.0;
    (*pLocalAtmosphere)->Eccentricity = 0.0;
    (*pLocalAtmosphere)->Samples = 100;
    (*pLocalAtmosphere)->Jitter = 0.0;
    (*pLocalAtmosphere)->AaThreshold = 0.3;
    (*pLocalAtmosphere)->AaLevel = 0;
    (*pLocalAtmosphere)->Color[0] = 0.0;
    (*pLocalAtmosphere)->Color[1] = 0.0;
    (*pLocalAtmosphere)->Color[2] = 0.0;
    (*pLocalAtmosphere)->Color[3] = 0.0;
    (*pLocalAtmosphere)->Color[4] = 0.0;
    OldToken = t; IsOldToken = 1;
  }
  if (NextToken(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch (t)
    {
      case Ttype:
        if (ParseFloat(&LocalFloat))
          return 1;
        (*pLocalAtmosphere)->Type = (int)LocalFloat;
        if ( ((*pLocalAtmosphere)->Type < 1) ||
             ((*pLocalAtmosphere)->Type > 5) )
        {
          WarningBox(_("Error"), _("%s:%d : The Atmosphere type must be between 1 and 5...\n"),
                     PovFile->Name, PovFile->Line);
          return 1;
        }
        break;
      case Tdistance:
        if (ParseFloat(&((*pLocalAtmosphere)->Distance)))
          return 1;
        break;
      case Tscattering:
        if (ParseFloat(&((*pLocalAtmosphere)->Scattering)))
          return 1;
        break;
      case Teccentricity:
        if (ParseFloat(&((*pLocalAtmosphere)->Eccentricity)))
          return 1;
        break;
      case Tsamples:
        if (ParseFloat(&LocalFloat))
          return 1;
        (*pLocalAtmosphere)->Samples = (int)LocalFloat;
        if ( (*pLocalAtmosphere)->Samples < 1)
        {
          WarningBox(_("Error"), _("%s:%d :  The Number of Samples must be > 0...\n"),
                     PovFile->Name, PovFile->Line);
          return 1;
        }
        break;
      case Tjitter:
        if (ParseFloat(&((*pLocalAtmosphere)->Jitter)))
          return 1;
        break;
      case Taa_threshold:
        if (ParseFloat(&((*pLocalAtmosphere)->AaThreshold)))
          return 1;
        break;
      case Taa_level:
        if (ParseFloat(&LocalFloat))
          return 1;
        (*pLocalAtmosphere)->AaLevel = (int)LocalFloat;
        if ( (*pLocalAtmosphere)->AaLevel < 0)
        {
          WarningBox(_("Error"), _("%s:%d : The aa level must be positive...\n"),
                     PovFile->Name, PovFile->Line);
          return 1;
        }
        break;
      case Tcolor:
        if (ParseColor((*pLocalAtmosphere)->Color))
          return 1;
        break;
      default:
        WarningBox(_("Error"), _("%s:%d : An Atmosphere statement was expected here...\n"),
                   PovFile->Name, PovFile->Line);
        return 1;
    }
    if (NextToken(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseRainbow
******************************************************************************/
int ParseRainbow(RainbowStruct **pLocalRainbow)
{
  int t;
  ParsedVector LocalVector;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d : a '{' was expected here...\n"),
               PovFile->Name, PovFile->Line);
    return 1;
  }
  if (NextToken(&t))
    return 1;
  if (t == TRainbowID)
  {
    *pLocalRainbow = CopyRainbow(LastRainbowConstant->Rainbow);
  } else
  {
    if (*pLocalRainbow == NULL)
    {
      *pLocalRainbow = g_malloc(sizeof(RainbowStruct));
      (*pLocalRainbow)->Next = NULL;
      (*pLocalRainbow)->Direction[0] = 0.0;
      (*pLocalRainbow)->Direction[1] = 0.0;
      (*pLocalRainbow)->Direction[2] = 0.0;
      (*pLocalRainbow)->Angle = 0.0;
      (*pLocalRainbow)->Width = 0.0;
      (*pLocalRainbow)->Distance = 1.0e7;
      (*pLocalRainbow)->ColorMap = NULL;
      (*pLocalRainbow)->Jitter = 0.0;
      (*pLocalRainbow)->Up[0] = 0.0;
      (*pLocalRainbow)->Up[1] = 1.0;
      (*pLocalRainbow)->Up[2] = 0.0;
      (*pLocalRainbow)->ArcAngle = 180.0;
      (*pLocalRainbow)->FalloffAngle = 180.0;
    }
    OldToken = t; IsOldToken = 1;
  }
  if (NextToken(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch (t)
    {
      case Tdirection:
        if (ParseVector(&LocalVector))
          return 1;
        (*pLocalRainbow)->Direction[0] = LocalVector.Val[0];
        (*pLocalRainbow)->Direction[1] = LocalVector.Val[1];
        (*pLocalRainbow)->Direction[2] = LocalVector.Val[2];
        break;
      case Tangle:
        if (ParseFloat(&((*pLocalRainbow)->Angle)))
          return 1;
        break;
      case Twidth:
        if (ParseFloat(&((*pLocalRainbow)->Width)))
          return 1;
        break;
      case Tdistance:
        if (ParseFloat(&((*pLocalRainbow)->Distance)))
          return 1;
        break;
      case Tcolor_map:
        if (ParseColorMap(&( (*pLocalRainbow)->ColorMap )))
          return 1;
        break;
      case Tjitter:
        if (ParseFloat(&((*pLocalRainbow)->Jitter)))
          return 1;
        break;
      case Tup:
        if (ParseVector(&LocalVector))
          return 1;
        (*pLocalRainbow)->Up[0] = LocalVector.Val[0];
        (*pLocalRainbow)->Up[1] = LocalVector.Val[1];
        (*pLocalRainbow)->Up[2] = LocalVector.Val[2];
        break;
      case Tarc_angle:
        if (ParseFloat(&((*pLocalRainbow)->ArcAngle)))
          return 1;
        break;
      case Tfalloff_angle:
        if (ParseFloat(&((*pLocalRainbow)->FalloffAngle)))
          return 1;
        break;
      default:
        WarningBox(_("Error"), _("%s:%d : A Rainbow statement was expected here...\n"),
                   PovFile->Name, PovFile->Line);
        return 1;
    }
    if (NextToken(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseFog
******************************************************************************/
int ParseFog(FogStruct **pLocalFog)
{
  int t;
  ParsedVector LocalVector;
  TransformStruct Trans;
  double LocalFloat;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d : a '{' was expected here...\n"),
               PovFile->Name, PovFile->Line);
    return 1;
  }
  if (NextToken(&t))
    return 1;
  if (t == TFogID)
  {
    *pLocalFog = CopyFog(LastFogConstant->Fog);
  } else
  {
    *pLocalFog = g_malloc(sizeof(FogStruct));
    (*pLocalFog)->FogType = 1;
    (*pLocalFog)->Distance = 0.0;
    (*pLocalFog)->Color[0] = 0.0;
    (*pLocalFog)->Color[1] = 0.0;
    (*pLocalFog)->Color[2] = 0.0;
    (*pLocalFog)->Color[3] = 0.0;
    (*pLocalFog)->Color[4] = 0.0;
    (*pLocalFog)->Turbulence[0] = 0.0;
    (*pLocalFog)->Turbulence[1] = 0.0;
    (*pLocalFog)->Turbulence[2] = 0.0;
    (*pLocalFog)->TurbDepth = 0.5;
    (*pLocalFog)->Omega = 0.0;
    (*pLocalFog)->Lambda = 0.0;
    (*pLocalFog)->Octaves = 0;
    (*pLocalFog)->FogOffset = 0.0;
    (*pLocalFog)->FogAlt = 0.0;
    (*pLocalFog)->Up[0] = 0.0;
    (*pLocalFog)->Up[1] = 1.0;
    (*pLocalFog)->Up[2] = 0.0;
    (*pLocalFog)->Trans = NULL;
    (*pLocalFog)->Next = NULL;
    OldToken = t; IsOldToken = 1;
  }
  if (NextToken(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch (t)
    {
      case Tfog_type:
        if (ParseFloat(&LocalFloat))
          return 1;
        (*pLocalFog)->FogType = (int)LocalFloat;
        if ( ((*pLocalFog)->FogType != 1) && ((*pLocalFog)->FogType != 2) )
        {
          WarningBox(_("Error"), _("%s:%d : The Fog type can only be 1 or 2...\n"),
                     PovFile->Name, PovFile->Line);
          return 1;
        }
        break;
      case Tdistance:
        if (ParseFloat(&((*pLocalFog)->Distance)))
          return 1;
        break;
      case Tcolor:
        if (ParseColor((*pLocalFog)->Color))
          return 1;
        break;
      case Tturbulence:
        if (ParseVector(&LocalVector))
          return 1;
        (*pLocalFog)->Turbulence[0] = LocalVector.Val[0];
        (*pLocalFog)->Turbulence[1] = LocalVector.Val[1];
        (*pLocalFog)->Turbulence[2] = LocalVector.Val[2];
        break;
      case Tturb_depth:
        if (ParseFloat(&((*pLocalFog)->TurbDepth)))
          return 1;
        break;
      case Tomega:
        if (ParseFloat(&((*pLocalFog)->Omega)))
          return 1;
        break;
      case Tlambda:
        if (ParseFloat(&((*pLocalFog)->Lambda)))
          return 1;
        break;
      case Toctaves:
        if (ParseFloat(&LocalFloat))
          return 1;
        (*pLocalFog)->Octaves = (int)LocalFloat;
        if ( ((*pLocalFog)->Octaves < 1) || ((*pLocalFog)->Octaves > 10) )
        {
          WarningBox(_("Error"), _("%s:%d : Octaves must be between 1 and 10...\n"),
                     PovFile->Name, PovFile->Line);
          return 1;
        }
        break;
      case Tfog_offset:
        if (ParseFloat(&((*pLocalFog)->FogOffset)))
          return 1;
        break;
      case Tfog_alt:
        if (ParseFloat(&((*pLocalFog)->FogAlt)))
          return 1;
        break;
      case Tup:
        if (ParseVector(&LocalVector))
          return 1;
        (*pLocalFog)->Up[0] = LocalVector.Val[0];
        (*pLocalFog)->Up[1] = LocalVector.Val[1];
        (*pLocalFog)->Up[2] = LocalVector.Val[2];
        break;
      case Ttranslate:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalFog)->Trans)
        {
          ComputeTranslateTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalFog)->Trans, &Trans);
         } else
        {
          (*pLocalFog)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeTranslateTrans((*pLocalFog)->Trans, LocalVector.Val);
        }
        break;
      case Trotate:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalFog)->Trans)
        {
          ComputeRotateTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalFog)->Trans, &Trans);
         } else
        {
          (*pLocalFog)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeRotateTrans((*pLocalFog)->Trans, LocalVector.Val);
        }
        break;
      case Tscale:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalFog)->Trans)
        {
          ComputeScaleTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalFog)->Trans, &Trans);
         } else
        {
          (*pLocalFog)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeScaleTrans((*pLocalFog)->Trans, LocalVector.Val);
        }
        break;
      case Tmatrix:
        if ((*pLocalFog)->Trans)
        {
          if (ParseMatrix(&Trans))
            return 1;
          ComposeTrans((*pLocalFog)->Trans, &Trans);
        } else
        {
          (*pLocalFog)->Trans = g_malloc(sizeof(TransformStruct));
          if (ParseMatrix((*pLocalFog)->Trans))
            return 1;
        }
        break;
      case Ttransform:
        /* FIXME */
        break;
      default:
        WarningBox(_("Error"), _("%s:%d : A Fog statement was expected here...\n"),
                   PovFile->Name, PovFile->Line);
        return 1;
    }
    if (NextToken(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  ParseSkySphere
******************************************************************************/
int ParseSkySphere(SkySphereStruct **pLocalSkySphere)
{
  int t;
  ParsedVector LocalVector;
  TransformStruct Trans;

  if (NextToken(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    WarningBox(_("Error"), _("%s:%d : a '{' was expected here...\n"),
               PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (NextToken(&t))
    return 1;
  if (t == TSkySphereID)
  {
    *pLocalSkySphere = CopySkySphere(LastSkySphereConstant->SkySphere);
  } else
  {
    if (*pLocalSkySphere)
      g_free((*pLocalSkySphere)->pPigmentList);
    else
      *pLocalSkySphere = g_malloc(sizeof(SkySphereStruct));
    (*pLocalSkySphere)->NbPigment = 0;
    (*pLocalSkySphere)->pPigmentList = NULL;
    (*pLocalSkySphere)->Trans = NULL;
    OldToken = t; IsOldToken = 1;
  }
  if (NextToken(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch (t)
    {
      case Tpigment:
        (*pLocalSkySphere)->NbPigment++;
        if ((*pLocalSkySphere)->pPigmentList)
        {
          (*pLocalSkySphere)->pPigmentList = 
            g_realloc((*pLocalSkySphere)->pPigmentList,
                      sizeof(PigmentStruct*)*(*pLocalSkySphere)->NbPigment);
        } else
        {
          (*pLocalSkySphere)->pPigmentList = g_malloc(sizeof(PigmentStruct*));
        }
        (*pLocalSkySphere)->pPigmentList[(*pLocalSkySphere)->NbPigment-1] = NULL;
        if (ParsePigment(&((*pLocalSkySphere)->pPigmentList[
                                        (*pLocalSkySphere)->NbPigment-1])))
          return 1;
        break;
      case Ttranslate:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalSkySphere)->Trans)
        {
          ComputeTranslateTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalSkySphere)->Trans, &Trans);
        } else
        {
          (*pLocalSkySphere)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeTranslateTrans((*pLocalSkySphere)->Trans, LocalVector.Val);
        }
        break;
      case Trotate:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalSkySphere)->Trans)
        {
          ComputeRotateTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalSkySphere)->Trans, &Trans);
        } else
        {
          (*pLocalSkySphere)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeRotateTrans((*pLocalSkySphere)->Trans, LocalVector.Val);
        }
        break;
      case Tscale:
        if (ParseVector(&LocalVector))
          return 1;
        if ((*pLocalSkySphere)->Trans)
        {
          ComputeScaleTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalSkySphere)->Trans, &Trans);
        } else
        {
          (*pLocalSkySphere)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeScaleTrans((*pLocalSkySphere)->Trans, LocalVector.Val);
        }
        break;
      case Ttransform:
        /* FIXME */
        break;
      case Tmatrix:
        if ((*pLocalSkySphere)->Trans)
        {
          if (ParseMatrix(&Trans))
            return 1;
          ComposeTrans((*pLocalSkySphere)->Trans, &Trans);
        } else
        {
          (*pLocalSkySphere)->Trans = g_malloc(sizeof(TransformStruct));
          if (ParseMatrix((*pLocalSkySphere)->Trans))
            return 1;
        }
        break;
      default:
        WarningBox(_("Error"), _("%s:%d : A Sky Sphere statement was expected here...\n"),
                   PovFile->Name, PovFile->Line);
        return 1;
    }
    if (NextToken(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  InitLoadPov
******************************************************************************/
void InitLoadPov(char *Name)
{
  IsOldChar = 0;
  IsOldToken = 0;
  PovVersionLangage = 3.0;
  LastString = NULL;
  PovClock = 0.0;
  NbPovRandomTable = 0;
  DontParse = 0;
  ParsingDirective = 0;
  PovRandomTable = NULL;
  CurrentFrame = NewFrame(Name);
  ConditionalStack = NULL;
}

/*****************************************************************************
*  ParsePov
******************************************************************************/
int ParsePov()
{
  LightSourceStruct *TmpLightSource;
  ObjectStruct *TmpObject;
  int t;

  if (NextToken(&t))
    return 1;
  while (t != TEOF)
  {
    switch (t)
    {
      case Tglobal_settings:
        if (ParseGlobalSettings())
          return 1;
        break;
      case Tcamera:
/*        if (CurrentFrame->Camera)
        {
          WarningBox(_("Error"), _("%s:%d : Only one camera is allowed per frame...\n"),
                     PovFile->Name, PovFile->Line);
          return 1;
        }*/
        if (ParseCamera(&(CurrentFrame->Camera)))
          return 1;
        break;
      case Tlight_source:
        if (ParseLightSource(&TmpLightSource))
          return 1;
        TmpLightSource->Next = CurrentFrame->FirstLightSource;
        CurrentFrame->FirstLightSource = TmpLightSource;
        break;
      case Tsky_sphere:
        if (CurrentFrame->SkySphere)
        {
          WarningBox(_("Warning"), _("Only one sky sphere is allowed, the last one will be use...\n"));
        }
        if (ParseSkySphere(&(CurrentFrame->SkySphere)))
          return 1;
        break;
      case Tbicubic_patch:
        if (ParseBicubicPatch(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tblob:
        if (ParseBlob(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tbox:
        if (ParseBox(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tcone:
        if (ParseCone(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tcubic:
        if (ParseCubic(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tcylinder:
        if (ParseCylinder(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tdifference:
        if (ParseCSG(CSG_DIFFERENCE, &TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tdisc:
        if (ParseDisc(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Theight_field:
        if (ParseHeightField(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tintersection:
        if (ParseCSG(CSG_INTERSECTION, &TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tjulia_fractal:
        if (ParseJuliaFractal(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tlathe:
        if (ParseLathe(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tmerge:
        if (ParseCSG(CSG_MERGE, &TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tmesh:
        if (ParseMesh(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tplane:
        if (ParsePlane(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tpoly:
        if (ParsePoly(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tpolygon:
        if (ParsePolygon(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tprism:
        if (ParsePrism(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tquadric:
        if (ParseQuadric(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tquartic:
        if (ParseQuartic(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tsmooth_triangle:
        if (ParseSmoothTriangle(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tsor:
        if (ParseSor(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tsphere:
        if (ParseSphere(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tsuperellipsoid:
        if (ParseSuperEllipsoid(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Ttext:
        if (ParseText(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Ttorus:
        if (ParseTorus(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Ttriangle:
        if (ParseTriangle(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tunion:
        if (ParseCSG(CSG_UNION, &TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
      case Tobject:
        OldToken = t; IsOldToken = 1;
        if (ParseObject(&TmpObject))
          return 1;
        TmpObject->Next = CurrentFrame->FirstObject;
        CurrentFrame->FirstObject = TmpObject;
        break;
    }
    if (NextToken(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  LoadPov
******************************************************************************/
void LoadPov(GtkWidget *w, GtkFileSelection *FS)
{
  ObjectStruct *TmpObject;
  GtkWidget *menu_item;
  char *Name;

  Name = g_strdup(gtk_file_selection_get_filename(GTK_FILE_SELECTION(FS)));
  
  InitLoadPov(Name);
  PovFile = g_malloc(sizeof(ListFile));
  PovFile->Prev = PovFile->Next = NULL;
  PovFile->Line = 1;
  PovFile->handle = LibOpen(Name);
  PovFile->Name = g_malloc(strlen(Name)+1);
  strcpy(PovFile->Name, Name);

  if (ParsePov())
  { /* There's an error in the parsing -> we must free the memory. */
    /* FIXME: DestroyFrame(CurrentFrame); */
    return;
  }

  for (TmpObject = CurrentFrame->FirstObject ; TmpObject ; 
       TmpObject = TmpObject->Next)
  {
    ComputeObjectTriangleMesh(TmpObject);
  }
  ComputeCamera(CurrentFrame->Camera);
  if (CSGTreeDialog)
  {
    menu_item = gtk_radio_menu_item_new_with_label(CSGTreeMenuGroup,
                                                   CurrentFrame->Name);
    gtk_signal_connect(GTK_OBJECT(menu_item), "activate",
                       (GtkSignalFunc)RebuildCSGTree, CurrentFrame);
    if (!CSGTreeMenuGroup)
      CSGTreeMenuGroup = gtk_radio_menu_item_group(GTK_RADIO_MENU_ITEM(menu_item));
    gtk_menu_append(GTK_MENU(CSGTreeMenu), menu_item);
    gtk_object_set_data(GTK_OBJECT(menu_item), "CSGMenu", CurrentFrame);
    gtk_widget_show(menu_item);
  }
  /* XXX: the created views should be chosen from the 'Preferences' */
  NewView(CurrentFrame, ORTHO_XY_CAMERA);
  NewView(CurrentFrame, ORTHO_XZ_CAMERA);
  NewView(CurrentFrame, ORTHO_ZY_CAMERA);
  NewViewCamera(CurrentFrame);
}

